#include "RendererGLUT.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "VKString.h"

int main(int argc, char** argv )
{
	if (argc<2)
	{
		std::cout<<"Usage: ./"<<argv[0]<<" inputDir [optional]"<<std::endl;
		std::cout<<"Optional:"<<std::endl;
		std::cout<<" -model modelName : only read single model"<<std::endl;
		std::cout<<" -dirType type : MyFormat (default), ColladaDirs, MeshFiles"<<std::endl;
		std::cout<<" -name dbName : specify name"<<std::endl;
		std::cout<<" -write writeDir : specify output directory"<<std::endl;
		std::cout<<"Reading Flags:"<<std::endl;
		std::cout<<" -manifold : read 3D meshes as manifold (polygon soup by default)"<<std::endl;
		std::cout<<"Create Flags:"<<std::endl;
		std::cout<<" -allVtx : mesh vertices"<<std::endl;
		std::cout<<" -allFace : mesh faces"<<std::endl;
		std::cout<<" -geod : geodesic distances"<<std::endl;
		std::cout<<" -eucl : euclidean distances"<<std::endl;
		std::cout<<" -vis : visualize"<<std::endl;
		exit(0);
	}
	
	bool visualize = false;
	ModelDatabase::InputDirFormat dirFormat = ModelDatabase::DBINPUTDIR_MyFormat;
	VKString dbDir = "none";
	VKString dbName = "none";
	int createFlags = 0;
	int readFlags = ModelDatabase::EVERYTHING;
	VKStringList readModels;
	
	// parse params

	for (int i=2; i<argc; i++)
	{
		if (strcmp(argv[i], "-name")==0)
			dbName = VKString(argv[++i]);
		else if (strcmp(argv[i], "-model")==0)
			readModels.push_back(argv[++i]);
		else if (strcmp(argv[i], "-dirType")==0)
		{
			i++;
			if (strcmp(argv[i], "MeshFiles")==0)
				dirFormat = ModelDatabase::DBINPUTDIR_MeshFiles;
			else if (strcmp(argv[i], "MyFormat")==0)
				dirFormat = ModelDatabase::DBINPUTDIR_MyFormat;
			else if (strcmp(argv[i], "ColladaDirs")==0)
				dirFormat = ModelDatabase::DBINPUTDIR_ColladaDirs;
			else
				dirFormat = ModelDatabase::DBINPUTDIR_Unknown;
		}
		else if (strcmp(argv[i], "-write")==0)
			dbDir = VKString(argv[++i]);
		else if (strcmp(argv[i], "-allVtx")==0)
			createFlags |= ModelDatabase::CREATE_PNTS_ALLVTX;
		else if (strcmp(argv[i], "-allFace")==0)
			createFlags |= ModelDatabase::CREATE_PNTS_ALLFACES;
		else if (strcmp(argv[i], "-geod")==0)
			createFlags |= ModelDatabase::CREATE_DIST_EUCLIDEAN;
		else if (strcmp(argv[i], "-eucl")==0)
			createFlags |= ModelDatabase::CREATE_DIST_GEODESIC;
		else if (strcmp(argv[i], "-manifold")==0)
		{
			readFlags |= ModelDatabase::READ_MESH_AS_MANIFOLD;
			readFlags &= ~(ModelDatabase::READ_MESH_AS_SOUP);
		}			
		else if (strcmp(argv[i], "-vis")==0)
			visualize = true;
		else
			std::cout<<"[WARNING] Unknown arg: "<<argv[i]<<std::endl;
	}
	

	ModelDatabase * db = ModelDatabase::CreateDBFromDir(VKString(argv[1]), dirFormat, readModels,
														readFlags, createFlags, dbDir, dbName);

	if (dbDir!="none")
		db->WriteDB();
	
	// todo: maybe create corrs
	DBCorrs * dbCorrs = NULL;
	std::cout<<std::flush;
	if (visualize)
	{	
#ifdef VISUALIZATION	
		int listFlags = DatabaseVisWindow::LIST_GROUPS_MODELS;
		
		if ((createFlags & ModelDatabase::CREATE_DIST_GEODESIC) 
			|| (createFlags & ModelDatabase::CREATE_DIST_GEODESIC) )
			listFlags |= DatabaseVisWindow::LISTGP_GeomDists;
		
		if (dbCorrs)
			listFlags = DatabaseVisWindow::LIST_GROUPS_CORRS;
		
		DatabaseVisWindow * dbVis = new DatabaseVisWindow(db, dbCorrs, listFlags);
		VisWindowRenderer * wndGLUT = new RendererGLUT(argc, argv, dbVis);
#else
		std::cout<<"[WARNING] Visualization disabled (compiling flag)"<<std::endl;
#endif
	}
	return 0;
}


