#include "FileManagement.h"
#include "VKString.h"
#include "TimeProfiler.h"

#include "RendererGLUT.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "ShapeDistance.h"

#include "ShapeTriangleSoup.h"
#include "Model3DColladaParser.h"

ShapePointCloud * ReadFuzzyFeatPnts(Model3D * m, const VKString & filename);
void WriteNewGTFeatPnts(ShapePointCloud * featPnts, const VKString & filename);

int main(int argc, char** argv )
{
	if (argc<2)
	{
		std::cout<<"Usage: ./"<<argv[0]<<" inputDir [params] [optional]\n";
		std::cout<<"Params:\n";
		std::cout<<" -dirType type : MyFormat (default), ColladaDirs, MeshFiles\n";
		std::cout<<" -writeDir outDirName : specify output directory\n";
		std::cout<<" -type (see options below)\n";
		std::cout<<"     MyFormat : simply convert to 'my database' format\n";
		std::cout<<"     ASCIIOff : convert to ascii off format\n";
		std::cout<<"     ASCIIObj : convert to ascii obj format\n";
		std::cout<<"     Screenshot : take screenshot of geometry\n";
		std::cout<<"     MidEdgeFlat : create flattened mid-edge mesh (for conformal maps)\n";
//		std::cout<<"     ReadSD : \n";
//		std::cout<<"     QSlim : \n";
		std::cout<<" -inGeomName geomName : process geometry with name\n";
		std::cout<<" -outGeomName geomName : write geometry with name\n";
		std::cout<<" -outGeomIsFile : outGeomName is a filename (otherwise writes in MyFormat)\n";
		std::cout<<"Optional:\n";
		std::cout<<" -verbose : be verbose\n";
		std::cout<<" -process : process only if file does not exist yet\n";
		std::cout<<" -write : write file\n";
		std::cout<<" -vis : visualize file\n";
		std::cout<<" -rearrangeAxes newx newy newz : change axis order (permutation of x y z)\n";
		std::cout<<" -html : print screenshots to html file\n";
		std::cout<<" -model modelName : model to process (use directory name for MyFormat and ColladaDirs)\n";
		std::cout<<" -mlabel modelLabels : read a separate segmentation file (Sid'11 style)\n";
		std::cout<<" -perFaceSeg segFile segNames : ascii file with integer ids for each face, segNames - can be none\n";
		std::cout<<" -fuzzyFeatPnts filename : fuzzy corrs filename\n";
		std::cout<<" -writeNewGT filename : write filename\n";
		exit(0);
	}
	bool verbose = false;
	bool outGeomInMyFormat = true;
	bool html = false;
	bool processIfNoFile = false;
	bool visualize = false;
	bool write = false;
	bool bRearrangeAxes = false;
	char newAxesOrder[3];
	int newAxesSign[3];
	for (int i=0; i<3; i++)
		newAxesOrder[i] = 'a';
	for (int i=0; i<3; i++)
		newAxesSign[i] = 0;
	VKString fuzzyFeatPnts = "none";
	ModelDatabase::InputDirFormat dirFormat = ModelDatabase::DBINPUTDIR_MyFormat;
	VKString dbDir(argv[1]);
	VKString outDir("./");
	VKString dbName = "none";
	VKString outGeomName = "none";
	VKString writeNewGT = "none";
	int createFlags = ModelDatabase::CREATE_DIST_EUCLIDEAN | ModelDatabase::CREATE_DIST_GEODESIC;
	int readFlags = ModelDatabase::READ_MESH_AS_SOUP | ModelDatabase::READ_MESH_AS_MANIFOLD;
	int writeFlags = ModelDatabase::WRITE_MODEL3D_FILES | ModelDatabase::WRITE_ARG_SINGLE_GEOM;	
		
	// these have to be specified
	bool keepLog = false;
	int nPnts = 0;
	VKString type = "none";
	VKString distName = "none";
	VKString inGeomName = Model3D::m_sMainGeomName;
	bool useStr = false;
	VKString modelToProcess="all";
	VKString mLabelFile = "none";
	
	VKString perFaceSegFile = "none";
	VKString segNamesFile = "none";
	
	// parse params	
	for (int i=2; i<argc; i++)
	{
		if (strcmp(argv[i], "-dirType")==0)
		{
			i++;
			if (strcmp(argv[i], "MeshFiles")==0)
				dirFormat = ModelDatabase::DBINPUTDIR_MeshFiles;
			else if (strcmp(argv[i], "MyFormat")==0)
				dirFormat = ModelDatabase::DBINPUTDIR_MyFormat;
			else if (strcmp(argv[i], "ColladaDirs")==0)
				dirFormat = ModelDatabase::DBINPUTDIR_ColladaDirs;
			else
				dirFormat = ModelDatabase::DBINPUTDIR_Unknown;
		}
		else if (strcmp(argv[i], "-writeDir")==0)
			outDir = VKString(argv[++i]);
		else if (strcmp(argv[i], "-type")==0)
			type = VKString(argv[++i]);
		else if (strcmp(argv[i], "-model")==0)
			modelToProcess = VKString(argv[++i]);
		else if (strcmp(argv[i], "-mlabel")==0)
			mLabelFile = VKString(argv[++i]);
		else if (strcmp(argv[i], "-perFaceSeg")==0)
		{
			perFaceSegFile = VKString(argv[++i]);
			segNamesFile = VKString(argv[++i]);;
		}
		else if (strcmp(argv[i], "-fuzzyFeatPnts")==0)
			fuzzyFeatPnts = VKString(argv[++i]);
		else if (strcmp(argv[i], "-process")==0)
			processIfNoFile = true;
		else if (strcmp(argv[i], "-vis")==0)
			visualize = true;
		else if (strcmp(argv[i], "-verbose")==0)
			verbose = true;
		else if (strcmp(argv[i], "-inGeomName")==0)
			inGeomName = VKString(argv[++i]);
		else if (strcmp(argv[i], "-outGeomName")==0)
			outGeomName = VKString(argv[++i]);
		else if (strcmp(argv[i], "-write")==0)
			write = true;
		else if (strcmp(argv[i], "-html")==0)
			html = true;
		else if (strcmp(argv[i], "-log")==0)
			keepLog = true;
		else if (strcmp(argv[i], "-outGeomIsFile")==0)
			outGeomInMyFormat = false;
		else if (strcmp(argv[i], "-rearrangeAxes")==0)
		{
			bRearrangeAxes = true;
			for (int d=1; d<=3; d++)
			{
				int neg = (int)(argv[i+d][0]=='-');
				newAxesOrder[d-1] = argv[i+d][neg];
				newAxesSign[d-1] = (int)(pow(-1., (double)neg));
			}				
			i += 3;
		}
		else if (strcmp(argv[i], "-manifold")==0)
		{
			readFlags |= ModelDatabase::READ_MESH_AS_MANIFOLD;
			readFlags &= ~(ModelDatabase::READ_MESH_AS_SOUP);	
		}
		else if (strcmp(argv[i], "-writeNewGT")==0)
			writeNewGT = VKString(argv[++i]);
		else
			std::cout<<"[WARNING] Unknown arg: "<<argv[i]<<std::endl;
	}

	if (outGeomName=="none")
	{
		std::cout<<"[ERROR] output geomName name not defined. Use -outGeomName name"<<std::endl;
		return 0;
	}
	
	if (dirFormat==ModelDatabase::DBINPUTDIR_ColladaDirs)
	{
#ifdef OPEN_COLLADA
		Model3D::AddParser(".dae", new Model3DColladaParserStatic());
#else
		std::cout<<"[ERROR] Collada parser not implemented (include OPEN_COLLADA in Makefile)"<<std::endl;
#endif
	}
	
	TimeProfiler timer;
	VKStringList allModels;
	if (modelToProcess=="all")
	{
		std::cout<<"Getting directory listing: "<<dbDir.c_str()<<std::endl;
		allModels = ModelDatabase::GetAllModelsInDir(dbDir, dirFormat);
	}
	else
		allModels.push_back(modelToProcess);
	
	VKStringList mName, mFile;
	ModelDatabase::GetModelFiles(dbDir, dirFormat, allModels, mName, mFile);
	assert(mName.count()==allModels.count() && mFile.count()==allModels.count());

	if (allModels.count()>1)
		std::cout<<"Creating geometries: "<<allModels.count()<<std::endl;
	else if (allModels.count()==0)
		std::cout<<"No geometries."<<std::endl;
	ModelDatabase * db = NULL;

	VKStringList screenshotModelFiles;
	VKStringList screenshotModelNames;
	std::vector<VKStringList> screenshotImageFiles;
	screenshotImageFiles.push_back(VKStringList());
	VKStringList htmlImages;
	VKStringList htmlText;
	if (html)
		screenshotImageFiles.push_back(VKStringList());
	for (int i=0; i<allModels.count(); i++)
	{
		if (allModels.count()>1)
			timer.WriteProgress("ProcModels", i, allModels.count());
		
		VKString mDir = ModelDatabase::GetModelDirname(outDir, mName[i]);
				
		VKString fileWrite;
		if (!outGeomInMyFormat)
			fileWrite = outDir+"/"+outGeomName;
		else if (type=="MyFormat")	// simply parse
			fileWrite = ModelDatabase::GetModelFilename(outDir, mName[i]);
		else if (type=="ASCIIOff")
		{
			VKString gDir = ModelDatabase::GetModelGeometryDir();
			VKString mDir = ModelDatabase::GetModelDirname(outDir,mName[i]);
			fileWrite = mDir+"/"+gDir+"/"+outGeomName;
			if (!outGeomName.endsWith(".off"))
				fileWrite+=".off";
		}
		else if (type=="ASCIIObj")
		{
			VKString gDir = ModelDatabase::GetModelGeometryDir();
			VKString mDir = ModelDatabase::GetModelDirname(outDir,mName[i]);
			fileWrite = mDir+"/"+gDir+"/"+outGeomName;
			if (!outGeomName.endsWith(".obj"))
				fileWrite+=".obj";
		}
		else if (type=="Screenshot")
		{
			fileWrite = mDir+"/"+ModelDatabase::GetModelGeometryDir()+"/"+outGeomName+".jpg";
			htmlImages.push_back(VKString("./Thumbnails/")+mName[i]+".jpg");
			htmlText.push_back(mName[i]);
		}
		else if (type=="MidEdgeFlat")
			fileWrite = mDir + "/"+ModelDatabase::GetModelGeometryDir()+"/"+outGeomName+".flat.ifs";
		else
		{
			std::cout<<"[ERROR] Unknown geometry type: "<<type.c_str()<<std::endl;
			exit(0);
		}

		if (processIfNoFile && FileManagement::FileExists(fileWrite.c_str()))
			continue;

		// load model
		Model3D * model = NULL;
		if (type!="Screenshot")
		{
			model = Model3D::CreateFromFile(mFile[i], readFlags, createFlags, mName[i]);
			if (bRearrangeAxes)
				model->RearrangeAxes(newAxesOrder, newAxesSign);
			
			if (mLabelFile!="none" || perFaceSegFile != "none")
			{
				GeomShape * shape = model->GetGeomShape(Model3D::m_sMainGeomName);
				assert(shape!=NULL);
				ShapeLabels * labels = new ShapeLabels(shape, ShapeValue::m_sDefaultSceneGraphLabels);
				shape->AddShapeVal(labels->GetShapeValueName(), labels);
				if (mLabelFile!="none")
					labels->ReadFromLabFile(mLabelFile);
				else
					labels->ReadFromSegFile(perFaceSegFile, segNamesFile);
			}
		}

		// fuzzy feature points
		ShapePointCloud * featPnts = NULL;
		if (fuzzyFeatPnts!="none")
			featPnts = ReadFuzzyFeatPnts(model, fuzzyFeatPnts);

		if (writeNewGT!="none")
			WriteNewGTFeatPnts(featPnts, writeNewGT);
		
		// create geometry
		if (type=="MyFormat")	// simply parse
		{
			model->WriteModelDBFormat(mDir, fileWrite);
			
			if (featPnts!=NULL)
			{
				VKString pntsFile = ModelDatabase::GetModelDirname(outDir, mName[i])+"/Geometry/main_featpnts.pnts";
				std::cout<<"Writing geometry: "<<pntsFile.c_str()<<std::endl;
				featPnts->WriteGeometryFile(pntsFile);
			}			
		}
		else if (type=="ASCIIOff")	// convert to plain ascii
		{
			GeomShape * shape = model->GetGeomShape(inGeomName);
			assert(shape!=NULL);
			ShapeManifoldMesh * manifold = shape->SafeCastToManifold();
			ShapeTriangleSoup * polysoup = shape->SafeCastToTriangleSoup();
			if (verbose)
				std::cout<<"Writing ASCII: "<<fileWrite.c_str()<<std::endl;
			if (manifold!=NULL)		// R3Mesh
				manifold->GetMesh()->WriteOffFile(fileWrite.c_str());
			else if (polysoup!=NULL)	// my triangle soup
				polysoup->WriteOff(fileWrite, ShapeTriangleSoup::OFF_ASCII);
			else
				assert(false);	// don't know how to convert
		}
		else if (type=="ASCIIObj")	// convert to plain ascii
		{
			GeomShape * shape = model->GetGeomShape(inGeomName);
			assert(shape!=NULL);
			ShapeManifoldMesh * manifold = shape->SafeCastToManifold();
			ShapeTriangleSoup * polysoup = shape->SafeCastToTriangleSoup();
			if (verbose)
				std::cout<<"Writing ASCII: "<<fileWrite.c_str()<<std::endl;
			if (manifold!=NULL)		// R3Mesh
				manifold->GetMesh()->WriteObjFile(fileWrite.c_str());
			else if (polysoup!=NULL)	// my triangle soup
				polysoup->WriteObj(fileWrite);
			else
				assert(false);	// don't know how to convert
		}
		else if (type=="Screenshot")
		{
			screenshotModelFiles.push_back(mFile[i]);
			screenshotModelNames.push_back(mName[i]);
			screenshotImageFiles[0].push_back(fileWrite);
			if (html)
				screenshotImageFiles[1].push_back(outDir+"/Thumbnails/"+mName[i]+".jpg");
		}
		else if (type=="MidEdgeFlat")
		{
			assert(false);
//			assert(write);
//			GeomShape * shape = model->GetGeomShape(inGeomName);
//			assert(shape!=NULL);
//			ShapeManifoldMesh * mesh = shape->SafeCastToManifold();
//			assert(mesh!=NULL);
//			ShapeManifoldMesh * flat = MapToPlane::CreateMidEdgeFlat(mesh);
//			assert(flat!=NULL);
//			flat->WriteGeometryFile(fileWrite);
//			delete flat;
		}
		else
		{
			std::cout<<"[ERROR] Unknown geometry type: "<<type.c_str()<<std::endl;
			exit(0);
		}
		
		// cleanup
		delete model;
	}
	
	
	
	if (visualize || screenshotModelFiles.count()>0)
	{	
#ifdef VISUALIZATION
		assert(db!=NULL);	// will fail - need to init  
		int listFlags = DatabaseVisWindow::LISTGP_Models | DatabaseVisWindow::LISTGP_Geoms;
		listFlags |= DatabaseVisWindow::LISTGP_MeshFlags;
		DatabaseVisWindow * dbVis = new DatabaseVisWindow(db, NULL, listFlags);
		
		VisWindowRenderer * wndGLUT = new RendererGLUT(argc, argv, dbVis);
#else
		std::cout<<"[WARNING] Visualization disabled (compiling flag)"<<std::endl;
#endif
	}
	return 0;
}

ShapePointCloud * ReadFuzzyFeatPnts(Model3D * m, const VKString & filename)
{
	// read points
	VKString tmpStr;
	
	ifstream textStream(filename.c_str());
	assert(textStream.is_open());
	tmpStr.readLine(textStream);	assert(tmpStr=="Format PointSet");
	tmpStr.readLine(textStream);
	tmpStr.readLine(textStream);
	tmpStr.readToken(textStream);	assert(tmpStr=="NPoints");
	int npnts;
	textStream>>npnts;
	tmpStr.readLine(textStream);
	
	
	GeomShape * shape = m->GetGeomShape(Model3D::m_sMainGeomName);
	assert(shape!=NULL && shape->SafeCastToMesh()!=NULL);
	ShapeMesh * mesh = shape->SafeCastToMesh();
	ShapePointCloud * pnts = new ShapePointCloud(mesh, "FeatPnts");
	
	for (int i=0; i<npnts; i++)
	{
		tmpStr.readLine(textStream);		assert(tmpStr=="ModelPoint 1");
		double x,y,z;
		int pid, vid, tid;
		textStream>>x>>y>>z;
		textStream>>pid>>vid>>tid;
		assert(vid>=0 && pid==-1 && tid==-1);
		tmpStr.readLine(textStream);
		tmpStr.readLine(textStream);
		tmpStr.readLine(textStream);
		pnts->AddPoint(ShapePoint(mesh, vid));
		
	}
	textStream.close();
	
	return pnts;
}

void WriteNewGTFeatPnts(ShapePointCloud * featPnts, const VKString & filename)
{
	FileManagement::CreateDirsOnPathToFile(filename.c_str());
	FILE * binStream = fopen((filename+".bin").c_str(), "wb");
	assert(binStream!=NULL);
	
	std::ofstream textStream((filename+".txt").c_str());
	assert(textStream.is_open());
	int fid;
	double b1, b2, b3;
	const ShapeMesh * mesh;
	textStream<<"VALID\n";
	int npnts =featPnts->NumPoints();
	char cv = 'V';
	fwrite(&cv, sizeof(char), 1, binStream);
	fwrite(&npnts, sizeof(int), 1, binStream);
	for (int i=0; i<featPnts->NumPoints(); i++)
	{
		R3Point fp = featPnts->PointPosition(i);
		double x = fp.X(), y = fp.Y(), z = fp.Z();
		featPnts->GetPoint(i).GetBarycentricCoords(&fid, &b1, &b2, &mesh);
		textStream<<fid<<" "<<b1<<" "<<b2<<" "<<(1.-b1-b2)<<" "<<x<<" "<<y<<" "<<z<<"\n";
		fwrite(&fid, sizeof(int), 1, binStream);
		fwrite(&b1, sizeof(double), 1, binStream);
		fwrite(&b2, sizeof(double), 1, binStream);
		fwrite(&b3, sizeof(double), 1, binStream);
		fwrite(&x, sizeof(double), 1, binStream);
		fwrite(&y, sizeof(double), 1, binStream);
		fwrite(&z, sizeof(double), 1, binStream);
	}
	
	textStream.close();
	fclose(binStream);
}






