#include "VKString.h"
#include "Sortable.h"
#include "FileManagement.h"

#include "RendererGLUT.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "ShapeTriangleSoup.h"

#include "STemplate.h"
#include "VisShapeTemplate.h"
#include "AppScreenshots.h"
#include "TMatch.h"
#include "RendererOSMesa.h"
#include "TMatchAlg.h"
#include "STemplateNode.h"
#include "InitWordNet.h"

#include "CreateTemplate.h"
#include "AutoSegment.h"

int main(int argc, char** argv )
{
	if (argc<2)
	{
		std::cout<<"Usage:"<<std::endl;
		std::cout<<"   "<<argv[0]<<" [args]"<<std::endl;;
		std::cout<<"Models to read:"<<std::endl;
		std::cout<<"      -readDB dbDir"<<std::endl;
		std::cout<<"      -model modelName"<<std::endl;
		std::cout<<"      -models modelList.txt"<<std::endl;
		
		std::cout<<"Creating/reading template:"<<std::endl;
		std::cout<<"      -initTmplt topologyFile geomName labelsName"<<std::endl;
		std::cout<<"      -initFromScratch "<<std::endl;
		std::cout<<"      -initFromPartDir dirName "<<std::endl;
		std::cout<<"      -initFromWordnet file "<<std::endl;
		std::cout<<"      -initFromGeometry modelName nsegs "<<std::endl;
		std::cout<<"      -train modelName trainGeomName trainLabelsName"<<std::endl;	// example segmentation
		std::cout<<"      -trainLabelsFile filename"<<std::endl;	// example segmentation
		std::cout<<"      -readTemplate templateFile"<<std::endl;						// pre-analyzed tempalte
		std::cout<<"      -readManifold - input meshes are manifolds"<<std::endl;

		std::cout<<"Fit template (segmentation, optimal t. configuration):"<<std::endl;
		std::cout<<"      -match modelName geomName"<<std::endl;
		std::cout<<"      -matchLabeled modelName geomName labelsName"<<std::endl;
		std::cout<<"      -matchLabeledAndE modelName geomName labelsName"<<std::endl;
		std::cout<<"      -matchLabeledAndEM modelName geomName labelsName"<<std::endl;
		std::cout<<"      -includeNull - include new nodes to template"<<std::endl;
		std::cout<<"      -includeNullOnly - only keep clusters of null nodes"<<std::endl;
		std::cout<<"      -updateNodes - update statistics from the segmentation"<<std::endl;
		std::cout<<"      -refGeom geomName - reference geometry name"<<std::endl;

		std::cout<<"Fit template parameters:"<<std::endl;
		std::cout<<"      -localAxes valName - local orientation at points "<<std::endl;
		std::cout<<"      -alignments [AllXYZ, AllXY, IdentityOnly] - initial alignments "<<std::endl;
		
		std::cout<<"Output:"<<std::endl;
		std::cout<<"      -writeTemplate templateFile"<<std::endl;
		std::cout<<"      -writeTemplateMetaData templateInfo.txt"<<std::endl;
		std::cout<<"      -writeMatching matching.txt"<<std::endl;
		std::cout<<"      -writeBinMatch file.match"<<std::endl;
		std::cout<<"      -writeSeedHierarchy hierarchy.txt"<<std::endl;
		std::cout<<"      -readBinMatch file.match"<<std::endl;
		std::cout<<"      -writeXFormedModel meshfile"<<std::endl;
        std::cout<<"      -writeLabelsExternal dirname"<<std::endl;
		std::cout<<"      -evalMatching eval.txt"<<std::endl;
		std::cout<<"      -equivLabels equivalence.txt"<<std::endl;
		std::cout<<"      -shotTemplate imgfile "<<std::endl;
		std::cout<<"      -shotCoAligned imgfile "<<std::endl;
		std::cout<<"      -readCoAligned binFile modelsFile threshold "<<std::endl;
		std::cout<<"      -useMeshThresh value "<<std::endl;
		std::cout<<"      -splitLabelsIntoCC "<<std::endl;
		std::cout<<"      -renameNodes perNodeVotesFile"<<std::endl;
		std::cout<<"      -vis - visualize"<<std::endl;

		exit(0);
	}
	
	VKString writeSeedHierarchy = "none";
	VKString refGeom = "none";
	bool interact = false;
	VKString initTmplt = "none";
	bool visualize = false;
	ModelDatabase::InputDirFormat dirFormat = ModelDatabase::DBINPUTDIR_MyFormat;
	VKString dbDir = "none";
	VKString dbName = "none";
	std::map<VKString, VKString> shotTypeToImage;
	VKStringList modelsToLoad;
	int createFlags = 0;
	int readFlags = 0;
	
	VKString matchingFile = "none";
	VKString evalFile = "none";
	VKString equivLabels = "none";
	VKString trainModel = "none";
	VKString trainGeomName="none", trainLabelsName="none", trainLabelsFile="none";
	
	VKString matchModel="none";
	VKString matchGeom="none", matchSegName="none";
	bool multiSeg = false;
	
	VKString axesValName = "none";
	
	VKString readTemplate = "none";
	VKString writeTemplate = "none";
	
	VKString readBinMatch = "none";
	VKString writeBinMatch = "none";
	VKString renameNodes = "none";
	
	bool includeNull = false;
	bool includeNullOnly = false;
	bool updateNodes = false;
	bool verbose = false;
	bool readManifold = false;
	
	bool trustLabels = false;	// vk: const
	VKString writeXFormedModel = "none";
	VKString writeTemplateMetaData = "none";
	VKString writeLabelsExternal = "none";
	
	TMatchAlgParams matchParams;
	double useMeshThresh = -1;	//never use mesh
	
	VKString initFromPartDir = "none";
	VKString initFromGeometry = "none";
	int nsegs = -1;
	VKString readCoAligned = "none";
	VKString coAlignedModelsFile = "none";
	double coAlignedThreshold = 0;

	VKString initFromWordnet = "none";
	
	char relabelType = '0';
	bool splitLabelsIntoCC = false;
	
	for (int i=1; i<argc; i++)
	{
		if (strcmp(argv[i], "-readDB")==0)
			dbDir = argv[++i];
		else if (strcmp(argv[i], "-model")==0)
		{
			VKString currM(argv[++i]);
			if (!modelsToLoad.contains(currM))
				modelsToLoad.push_back(currM);
		}
		else if (strcmp(argv[i], "-models")==0)
			ReadModelNames(modelsToLoad, argv[++i]);
		else if (strcmp(argv[i], "-train")==0)
		{
			trainModel = VKString(argv[++i]);
			trainGeomName = argv[++i];
			trainLabelsName = argv[++i];
		}
		else if (strcmp(argv[i], "-trainLabelsFile")==0)
			trainLabelsFile = argv[++i];
		else if (strcmp(argv[i], "-initTmplt")==0)
		{
			initTmplt = VKString(argv[++i]);
			trainGeomName = VKString(argv[++i]);
			trainLabelsName = VKString(argv[++i]);
		}
		else if (strcmp(argv[i], "-initFromGeometry")==0)
		{
			initFromGeometry = VKString(argv[++i]);
			nsegs = atoi(argv[++i]);
		}
		else if (strcmp(argv[i], "-splitLabelsIntoCC")==0)
			splitLabelsIntoCC = true;
		else if (strcmp(argv[i], "-initFromScratch")==0)
			interact = true;
		else if (strcmp(argv[i], "-initFromPartDir")==0)
			initFromPartDir = VKString(argv[++i]);
		else if (strcmp(argv[i], "-initFromWordnet")==0)
			initFromWordnet = VKString(argv[++i]);
		else if (strcmp(argv[i], "-refGeom")==0)
			refGeom = VKString(argv[++i]);
		else if (strcmp(argv[i], "-renameNodes")==0)
			renameNodes = VKString(argv[++i]);
		else if (strcmp(argv[i], "-match")==0)
		{
			matchModel = argv[++i];
			matchGeom = argv[++i];
		}
		else if (strcmp(argv[i], "-matchLabeled")==0
				 || strcmp(argv[i], "-matchLabeledAndE")==0 || strcmp(argv[i], "-matchLabeledAndEM")==0)
		{
			if (strcmp(argv[i], "-matchLabeledAndE")==0)
				relabelType = 'e';
			else if (strcmp(argv[i], "-matchLabeledAndEM")==0)
				relabelType = 'm';
			matchModel = argv[++i];
			matchGeom = argv[++i];
			matchSegName = argv[++i];
		}
		else if (strcmp(argv[i], "-matchGeomName")==0)
			matchGeom = argv[++i];
		else if (VKString(argv[i]).startsWith("-shot"))
		{
			VKString shotType(argv[i]);
			shotType.replace("-shot", "");
			shotTypeToImage[shotType] = VKString(argv[++i]);
		}
		else if (strcmp(argv[i], "-readBinMatch")==0)
			readBinMatch = argv[++i];
		else if (strcmp(argv[i], "-readManifold")==0)
			readManifold = true;
		else if (strcmp(argv[i], "-writeBinMatch")==0)
			writeBinMatch = argv[++i];
		else if (strcmp(argv[i], "-includeNull")==0)
			includeNull = true;
		else if (strcmp(argv[i], "-includeNullOnly")==0)
			includeNullOnly = true;
		else if (strcmp(argv[i], "-updateNodes")==0)
			updateNodes = true;
		else if (strcmp(argv[i], "-readTemplate")==0)
			readTemplate = argv[++i];
		else if (strcmp(argv[i], "-writeTemplate")==0)
			writeTemplate = argv[++i];
		else if (strcmp(argv[i], "-writeMatching")==0)
			matchingFile = argv[++i];
		else if (strcmp(argv[i], "-evalMatching")==0)
			evalFile = argv[++i];
		else if (strcmp(argv[i], "-equivLabels")==0)
			equivLabels = argv[++i];
		else if (strcmp(argv[i], "-localAxes")==0)
			axesValName = argv[++i];
		else if (strcmp(argv[i], "-writeSeedHierarchy")==0)
			writeSeedHierarchy = argv[++i];
		else if (strcmp(argv[i], "-alignments")==0)
		{
			i++;
			if (strcmp(argv[i], "IdentityOnly")==0)
				matchParams.m_InitAlignments = TMatchAlgParams::ALIGN_IdentityOnly;
			else if (strcmp(argv[i], "AllXYZ")==0)
				matchParams.m_InitAlignments = TMatchAlgParams::ALIGN_AllXYZ;
			else if (strcmp(argv[i], "AllXY")==0)
				matchParams.m_InitAlignments = TMatchAlgParams::ALIGN_AllXY;
			else if (strcmp(argv[i], "AllPCA")==0)
				matchParams.m_InitAlignments = TMatchAlgParams::ALIGN_AllPCA;
			else
				assert(false);
		}
		else if (strcmp(argv[i], "-vis")==0)
			visualize = true;
		else if (strcmp(argv[i], "-verb")==0)
			verbose = true;
		else if (strcmp(argv[i], "-writeXFormedModel")==0)
			writeXFormedModel = argv[++i];
		else if (strcmp(argv[i], "-writeTemplateMetaData")==0)
			writeTemplateMetaData = argv[++i];
		else if (strcmp(argv[i], "-writeLabelsExternal")==0)
			writeLabelsExternal = argv[++i];
		else if (strcmp(argv[i], "-readCoAligned")==0)
		{
			readCoAligned = argv[++i];
			coAlignedModelsFile = argv[++i];
			coAlignedThreshold = atof(argv[++i]);
		}
		else if (strcmp(argv[i], "-useMeshThresh")==0)
			useMeshThresh = atof(argv[++i]);
		else
			std::cout<<"[WARNING] Unknown arg: "<<argv[i]<<" in CreateTemplate.cpp"<<std::endl;
	}
//	std::cout<<"Reading match: "<<dbDir.c_str()<<" "<<modelsToLoad[0].c_str()<<std::endl;
		
	DBCorrs * dbCorrs = NULL;	// todo
	ModelDatabase * db = NULL;
	if (modelsToLoad.count()==0)
		modelsToLoad.push_back("LOAD_NO_MODELS");
	if (dbDir!="none")
		db = ModelDatabase::CreateDBFromDir(dbDir, dirFormat, modelsToLoad, readFlags, createFlags, dbDir, dbName);
	else
		db = new ModelDatabase();
		
	STemplate * newTemplate = NULL;
	std::cout<<"Reading / Creating template"<<std::endl;
	if (trainGeomName!="none" && trainLabelsName!="none" && trainModel!="none")
 		newTemplate = SingleTemplate(db, trainModel, trainGeomName, trainLabelsName,
									 trustLabels, useMeshThresh, splitLabelsIntoCC, trainLabelsFile);
	else if (readTemplate!="none")
		newTemplate = STemplate::ReadTemplate(readTemplate, db);
	else if (initTmplt!="none")
		newTemplate = InitTemplate(initTmplt, trainGeomName, trainLabelsName, db);
	else if (initFromPartDir!="none")
		newTemplate = ReadTemplateFromDir(initFromPartDir);
	else if (initFromWordnet!="none")
		newTemplate = CreateTemplateFromWordnetAnalysis(db, initFromPartDir);
	else if (initFromGeometry!="none")
		newTemplate = AutoSegment(db, initFromGeometry, readManifold, nsegs);
//	else
//	{
//		std::cout<<"[WARNING] Unknown template initialization method: "<<std::endl;
//	}
		
	std::vector<TMatchEnergy*> allConfigsEnergies;
	if (readBinMatch!="none")
	{
		std::cout<<"Reading matches"<<std::endl;
		assert(newTemplate!=NULL && db!=NULL);
		FILE * matchBin = fopen(readBinMatch.c_str(), "rb");
		if(matchBin==NULL)
			std::cout<<"[ERROR] Cannot read "<<readBinMatch.c_str()<<std::endl;
		assert(matchBin!=NULL); 
		assert(db->NumModels()<=2);
		// TMatch reader assumes that geometry was loaded
		assert(matchGeom!="none");	// although saved
		for (int i=0; i<db->NumModels(); i++)
			ReadModelGeometry(db, db->GetModel(i)->ModelName(), matchGeom, readManifold);
		allConfigsEnergies.push_back(TMatchEnergy::ReadEnergyAndMatch(matchBin, newTemplate, db));
		allConfigsEnergies[0]->GetMatch()->CreateTemplateSamples();
		fclose(matchBin);
	}
	//std::cout<<"energies = "<<allConfigsEnergies.size()<<std::endl;
	
	// check if template exists
	if (matchModel!="none")
	{
		assert(newTemplate!=NULL);
//		if (verbose)
			std::cout<<"Matching Models: "<<matchModel.c_str()<<std::endl;

		if (matchSegName=="none")
		{ 
			GeomShape * shape = ReadModelGeometry(db, matchModel, matchGeom, readManifold);
			if (readBinMatch!="none")	// start with initial guess
			{
				assert(allConfigsEnergies.size()==1);
				TMatch * match = allConfigsEnergies[0]->GetMatch();
				allConfigsEnergies[0] = TMatchAlg::OptimizeConfigured(newTemplate, match, matchParams);
			}
			else	// new match
				TMatchAlg::Optimize(newTemplate, shape, matchParams, allConfigsEnergies);
		}
		else
		{
			assert(false);
			GeomShape * mesh = ReadModelGeometry(db, matchModel, Model3D::m_sMainGeomName, readManifold, matchSegName);
			ShapeValue * vals = mesh->GetShapeVal(matchSegName);
			assert(vals!=NULL && vals->SafeCastToLabels()!=NULL);
			GeomShape * pnts = ReadModelGeometry(db, matchModel, matchGeom, readManifold);
			ShapeLabels * labels = ShapeLabels::CopyLabels(vals->SafeCastToLabels(), pnts);
			TMatchAlg::OptimizeLabeled(newTemplate, labels, matchParams, allConfigsEnergies, relabelType);
		}
		assert(allConfigsEnergies.size()>0);
		
		// write the best matching
		TMatchEnergy * bestEnergy = allConfigsEnergies[0];
		if (matchingFile!="none")
		{
			// should include: energy, per-point labels, node deformation
			FileManagement::CreateDirsOnPathToFile(matchingFile.c_str());
			std::ofstream textStream(matchingFile.c_str());
			assert(textStream.is_open());
			textStream<<"Score="<<bestEnergy->Energy()<<"\n";
			textStream<<"Time="<<TMatchAlg::s_sProfiler.ProcessLengthStr("1_OptimizeMeasure").c_str()<<"\n";
			textStream.close();
		}
		else
		{
			std::cout<<"Best Energy: "<<bestEnergy->EnergyStr().c_str()<<std::endl;
			std::cout<<"Total Time: "<<TMatchAlg::s_sProfiler.ProcessLengthStr("1_OptimizeMeasure").c_str()<<std::endl;
			std::cout<<TMatchAlg::s_sProfiler.profile().c_str()<<std::endl;
		}
		
		// binary best matching
		if (writeBinMatch!="none")
		{
			assert(bestEnergy!=NULL);
			FILE * matchBin = fopen(writeBinMatch.c_str(), "wb");
			if(matchBin==NULL)
				std::cout<<"[ERROR] Cannot write "<<writeBinMatch.c_str()<<std::endl;
			assert(matchBin!=NULL);
			bestEnergy->WriteEnergyAndMatch(matchBin);
			fclose(matchBin);
		}
		
		// write evaluation (treats SceneGraphNodes labels as ground truth)
		if (evalFile!="none")
			EvalMatching(db, bestEnergy->GetMatch(), evalFile, equivLabels);
				
		// write aligned model
		if (writeXFormedModel!="none")
			WriteAlignedModel(bestEnergy->GetMatch(), writeXFormedModel);
	}
	
	if (renameNodes!="none")
	{
		std::cout<<"Renaming template based on "<<renameNodes.c_str()<<std::endl;
		std::ifstream renameNodesStream(renameNodes.c_str());
		assert(renameNodesStream.is_open());
		std::vector<double> bestScores;
		std::vector<VKString> bestLabels;
		for (int i=0; i<newTemplate->NNodes(); i++)
		{
			bestScores.push_back(-1);
			bestLabels.push_back("keep");			
		}
			
		while (!renameNodesStream.eof())
		{
			bool ok;
			VKString renameLine;
			renameLine.readLine(renameNodesStream);
			//std::cout<<"LINE: "<<renameLine.c_str()<<std::endl;
			VKStringList renameLineList = renameLine.split(" ");
			if (renameLineList.count()>=2)
			{
				VKStringList underscoreSplit = renameLineList[0].split("_");
				assert(underscoreSplit.count()>=2);
				int nodeID = underscoreSplit[0].toInt(&ok);
				VKString mergedNodeName = "";
				for (int i=1; i<underscoreSplit.count(); i++)
					mergedNodeName += underscoreSplit[i];
				assert(ok);
				assert(nodeID < newTemplate->NNodes());
				double score = renameLineList[1].toDouble(&ok);
				//std::cout<<"Node: "<<nodeID<<" -> "<<underscoreSplit[1].c_str()<<" with score="<<score<<std::endl;
				assert(ok);
				if (bestScores[nodeID]==-1 || bestScores[nodeID] < score)
				{
					bestScores[nodeID] = score;
					bestLabels[nodeID] = mergedNodeName;
				}
			}
			else
				continue;
		}
		renameNodesStream.close();
		std::cout<<"Best Names:"<<std::endl;
		for (int i=0; i<(int)bestScores.size(); i++)
		{
			std::cout<<i<<": "<<newTemplate->GetNode(i)->NodeName().c_str()<<" -> "<<bestLabels[i].c_str()<<std::endl;
			if (bestLabels[i] != "keep" && bestLabels[i] != newTemplate->GetNode(i)->NodeName())
			{
				std::cout<<"Changing template node name: "<<i<<": "<<newTemplate->GetNode(i)->NodeName().c_str();
				std::cout<<" -> "<<bestLabels[i].c_str()<<std::endl;
				((STemplateNode*)(newTemplate->GetNode(i)))->SetName(bestLabels[i]);
			}
		}
	}
		
	// write template file
	if (writeTemplate!="none" && newTemplate!=NULL)
	{
		if (verbose)
			std::cout<<"Writing template: "<<writeTemplate.c_str()<<std::endl;
		
		FileManagement::CreateDirsOnPathToFile(writeTemplate.c_str());
//		std::cout<<"Write Template: "<<writeTemplate.c_str()<<std::endl;
		newTemplate->WriteTemplate(writeTemplate);
	}
	
	if (writeSeedHierarchy!="none" && newTemplate!=NULL)
	{
		FileManagement::CreateDirsOnPathToFile(writeSeedHierarchy.c_str());
		FILE * seedHierarchyFile = fopen(writeSeedHierarchy.c_str(), "wb");
		assert(seedHierarchyFile!=NULL);
		int nConfigs = (int)newTemplate->GetConfigHierarchy().size();
		fwrite(&nConfigs, sizeof(int), 1, seedHierarchyFile);
		int * configHierarchy = new int[nConfigs];
		for (int i=0; i<nConfigs; i++)
			configHierarchy[i] = newTemplate->GetConfigHierarchy()[i];
		fwrite(configHierarchy, sizeof(int), nConfigs, seedHierarchyFile);
		fclose(seedHierarchyFile);
		delete [] configHierarchy;
	}
	
	if (writeTemplateMetaData!="none")
	{
		if (verbose)
			std::cout<<"Writing template meta data: "<<writeTemplateMetaData.c_str()<<std::endl;
		assert(newTemplate!=NULL);
		WriteTemplateMetaData(newTemplate, writeTemplateMetaData);
	}
	
	if (writeLabelsExternal!="none" && allConfigsEnergies.size()>0)
		WriteLabelsExternal(allConfigsEnergies[0]->GetMatch(), writeLabelsExternal);
	
	if (interact)
		visualize = true;
		
	if (visualize || shotTypeToImage.size()>0)
	{	
#ifdef VISUALIZATION		
		if (verbose)
			std::cout<<"Init visualization"<<std::endl;
		int listFlags = DatabaseVisWindow::LIST_GROUPS_CORRS;
		listFlags |= DatabaseVisWindow::LIST_GROUPS_MODELS;

		int nWnd = (allConfigsEnergies.size()==0 ? 1 : 2);
		DatabaseVisWindow * dbVis = new DatabaseVisWindow(db, dbCorrs, listFlags, nWnd, 1, 0, 0, 1);
		if (db->NumModels()==1 && allConfigsEnergies.size()>0)
			dbVis->WindowToModel()[1] = 0;
		VisShapeTemplate * visTemplate = new VisShapeTemplate(db);
		if (readCoAligned!="none")
		{
			std::vector<double> scores;
			std::ifstream fStream(coAlignedModelsFile.c_str());
			assert(fStream.is_open());
			while(!fStream.eof())
			{
				VKString line;
				line.readLine(fStream);
				VKStringList lineSplit = line.split(" ");
				if (lineSplit.count()>1)
				{
					bool ok;
					scores.push_back(lineSplit[1].toDouble(&ok));
					assert(ok);
				}
			}

			visTemplate->SetCoAlignedPoints(readCoAligned, scores, coAlignedThreshold);
		}
		if (interact)
		{
			visTemplate->SetInteractionCreateTemplate(writeTemplate, newTemplate);
			if (refGeom!="none")
			{
				assert(db->NumModels()==1);
				for (int i=0; i<db->NumModels(); i++)
					ReadModelGeometry(db, db->GetModel(i)->ModelName(), refGeom, readManifold);
				visTemplate->SetReferenceModel(db->GetModel(0)->ModelName(), refGeom);
			}
		}
		else
		{
			assert(newTemplate!=NULL);
			if (allConfigsEnergies.size()==0)
			{
				for (int i=0; i<newTemplate->NSeedConfigs(); i++)
				{
					TMatch * seed = newTemplate->CreateSeedConfig(i, NULL, std::vector<int>(), R3identity_affine);
					visTemplate->AddMatchedModel(seed, new TMatchEnergy(seed));
				}
			}
			else
			{
				for (int i=0; i<(int)allConfigsEnergies.size(); i++)
				{
					TMatch * matched = allConfigsEnergies[i]->GetMatch();
					const STemplate * st = matched->GetSTemplate();
					TMatch * seed = st->CreateSeedConfig(matched->SeedID(), NULL, std::vector<int>(), R3identity_affine);
					visTemplate->AddMatchedModel(seed, allConfigsEnergies[i]);
				}
			}
		}
		
		if (shotTypeToImage.size()>0)
		{
			AppScreenshots * takeShots = new AppScreenshots(!visualize, db);
			dbVis->SetApp(takeShots); 

			for (std::map<VKString, VKString>::iterator iter = shotTypeToImage.begin();
				 iter != shotTypeToImage.end(); iter++)
			{
				if (iter->first == "Template")
				{
					assert(iter->second.endsWith(".jpg"));
					assert(newTemplate!=NULL);
					
					for (int i=0; i<newTemplate->NSeedConfigs(); i++)
					{
						VKString outFName = iter->second;
						outFName.replace(".jpg", VKString::number(i)+".jpg");
						visTemplate->SetTemplate(newTemplate);
						takeShots->TakeShotOfApp(visTemplate, iter->first, i, outFName);
					}
				}
				else if (iter->first == "CoAligned")
				{
					assert(newTemplate!=NULL);
					visTemplate->SetTemplate(newTemplate);
					takeShots->TakeShotOfApp(visTemplate, iter->first, 0, iter->second);
					takeShots->TakeShotOfApp(visTemplate, iter->first, 0, iter->second);
				}
				else
					std::cout<<"[WARNING] Unknown screenshot type |"<<iter->first.c_str()<<"|"<<std::endl;
			}
		}
		else
			dbVis->SetApp(visTemplate);

#	ifdef MESA_RENDERING
		std::cout<<"[WARNING] Mesa renderer"<<std::endl;
		VisWindowRenderer * wndMesa = new RendererOSMesa(dbVis);
		wndMesa->UpdateScreen();
#	else
		if (verbose)
			std::cout<<"Running App"<<std::endl;
		VisWindowRenderer * wndGLUT = new RendererGLUT(argc, argv, dbVis);
#	endif
#else
		std::cout<<"[WARNING] Visualization disabled (compiling flag)"<<std::endl;
#endif
	}
	
	return 0;
}

void ReadModelNames(VKStringList & store, const VKString & filename)
{
	std::ifstream textStream(filename.c_str());
	assert(textStream.is_open());
	while (!textStream.eof())
	{
		VKString mName;
		mName.readToken(textStream);
		if (!textStream.eof())
			store.push_back(mName);
	}
	textStream.close();
}

STemplate * SingleTemplate(ModelDatabase * db, const VKString & mName, const VKString & geomName,
						   const VKString & labelsName, bool trustLabels, double useMeshThresh,
						   bool splitLabelsIntoCC, const VKString & geomValueFile)
{
	GeomShape * shape = NULL;
	ShapeLabels * labels = NULL;
	ShapeLabels * meshLabels = NULL;
	
	if (geomName.endsWith(".pnts"))
	{
		GeomShape * mesh = ReadModelGeometry(db, mName, Model3D::m_sMainGeomName, false, labelsName, geomValueFile);
		ShapeValue * vals = mesh->GetShapeVal(labelsName);
		assert(vals!=NULL && vals->SafeCastToLabels()!=NULL);
		if (splitLabelsIntoCC)
			vals = ShapeLabels::SplitLabelsIntoCC(vals->SafeCastToLabels());
		
		GeomShape * pnts = ReadModelGeometry(db, mName, geomName, false);
		labels = ShapeLabels::CopyLabels(vals->SafeCastToLabels(), pnts);
		meshLabels = vals->SafeCastToLabels();
	}
	else
	{
		shape = ReadModelGeometry(db, mName, geomName, false, labelsName, geomValueFile);
		assert(shape!=NULL && shape->GetShapeVal(labelsName)!=NULL);
		labels = shape->GetShapeVal(labelsName)->SafeCastToLabels();
		assert(labels!=NULL);
		meshLabels = labels;
	}
	STemplate * st = STemplate::CreateTemplate(labels, trustLabels, false, useMeshThresh, meshLabels);
	return st;
}

GeomShape * ReadModelGeometry(ModelDatabase * db, const VKString & mName, const VKString & geomName,
							  bool readManifold, const VKString & geomValue, const VKString & geomValueFile)
{
	assert(db!=NULL);
	Model3D * m = db->GetModel(mName);
	if (m==NULL)
		std::cout<<"[ERROR] Cannot find model: "<<mName.c_str()<<std::endl;
	assert(m!=NULL);
	VKString geomFile = db->GetModelGeometryFilename(m->ModelName(), geomName);
	VKString mDir = db->GetDBDirname()+"/Models/";

	if (m->GetGeomShape(geomName)==NULL)
	{
		if (geomName.endsWith(".pnts"))
		{
			assert(geomName.startsWith(Model3D::m_sMainGeomName));
			ReadModelGeometry(db, mName, Model3D::m_sMainGeomName, readManifold);
			m->ReadGeometry(GeomShape::SHAPE_TYPE_POINT_CLOUD, geomName, geomFile, Model3D::m_sMainGeomName);
		}
		else
		{
			VKString geomFile = db->GetModelGeometryFilename(m->ModelName(), geomName)+".polysoup.off";
			GeomShape::GeomShapeTypes mt = GeomShape::SHAPE_TYPE_TRIANGLE_SOUP;
			if (readManifold)
				mt = GeomShape::SHAPE_TYPE_MANIFOLD;
			m->ReadGeometry(mt, geomName, geomFile);
		}
	}
	GeomShape * shape = m->GetGeomShape(geomName);
	assert(shape!=NULL);
	if (geomValue!="none")
	{
		VKString shapeValFile = geomValueFile=="none"?mDir+db->ShapeValueFilename(m->ModelName(), geomName, geomValue):geomValueFile;
		if (!FileManagement::FileExists(shapeValFile.c_str()))
			std::cout<<"[ERROR] Cannot find file: "<<shapeValFile.c_str();
		ShapeLabels * labels = NULL;
		if (shape->GetShapeVal(geomValue)!=NULL)
			labels = shape->GetShapeVal(geomValue)->SafeCastToLabels();
		if (labels==NULL)
			labels = m->ReadGeomValue("ShapeLabels", shape, geomValue, shapeValFile)->SafeCastToLabels();
		assert(labels!=NULL);
		
		// SUPPOSE PERSON PROVIDED LABELS
		// majority vote w.r.t. ground truth to get actual labels (suppose person provides the labels)
		VKString gtValsFile = mDir+db->ShapeValueFilename(m->ModelName(), geomName, "SceneGraphNodes");
		if (FileManagement::FileExists(gtValsFile.c_str()) && gtValsFile!=shapeValFile)
		{
			ShapeLabels * gtLabels = m->ReadGeomValue("ShapeLabels", shape, "tempgt", gtValsFile)->SafeCastToLabels();
			assert(gtLabels!=NULL);
			std::map<VKString, double> name2gtArea;
			ShapeMesh * mesh = shape->SafeCastToMesh();
			assert(mesh!=NULL);
			double meshArea = 0;
			for (int i=0; i<mesh->NFaces(); i++)
			{
				VKString label = gtLabels->NodeLabel((int)gtLabels->Value(ShapePoint(mesh, i, .3, .3), 0));
				name2gtArea[label] += mesh->FaceArea(i);
				meshArea += mesh->FaceArea(i);
			}
			
			for (int i=1; i<labels->NNodes(); i++)
			{
				const std::vector<int> & facesAtNode = labels->NodeFacesAtNode(i);
				std::map<VKString, double> name2votes;
				double segArea = 0;
				for (int f=0; f<(int)facesAtNode.size(); f++)
				{
					double vote = mesh->FaceArea(facesAtNode[f]);
					int gtNode = (int)gtLabels->Value(ShapePoint(mesh, facesAtNode[f], .3, .3), 0);
					VKString currLabel = gtLabels->NodeLabel(gtNode);
					if (name2votes.find(currLabel)==name2votes.end())
						name2votes[currLabel] = vote;
					else
						name2votes[currLabel] += vote;
					segArea += vote;
				}
								
				VKString gtLabel = labels->NodeLabel(i);
				double bestVal = -1;
				bool warning = false;
				for (std::map<VKString, double>::iterator iter = name2votes.begin(); iter != name2votes.end(); iter++)
				{
					double precision = iter->second / segArea;
					double recall = iter->second / name2gtArea[iter->first];
					if (iter->first=="leg")
					{
						warning = true;
						recall *= 4.;	// four legs in our classes
					}
					//double score = precision + recall;
					double score = precision;
					if (score > bestVal)
					{
						bestVal = score;
						gtLabel = iter->first;
					}
				}
				
				if (gtLabel=="flame")	// not a part
				{
					warning = true;
					gtLabel = "candle";
				}
				if (warning)
					std::cout<<"[WARNING] special label: "<<gtLabel.c_str()<<std::endl;
				labels->SetNodeLabel(i, gtLabel);
				
			}
		}
	}
	return shape;
}

void EvalMatching(ModelDatabase * db, const TMatch * match, const VKString & filename, const VKString & equivLabels)
{
	std::vector<VKStringList> equivalenceList;
	std::map<VKString, int> label2ListID;
	if (equivLabels!="none")
		ReadEquivalentLabels(equivLabels, equivalenceList, label2ListID);
		
	if (filename!="print")
		FileManagement::CreateDirsOnPathToFile(filename.c_str());
	assert(db->NumModels()>=1);
	
	assert(match!=NULL);
	const ShapePointCloud * pnts = match->GetMatchedGeomAsPnts();
	const ShapeLabels * labels = match->GetPerPointLabels();
	VKString nullLabel = "XXX_UNKNOWN_XXX";
	for (int i=0; i<match->NTemplateNodes(); i++)
		if (match->NodeEllipsoid(i).IsType(PartBox::BOXTYPE_NullBox))
			nullLabel = match->GetSTemplate()->GetNode(match->Deformation(i).NodeID())->NodeName();
	assert(pnts!=NULL && labels!=NULL);
	
	// get ground truth (mesh labels)
	VKString gtLabelsName = ShapeValue::m_sDefaultSceneGraphLabels;
	VKString mName = pnts->GetModel()->ModelName();
	const GeomShape * gtShape = ReadModelGeometry(db, mName, Model3D::m_sMainGeomName, true,
												  gtLabelsName);
	const ShapeMesh * gtMesh = (gtShape!=NULL ? gtShape->SafeCastToMesh() : NULL);
	const ShapeValue * gtVals = (gtMesh!=NULL ? gtMesh->GetShapeVal(gtLabelsName) : NULL);
	const ShapeLabels * gtLabels = (gtVals!=NULL ? gtVals->SafeCastToLabels() : NULL);
	assert(gtLabels!=NULL);
	
	// OVERALL: 
	//		- fraction classified correctly 
	//		- fraction mis-classified (non NULL)
	// PER LABEL:
	//		- precision:	(# correctly classified as label) / (# classified as label in prediction)
	//		- recall:		(# correctly classified as label) / (# classified as label in gt)
	int overallCorrect = 0;
	int overallCorrectWithNull = 0;
	int overallLabels = 0;
	std::map<VKString, int> perClassCorrect;
	std::map<VKString, int> perClassClassifiedAs;
	std::map<VKString, int> perClassInGT;
	std::vector<std::map<VKString, int> * > allStatsVars;
	allStatsVars.push_back(&perClassCorrect);
	allStatsVars.push_back(&perClassClassifiedAs);
	allStatsVars.push_back(&perClassInGT);
	
	VKStringList templateLabels;
	for (int i=0; i<match->GetSTemplate()->NNodes(); i++)
		if (!templateLabels.contains(match->GetSTemplate()->GetNode(i)->NodeName()))
			templateLabels.push_back(match->GetSTemplate()->GetNode(i)->NodeName());
	
	// go over all points
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		int segID = (int)labels->Value(pnts->GetPoint(i), 0);
		int gtSegID = (int)gtLabels->Value(pnts->GetPoint(i), 0);
		VKString label = MainName(labels->NodeLabel(segID), equivalenceList, label2ListID);
		VKString gtLabel = MainName(gtLabels->NodeLabel(gtSegID), equivalenceList, label2ListID);
	
		overallLabels++;
		
		// include all labels
		for (int s=0; s<(int)allStatsVars.size(); s++)
		{
			if (allStatsVars[s]->find(label)==allStatsVars[s]->end())
				(*allStatsVars[s])[label] = 0;
			if (allStatsVars[s]->find(gtLabel)==allStatsVars[s]->end())
				(*allStatsVars[s])[gtLabel] = 0;
		}
		
		if (Equivalent(label, gtLabel, label2ListID))
		{
			overallCorrectWithNull++;
			overallCorrect++;
			perClassCorrect[label]++;
		}
		else if (nullLabel == label)	// if labeled as null - still might be accepted if not in template
		{
			// is there a potentially correct label?
			bool labelExistsInTemplate = false;
			for (int tid = 0; tid < templateLabels.count() && !labelExistsInTemplate; tid++)
				labelExistsInTemplate = Equivalent(templateLabels[tid], gtLabel, label2ListID);
			if (!labelExistsInTemplate)
				overallCorrectWithNull++;
		}
			
		
		perClassClassifiedAs[label]++;
		perClassInGT[gtLabel]++;
	}
	
	// print to file
	std::ofstream textStream;
	if (filename!="print")
	{
		//assert(FileManagement::FileExists(filename.c_str()));
		textStream.open(filename.c_str());
		assert(textStream.is_open());
		textStream<<"All "<<overallCorrect<<" "<<overallCorrectWithNull<<" "<<overallLabels<<"\n";
	}
	std::cout<<"----- Evaluation Results. Model: "<<match->ModelName().c_str()<<" -----"<<std::endl;
	std::cout<<"ALL (NULL is always wrong): "<<(100*overallCorrect / overallLabels)<<"%"<<std::endl;
	std::cout<<"ALL (w/ NULL for unknown) : "<<(100*overallCorrectWithNull / overallLabels)<<"%"<<std::endl;
	std::cout<<"Part           Pr     Rc     Lb     Gt"<<std::endl;
	for (std::map<VKString, int>::iterator iter = perClassInGT.begin();
		 iter != perClassInGT.end(); iter++)
	{
		bool inGT = perClassInGT.find(iter->first)!=perClassInGT.end();
		bool inLabel = perClassClassifiedAs.find(iter->first)!=perClassClassifiedAs.end();
		int nCorrect = perClassCorrect[iter->first];
		int nLabeled = inLabel ? perClassClassifiedAs[iter->first] : 0;
		int nGT = inGT ? perClassInGT[iter->first] : 0;
		if (filename!="print")
			textStream<<iter->first.c_str()<<" "<<nCorrect<<" "<<nLabeled<<" "<<nGT<<"\n";
		
		VKString name = iter->first;
		while (name.length()<15)
			name = name+" ";
		if (name.length()>15)
			name = name.left(14)+".";
		if (nLabeled==0)
			nLabeled = 1;
		if (nGT==0)
			nGT = 1;
		
		VKString pStr = VKString::number((100*nCorrect/nLabeled))+"%";
		while (pStr.length()<7)
			pStr+=" ";
		std::cout<<name.c_str()<<pStr.c_str();
		
		pStr = VKString::number((100*nCorrect/nGT))+"%";
		while (pStr.length()<7)
			pStr+=" ";
		std::cout<<pStr.c_str();
		
		pStr = VKString::number((100*nLabeled/overallLabels))+"%";
		while (pStr.length()<7)
			pStr+=" ";
		std::cout<<pStr.c_str();

		pStr = VKString::number((100*nGT/overallLabels))+"%";
		while (pStr.length()<7)
			pStr+=" ";
		std::cout<<pStr.c_str();
		
		std::cout<<std::endl;
	}
	std::cout<<"-----------------------------"<<std::endl;
	
	if (filename!="print")
		textStream.close();
	
	
}

VKString MainName(const VKString & label, 
				  std::vector<VKStringList> & equivalenceList, std::map<VKString, int> & label2ListID)
{
	std::map<VKString, int>::iterator iter = label2ListID.find(label);
	if (iter==label2ListID.end())
		return label;
	assert(iter->second < (int)equivalenceList.size() && equivalenceList[iter->second].count()>0);
	return equivalenceList[iter->second][0];
}

bool Equivalent(const VKString & label1, const VKString & label2, std::map<VKString, int> & label2ListID)
{
	if (label1==label2)
		return true;
	std::map<VKString, int>::iterator iter1 = label2ListID.find(label1);
	std::map<VKString, int>::iterator iter2 = label2ListID.find(label2);
	if (iter1==label2ListID.end() || iter2==label2ListID.end())
		return false;
	return (iter1->second==iter2->second);
}

void ReadEquivalentLabels(const VKString & equivFile, 
						  std::vector<VKStringList> & equivalenceList, std::map<VKString, int> & label2ListID)
{
	std::ifstream textStream(equivFile.c_str());
	assert(textStream.is_open());
	VKString tempStr;
	while (!textStream.eof())
	{
		tempStr.readLine(textStream);
		int currID = (int)equivalenceList.size();
		equivalenceList.push_back(tempStr.split(" "));
		for (int i=0; i<equivalenceList[currID].count(); i++)
			label2ListID[equivalenceList[currID][i]] = currID;
	}
	
	textStream.close();
}

void WriteAlignedModel(const TMatch * match, const VKString & filename)
{
	FileManagement::CreateDirsOnPathToFile(filename.c_str());
	const R3Affine & xForm = match->GlobalXForm();
	const ShapePointCloud * pnts = match->GetMatchedGeomAsPnts();
	const ShapeMesh * mesh = pnts->GetMesh();
	assert(mesh!=NULL);
	
	ShapeTriangleSoup * xformed = new ShapeTriangleSoup();

	for (int i=0; i<mesh->NVertices(); i++)
	{
		R3Point p = mesh->VertexPosition(i);
		xForm.Apply(p);
		xformed->AddVertex(p[0], p[1], p[2]);
	}
	
	for (int i=0; i<mesh->NFaces(); i++)
		xformed->AddFace(mesh->VertexOnFace(i, 0), mesh->VertexOnFace(i, 1), mesh->VertexOnFace(i, 2));
	
	xformed->WriteOff(filename, ShapeTriangleSoup::OFF_ASCII);
	
}

void WriteTemplateMetaData(const STemplate * tmplt, const VKString & filename)
{
	std::ofstream textStream(filename.c_str());
	assert(textStream.is_open());
	
	textStream<<"Parts\n";
	for (int i=0; i<tmplt->NNodes(); i++)
	{
		VKString nodeName = tmplt->GetNode(i)->NodeName();
		VKString nodeType = tmplt->GetNode(i)->GetPart(0)->GetBoxTypeStr();
		double nodeErr = tmplt->GetNode(i)->GetPart(0)->GetFittingErr();
		textStream<<nodeName.c_str()<<" "<<nodeType.c_str()<<" "<<nodeErr<<"\n";
	}
	
	textStream.close();
}

void WriteLabelsExternal(const TMatch * match, const VKString & dirName)
{
    const ShapePointCloud * pnts = match->GetMatchedGeomAsPnts();
    assert(pnts!=NULL);
    VKString modelFile = dirName + "/"+pnts->GetModel()->ModelName()+".off";
	WriteAlignedModel(match, modelFile);
    VKString labelsFile = dirName+"/"+pnts->GetModel()->ModelName()+".pnt.labels";
    // n nodes
    std::ofstream textStream(labelsFile.c_str());
    assert(textStream.is_open());
    textStream<<"NNodes "<<match->NTemplateNodes()<<"\n";
    for (int i=0; i<match->NTemplateNodes(); i++)
    {
        const TemplateNodeDeform & deform = match->Deformation(i);
        textStream<<match->GetSTemplate()->GetNode(deform.NodeID())->NodeName().c_str()<<" ";
        textStream<<deform.NodeID()<<" "<<deform.NParts()<<" "<<deform.PartID()<<"\n";
    }
    
    // node names
    // n points
    textStream<<"NPoints "<<pnts->NumPoints()<<"\n";
    for (int i=0; i<pnts->NumPoints(); i++)
    {
        R3Point pos = pnts->PointPosition(i);
        int nodeID = match->LocalNodeID(i);
        textStream<<pos.X()<<" "<<pos.Y()<<" "<<pos.Z()<<" "<<nodeID<<"\n";
    }
    textStream.close();
}

STemplate * InitTemplate(const VKString & topologyFile, const VKString & trainGeomName,
						 const VKString & trainLabelsName, ModelDatabase * db)
{
	VKString flatSegName = "MinDepthSeg";
	// read topology file
	std::ifstream topologyFStream(topologyFile.c_str());
	assert(topologyFStream.is_open());
	VKString rootStr;
	rootStr.readLine(topologyFStream);
	std::map<VKString, int> children;
	std::map<VKString, VKString> equivalence;
	while (!topologyFStream.eof())
	{
		VKString line;
		line.readLine(topologyFStream);
		if (line=="")
			break;
		
		line = line.replace(" - ", "");
		VKStringList lines = line.split(", ");
		if (lines.count()>0)
		{
			for (int i=0; i<lines.count(); i++)
			{
				//std::cout<<"Keyword: |"<<lStr.c_str()<<"|"<<std::endl;
				children[lines[i]] = (int)children.size();
				equivalence[lines[i]] = lines[0];
			}
		}
	}
	
	// find appropriate segmentations (at level of labels)
	for (int i=0; i<db->NumModels(); i++)
	{
		// read segmentations
		Model3D * m = db->GetModel(i);
		assert(m!=NULL);
		VKString mName = m->ModelName();
		GeomShape * g = ReadModelGeometry(db, mName, trainGeomName, false);
		assert(g!=NULL);
		GeomShape * mesh = ReadModelGeometry(db, mName, Model3D::m_sMainGeomName, false, trainLabelsName);
		ShapeValue * vals = mesh->GetShapeVal(trainLabelsName);
		assert(vals!=NULL && vals->SafeCastToLabels()!=NULL);
		ShapePointCloud * pnts = g->SafeCastToPointCloud();
		assert(pnts!=NULL);
		ShapeLabels * labels = ShapeLabels::CopyLabels(vals->SafeCastToLabels(), pnts);
		assert(labels!=NULL);

		// find appropriate depth of segmentation
		// for every label - find depth
		int minDepth = -1;
		for (int l=0; l<labels->NNodes(); l++)
		{
			const VKString & nl = labels->NodeLabel(l);
			int depth = labels->NodeDepth(l);
			if ((minDepth < 0 || minDepth > depth) && equivalence.find(nl)!=equivalence.end())
				minDepth = depth; 
		}
		
		// expand to minimal depth: create flat segmentation
		std::vector<SceneGraphLabel*> allLabels;
		ShapeLabels * newLabels = new ShapeLabels(pnts, flatSegName);
		SceneGraphLabel * root = new SceneGraphLabel();
		root->m_sLabel = rootStr;
		root->m_iID = (int)allLabels.size();
		root->m_pParent = NULL;
		allLabels.push_back(root);
				
		std::vector<int> pnts2segs;
		for (int p=0; p<pnts->NumPoints(); p++)
			pnts2segs.push_back(0);

		for (int l=0; l<labels->NNodes(); l++)
		{
			if (labels->NodeDepth(l) == minDepth || labels->NodeNChildren(l) == 0)
			{
				SceneGraphLabel * newNode = new SceneGraphLabel();
				std::map<VKString, VKString>::iterator iter = equivalence.find(labels->NodeLabel(l));
				if (iter==equivalence.end())
					newNode->m_sLabel = labels->NodeLabel(l);
				else
					newNode->m_sLabel = iter->second;
				newNode->m_iID = (int)allLabels.size();
				newNode->m_pParent = root;
				root->m_apChildren.push_back(newNode);
				allLabels.push_back(newNode);
				
				std::vector<int> allPntIDs;
				labels->NodeAllPoints(l, allPntIDs);
				for (int p=0; p<(int)allPntIDs.size(); p++)
				{
					int pid = allPntIDs[p];
					assert(pid >= 0 && pid < (int)pnts2segs.size());
					pnts2segs[pid] = newNode->m_iID;
				}
			}
		}
		
		newLabels->SetAllLabels(allLabels);
		newLabels->SetPntsSegments(pnts2segs);
		pnts->AddShapeVal(newLabels->GetShapeValueName(), newLabels);
	}
	
	// find pairs that share at least one label
//	VKGraph graph(db->NumModels();
//	for (int i=0; i<db->NumModels(); i++)
//	for (int j=i+1; j<db->NumModels(); j++)
//	{
//	}
	
	// do consistent segmentation?
	std::cout<<"NOT FINISHED IMPLEMENTATION YET: InitTemplate"<<std::endl;
	exit(0);
}





