#include "EvalCorrs.h"
#include "TemplateMappings.h"

void ReadCoAlignedPoints(const VKString & coAlignedPntsFile, std::vector<std::vector<R3Point> > & coAlignedPoints)
{
	FILE * file = fopen(coAlignedPntsFile.c_str(), "rb");
	if (file==NULL)
		std::cout<<"[ERROR] Cannot read file: "<<coAlignedPntsFile.c_str()<<std::endl;
	assert(file!=NULL);
	int mid, nseeds, npnts;
	
	int counter = 0;
	while (!feof(file))
	{
		fread(&mid, sizeof(int), 1, file);
		//std::cout<<"mid="<<mid<<std::endl;
		fread(&nseeds, sizeof(int), 1, file);
		fread(&npnts, sizeof(int), 1, file);
		if (feof(file))
			break;
		//assert(counter==mid);
		
		coAlignedPoints.push_back(std::vector<R3Point>());
		assert(npnts<1000000);
		for (int i=0; i<npnts; i++)
		{
			double xyz[3];
			fread(xyz, sizeof(double), 3, file);
			coAlignedPoints[counter].push_back(R3Point(xyz[0], xyz[1], xyz[2]));
		}
		counter++;
	}
	
	fclose(file);
}

TMatchEnergy * ReadMatch(STemplate * st, ModelDatabase * db, const VKString & mName, const VKString & dbDir, const VKString & methodDir)
{
	// read geometries
	//	assert(db->NumModels()==0);
	Model3D * m = new Model3D(mName, db->NumModels());
	db->AddModel(m);
	VKString meshFile = dbDir + "/Models/"+mName+"/Geometry/main.polysoup.off";
	VKString pntsFile = dbDir + "/Models/"+mName+"/Geometry/main_Even1000.pnts";
	m->ReadGeometry(GeomShape::SHAPE_TYPE_TRIANGLE_SOUP, Model3D::m_sMainGeomName, meshFile);
	m->ReadGeometry(GeomShape::SHAPE_TYPE_POINT_CLOUD, "main_Even1000.pnts", pntsFile, Model3D::m_sMainGeomName);
	
	// read match
	FILE * mFile = fopen((methodDir+"/"+mName+"/match.match.bin").c_str(), "rb");
	assert(mFile!=NULL);
	
	//	std::cout<<"File: "<<(methodDir+"/"+mName+"/match.match.bin").c_str()<<std::endl;
	//	TMatch * match = TMatch::ReadMatch(mFile, st, db);
	TMatchEnergy * energy = TMatchEnergy::ReadEnergyAndMatch(mFile, st, db);
	assert(energy!=NULL);
	assert(energy->GetMatch()->GetMatchedGeomAsPnts()!=NULL);
	fclose(mFile);
	return energy;
}

void EvalCorrs(STemplate * st, const VKString & fileList, const VKString & methodDir,
			   const VKString & dbDir, const VKString & outIntegratedCorrs, bool debugVis)
{
	assert(FileManagement::FileExists(fileList.c_str()));
	std::ifstream textStream(fileList.c_str());
	VKStringList matchesToProcess;
	assert(textStream.is_open());
	while (!textStream.eof())
	{
		VKString currLn;
		currLn.readLine(textStream);
		VKStringList lnSplit = currLn.split(" ");
		if (lnSplit.count()>0)
		{
			VKString fName = lnSplit[0];
			VKString gtFile = dbDir+"/../gt/"+fName+".bin";
			if (fName != "" && FileManagement::FileExists(gtFile.c_str()))	// gt file exists
			{
				FILE * tmpFile = fopen(gtFile.c_str(), "rb");
				assert(tmpFile!=NULL);
				char type;
				fread(&type, sizeof(char), 1, tmpFile);
				if (type=='V')
					matchesToProcess.push_back(fName);
				fclose(tmpFile);
			}
		}
	}
	textStream.close();
	
	// match template -> template
	std::vector<TMatch*> seedMatches;
	std::vector<std::vector<Tmplt2TmpltMapping*> > t2tmapping;
	for (int i=0; i<st->NSeedConfigs(); i++)
	{
		if (i==0 && st->NSeedConfigs()>1)
			seedMatches.push_back(NULL);
		else
			seedMatches.push_back(st->CreateSeedConfig(i, NULL, std::vector<int>(), R3identity_affine));
		t2tmapping.push_back(std::vector<Tmplt2TmpltMapping*>());
	}
	for (int i=0; i<(int)seedMatches.size(); i++)
		for (int j=0; j<(int)seedMatches.size(); j++)
		{
			if (seedMatches[i]==NULL || seedMatches[j]==NULL)
				t2tmapping[i].push_back(NULL);
			else
				t2tmapping[i].push_back(new Tmplt2TmpltMapping(st, seedMatches[i], seedMatches[j]));
		}
	
	// match feat pnts -> template -> model
	std::vector<Model2TmpltMapping*> perModelAlign;
	ModelDatabase * db = new ModelDatabase::ModelDatabase();
	int nmaxPnts = 0;
	for (int i=0; i<matchesToProcess.count(); i++)
	{
		TMatchEnergy * energy = ReadMatch(st, db, matchesToProcess[i], dbDir, methodDir);
		//int mid = db->NumModels()-1;
		int mid = db->GetModel(matchesToProcess[i])->ModelID();
		assert(mid>=0);
		std::vector<ShapePoint *> gtpnts;
		assert(db->GetModel(mid)->GetGeomShape(Model3D::m_sMainGeomName)!=NULL);
		ShapeMesh * mesh = db->GetModel(mid)->GetGeomShape(Model3D::m_sMainGeomName)->SafeCastToMesh();
		assert(mesh!=NULL);
		EvalCorrsReadFeatPnts(mesh, dbDir+"/../gt/"+matchesToProcess[i]+".bin", gtpnts);
		nmaxPnts = vkMax(nmaxPnts, (int)gtpnts.size());
		perModelAlign.push_back(new Model2TmpltMapping(energy->GetMatch(), gtpnts));
	}
	
	// measure errors as cdf
	int nsteps = 200;
	double minErr = 0;
	double maxErr = .25;
	double * pdf = new double[nsteps];
	for (int i=0; i<nsteps; i++)
		pdf[i] = 0;
	
	double * pdfShape = new double[nsteps];
	for (int i=0; i<nsteps; i++)
		pdfShape[i] = 0;
	
	// pairwise errors
	bool valid;
	int npairs = 0;
	int nShapePairs = 0;
	std::vector<double> aveEuclError;
	std::vector<int> aveEuclCounter;
	for (int i=0; i<(int)perModelAlign.size(); i++)
	{
		aveEuclError.push_back(0);
		aveEuclCounter.push_back(0);
	}
	std::vector<Sortable> maxPairwiseErr;
	double aveErr = 0;
	for (int i=0; i<(int)perModelAlign.size(); i++)
	for (int j=0; j<(int)perModelAlign.size(); j++)
	{
		int pairid = i * (int)perModelAlign.size() + j;
		maxPairwiseErr.push_back(Sortable(0, NULL, pairid));
		if (i==j)
			continue;
		
		int c1 = perModelAlign[i]->ConfigID();
		int c2 = perModelAlign[j]->ConfigID();
		
		double meshMaxErr = 0;
		for (int p=0; p<nmaxPnts; p++)
		{
			// feature point -> template ID
			UnqiueTemplatePoint t1 = perModelAlign[i]->FromFeature2Tmplt(p, valid);
			if (!valid) continue;	// ignore point: no gt
			R3Point p2GT = perModelAlign[j]->GetGroundTruthPoint(p, valid);
			if (!valid) continue;	// ignore point: no gt

			R3Point p2 = MapPoint(p, &t2tmapping, &perModelAlign, i, j, false, valid);
			assert(valid);

//			// template ID -> template ID (in different configurations)
//			assert(c1 < (int)t2tmapping.size() && c2 < (int)t2tmapping.size() && c1>0 && c2>0);
//			assert(t2tmapping[c1][c2]!=NULL);
//			UnqiueTemplatePoint t2 = t2tmapping[c1][c2]->MapConfig1ToConfig2(t1);
//			
//			// template ID -> model point
//			R3Point p2 = perModelAlign[j]->FromTemplate2Model(t2);
			
			double err = R3Distance(p2GT, p2) / perModelAlign[j]->Radius();
			meshMaxErr = vkMax(meshMaxErr, err);
			EvalCorrsAddPDF(err, pdf, nsteps, minErr, maxErr);
			npairs ++ ;
			aveEuclError[i] += err;
			aveEuclCounter[i]++;
			aveErr += err;
		}
		maxPairwiseErr[pairid].value = -meshMaxErr;
		EvalCorrsAddPDF(meshMaxErr, pdfShape, nsteps, minErr, maxErr);
		nShapePairs++;
	}
	std::sort(maxPairwiseErr.begin(), maxPairwiseErr.end());
	
	double * cdf = EvalCorrsCDF(pdf, nsteps);
	double * cdfShape = EvalCorrsCDF(pdfShape, nsteps);
	
	if (outIntegratedCorrs!="none")
	{
		std::cout<<"Writing Results: "<<outIntegratedCorrs.c_str()<<std::endl;
		VKString outName = outIntegratedCorrs;
		if (outName.endsWith(".txt"))
			outName.replace(".txt", "");
		std::ofstream outFileStream((outName+".perpoint.txt").c_str());
		assert(outFileStream.is_open());
		for (int i=0; i<nsteps; i++)
			outFileStream<<(cdf[i]/npairs)<<" ";
		outFileStream.close();
		
		std::ofstream outFileStream2((outName+".pershape.txt").c_str());
		assert(outFileStream2.is_open());
		for (int i=0; i<nsteps; i++)
			outFileStream2<<(cdfShape[i]/nShapePairs)<<" ";
		outFileStream2.close();
		
		std::ofstream outFileStream3((outName+".euclerr.txt").c_str());
		assert(outFileStream3.is_open());
		outFileStream3<<(aveErr/npairs)<<"\n";
		assert((int)aveEuclError.size()==matchesToProcess.count());
		for (int i=0; i<(int)aveEuclError.size(); i++)
		{
			double err = aveEuclError[i];
			int norm = aveEuclCounter[i];
			outFileStream3<<matchesToProcess[i].c_str()<<" "<<(err == 0 ? 0 : (err/norm))<<"\n";
		}
		outFileStream3.close();
	}
	std::cout<<"Ave Err: "<<(aveErr/npairs)<<std::endl;
	if (debugVis)
	{
#ifdef VISUALIZATION
		TMapDebugApp * debugApp = new TMapDebugApp(db, t2tmapping, perModelAlign);
		debugApp->SetMaxPairwiseErr(&maxPairwiseErr);
		int listFlags = DatabaseVisWindow::LISTGP_Models | DatabaseVisWindow::LISTGP_Geoms;
		listFlags |= DatabaseVisWindow::LISTGP_MeshFlags;
		DatabaseVisWindow * dbVis = new DatabaseVisWindow(db, NULL, listFlags, 1, 1, 1, 1, .5);
		dbVis->SetApp(debugApp);
		int argc = 0;
		char ** argv;
		VisWindowRenderer * wndGLUT = new RendererGLUT(argc, argv, dbVis);
#else
		std::cout<<"[WARNING] Visualization disabled (compiling flag)"<<std::endl;
#endif
	}
	
	//	delete [] pdf;
	//	delete [] cdf;
}

double * EvalCorrsCDF(const double * evalCorrsPDF, int nSteps)
{
	double * cdf = new double[nSteps];
	double cdfVal = 0;
	for (int i=0; i<nSteps; i++)
	{
		cdf[i] = evalCorrsPDF[i] + cdfVal;
		cdfVal += evalCorrsPDF[i];
	}
	return cdf;
}

void EvalCorrsReadFeatPnts(ShapeMesh * mesh, const VKString & filename, std::vector<ShapePoint *> & gt)
{
	FILE * binFile = fopen(filename.c_str(), "rb");
	assert(binFile!=NULL);
	char type;
	int npnts, fid;
	double b1, b2, b3, x, y, z;
	fread(&type, sizeof(char), 1, binFile);
	if (type=='V')
	{
		fread(&npnts, sizeof(int), 1, binFile);
		for (int i=0; i<npnts; i++)
		{
			fread(&fid, sizeof(int), 1, binFile);
			if (fid==-1)
				gt.push_back(NULL);
			else
			{
				fread(&b1, sizeof(double), 1, binFile);
				fread(&b2, sizeof(double), 1, binFile);
				fread(&b3, sizeof(double), 1, binFile);
				fread(&x, sizeof(double), 1, binFile);
				fread(&y, sizeof(double), 1, binFile);
				fread(&z, sizeof(double), 1, binFile);
				gt.push_back(new ShapePoint(mesh, fid, b1, b2));
			}
		}
	}
	fclose(binFile);
}

void EvalCorrsAddPDF(double error, double * pdf, int nSteps, double minErr, double maxErr)
{
	int bin = nSteps * (error-minErr) / (maxErr - minErr);
	if (bin >= 0 && bin < nSteps)
		pdf[bin] += 1.;
}

////////////// Fuzzy Correspondences - quick test
void EvalFuzzyCorrs(const VKString & fuzzyCorrsFile, const VKString & fileList,
					const VKString & dbDir, const VKString & outIntegratedCorrs, bool debugVis)
{
	if (!FileManagement::FileExists(fileList.c_str()))
		std::cout<<"Missing File: "<<fileList.c_str()<<std::endl;
	assert(FileManagement::FileExists(fileList.c_str()));
	std::ifstream textStream(fileList.c_str());
	VKStringList matchesToProcess;
	assert(textStream.is_open());
	while (!textStream.eof())
	{
		VKString currLn;
		currLn.readLine(textStream);
		VKStringList lnSplit = currLn.split(" ");
		if (lnSplit.count()>0)
		{
			VKString fName = lnSplit[0];
			VKString gtFile = dbDir+"/../gt/"+fName+".bin";
			if (fName != "" && FileManagement::FileExists(gtFile.c_str()))	// gt file exists
			{
				FILE * tmpFile = fopen(gtFile.c_str(), "rb");
				assert(tmpFile!=NULL);
				char type;
				fread(&type, sizeof(char), 1, tmpFile);
				if (type=='V')
					matchesToProcess.push_back(fName);
				fclose(tmpFile);
			}
		}
	}
	textStream.close();

	int nmodels = matchesToProcess.count();
	// read models
	ModelDatabase * db = new ModelDatabase::ModelDatabase();
	std::vector<ShapeMesh*> inputMeshes;
	for (int i=0; i<nmodels; i++)
	{
		VKString mName = matchesToProcess[i];
		Model3D * m = new Model3D(mName, db->NumModels());
		db->AddModel(m);
		VKString meshFile = dbDir + "/Models/"+mName+"/Geometry/main.polysoup.off";
		GeomShape * shape = m->ReadGeometry(GeomShape::SHAPE_TYPE_TRIANGLE_SOUP, Model3D::m_sMainGeomName, meshFile);
		assert(shape!=NULL && shape->SafeCastToMesh()!=NULL);
		VKString pntsFile = dbDir + "/Models/"+mName+"/Geometry/main_Even1000.pnts";
		m->ReadGeometry(GeomShape::SHAPE_TYPE_POINT_CLOUD, "main_Even1000.pnts", pntsFile, Model3D::m_sMainGeomName);
		inputMeshes.push_back(shape->SafeCastToMesh());
	}

	// read fuzzy corrs file
	std::cout<<"Reading fuzzy corrs"<<std::endl;
	std::vector<ShapePointCloud *> fuzzyPnts;
	std::map<int, R3Point> * v2vmaps = new std::map<int, R3Point>[nmodels * nmodels];
	ReadFuzzyCorrsFile(fuzzyCorrsFile, inputMeshes, fuzzyPnts, v2vmaps);
	assert((int)fuzzyPnts.size()==nmodels);
	std::vector<double> meshradius;
	std::cout<<"Computing radius"<<std::endl;
	TimeProfiler timerRadius;
	for (int i=0; i<(int)fuzzyPnts.size(); i++)
	{
		timerRadius.WriteProgress("prog", i, fuzzyPnts.size());
		meshradius.push_back(GetMeshRadius(inputMeshes[i]));
	}
	std::cout<<" Successful!"<<std::endl;
	
	// match feat pnts -> template -> model
	std::vector<std::vector<ShapePoint*> > gtpnts;
	int nmaxPnts = 0;
	std::cout<<"Reading feature points"<<std::endl;
	for (int i=0; i<matchesToProcess.count(); i++)
	{
		Model3D * m = db->GetModel(matchesToProcess[i]);
		assert(m!=NULL && m->GetGeomShape(Model3D::m_sMainGeomName)!=NULL);
		ShapeMesh * mesh = m->GetGeomShape(Model3D::m_sMainGeomName)->SafeCastToMesh();
		assert(mesh!=NULL);
		int lid = (int)gtpnts.size();
		gtpnts.push_back(std::vector<ShapePoint*>());
		EvalCorrsReadFeatPnts(mesh, dbDir+"/../gt/"+matchesToProcess[i]+".bin", gtpnts[lid]);
		nmaxPnts = vkMax(nmaxPnts, (int)gtpnts.size());
	}
	
	std::cout<<"Reading errors"<<std::endl;
	// measure errors as cdf
	int nsteps = 200;
	double minErr = 0;
	double maxErr = .25;
	double * pdf = new double[nsteps];
	for (int i=0; i<nsteps; i++)
		pdf[i] = 0;
	
	double * pdfShape = new double[nsteps];
	for (int i=0; i<nsteps; i++)
		pdfShape[i] = 0;
	
	// pairwise errors
	bool valid;
	int npairs = 0;
	int nShapePairs = 0;
	std::vector<double> aveEuclError;
	std::vector<int> aveEuclCounter;
	for (int i=0; i<nmodels; i++)
	{
		aveEuclError.push_back(0);
		aveEuclCounter.push_back(0);
	}
	double aveErr = 0;
	TimeProfiler timer;
	for (int i=0; i<(int)matchesToProcess.count(); i++)
	for (int j=0; j<(int)matchesToProcess.count(); j++)
	{
		if (j==0)
			timer.WriteProgress("prg", i, matchesToProcess.count());
		if (i==j)
			continue;
		
		double meshMaxErr = 0;
		for (int p=0; p<nmaxPnts; p++)
		{
			// feature point -> feature point ground truth
			valid = p < (int)gtpnts[i].size() && gtpnts[i][p]!=NULL;
			if (!valid) continue;	// ignore point: no gt
			R3Point p1 =gtpnts[i][p]->GetPosition();
			
			valid = p < (int)gtpnts[j].size() && gtpnts[j][p]!=NULL;
			if (!valid) continue;	// ignore point: no gt
			R3Point p2GT = gtpnts[j][p]->GetPosition();
			
			// predict p1 position (p2)
			// map p1 -> nearest point in fuzzyPnts[i]
			const PointInSet * closest = fuzzyPnts[i]->GetKdTree()->FindClosest(p1);
//			std::cout<<"p1: ["<<p1.X()<<" "<<p1.Y()<<" "<<p1.Z()<<"] "<<std::endl;
			assert(closest!=NULL);
//			bool atvtx = false;
//			int vid = fuzzyPnts[i]->GetPoint(closest->id).VertexID(&atvtx);
//			assert(atvtx);
			
			// use v2vmaps[i * nmodels + j] to find i -> j map (np1 -> p2)
			//std::cout<<"MAP: "<<i<<" -> "<<j<<": vid="<<vid<<std::endl;
			std::map<int, R3Point>::iterator iter = v2vmaps[i * nmodels + j].find(closest->id);
			assert(iter != v2vmaps[i * nmodels + j].end());
//			assert(fuzzyPnts[j]->GetMesh()!=NULL);
//			R3Point p2 = fuzzyPnts[j]->GetMesh()->VertexPosition(iter->second);
			R3Point p2 = iter->second;
			double err = R3Distance(p2GT, p2) / meshradius[j];
			meshMaxErr = vkMax(meshMaxErr, err);
			EvalCorrsAddPDF(err, pdf, nsteps, minErr, maxErr);
			npairs ++ ;
			aveEuclError[i] += err;
			aveEuclCounter[i]++;
			aveErr += err;
		}
		EvalCorrsAddPDF(meshMaxErr, pdfShape, nsteps, minErr, maxErr);
		nShapePairs++;
	}
	std::cout<<" Done!"<<std::endl;
	
	double * cdf = EvalCorrsCDF(pdf, nsteps);
	double * cdfShape = EvalCorrsCDF(pdfShape, nsteps);
	
	std::cout<<"Writing Results: "<<outIntegratedCorrs.c_str()<<std::endl;
	VKString outName = outIntegratedCorrs;
	if (outName.endsWith(".txt"))
		outName.replace(".txt", "");
	std::ofstream outFileStream((outName+".perpoint.txt").c_str());
	assert(outFileStream.is_open());
	for (int i=0; i<nsteps; i++)
		outFileStream<<(cdf[i]/npairs)<<" ";
	outFileStream.close();
	
	std::ofstream outFileStream2((outName+".pershape.txt").c_str());
	assert(outFileStream2.is_open());
	for (int i=0; i<nsteps; i++)
		outFileStream2<<(cdfShape[i]/nShapePairs)<<" ";
	outFileStream2.close();
	
	std::ofstream outFileStream3((outName+".euclerr.txt").c_str());
	assert(outFileStream3.is_open());
	outFileStream3<<(aveErr/npairs)<<"\n";
	assert((int)aveEuclError.size()==matchesToProcess.count());
	for (int i=0; i<(int)aveEuclError.size(); i++)
	{
		double err = aveEuclError[i];
		int norm = aveEuclCounter[i];
		outFileStream3<<matchesToProcess[i].c_str()<<" "<<(err == 0 ? 0 : (err/norm))<<"\n";
	}
	outFileStream3.close();
	
	if (debugVis)
	{
#ifdef VISUALIZATION
		TMapDebugApp * debugApp = new TMapDebugApp(db, gtpnts, fuzzyPnts, v2vmaps);
		int listFlags = DatabaseVisWindow::LISTGP_Models | DatabaseVisWindow::LISTGP_Geoms;
		listFlags |= DatabaseVisWindow::LISTGP_MeshFlags;

		DatabaseVisWindow * dbVis = new DatabaseVisWindow(db, NULL, listFlags, 1, 1, 1, 1, .5);
		dbVis->SetApp(debugApp);
		int argc = 0;
		char ** argv;
		VisWindowRenderer * wndGLUT = new RendererGLUT(argc, argv, dbVis);
#else
		std::cout<<"[WARNING] Visualization disabled (compiling flag)"<<std::endl;
#endif
	}
}

void ReadFuzzyCorrsFile(const VKString & fuzzyCorrsFile, const std::vector<ShapeMesh * > & inputMeshes,
						std::vector<ShapePointCloud *> & fuzzyPnts, std::map<int, R3Point> * v2vmaps)
{
	int nmodels = (int)inputMeshes.size();
	double * corrVals = new double[nmodels];
	int * corrIDs = new int[nmodels];
	int corr_mid;
	int corr_vid;
	double corr_val;

	for (int i=0; i<(int)inputMeshes.size(); i++)
	{
		ShapePointCloud * newFuzzPnts = new ShapePointCloud(inputMeshes[i], "fuzzypnts");
		((Model3D*)inputMeshes[i]->GetModel())->AddGeomShape("fuzzypnts", newFuzzPnts);
		fuzzyPnts.push_back(newFuzzPnts);
	}
	
	FILE * fuzzyCorrsBin = fopen(fuzzyCorrsFile.c_str(), "rb");
	assert(fuzzyCorrsBin!=NULL);
	
	int nmodelsread;
	fread(&nmodelsread, sizeof(int), 1, fuzzyCorrsBin);
	if (nmodelsread!=nmodels)
		std::cout<<"[ERROR] Invalid # of models: "<<nmodelsread<<" vs "<<nmodels<<std::endl;
	assert(nmodelsread==nmodels);
	TimeProfiler timerCorrs;
	for (int i=0; i<nmodels; i++)
	{
		timerCorrs.WriteProgress("prg", i, nmodels);
		int mid = 0;
		fread(&mid, sizeof(int), 1, fuzzyCorrsBin);
		assert(mid==i);
		int npnts;
		fread(&npnts, sizeof(int), 1, fuzzyCorrsBin);
		for (int p=0; p<npnts; p++)
		{
			double xyz[3];
			fread(xyz, sizeof(double), 3, fuzzyCorrsBin);
			fuzzyPnts[i]->AddPoint(ShapePoint(R3Point(xyz[0], xyz[1], xyz[2])));
			
			int ncorrs;
			fread(&ncorrs, sizeof(int), 1, fuzzyCorrsBin);
			assert(ncorrs==nmodels);
			for (int m=0; m<nmodels; m++)
			{
				corrVals[m] = -1;
				corrIDs[m] = -1;
			}
			for (int c=0; c<ncorrs; c++)
			{
				fread(xyz, sizeof(double), 3, fuzzyCorrsBin);
				v2vmaps[i * nmodels + c][p] = R3Point(xyz[0], xyz[1], xyz[2]);
			}
		}
	}
	std::cout<<"Successful!"<<std::endl;
	int checksum;
	fread(&checksum, sizeof(int), 1, fuzzyCorrsBin);
	assert(checksum==-29349234);
	fclose(fuzzyCorrsBin);
}

double GetMeshRadius(const ShapeMesh * mesh)
{
	const GeomShape * shape = mesh->GetModel()->GetGeomShape("main_Even1000.pnts");
	//const GeomShape * shape = mesh->GetModel()->GetGeomShape("fuzzypnts");
	assert(shape!=NULL && shape->SafeCastToPointCloud()!=NULL);
	
	const ShapePointCloud * pnts = shape->SafeCastToPointCloud();
	double radius = 0;
	for (int i=0; i<(int)pnts->NumPoints(); i++)
	for (int j=i+1; j<(int)pnts->NumPoints(); j++)
		radius = vkMax(radius, R3Distance(pnts->PointPosition(i), pnts->PointPosition(j)));
	return radius;
}

