
#ifndef __TEMPLATE_MAPPING_H
#define __TEMPLATE_MAPPING_H

#include "VKString.h"
#include "Sortable.h"
#include "FileManagement.h"  
#include "BasicNumerical.h"

#include "RendererOSMesa.h"
#include "RendererGLUT.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "AppScreenshots.h"
#include "VisShapeTemplate.h"

#include "STemplate.h"
#include "TMatch.h"
#include "TMatchEnergy.h"
#include "TMatchAlg.h"
#include "PartOfEdge.h"
#include "ConsistentPartsAlg.h"
#include "STemplateNode.h"
#include "EvalCorrs.h"

class UnqiueTemplatePoint
{
	public:
		static bool m_bKeepOffset;
	
		UnqiueTemplatePoint(const R3Point & pnt, const R3Vector & offset, int uniqueID = -1);
		UnqiueTemplatePoint(const UnqiueTemplatePoint & pnt);
		
		int m_iUnqiuePointID;
		R3Point m_PntPos;	// always in undeformed space
		
		// offset for a point (i.e. relative to unqiue template point)
		R3Vector m_Offset;	// vector from template to point (aligned to template)
};

class Tmplt2TmpltMapping
{
	public:
		Tmplt2TmpltMapping(STemplate * st, TMatch * config1, TMatch * config2);
		UnqiueTemplatePoint MapConfig1ToConfig2(const UnqiueTemplatePoint & pntID) const;
		bool Verify(int config1, int config2) const;
	
//	protected:
		std::vector<UnqiueTemplatePoint> m_Config1ToConfig2;
		int m_iConfig1;
		int m_iConfig2;
	
	// for debug only
		TMatch * m_pMatch1;
		TMatch * m_pMatch2;
};

class Model2TmpltMapping
{
	public:
		Model2TmpltMapping(TMatch * match, std::vector<ShapePoint *> & featPnts);
		UnqiueTemplatePoint FromFeature2Tmplt(int featPntID, bool & valid) const;
		R3Point FromTemplate2Model(const UnqiueTemplatePoint & uniquePnt, int * optTmpltID = NULL ) const;
		R3Point GetGroundTruthPoint(int featPntID, bool & valid) const;
		double Radius() const;
		int ConfigID() const;

//	protected:
		std::vector<int> m_TemplateAllID2IncludedMap;
		std::vector<int> m_TemplateInlcuded2AllIDMap;
	
		double m_fRadius;
		R3Point m_MeshNormCenter;
		double m_fMeshNormRadius;
		std::vector<UnqiueTemplatePoint> m_FeatPnt2Template;
		ShapePointCloud * m_pUndeformedTemplatePnts;
		const R3Kdtree<const PointInSet *> * m_pKdTreeUndeformedTemplate;
	 
		// NOTE: template ids are local (i.e. ignoring excluded nodes)
		// If offset is non-zero need to have the model
		std::vector<R3Point> m_Tmplt2Model;
		std::vector<bool> m_bValidFeatPnt;
		std::vector<R3Point> m_GroundTruthPoint;
		int m_iConfigID;
	
	// for debug only
		TMatch * m_pMatch;
};

class TMapDebugApp : public DBApp
{
	public:	
		TMapDebugApp(ModelDatabase * db, std::vector<std::vector<Tmplt2TmpltMapping*> > & t2tmapping,
					 std::vector<Model2TmpltMapping*> & perModelAlign);
		TMapDebugApp(ModelDatabase * db, std::vector<std::vector<ShapePoint*> > & gtpnts,
					 std::vector<ShapePointCloud *> & fuzzyPnts, std::map<int, R3Point> * v2vmaps);

		virtual bool KeyPressed(DatabaseVisWindow * window, const char key, int mouseX, int mouseY);
		virtual bool Draw(DatabaseVisWindow * window);
		virtual void DrawTemplate(DatabaseVisWindow * window);
		virtual void DrawFuzzy(DatabaseVisWindow * window);
		
//	protected:
		std::vector<std::vector<Tmplt2TmpltMapping*> > * m_pT2TMapping;
		std::vector<Model2TmpltMapping*> * m_pMdl2TmpltMapping;
		std::vector<std::vector<ShapePoint*> > * m_pGtPnts;
		std::vector<ShapePointCloud *> * m_pFuzzyPnts;
		std::map<int, R3Point> * m_pV2VMap;
	
		void SetMaxPairwiseErr(std::vector<Sortable> * maxpairerr);
		std::vector<Sortable> * m_pMaxPairwiseErr;
		int m_iCurrPair;
		
	
};

R3Point MapPoint(int gtid, std::vector<std::vector<Tmplt2TmpltMapping*> > * tmplt2tmpltMapping,
				 std::vector<Model2TmpltMapping*> * mdl2tmpltMapping, int fromModelID, int toModelID,
				 bool verb, bool & valid, int *opt_t1 = NULL, int * opt_t2 = NULL);

#endif

