
#include "MatchHeaders.h"
/// Functions defined in MatchTemplate

void EvalFaces(ModelDatabase * db, const ShapeLabels * labels, const VKString & filename, const VKString & equivLabels,
			   const VKString & evalPerNodeFile, TMatch * match)
{
	std::map<VKString, double> nodeidAndLabel2Area;
	std::vector<VKStringList> equivalenceList;
	std::map<VKString, int> label2ListID;
	if (equivLabels!="none")
		ReadEquivalentLabels(equivLabels, equivalenceList, label2ListID);
	
	if (filename!="print")
		FileManagement::CreateDirsOnPathToFile(filename.c_str());
	
	assert(labels!=NULL);
	const ShapeMesh * mesh = labels->GetShape()->SafeCastToMesh();
	assert(mesh!=NULL);
	
	// get ground truth (mesh labels)
	VKString gtLabelsName = ShapeValue::m_sDefaultSceneGraphLabels;
	VKString mName = mesh->GetModel()->ModelName();
	const GeomShape * gtShape = ReadModelGeometry(db, mName, Model3D::m_sMainGeomName, true, gtLabelsName);
	const ShapeMesh * gtMesh = (gtShape!=NULL ? gtShape->SafeCastToMesh() : NULL);
	const ShapeValue * gtVals = (gtMesh!=NULL ? gtMesh->GetShapeVal(gtLabelsName) : NULL);
	const ShapeLabels * gtLabels = (gtVals!=NULL ? gtVals->SafeCastToLabels() : NULL);
	assert(gtLabels!=NULL);
	
	// OVERALL:
	//		- fraction classified correctly
	//		- fraction mis-classified (non NULL)
	// PER LABEL:
	//		- precision:	(# correctly classified as label) / (# classified as label in prediction)
	//		- recall:		(# correctly classified as label) / (# classified as label in gt)
	double overallCorrect = 0;
	double overallCorrectWithNull = 0;
	double overallLabels = 0;
	std::map<VKString, double> perClassCorrect;
	std::map<VKString, double> perClassClassifiedAs;
	std::map<VKString, double> perClassInGT;
	std::vector<std::map<VKString, double> * > allStatsVars;
	allStatsVars.push_back(&perClassCorrect);
	allStatsVars.push_back(&perClassClassifiedAs);
	allStatsVars.push_back(&perClassInGT);
	
	// go over all points
	for (int i=0; i<mesh->NFaces(); i++)
	{
		double farea = mesh->FaceArea(i);
		ShapePoint pntOnFace(mesh, i, 1./3., 1./3.);
		int segID = (int)labels->Value(pntOnFace, 0);
		int gtSegID = (int)gtLabels->Value(pntOnFace, 0);
		VKString label = MainName(labels->NodeLabel(segID), equivalenceList, label2ListID);
		VKString gtLabel = MainName(gtLabels->NodeLabel(gtSegID), equivalenceList, label2ListID);
		
		int localnodeid = segID - 1 ;
		VKString key = VKString::number(match->Deformation(localnodeid).NodeID())+"_"+gtLabel;
		if (nodeidAndLabel2Area.find(key) == nodeidAndLabel2Area.end())
			nodeidAndLabel2Area[key] = 0;
		nodeidAndLabel2Area[key] += farea;
		
		overallLabels += farea;
		
		// include all labels
		for (int s=0; s<(int)allStatsVars.size(); s++)
		{
			if (allStatsVars[s]->find(label)==allStatsVars[s]->end())
				(*allStatsVars[s])[label] = 0;
			if (allStatsVars[s]->find(gtLabel)==allStatsVars[s]->end())
				(*allStatsVars[s])[gtLabel] = 0;
		}
		
		if (Equivalent(label, gtLabel, label2ListID))
		{
			overallCorrectWithNull += farea;
			overallCorrect+= farea;
			perClassCorrect[label]+= farea;
		}
		
		perClassClassifiedAs[label] += farea;
		perClassInGT[gtLabel] += farea;
	}
	
	if (evalPerNodeFile!="none")
	{
		std::ofstream textStream(evalPerNodeFile.c_str());
		assert(textStream.is_open());
		for (std::map<VKString, double>::iterator iter = nodeidAndLabel2Area.begin();
			 iter != nodeidAndLabel2Area.end(); iter++)
			textStream<<iter->first.c_str()<<" "<<(iter->second / overallLabels)<<"\n";
		textStream.close();
		std::cout<<"Write Eval Per Node File: "<<evalPerNodeFile.c_str()<<std::endl;
	}
	
	// print to file
	std::ofstream textStream;
	if (filename!="print")
	{
		//assert(FileManagement::FileExists(filename.c_str()));
		textStream.open(filename.c_str());
		assert(textStream.is_open());
		textStream<<"All "<<overallCorrect<<" "<<overallCorrectWithNull<<" "<<overallLabels<<"\n";
	}
	std::cout<<"----- Evaluation Results. Model: "<<mName.c_str()<<" -----"<<std::endl;
	std::cout<<"ALL (NULL is always wrong): "<<((int)(100*overallCorrect / overallLabels))<<"%"<<std::endl;
//	std::cout<<"ALL (w/ NULL for unknown) : "<<((int)(100*overallCorrectWithNull / overallLabels))<<"%"<<std::endl;
	std::cout<<"Part           Pr     Rc     Lb     Gt"<<std::endl;
	for (std::map<VKString, double>::iterator iter = perClassInGT.begin();
		 iter != perClassInGT.end(); iter++)
	{
		bool inGT = perClassInGT.find(iter->first)!=perClassInGT.end();
		bool inLabel = perClassClassifiedAs.find(iter->first)!=perClassClassifiedAs.end();
		double nCorrect = perClassCorrect[iter->first];
		double nLabeled = inLabel ? perClassClassifiedAs[iter->first] : 0;
		double nGT = inGT ? perClassInGT[iter->first] : 0;
		if (filename!="print")
			textStream<<iter->first.c_str()<<" "<<nCorrect<<" "<<nLabeled<<" "<<nGT<<"\n";
		
		VKString name = iter->first;
		while (name.length()<15)
			name = name+" ";
		if (name.length()>15)
			name = name.left(14)+".";
		if (nLabeled==0)
			nLabeled = 1;
		if (nGT==0)
			nGT = 1;
		
		VKString pStr = VKString::number((int)(100*nCorrect/nLabeled))+"%";
		while (pStr.length()<7)
			pStr+=" ";
		std::cout<<name.c_str()<<pStr.c_str();
		
		pStr = VKString::number((int)(100*nCorrect/nGT))+"%";
		while (pStr.length()<7)
			pStr+=" ";
		std::cout<<pStr.c_str();
		
		pStr = VKString::number((int)(100*nLabeled/overallLabels))+"%";
		while (pStr.length()<7)
			pStr+=" ";
		std::cout<<pStr.c_str();
		
		pStr = VKString::number((int)(100*nGT/overallLabels))+"%";
		while (pStr.length()<7)
			pStr+=" ";
		std::cout<<pStr.c_str();
		
		std::cout<<std::endl;
	}
	std::cout<<"-----------------------------"<<std::endl;
	
	if (filename!="print")
		textStream.close();

}

void EvalMatching(ModelDatabase * db, const TMatch * match, const VKString & filename, const VKString & equivLabels)
{
	std::vector<VKStringList> equivalenceList;
	std::map<VKString, int> label2ListID;
	if (equivLabels!="none")
		ReadEquivalentLabels(equivLabels, equivalenceList, label2ListID);
	
	if (filename!="print")
		FileManagement::CreateDirsOnPathToFile(filename.c_str());
	assert(db->NumModels()>=1);
	
	assert(match!=NULL);
	const ShapePointCloud * pnts = match->GetMatchedGeomAsPnts();
	const ShapeLabels * labels = match->GetPerPointLabels();
	VKString nullLabel = "XXX_UNKNOWN_XXX";
	for (int i=0; i<match->NTemplateNodes(); i++)
		if (match->NodeEllipsoid(i).IsType(PartBox::BOXTYPE_NullBox))
			nullLabel = match->GetSTemplate()->GetNode(match->Deformation(i).NodeID())->NodeName();
	assert(pnts!=NULL && labels!=NULL);
	
	// get ground truth (mesh labels)
	VKString gtLabelsName = ShapeValue::m_sDefaultSceneGraphLabels;
	VKString mName = pnts->GetModel()->ModelName();
	const GeomShape * gtShape = ReadModelGeometry(db, mName, Model3D::m_sMainGeomName, true,
												  gtLabelsName);
	const ShapeMesh * gtMesh = (gtShape!=NULL ? gtShape->SafeCastToMesh() : NULL);
	const ShapeValue * gtVals = (gtMesh!=NULL ? gtMesh->GetShapeVal(gtLabelsName) : NULL);
	const ShapeLabels * gtLabels = (gtVals!=NULL ? gtVals->SafeCastToLabels() : NULL);
	assert(gtLabels!=NULL);
	
	// OVERALL:
	//		- fraction classified correctly
	//		- fraction mis-classified (non NULL)
	// PER LABEL:
	//		- precision:	(# correctly classified as label) / (# classified as label in prediction)
	//		- recall:		(# correctly classified as label) / (# classified as label in gt)
	int overallCorrect = 0;
	int overallCorrectWithNull = 0;
	int overallLabels = 0;
	std::map<VKString, int> perClassCorrect;
	std::map<VKString, int> perClassClassifiedAs;
	std::map<VKString, int> perClassInGT;
	std::vector<std::map<VKString, int> * > allStatsVars;
	allStatsVars.push_back(&perClassCorrect);
	allStatsVars.push_back(&perClassClassifiedAs);
	allStatsVars.push_back(&perClassInGT);
	
	VKStringList templateLabels;
	for (int i=0; i<match->GetSTemplate()->NNodes(); i++)
		if (!templateLabels.contains(match->GetSTemplate()->GetNode(i)->NodeName()))
			templateLabels.push_back(match->GetSTemplate()->GetNode(i)->NodeName());
	
	// go over all points
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		int segID = (int)labels->Value(pnts->GetPoint(i), 0);
		int gtSegID = (int)gtLabels->Value(pnts->GetPoint(i), 0);
		VKString label = MainName(labels->NodeLabel(segID), equivalenceList, label2ListID);
		VKString gtLabel = MainName(gtLabels->NodeLabel(gtSegID), equivalenceList, label2ListID);
		
		overallLabels++;
		
		// include all labels
		for (int s=0; s<(int)allStatsVars.size(); s++)
		{
			if (allStatsVars[s]->find(label)==allStatsVars[s]->end())
				(*allStatsVars[s])[label] = 0;
			if (allStatsVars[s]->find(gtLabel)==allStatsVars[s]->end())
				(*allStatsVars[s])[gtLabel] = 0;
		}
		
		if (Equivalent(label, gtLabel, label2ListID))
		{
			overallCorrectWithNull++;
			overallCorrect++;
			perClassCorrect[label]++;
		}
		else if (nullLabel == label)	// if labeled as null - still might be accepted if not in template
		{
			// is there a potentially correct label?
			bool labelExistsInTemplate = false;
			for (int tid = 0; tid < templateLabels.count() && !labelExistsInTemplate; tid++)
				labelExistsInTemplate = Equivalent(templateLabels[tid], gtLabel, label2ListID);
			if (!labelExistsInTemplate)
				overallCorrectWithNull++;
		}
		
		
		perClassClassifiedAs[label]++;
		perClassInGT[gtLabel]++;
	}
	
	// print to file
	std::ofstream textStream;
	if (filename!="print")
	{
		//assert(FileManagement::FileExists(filename.c_str()));
		textStream.open(filename.c_str());
		assert(textStream.is_open());
		textStream<<"All "<<overallCorrect<<" "<<overallCorrectWithNull<<" "<<overallLabels<<"\n";
	}
	std::cout<<"----- Evaluation Results. Model: "<<match->ModelName().c_str()<<" -----"<<std::endl;
	std::cout<<"ALL (NULL is always wrong): "<<(100*overallCorrect / overallLabels)<<"%"<<std::endl;
	std::cout<<"ALL (w/ NULL for unknown) : "<<(100*overallCorrectWithNull / overallLabels)<<"%"<<std::endl;
	std::cout<<"Part           Pr     Rc     Lb     Gt"<<std::endl;
	for (std::map<VKString, int>::iterator iter = perClassInGT.begin();
		 iter != perClassInGT.end(); iter++)
	{
		bool inGT = perClassInGT.find(iter->first)!=perClassInGT.end();
		bool inLabel = perClassClassifiedAs.find(iter->first)!=perClassClassifiedAs.end();
		int nCorrect = perClassCorrect[iter->first];
		int nLabeled = inLabel ? perClassClassifiedAs[iter->first] : 0;
		int nGT = inGT ? perClassInGT[iter->first] : 0;
		if (filename!="print")
			textStream<<iter->first.c_str()<<" "<<nCorrect<<" "<<nLabeled<<" "<<nGT<<"\n";
		
		VKString name = iter->first;
		while (name.length()<15)
			name = name+" ";
		if (name.length()>15)
			name = name.left(14)+".";
		if (nLabeled==0)
			nLabeled = 1;
		if (nGT==0)
			nGT = 1;
		
		VKString pStr = VKString::number((100*nCorrect/nLabeled))+"%";
		while (pStr.length()<7)
			pStr+=" ";
		std::cout<<name.c_str()<<pStr.c_str();
		
		pStr = VKString::number((100*nCorrect/nGT))+"%";
		while (pStr.length()<7)
			pStr+=" ";
		std::cout<<pStr.c_str();
		
		pStr = VKString::number((100*nLabeled/overallLabels))+"%";
		while (pStr.length()<7)
			pStr+=" ";
		std::cout<<pStr.c_str();
		
		pStr = VKString::number((100*nGT/overallLabels))+"%";
		while (pStr.length()<7)
			pStr+=" ";
		std::cout<<pStr.c_str();
		
		std::cout<<std::endl;
	}
	std::cout<<"-----------------------------"<<std::endl;
	
	if (filename!="print")
		textStream.close();
	
	
}

VKString MainName(const VKString & label,
				  std::vector<VKStringList> & equivalenceList, std::map<VKString, int> & label2ListID)
{
	std::map<VKString, int>::iterator iter = label2ListID.find(label);
	if (iter==label2ListID.end())
		return label;
	assert(iter->second < (int)equivalenceList.size() && equivalenceList[iter->second].count()>0);
	return equivalenceList[iter->second][0];
}

bool Equivalent(const VKString & label1, const VKString & label2, std::map<VKString, int> & label2ListID)
{
	if (label1==label2)
		return true;
	std::map<VKString, int>::iterator iter1 = label2ListID.find(label1);
	std::map<VKString, int>::iterator iter2 = label2ListID.find(label2);
	if (iter1==label2ListID.end() || iter2==label2ListID.end())
		return false;
	return (iter1->second==iter2->second);
}

void ReadEquivalentLabels(const VKString & equivFile,
						  std::vector<VKStringList> & equivalenceList, std::map<VKString, int> & label2ListID)
{
	std::ifstream textStream(equivFile.c_str());
	assert(textStream.is_open());
	VKString tempStr;
	while (!textStream.eof())
	{
		tempStr.readLine(textStream);
		int currID = (int)equivalenceList.size();
		equivalenceList.push_back(tempStr.split(" "));
		for (int i=0; i<equivalenceList[currID].count(); i++)
			label2ListID[equivalenceList[currID][i]] = currID;
	}
	
	textStream.close();
}


ShapeValue * TransferWithGraphCut(TMatch * match, GeomShape * toShape)
{
	//double wEdge = .2;
	// current scenario: pnts -> manifold mesh
	const ShapePointCloud * fromPnts = match->GetMatchedGeomAsPnts();
	assert(fromPnts!=NULL);
	ShapeMesh * toMesh = toShape->SafeCastToMesh();
	assert(toMesh!=NULL);
	
	// initialize MRF problem
	int nLabels = match->NTemplateNodes();
	int nFaces = toMesh->NFaces();
	
	// initialize edge energy based on dihedral angle
	VKGraph * smoothness = new VKGraph(nFaces);
	double pi = 3.14159265359;
	double epsilon = 0.0000001;
	double aveRad = toMesh->Area();
	bool manifold = toMesh->SafeCastToManifold()!=NULL;
	double wEdge = 50;	// vk: const, default = 70
	if (!manifold)
		wEdge = 1;
//	std::cout<<"wEdge: "<<wEdge<<std::endl;
	
	for (int f=0; f<nFaces; f++)
	{
		R3Vector n1 = toMesh->FaceNormal(f);
		std::vector<int> adj;
		toMesh->AdjacentFaces(f, adj);
		for (int fn=0; fn<(int)adj.size(); fn++)
		{
			if (f < adj[fn])	// ensure that only one edge is added
			{
				R3Vector n2 = toMesh->FaceNormal(fn);
				double dot = vkAbs(n1.Dot(n2));
				if (dot>=1.)
					dot = 1.;
				if (dot<=-1.)
					dot = -1.;
				double angle = acos(dot);
				int v1, v2;
				toMesh->EdgeBetweenFaces(f, adj[fn], v1, v2);
				double length = R3Distance(toMesh->VertexPosition(v1), toMesh->VertexPosition(v2))/aveRad;
				double w = -log(angle/pi+epsilon) * length;
				smoothness->AddEdge(f, adj[fn], wEdge * w);
				assert(w>=0);
			}
		}
	}
	
	// initialize face energy (based on points beloning to a face)
	double * facesEnergy = new double[nLabels * nFaces];
	int * votesPerFace = new int[nFaces];
	for (int i=0; i<nFaces; i++)
		votesPerFace[i] = 0;
	for (int i=0; i<nFaces*nLabels; i++)	// initialize to 0
		facesEnergy[i] = 0;
	int fid;
	double b1, b2;
	const ShapeMesh * mesh=NULL;
	// count votes
	for (int i=0; i<fromPnts->NumPoints(); i++)
	{
		fromPnts->GetPoint(i).GetBarycentricCoords(&fid, &b1, &b2, &mesh);
		assert(mesh == toMesh && fid>=0 && fid < nFaces);
		int lid = match->LocalNodeID(i);
		if (!match->GetSTemplate()->GetNode(match->Deformation(lid).NodeID())->IsNull())
		{
			votesPerFace[fid]++;
			facesEnergy[fid * nLabels + lid] += 1.;
		}
	}
	
	// convert votes into penalties
	for (int fid=0; fid<nFaces; fid++)
	{
		if (votesPerFace[fid]>0)
		{
			for (int lid=0; lid < nLabels; lid++)
			{
				int nvotes = facesEnergy[fid * nLabels + lid];
				assert(1. - nvotes / votesPerFace[fid]>=0);
				facesEnergy[fid * nLabels + lid] = 1. - (double)nvotes / (double)votesPerFace[fid];
			}
		}
	}
	
	// if not a manifold - assign small weights to all faces (to avoid isolated faces problem)
	if (!manifold)
	{
		double tArea = mesh->Area();
		int nPoints = 10000;
		
		for (int fid=0; fid<nFaces; fid++)
		{
			for (int lid=0; lid < nLabels; lid++)
				facesEnergy[fid * nLabels + lid] = .1;
		
			double fArea = mesh->FaceArea(fid);
			int currNpoints = fArea / tArea * nPoints;
			if (currNpoints==0)
				currNpoints = 1;
			
			for (int vid = 0; vid < currNpoints; vid++)
			{
				double b1 = vkAbs((double)rand() / (double)RAND_MAX);
				double b2 = vkAbs((double)rand() / (double)RAND_MAX);
				double b3 = vkAbs((double)rand() / (double)RAND_MAX);
				double sum = b1 + b2 + b3;
				R3Point p = ShapePoint(mesh, fid, b1 / sum, b2 / sum).GetPosition();
//				R3Point p = mesh->VertexPosition(mesh->VertexOnFace(fid, vid));
				const PointInSet * closest = fromPnts->GetKdTree()->FindClosest(p);
				assert(closest!=NULL);
				int lid = match->LocalNodeID(closest->id);
				facesEnergy[fid * nLabels + lid] -= .1 * 1. / (double)currNpoints;
				if (facesEnergy[fid * nLabels + lid]<0)
					facesEnergy[fid * nLabels + lid] = 0;
			}
		}
	}

	// label co-occurance
	double * labelsEnergy = new double[nLabels * nLabels];
	for (int i=0; i<nLabels; i++)
	for (int j=0; j<nLabels; j++)
	{
		if (i!=j)
			labelsEnergy[i*nLabels+j] = 1.;
		else
			labelsEnergy[i*nLabels+j] = 0;
	}
	
	// optimize MRF
	MRFProblemArray * problem = new MRFProblemArray(nFaces, nLabels, smoothness, facesEnergy, labelsEnergy, 1000);
	MRFOptimizer * optimizer = new MRFOptimizer(problem);
	int data, smooth;
	std::cout<<"Transfer Labels with Graph Cuts:"<<std::endl;
	std::cout<<"    Einit  = "<<optimizer->Energy(&data, &smooth)<<" = ";
	std::cout<<""<<(((double)data)/1000.)<<" + "<<((double)smooth/1000.)<<std::endl;
	optimizer->Optimize(10000);
	std::cout<<"    Efinal = "<<optimizer->Energy(&data, &smooth)<<" = ";
	std::cout<<""<<(((double)data)/1000.)<<" + "<<((double)smooth/1000.)<<std::endl;
	
	// create segmentation
	ShapeLabels * faceSeg = ShapeLabels::CopyLabels((ShapeLabels*)match->GetPerPointLabels(), toShape);
	std::vector<int> segIDs;
	for (int f=0; f<nFaces; f++)
	{
		int lid = optimizer->Label(f);
		segIDs.push_back(lid+1);
	}
	faceSeg->SetFaceSegments(segIDs);
	
	// return valid segmentation
	return faceSeg;
}

ShapeValue * TransferFromPly(TMatch * match, GeomShape * toShape, const VKString & plyFile)
{
	R3Mesh * plyMesh = new R3Mesh();
	plyMesh->ReadPlyFile(plyFile.c_str());
	ShapeLabels * faceSeg = ShapeLabels::CopyLabels((ShapeLabels*)match->GetPerPointLabels(), toShape);
	std::vector<int> segIDs;
	for (int f=0; f<plyMesh->NFaces(); f++)
	{
		int lid = plyMesh->FaceSegment(plyMesh->Face(f));
		segIDs.push_back(lid+1);
	}
	faceSeg->SetFaceSegments(segIDs);
	return faceSeg;
}
