#include "MatchHeaders.h"


int main(int argc, char** argv )
{
	if (argc<2)
	{
		std::cout<<"Usage:"<<std::endl;
		std::cout<<"   "<<argv[0]<<" dbName [args]"<<std::endl;;
		std::cout<<"Models to read:"<<std::endl;
		std::cout<<"    -model modelName"<<std::endl;
		std::cout<<"    -models modelList.txt"<<std::endl;
		
		std::cout<<"Creating/reading template:"<<std::endl;
		std::cout<<"    -initTmplt topologyFile geomName labelsName"<<std::endl;
		std::cout<<"    -initFromScratch "<<std::endl;
		std::cout<<"    -initFromPartDir dirName "<<std::endl;
		std::cout<<"    -train modelName trainGeomName trainLabelsName"<<std::endl;	// example segmentation
		std::cout<<"    -readTemplate templateFile"<<std::endl;						// pre-analyzed tempalte
		std::cout<<"    -readManifold - input meshes are manifolds"<<std::endl;
		std::cout<<"    -createFaceSegments outSegFile - write per face values"<<std::endl;
		std::cout<<"    -readFaceSegmentsFromPly plyFile outSegFile - per face values were projected externally"<<std::endl;

		std::cout<<"Fit template (segmentation, optimal t. configuration):"<<std::endl;
		std::cout<<"    -match modelName geomName"<<std::endl;
		std::cout<<"    -matchLabeled modelName geomName labelsName"<<std::endl;
		std::cout<<"    -matchLabeledAndE modelName geomName labelsName"<<std::endl;
		std::cout<<"    -matchLabeledAndEM modelName geomName labelsName"<<std::endl;
		std::cout<<"    -includeNull - include new nodes to template"<<std::endl;
		std::cout<<"    -includeNullOnly - only keep clusters of null nodes"<<std::endl;
		std::cout<<"    -updateNodes - update statistics from the segmentation"<<std::endl;
		std::cout<<"    -refGeom geomName - reference geometry name"<<std::endl;

		std::cout<<"Fit template parameters:"<<std::endl;
		std::cout<<"    -localAxes valName - local orientation at points "<<std::endl;
		std::cout<<"    -alignments [AllXYZ, AllXY, IdentityOnly] - initial alignments "<<std::endl;
		
		std::cout<<"Output:"<<std::endl;
		std::cout<<"    -writeTemplate templateFile"<<std::endl;
		std::cout<<"    -writeTemplateMetaData templateInfo.txt"<<std::endl;
		std::cout<<"    -writeMatching matching.txt"<<std::endl;
		std::cout<<"    -writeBinMatch file.match"<<std::endl;
		std::cout<<"    -readBinMatch file.match"<<std::endl;
		std::cout<<"    -writeXFormedModel meshfile"<<std::endl;
        std::cout<<"    -writeLabelsExternal dirname"<<std::endl;
		std::cout<<"    -evalMatching eval.txt"<<std::endl;
		std::cout<<"    -evalFaces eval.txt"<<std::endl;
		std::cout<<"    -equivLabels equivalence.txt"<<std::endl;
		std::cout<<"    -evalPerNode perNodeFile.txt"<<std::endl;
		std::cout<<"    -shotMatchedSeed imgfile "<<std::endl;
		std::cout<<"    -shotMatchedTmpltOnly imgfile "<<std::endl;
		std::cout<<"    -shotMatchedSegmentation imgfile "<<std::endl;
		std::cout<<"    -shotMatchedSegMesh imgfile "<<std::endl;
		std::cout<<"    -shotMatchedDeformed imgfile "<<std::endl;
		std::cout<<"    -shotMatchedDeformedMesh imgfile "<<std::endl;
		std::cout<<"    -shotTextured imgfile "<<std::endl;
		std::cout<<"    -shotPartMask imgfile "<<std::endl;
		std::cout<<"    -useMeshThresh value "<<std::endl;
		std::cout<<"    -writeCoAligned file.txt configID - co align points file "<<std::endl;
		std::cout<<"    -vis - visualize"<<std::endl;

		exit(0);
	}
	
	VKString createFaceSegments = "none";
	VKString readFaceSegmentsFromPly = "none";
	VKString refGeom = "none";
	bool interact = false;
	VKString initTmplt = "none";
	bool visualize = false;
	ModelDatabase::InputDirFormat dirFormat = ModelDatabase::DBINPUTDIR_MyFormat;
	VKString dbDir(argv[1]);
	VKString dbName = "none";
	std::map<VKString, VKString> shotTypeToImage;
	VKStringList modelsToLoad;
	int createFlags = 0;
	int readFlags = 0;
	
	VKString matchingFile = "none";
	VKString evalFile = "none";
	VKString evalFaces = "none";
	VKString evalPerNode= "none";
	VKString equivLabels = "none";
	VKString trainModel = "none";
	VKString trainGeomName="none", trainLabelsName="none";
	
	VKString matchModel="none";
	VKString matchGeom="none", matchSegName="none";
	bool multiSeg = false;
	
	VKString axesValName = "none";
	
	VKString readTemplate = "none";
	VKString writeTemplate = "none";
	
	VKString readBinMatch = "none";
	VKString writeBinMatch = "none";
	
	bool deformOnly = false;	// m-step only
	bool includeNull = false;
	bool includeNullOnly = false;
	bool updateNodes = false;
	bool verbose = false;
	bool readManifold = false;
	
	bool trustLabels = false;	// vk: const
	VKString writeXFormedModel = "none";
	VKString writeTemplateMetaData = "none";
	VKString writeLabelsExternal = "none";
	
	TMatchAlgParams matchParams;
	double useMeshThresh = -1;	//never use mesh
	
	VKString initFromPartDir = "none";
	VKString writeCoAligned = "none";
	int coAlignedID = -1;
	
	char relabelType = '0';
	
	for (int i=2; i<argc; i++)
	{
		if (strcmp(argv[i], "-model")==0)
		{
			VKString currM(argv[++i]);
			if (!modelsToLoad.contains(currM))
				modelsToLoad.push_back(currM);
		}
		else if (strcmp(argv[i], "-models")==0)
			ReadModelNames(modelsToLoad, argv[++i]);
		else if (strcmp(argv[i], "-train")==0)
		{
			trainModel = VKString(argv[++i]);
			trainGeomName = argv[++i];
			trainLabelsName = argv[++i];
		}
		else if (strcmp(argv[i], "-initTmplt")==0)
		{
			initTmplt = VKString(argv[++i]);
			trainGeomName = VKString(argv[++i]);
			trainLabelsName = VKString(argv[++i]);
		}
		else if (strcmp(argv[i], "-initFromScratch")==0)
			interact = true;
		else if (strcmp(argv[i], "-singleConfig")==0)
			matchParams.m_iExploreConfigs = atoi(argv[++i]);
		else if (strcmp(argv[i], "-initFromPartDir")==0)
			initFromPartDir = VKString(argv[++i]);
		else if (strcmp(argv[i], "-refGeom")==0)
			refGeom = VKString(argv[++i]);
		else if (strcmp(argv[i], "-match")==0)
		{
			matchModel = argv[++i];
			matchGeom = argv[++i];
		}
		else if (strcmp(argv[i], "-deformOnly")==0)
			deformOnly = true;
		else if (strcmp(argv[i], "-matchLabeled")==0
				 || strcmp(argv[i], "-matchLabeledAndE")==0
				 || strcmp(argv[i], "-matchLabeledAndEM")==0)
		{
			if (strcmp(argv[i], "-matchLabeledAndE")==0)
				relabelType = 'e';
			else if (strcmp(argv[i], "-matchLabeledAndEM")==0)
				relabelType = 'm';
			matchModel = argv[++i];
			matchGeom = argv[++i];
			matchSegName = argv[++i];
		}
		else if (strcmp(argv[i], "-matchGeomName")==0)
			matchGeom = argv[++i];
		else if (VKString(argv[i]).startsWith("-shot"))
		{
			VKString shotType(argv[i]);
			shotType.replace("-shot", "");
			shotTypeToImage[shotType] = VKString(argv[++i]);
		}
		else if (strcmp(argv[i], "-createFaceSegments")==0)
			createFaceSegments = argv[++i];
		else if (strcmp(argv[i], "-readFaceSegmentsFromPly")==0)
		{
			readFaceSegmentsFromPly = argv[++i];
			createFaceSegments = argv[++i];
		}
		else if (strcmp(argv[i], "-readBinMatch")==0)
			readBinMatch = argv[++i];
		else if (strcmp(argv[i], "-readManifold")==0)
			readManifold = true;
		else if (strcmp(argv[i], "-writeBinMatch")==0)
			writeBinMatch = argv[++i];
		else if (strcmp(argv[i], "-includeNull")==0)
			includeNull = true;
		else if (strcmp(argv[i], "-includeNullOnly")==0)
			includeNullOnly = true;
		else if (strcmp(argv[i], "-updateNodes")==0)
			updateNodes = true;
		else if (strcmp(argv[i], "-readTemplate")==0)
			readTemplate = argv[++i];
		else if (strcmp(argv[i], "-writeTemplate")==0)
			writeTemplate = argv[++i];
		else if (strcmp(argv[i], "-writeMatching")==0)
			matchingFile = argv[++i];
		else if (strcmp(argv[i], "-evalMatching")==0)
			evalFile = argv[++i];
		else if (strcmp(argv[i], "-evalFaces")==0)
			evalFaces = argv[++i];
		else if (strcmp(argv[i], "-evalPerNode")==0)
			evalPerNode = argv[++i];
		else if (strcmp(argv[i], "-equivLabels")==0)
			equivLabels = argv[++i];
		else if (strcmp(argv[i], "-localAxes")==0)
			axesValName = argv[++i];
		else if (strcmp(argv[i], "-writeCoAligned")==0)
		{
			writeCoAligned = argv[++i];
			coAlignedID = atoi(argv[++i]);
		}
		else if (strcmp(argv[i], "-alignments")==0)
		{
			i++;
			if (strcmp(argv[i], "IdentityOnly")==0)
				matchParams.m_InitAlignments = TMatchAlgParams::ALIGN_IdentityOnly;
			else if (strcmp(argv[i], "AllXYZ")==0)
				matchParams.m_InitAlignments = TMatchAlgParams::ALIGN_AllXYZ;
			else if (strcmp(argv[i], "AllXY")==0)
				matchParams.m_InitAlignments = TMatchAlgParams::ALIGN_AllXY;
			else if (strcmp(argv[i], "AllPCA")==0)
				matchParams.m_InitAlignments = TMatchAlgParams::ALIGN_AllPCA;
			else
				assert(false);
		}
		
		else if (strcmp(argv[i], "-vis")==0)
			visualize = true;
		else if (strcmp(argv[i], "-verb")==0)
			verbose = true;
		else if (strcmp(argv[i], "-writeXFormedModel")==0)
			writeXFormedModel = argv[++i];
		else if (strcmp(argv[i], "-writeTemplateMetaData")==0)
			writeTemplateMetaData = argv[++i];
		else if (strcmp(argv[i], "-writeLabelsExternal")==0)
			writeLabelsExternal = argv[++i];
		else if (strcmp(argv[i], "-useMeshThresh")==0)
			useMeshThresh = atof(argv[++i]);
		else
			std::cout<<"[WARNING] Unknown arg: "<<argv[i]<<" in MatchTemplate.cpp"<<std::endl;
	}
		
	DBCorrs * dbCorrs = NULL;	// todo
	ModelDatabase * db = NULL;
	if (modelsToLoad.count()==0)
		modelsToLoad.push_back("LOAD_NO_MODELS");
	db = ModelDatabase::CreateDBFromDir(dbDir, dirFormat, modelsToLoad, readFlags, createFlags, dbDir, dbName);
		
	STemplate * newTemplate = NULL;
	std::cout<<"Reading / Creating template"<<std::endl;
	if (trainGeomName!="none" && trainLabelsName!="none" && trainModel!="none")
 		newTemplate = SingleTemplate(db, trainModel, trainGeomName, trainLabelsName, trustLabels, useMeshThresh);
	else if (readTemplate!="none")
		newTemplate = STemplate::ReadTemplate(readTemplate, db);
	else if (initTmplt!="none")
		newTemplate = InitTemplate(initTmplt, trainGeomName, trainLabelsName, db);
	else if (initFromPartDir!="none")
		newTemplate = ReadTemplateFromDir(initFromPartDir);
		
	std::vector<TMatchEnergy*> allConfigsEnergies;
	if (readBinMatch!="none")
	{
		std::cout<<"Reading matches"<<std::endl;
		assert(newTemplate!=NULL && db!=NULL);
		FILE * matchBin = fopen(readBinMatch.c_str(), "rb");
		if(matchBin==NULL)
			std::cout<<"[ERROR] Cannot read "<<readBinMatch.c_str()<<std::endl;
		assert(matchBin!=NULL); 
		assert(db->NumModels()<=2);
		// TMatch reader assumes that geometry was loaded
		assert(matchGeom!="none");	// although saved
		for (int i=0; i<db->NumModels(); i++)
			ReadModelGeometry(db, db->GetModel(i)->ModelName(), matchGeom, readManifold);
		allConfigsEnergies.push_back(TMatchEnergy::ReadEnergyAndMatch(matchBin, newTemplate, db));
		allConfigsEnergies[0]->GetMatch()->CreateTemplateSamples();
		fclose(matchBin);
	}
	
	// check if template exists
	if (matchModel!="none")
	{
		assert(newTemplate!=NULL);
		if (verbose)
			std::cout<<"Matching Models: "<<matchModel.c_str()<<std::endl;

		if (matchSegName=="none")
		{ 
			GeomShape * shape = ReadModelGeometry(db, matchModel, matchGeom, readManifold);
			if (readBinMatch!="none")	// start with initial guess
			{
				TMatchAlg::s_sProfiler.startedProcess("1_Optimize");
				TMatchAlg::s_sProfiler.startedProcess("1_OptimizeMeasure");
				assert(allConfigsEnergies.size()==1);
				TMatch * match = allConfigsEnergies[0]->GetMatch();
				int niter = 0;
				if (deformOnly)
				{
					int oldTerms = allConfigsEnergies[0]->GetTerms();
					int newTerms = oldTerms & (~(TMatchEnergy::ET_NodeXForm | TMatchEnergy::ET_NodeShape));
					allConfigsEnergies[0]->SetTerms(newTerms);
					TMatchAlgParams params;
					int globFlags = TMatchAlgParams::OPT_GlobalScale | TMatchAlgParams::OPT_GlobalRotate;
					globFlags |= TMatchAlgParams::OPT_GlobalTrans;
					params.m_iOptimFlags &= ~globFlags;
					TMatchAlg::MStep_DeformTemplate(allConfigsEnergies[0], params, niter);
					//allConfigsEnergies[0]->SetTerms(oldTerms);
				}
				else
					allConfigsEnergies[0] = TMatchAlg::OptimizeConfigured(newTemplate, match, matchParams);
				TMatchAlg::s_sProfiler.finishedProcess("1_OptimizeMeasure");
				TMatchAlg::s_sProfiler.finishedProcess("1_Optimize");
			}
			else	// new match
				TMatchAlg::Optimize(newTemplate, shape, matchParams, allConfigsEnergies);
		}
		else
		{
			assert(false);
			GeomShape * mesh = ReadModelGeometry(db, matchModel, Model3D::m_sMainGeomName, readManifold, matchSegName);
			ShapeValue * vals = mesh->GetShapeVal(matchSegName);
			assert(vals!=NULL && vals->SafeCastToLabels()!=NULL);
			GeomShape * pnts = ReadModelGeometry(db, matchModel, matchGeom, readManifold);
			ShapeLabels * labels = ShapeLabels::CopyLabels(vals->SafeCastToLabels(), pnts);
			TMatchAlg::OptimizeLabeled(newTemplate, labels, matchParams, allConfigsEnergies, relabelType);
		}
	}
	
	assert(allConfigsEnergies.size()>0);
	// best energy
	TMatchEnergy * bestEnergy = allConfigsEnergies[0];
	
	// project points to faces
	ShapeLabels * faceSegments = NULL;
	
	std::cout<<"Create Face Segments: "<<createFaceSegments.c_str()<<std::endl;
	if (createFaceSegments!="none")
	{
		VKString mName = bestEnergy->GetMatch()->ModelName();
		GeomShape * mainGeom = db->GetModel(mName)->GetGeomShape(Model3D::m_sMainGeomName);
		ShapeValue * newLabels = NULL;
		if (readFaceSegmentsFromPly!="none")	// read ply file - someone else projected faces
			newLabels = TransferFromPly(bestEnergy->GetMatch(), mainGeom, readFaceSegmentsFromPly);
		else
			newLabels = TransferWithGraphCut(bestEnergy->GetMatch(), mainGeom);
		assert(newLabels!=NULL);
		faceSegments = newLabels->SafeCastToLabels();
		mainGeom->AddShapeVal("FaceSegments", faceSegments);
		newLabels->WriteShapeValues(createFaceSegments);
	}
	
	// write the best matching
	
	if (matchingFile!="none")
	{
		// should include: energy, per-point labels, node deformation
		FileManagement::CreateDirsOnPathToFile(matchingFile.c_str());
		std::ofstream textStream(matchingFile.c_str());
		assert(textStream.is_open());
		int origFlags = bestEnergy->GetTerms();
		char * partFlags = new char[newTemplate->NNodes()+1];
		partFlags[newTemplate->NNodes()] = 0;
		
		for (int i=0; i<newTemplate->NNodes(); i++)
			partFlags[i] = '0';
		
		for (int i=0; i<bestEnergy->GetMatch()->NTemplateNodes(); i++)
		{
			const TemplateNodeDeform & deform = bestEnergy->GetMatch()->Deformation(i);
			int nodeid = deform.NodeID();
			int partid = deform.PartID();
			int nparts = deform.NParts();
			assert(partid==0 && nparts==1 && nodeid < newTemplate->NNodes() && nodeid>=0);
			if (deform.IsIncluded())
				partFlags[nodeid] = '1';
		}
		
		bestEnergy->SetTerms(TMatchEnergy::DefaultMatchQualityEnergy());
		textStream<<"Score="<<bestEnergy->Energy()<<"\n";	textStream.flush();
		bestEnergy->SetTerms(origFlags);
		textStream<<"Time="<<TMatchAlg::s_sProfiler.ProcessLengthStr("1_OptimizeMeasure").c_str()<<"\n";  textStream.flush();
		textStream<<"SeedID="<<bestEnergy->GetMatch()->SeedID()<<"\n"; textStream.flush();
		textStream.flush();
		textStream<<"Energy="<<bestEnergy->Energy()<<"\n";	textStream.flush();
		textStream.flush();
		int pid = bestEnergy->GetMatch()->SeedID();
		while (pid > 0)
			pid = newTemplate->ConfigParent(pid);
		textStream<<"BasicSeedID="<<pid<<"\n";  textStream.flush();
		textStream.flush();
		textStream<<"PartFlags="<<partFlags<<"\n";  textStream.flush();
		textStream<<"TmpltSamples="<<newTemplate->NSamples(bestEnergy->GetMatch()->SeedID())<<"\n";  textStream.flush();
		std::vector<int> subset;
		TMatch * s = newTemplate->CreateSeedConfig(bestEnergy->GetMatch()->SeedID(), NULL, subset, R3identity_affine);
		bool sameParts, sameIncluded;
		
		const std::vector<TemplateNodeDeform *> emptySet;
		TMatchEnergy * deformed = new TMatchEnergy(bestEnergy->GetMatch()->CreateCopy(emptySet));
		
		int oldTerms = deformed->GetTerms();
		int newTerms = oldTerms & (~(TMatchEnergy::ET_NodeXForm | TMatchEnergy::ET_NodeShape));
		deformed->SetTerms(newTerms);
		TMatchAlgParams params;
		int globFlags = TMatchAlgParams::OPT_GlobalScale | TMatchAlgParams::OPT_GlobalRotate;
		globFlags |= TMatchAlgParams::OPT_GlobalTrans;
		params.m_iOptimFlags &= ~globFlags;
		int niter = 0;
		TMatchAlg::MStep_DeformTemplate(deformed, params, niter);

		double dist = TMatchEnergy::DeformationDistance(deformed->GetMatch(), s, sameParts, sameIncluded);
		if (!sameParts && !sameIncluded)
			dist = 10000;

		textStream<<"DeformDist="<<dist<<"\n";  textStream.flush();
		textStream.close();
		delete [] partFlags;
		
		VKString timingFile = matchingFile;
		timingFile = timingFile.replace(".txt", "").replace(".matching", ".timing")+".txt";
		std::ofstream textStream2(timingFile.c_str());
		assert(textStream2.is_open());
		textStream2<<TMatchAlg::s_sProfiler.profile().c_str();
		textStream2.close();
	}
	else
	{
		std::cout<<"Best Energy: "<<bestEnergy->EnergyStr().c_str()<<std::endl;
		std::cout<<"Total Time: "<<TMatchAlg::s_sProfiler.ProcessLengthStr("1_Optimize").c_str()<<std::endl;
		std::cout<<TMatchAlg::s_sProfiler.profile().c_str()<<std::endl;
	}
	
	// binary best matching
	if (writeBinMatch!="none")
	{
		assert(bestEnergy!=NULL);
		FILE * matchBin = fopen(writeBinMatch.c_str(), "wb");
		if(matchBin==NULL)
			std::cout<<"[ERROR] Cannot write "<<writeBinMatch.c_str()<<std::endl;
		assert(matchBin!=NULL);
		bestEnergy->WriteEnergyAndMatch(matchBin);
		fclose(matchBin);
	}
	
	// write evaluation (treats SceneGraphNodes labels as ground truth)
	if (evalFile!="none")
		EvalMatching(db, bestEnergy->GetMatch(), evalFile, equivLabels);
	if (evalFaces!="none")
		EvalFaces(db, faceSegments, evalFaces, equivLabels, evalPerNode, bestEnergy->GetMatch());
			
	// write aligned model
	if (writeXFormedModel!="none")
		WriteAlignedModel(bestEnergy->GetMatch(), writeXFormedModel);
		
	// write template file
	if (writeTemplate!="none" && newTemplate!=NULL)
	{
		if (verbose)
			std::cout<<"Writing template: "<<writeTemplate.c_str()<<std::endl;
		
		FileManagement::CreateDirsOnPathToFile(writeTemplate.c_str());
		newTemplate->WriteTemplate(writeTemplate);
	}
	
	if (writeTemplateMetaData!="none")
	{
		if (verbose)
			std::cout<<"Writing template meta data: "<<writeTemplateMetaData.c_str()<<std::endl;
		assert(newTemplate!=NULL);
		WriteTemplateMetaData(newTemplate, writeTemplateMetaData);
	}
	
	if (writeLabelsExternal!="none" && allConfigsEnergies.size()>0)
	{
		WriteLabelsExternal(db, allConfigsEnergies[0]->GetMatch(), writeLabelsExternal, equivLabels);
	}
	
	if (writeCoAligned!="none")
	{
		std::cout<<"Writing "<<writeCoAligned.c_str()<<std::endl;
		FileManagement::CreateDirsOnPathToFile(writeCoAligned.c_str());
		FILE * file = fopen(writeCoAligned.c_str(), "wb");
		assert(file!=NULL);
		assert(db->NumModels()==1 && allConfigsEnergies.size()>0);
		TMatch * match = allConfigsEnergies[0]->GetMatch();
		int seedID = match->SeedID();
		int npnts = match->NShapePrimitives();
		fwrite(&seedID, sizeof(int), 1, file);
		
		const STemplate * st = match->GetSTemplate();
		std::vector<R3Point> pntsToWrite;
		for (int i=0; i<npnts; i++)
		{
			//std::cout<<"Pnt: "<<i<<std::endl;
//			// DEFORM SPACE:
//			R3Point pos = match->AlignedToNodePointPosition(match->LocalNodeID(i), i, coAlignedID);
			
			// PROJECT WITH OFFSET
			if (st->GetNode(match->Deformation(match->LocalNodeID(i)).NodeID())->IsNull())
				continue;
			R3Point p0 = match->AlignedPointPosition(i);
			R3Point pos = p0;
			if (coAlignedID>=0)
			{
				int tid = match->NearestTemplatePointGlobalID(match->LocalNodeID(i), p0);
				R3Point tp = match->DeformedTemplateSample(tid);
				pos = st->TemplatePointInOtherConfig(coAlignedID, match, tid) + (p0 - tp);
			}
			
			//std::cout<<"Pnt: "<<i<<" - GOT IT"<<std::endl;
			// remember position to write
			pntsToWrite.push_back(pos);
		}
		
		npnts = (int)pntsToWrite.size();
		fwrite(&npnts, sizeof(int), 1, file);
		for (int i=0; i<npnts; i++)
		{
			// write output
			double xyz[3];
			xyz[0] = pntsToWrite[i].X();
			xyz[1] = pntsToWrite[i].Y();
			xyz[2] = pntsToWrite[i].Z();
			fwrite(xyz, sizeof(double), 3, file);
		}
		
		fclose(file);
	}
	
	if (interact)
		visualize = true;
		
	if (visualize || shotTypeToImage.size()>0)
	{	
#ifdef VISUALIZATION		
		if (verbose)
			std::cout<<"Init visualization"<<std::endl;
		int listFlags = DatabaseVisWindow::LIST_GROUPS_CORRS;
		listFlags |= DatabaseVisWindow::LIST_GROUPS_MODELS;

		int nWnd = (allConfigsEnergies.size()==0 ? 1 : 2);
		DatabaseVisWindow * dbVis = new DatabaseVisWindow(db, dbCorrs, listFlags, nWnd, 1, 0, 0, 1);
		if (db->NumModels()==1 && allConfigsEnergies.size()>0)
			dbVis->WindowToModel()[1] = 0;
		VisShapeTemplate * visTemplate = new VisShapeTemplate(db);
		if (interact)
		{
			visTemplate->SetInteractionCreateTemplate(writeTemplate, newTemplate);
			if (refGeom!="none")
			{
				assert(db->NumModels()==1);
				for (int i=0; i<db->NumModels(); i++)
					ReadModelGeometry(db, db->GetModel(i)->ModelName(), refGeom, readManifold);
				visTemplate->SetReferenceModel(db->GetModel(0)->ModelName(), refGeom);
			}
		}
		else
		{
			assert(newTemplate!=NULL);
			if (allConfigsEnergies.size()==0)
			{
				for (int i=0; i<newTemplate->NSeedConfigs(); i++)
				{
					TMatch * seed = newTemplate->CreateSeedConfig(i, NULL, std::vector<int>(), R3identity_affine);
					visTemplate->AddMatchedModel(seed, new TMatchEnergy(seed));
				}
			}
			else
			{
				for (int i=0; i<(int)allConfigsEnergies.size(); i++)
				{
					TMatch * matched = allConfigsEnergies[i]->GetMatch();
					const STemplate * st = matched->GetSTemplate();
					TMatch * seed = st->CreateSeedConfig(matched->SeedID(), NULL, std::vector<int>(), R3identity_affine);
					visTemplate->AddMatchedModel(seed, allConfigsEnergies[i]);
				}
			}
		}

		if (shotTypeToImage.size()>0)
		{
			AppScreenshots * takeShots = new AppScreenshots(!visualize, db);
			dbVis->SetApp(takeShots);

			std::map<VKString, int> visConfigs;
			int visbox = (VisShapeTemplate::VISTMPLT_DrawTemplateAsBoxes 
						  | VisShapeTemplate::VISTMPLT_DrawTemplateAsAxes);
			visConfigs["MatchedSegmentation"] = VisShapeTemplate::VISTMPLT_MatchedSegPnts;
			visConfigs["MatchedSeed"] = (VisShapeTemplate::VISTMPLT_MatchedSegPnts
										 | VisShapeTemplate::VISTMPLT_SeedOverMatched | visbox);
			visConfigs["MatchedDeformed"] = (VisShapeTemplate::VISTMPLT_MatchedSegPnts
											| VisShapeTemplate::VISTMPLT_DeformedOverMatched | visbox);
			visConfigs["MatchedDeformedMesh"] = (VisShapeTemplate::VISTMPLT_DrawSegmentedMesh
											 | VisShapeTemplate::VISTMPLT_DeformedOverMatched | visbox);
			visConfigs["MatchedTmpltOnly"] = VisShapeTemplate::VISTMPLT_DeformedOverMatched | visbox;
			visConfigs["Textured"] = VisShapeTemplate::VISTMPLT_DrawMeshTextures;
			visConfigs["MatchedSegMesh"] = VisShapeTemplate::VISTMPLT_DrawSegmentedMesh;	// TODO:
			visConfigs["PartMask"] = 0;	// TODO:
			 
			for (std::map<VKString, VKString>::iterator iter = shotTypeToImage.begin();
				 iter != shotTypeToImage.end(); iter++)
			{
				if (visConfigs.find(iter->first)==visConfigs.end())
					std::cout<<"[WARNING] Unknown screenshot type "<<iter->first.c_str()<<std::endl;
				else
					takeShots->TakeShotOfApp(visTemplate, iter->first, visConfigs[iter->first], iter->second);
			}
		}
		else
			dbVis->SetApp(visTemplate);

#	ifdef MESA_RENDERING
		std::cout<<"[WARNING] Mesa renderer"<<std::endl;
		VisWindowRenderer * wndMesa = new RendererOSMesa(dbVis);
		wndMesa->UpdateScreen();
#	else
		if (verbose)
			std::cout<<"Running App"<<std::endl;
		VisWindowRenderer * wndGLUT = new RendererGLUT(argc, argv, dbVis);
#	endif
#else
		std::cout<<"[WARNING] Visualization disabled (compiling flag)"<<std::endl;
#endif
	}
	
	return 0;
}



