#include "LinAlgMatrixSparseReal.h"
#include "LinAlgVectorReal.h"
#include "LinAlgMatrixReal.h"
#include <stdlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <sstream>
#include <fstream>
#include <iostream>

//////////    LinAlgMatrixSparseReal - PARENT CLASS /////////////
LinAlgMatrixSparseReal::LinAlgMatrixSparseReal(int m, int n, SparseMtxDataTypes tp)
{
	this->m = m;
	this->n = n;
	m_SparseMtxType = tp;
}

LinAlgMatrixSparseReal::~LinAlgMatrixSparseReal()
{
	
}

SparseMtxDataTypes LinAlgMatrixSparseReal::GetMatrixDataType() const
{
	return m_SparseMtxType;
}

int LinAlgMatrixSparseReal::Rows() const
{
	return m;
}

int LinAlgMatrixSparseReal::Cols() const
{
	return n;
}

//////////    LinAlgMatrixSparseRealCXSparse - CXSparse /////////////
#ifdef WITH_CXSPARSE 
LinAlgMatrixSparseRealCXSparse::LinAlgMatrixSparseRealCXSparse(int m, int n, int maxNonzero)
: LinAlgMatrixSparseReal(m, n, SPARSE_MTX_CXSPARSE)
{
	m_aPreAllocatedX = NULL;
	m_pPreFactorizedData = NULL;
	m_pPreFactorizedS = NULL;
	m_pPreFactorizedN = NULL;
	m_aPreFactorizedWorkspace = NULL;
	m_PreFactorized = LinAlg::MAT_FACT_NONE;
	m_pData = NULL;

	int values = 1;	//init values
	int triplet = 1; //triplet
	if (maxNonzero<0)
		maxNonzero = m * 10;
	m_pData = cs_di_spalloc (m, n, maxNonzero, values, triplet);
}

LinAlgMatrixSparseRealCXSparse::~LinAlgMatrixSparseRealCXSparse()
{
	
	cs_di_spfree(m_pData);
	if (m_pPreFactorizedData!=NULL)
		cs_di_spfree(m_pPreFactorizedData);
	if (m_aPreAllocatedX!=NULL)
		delete [] m_aPreAllocatedX;
	
	if (m_aPreFactorizedWorkspace!=NULL)
		delete [] m_aPreFactorizedWorkspace;
	
	if (m_pPreFactorizedS!=NULL)
		cs_sfree (m_pPreFactorizedS) ;
	if (m_pPreFactorizedN!=NULL)
		cs_nfree (m_pPreFactorizedN) ;

}

void LinAlgMatrixSparseRealCXSparse::AddVal(int i, int j, double val)
{
	cs_di_entry (m_pData, i, j, val) ;
}

// stores factorization result in the matrix
void LinAlgMatrixSparseRealCXSparse::Factorize(LinAlg::MatrixFactorizationType factType)
{
	m_PreFactorized = factType;
	std::cout<<"Factorize ("<<m<<"x"<<n<<", nz="<<")"<<std::flush;
	assert(m_pPreFactorizedData==NULL);
	assert(m_pPreFactorizedS==NULL);
	assert(m_pPreFactorizedN==NULL);
	assert(m_aPreFactorizedWorkspace==NULL);
		
	// should have more
	assert(m >= n);
	cs_di * A = cs_di_compress(m_pData);
	assert(A!=NULL);
	bool couldDupl = cs_dupl(A);
	assert(couldDupl);
	
	// A = At * A (make matrix square and symmetric)
	if (m>n || m_PreFactorized==LinAlg::MAT_FACT_CHOLESKY_SYMM)
	{
		assert(m_PreFactorized!=LinAlg::MAT_FACT_CHOLESKY);
		cs_di * AT = cs_transpose (A, 1) ;
		assert(AT!=NULL);
		m_pPreFactorizedData = cs_di_multiply (AT, A) ;		
		m_pData = AT;
		cs_spfree(A);
	}
	else	// matrix is already square
	{
		m_pPreFactorizedData = A;
	}
	
	
	std::cout<<"FactorizeatioN: At*A="<<m_pPreFactorizedData->m;
	std::cout<<" x "<<m_pPreFactorizedData->n<<std::endl;
	assert(m_pPreFactorizedData->m==m_pPreFactorizedData->n);
	
	// create temp data: compressed A
	int order = 1;	// order: 0, 1, 2, 3 (?)
	double tol = 0;
	int ok;
	
	assert(m_pPreFactorizedData!=NULL);
	int allocSize = 1;
	
	if (m_PreFactorized==LinAlg::MAT_FACT_QR)		// pre-factorize: QR version
	{
		// ordering and symbolic analysis
		m_pPreFactorizedS = cs_di_sqr (order, m_pPreFactorizedData, 1);	
		// numeric QR factorization
		m_pPreFactorizedN = cs_di_qr (m_pPreFactorizedData, m_pPreFactorizedS);
		if (m_pPreFactorizedS!=NULL)
			allocSize=m_pPreFactorizedS->m2;
	}
	else if (m_PreFactorized==LinAlg::MAT_FACT_CHOLESKY_SYMM 
			 || m_PreFactorized==LinAlg::MAT_FACT_CHOLESKY)		// pre-factorize: Cholesky version
	{
		// ordering and symbolic analysis 
		m_pPreFactorizedS = cs_di_schol (order, m_pPreFactorizedData);
		// numeric Cholesky factorization
		m_pPreFactorizedN = cs_di_chol (m_pPreFactorizedData, m_pPreFactorizedS);
		if (m_pPreFactorizedS!=NULL)
			allocSize=m_pPreFactorizedData->n;
	}
	else
		assert(false);
	
	// allocate workspace
	m_aPreFactorizedWorkspace =(double*) cs_calloc(allocSize, sizeof (double));	
	
	assert(m_pPreFactorizedS!=NULL);
	assert(m_pPreFactorizedN!=NULL);
	assert(m_aPreFactorizedWorkspace!=NULL);
}

void LinAlgMatrixSparseRealCXSparse::SolveFactorized(const LinAlgVectorReal & b, 
													 LinAlgVectorReal &x) const
{
	// create temp data: b
	LinAlgMatrixSparseRealCXSparse * ncThis= (LinAlgMatrixSparseRealCXSparse*)this;
	if (ncThis->m_aPreAllocatedX==NULL)
		ncThis->m_aPreAllocatedX = new double[x.Rows()];
	
	assert(m>=n);
	int newM = n;
	
	// b = m_pData * b		(A^t * b)
	if (m > n || m_PreFactorized==LinAlg::MAT_FACT_CHOLESKY_SYMM)
	{
		for (int i=0; i<x.Rows(); i++)
			x.m_aRealData[i] = 0;
		cs_di_gaxpy(m_pData, b.m_aRealData, x.m_aRealData);			
	}
	else
		x = (LinAlgVectorReal&)b;
	
	if (m_PreFactorized==LinAlg::MAT_FACT_QR)	// QR-solver
	{
		cs_ipvec (m_pPreFactorizedS->pinv, x.m_aRealData, 
				  m_aPreAllocatedX, newM) ;   // x(0:m-1) = b(p(0:m-1)
		for (int k = 0 ; k < n ; k++)       // apply Householder refl. to x
			cs_happly (m_pPreFactorizedN->L, k, m_pPreFactorizedN->B [k], m_aPreAllocatedX) ;
		cs_usolve (m_pPreFactorizedN->U, m_aPreAllocatedX) ;           // x = R\x
		cs_ipvec (m_pPreFactorizedS->q, m_aPreAllocatedX, 
				  x.m_aRealData, n) ;      // x(q(0:n-1)) = x(0:n-1) 
	}
	else if (m_PreFactorized==LinAlg::MAT_FACT_CHOLESKY_SYMM
			 || m_PreFactorized==LinAlg::MAT_FACT_CHOLESKY)	// Cholesky solver
	{
		cs_ipvec (m_pPreFactorizedS->pinv, x.m_aRealData, m_aPreAllocatedX, newM) ;   // x = P*b
		cs_lsolve (m_pPreFactorizedN->L, m_aPreAllocatedX) ;           // x = L\x 
		cs_ltsolve (m_pPreFactorizedN->L, m_aPreAllocatedX) ;          // x = L'\x 
		cs_pvec (m_pPreFactorizedS->pinv, m_aPreAllocatedX, x.m_aRealData, n) ;    // b = P'*x 
	}
}

// note: matrix content might be destroyed
void LinAlgMatrixSparseRealCXSparse::Solve(const LinAlgVectorReal & b,
										   LinAlgVectorReal &x)
{
	// only for square input matrix
	assert(b.Rows()==m && b.Rows()==n 
		   && b.Rows()==x.Rows());
	
	// create temp data: b
	if (m_aPreAllocatedX==NULL)
		m_aPreAllocatedX = new double[b.Rows()];
	
	for (int i=0; i<b.Rows(); i++)
		m_aPreAllocatedX[i] = b.GetValue(i);
	
	// create temp data: compressed A
	int order = 1;	// order: 0, 1, 2, 3 (?)
	double tol = 0;
	int ok;
	cs_di * deleteMeData = cs_di_compress(m_pData);
	assert(deleteMeData!=NULL);
	bool couldDupl = cs_dupl(deleteMeData);
	assert(couldDupl);
	
	// solve: A x = b
	ok = cs_lusol (order, deleteMeData, m_aPreAllocatedX, tol);
	
	if (!ok)
	{
		std::cout<<"[WARNING] Solution did not succeed "<<std::endl;
	}
	//assert(ok);
	
	// copy x 
	for (int i=0; i<m; i++)
		x(i) = m_aPreAllocatedX[i];
	
	//delete [] deleteMe;
	
	// free copied matrix
	cs_di_spfree(deleteMeData);
}
#endif

//////////    LinAlgMatrixSparseLibSVDC - LIBSVDC (included) /////////////
LinAlgMatrixSparseRealLibSVDC::LinAlgMatrixSparseRealLibSVDC(int m, int n, int maxNonzero)
: LinAlgMatrixSparseReal(m, n, SPARSE_MTX_LIBSVDC)
{
	assert(maxNonzero>0);	// need to know exact number of non-zero entries
	m_pSVDLibData = svdNewSMat(m, n, maxNonzero);
	m_iCurrSVDLibDataCol = -1;
	m_iCurrSVDLibDataRow = -1;
	m_iCurrentSVDLibDataN = 0;

}

LinAlgMatrixSparseRealLibSVDC::~LinAlgMatrixSparseRealLibSVDC()
{
	svdFreeSMat(m_pSVDLibData);
}

void LinAlgMatrixSparseRealLibSVDC::AddVal(int i, int j, double val)
{
	assert(m_pSVDLibData!=NULL);
	if (j < m_iCurrSVDLibDataCol )
	{
		std::cout<<"[ERROR] Sparse Matrix (LibSVDC): Columns can only increase."<<std::endl;
		assert(false);
	}
	else if (j == m_iCurrSVDLibDataCol && i <= m_iCurrSVDLibDataRow)
	{
		std::cout<<"[ERROR] Sparse Matrix (LibSVDC): Rows can only increase for the same column.";
		assert(false);
	}
	
	assert(m_iCurrentSVDLibDataN < m_pSVDLibData->vals);	// can fit non-zero entries
	if (j > m_iCurrSVDLibDataCol)	// column increased
	{
		for (int c=m_iCurrSVDLibDataCol+1; c<=j; c++)
			m_pSVDLibData->pointr[c] = m_iCurrentSVDLibDataN;
		m_iCurrSVDLibDataCol = j;
		m_iCurrSVDLibDataRow = i;
	}
	
	// add entry
	m_pSVDLibData->rowind[m_iCurrentSVDLibDataN] = i;
	m_pSVDLibData->value[m_iCurrentSVDLibDataN] = val;
	m_iCurrSVDLibDataRow = i;
	m_iCurrentSVDLibDataN++;	
}

// NOTE: only for symmetric matrices
void LinAlgMatrixSparseRealLibSVDC::EigsSymmetric(LinAlgVectorReal & eigenValues, 
											  LinAlgMatrixReal & eigenVectors, 
											  int dimensions)
{
	assert(m_pSVDLibData!=NULL);
	assert(m_iCurrentSVDLibDataN==m_pSVDLibData->vals);	// found all
	
	std::cout<<"Running Lanczos iteration. Note: the matrix must be symmetric (not checked)"<<std::endl;
	//	SVDRec svdRec = svdLAS2A(m_pSVDLibData, dimensions);
	//	double end[2] = {-1.0e-30, 1.0e-30};
	//	double kappa = 1e-30;
	double end[2] = {0, 0};
	double kappa = 0;
	SVDRec svdRec = svdLAS2(m_pSVDLibData, dimensions, dimensions * 1000, end, kappa);
	
	int d = svdRec->d;	// dimensionality (rank)
	//	DMat Ut = svdRec->Ut;    // Transpose of left singular vectors. (d by m)  The vectors are the rows of Ut.
	double *S = svdRec->S;  // Array of singular values. (length d) 
	DMat Vt = svdRec->Vt;    // Transpose of right singular vectors. (d by n) The vectors are the rows of Vt. 
	
	if (eigenValues.Rows()!=dimensions)
	{
		std::cout<<"[ERROR] Cannot fit eigenvalues: "<<eigenValues.Rows();
		std::cout<<" vs "<<dimensions<<std::endl;
		assert(false);
	}
	if (d!=dimensions)
	{
		std::cout<<"[WARNING] Unexpected rank: "<<d<<" (expecting: "<<dimensions<<")"<<std::endl;
		std::cout<<"\tEigenvectors / eigenvalues not filled"<<std::endl;
		return;
	}
	
	for (int i=0; i<d; i++)
		eigenValues(i) = S[i];
	
	int r = Vt->rows;
	int c = Vt->cols;	
	assert(r==d);
	assert(eigenVectors.Cols()==r && eigenVectors.Rows()==c);
	for (int i=0; i<d; i++)
		for (int j=0; j<c; j++)
			eigenVectors(j,i) = Vt->value[i][j];	
}

//////////    LinAlgMatrixSparseRealGeneral - matlab /////////////
LinAlgMatrixSparseRealGeneral::LinAlgMatrixSparseRealGeneral(int m, int n)
: LinAlgMatrixSparseReal(m, n, SPARSE_MTX_GENERAL)
{
	m_ArrayOfColsMapRows = new std::map<int, double>[n]; 
}

LinAlgMatrixSparseRealGeneral::LinAlgMatrixSparseRealGeneral
								(const LinAlgMatrixSparseRealGeneral & M)
: LinAlgMatrixSparseReal(M.m, M.n, SPARSE_MTX_GENERAL)
{
	m_ArrayOfColsMapRows = new std::map<int, double>[n]; 
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = M.m_ArrayOfColsMapRows[i].begin();
			 iter != M.m_ArrayOfColsMapRows[i].end(); iter++)
			m_ArrayOfColsMapRows[i][iter->first] = iter->second;
	}
}

double LinAlgMatrixSparseRealGeneral::GetVal(int i, int j)
{
	std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[j].find(i);
	if (iter==m_ArrayOfColsMapRows[j].end())
		return 0;
	else
		return iter->second;
}

void LinAlgMatrixSparseRealGeneral::AddVal(int i, int j, double val)
{
	std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[j].find(i);
	if (iter == m_ArrayOfColsMapRows[j].end())
		(m_ArrayOfColsMapRows[j])[i] = val;
	else
	{
		iter->second += val;
		if (iter->second==0)	// can use epsilon?
			m_ArrayOfColsMapRows[j].erase(iter);
	}

}

LinAlgMatrixSparseRealGeneral::~LinAlgMatrixSparseRealGeneral()
{
	delete [] m_ArrayOfColsMapRows;
}

void LinAlgMatrixSparseRealGeneral::Symmetrize()
{
	for (int i=0; i<m; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
		{
			double & v1 = iter->second;
			double & v2 = m_ArrayOfColsMapRows[iter->first][i];
			
			double v = .5 * (v1+v2);
			v1 = v;
			v2 = v;
		}
	}
}

void LinAlgMatrixSparseRealGeneral::RowNormalize(LinAlgVectorReal * saveRows, double norm)
{
	bool deleteRowNorms = saveRows==NULL;
	double * rowNorms = NULL;
	
	if (saveRows!=NULL)
		rowNorms = saveRows->m_aRealData;
	else
		rowNorms = new double[m];
	
	for (int i=0; i<m; i++)
		rowNorms[i] = 0;
	
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
		{
			assert(iter->first < m);
			rowNorms[iter->first] += iter->second;
		}
	}		
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
			iter->second = norm * iter->second / rowNorms[iter->first];
	}		
	
	if (deleteRowNorms)
		delete [] rowNorms;
}

void LinAlgMatrixSparseRealGeneral::ColNormalize(LinAlgVectorReal * saveCols, double norm)
{
	bool deleteColNorms = saveCols==NULL;
	double * colNorms = NULL;
	
	if (saveCols!=NULL)
		colNorms = saveCols->m_aRealData;
	else
		colNorms = new double[m];
	
	for (int i=0; i<m; i++)
		colNorms[i] = 0;
	
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
		{
			assert(iter->first < m);
			colNorms[i] += iter->second;
		}
	}		
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
			iter->second = norm * iter->second / colNorms[i];
	}		
	
	if (deleteColNorms)
		delete [] colNorms;
}

void LinAlgMatrixSparseRealGeneral::Eigs(LinAlgVectorReal & eigenValues, 
										 LinAlgMatrixReal & eigenVectors, 
										 int dimensions, 
										 const char * uniquePrefixChar,
										 double * timeInSec)
{
	std::cout<<"RUNNING MATLAB: sparse eigs"<<std::endl;
	//const char * matlabExe="/Applications/MATLAB_R2010a.app/bin/matlab -maci";	// mac OSX
	const char * matlabExe = "matlab";
	
	// MATLAB VERSION
	std::ostringstream uniquePrefixStream;
	if (uniquePrefixChar==NULL)
		uniquePrefixStream<<rand()<<time(NULL)<<(long int)this;
	else
		uniquePrefixStream<<uniquePrefixChar;
	std::string uniquePrefix = uniquePrefixStream.str();
	std::string mtxName = (uniquePrefix+".matrix.bin");
	std::string evecsName = (uniquePrefix+".eigenvectors.bin");
	std::string evalsName = (uniquePrefix+".eigenvalues.bin");
	std::string timerName = (uniquePrefix+".time.bin");
	assert(!LinAlg::FileExists(mtxName.c_str()));
	WriteMatrixBin(mtxName.c_str());
	
	assert(!LinAlg::FileExists(evecsName.c_str()));
	assert(!LinAlg::FileExists(evalsName.c_str()));
	assert(LinAlg::FileExists(mtxName.c_str()));
	
	std::string cmd = "findAndWriteEigenvectorsSparse('";
	char tempChar[10];
	sprintf(tempChar, "%d", dimensions);
	cmd = cmd + uniquePrefix + "', " + std::string(tempChar);
	sprintf(tempChar, "%d", m);
	cmd = cmd + ", " +  std::string(tempChar) + ", ";
	sprintf(tempChar, "%d", n);
	cmd = cmd + std::string(tempChar) + ");";
	bool bReady = LinAlg::RunMatlab(cmd.c_str());
	if (!bReady)
	{
		LinAlg::WaitForFile(evecsName.c_str(), 5, 8000);
		LinAlg::WaitForFile(evalsName.c_str(), 5, 8000);
	}
	bool readEigenval = eigenValues.ReadVectorBin(evalsName.c_str());
	bool readEigenvec = eigenVectors.ReadMatrixBin(evecsName.c_str());
	
	assert(readEigenval && readEigenvec);
	if (timeInSec!=NULL)
	{
		FILE * timeFile = fopen(timerName.c_str(), "rb");
		if (timeFile!=NULL)
		{
			fread(timeInSec, sizeof(double), 1, timeFile);
			fclose(timeFile);
		}
		else
		{
			std::cout<<"[WARNING] Could not read timing: ";
			std::cout<<timerName.c_str()<<std::endl;
			*timeInSec = 0;
		}
	}
	if(uniquePrefixChar==NULL)	// delete temp matrix
	{
		int ok = remove(mtxName.c_str());
		assert(ok==0);
		ok = remove(evecsName.c_str());
		assert(ok==0);
		ok = remove(evalsName.c_str());
		assert(ok==0);
		ok = remove(timerName.c_str());
		assert(ok==0);
	}
	else	// move temp data
	{
		std::string mtxName2	= mtxName+".copy";
		std::string evecsName2	= evecsName+".copy";
		std::string evalsName2	= evalsName+".copy";
		std::string timerName2	= timerName+".copy";
		
		int ok = rename(mtxName.c_str(), mtxName2.c_str());
		assert(ok==0);
		ok = rename(evecsName.c_str(), evecsName2.c_str());
		assert(ok==0);
		ok = rename(evalsName.c_str(), evalsName2.c_str());
		assert(ok==0);
		ok = rename(timerName.c_str(), timerName2.c_str());
		assert(ok==0);
	}

}

void LinAlgMatrixSparseRealGeneral::WriteMatrixBin(const char * filename)
{
	FILE * file = fopen(filename, "wb");
	assert(file!=NULL);
	WriteMatrixBin(file);
	fclose(file);
}

void LinAlgMatrixSparseRealGeneral::ReadMatrixBin(const char * filename)
{
	FILE * file = fopen(filename, "rb");
	assert(file!=NULL);
	ReadMatrixBin(file);
	fclose(file);	
}

void LinAlgMatrixSparseRealGeneral::WriteMatrixBin(FILE * file)
{
	int nz=0;
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
			nz++;
	}
	int * rows = new int[nz];
	int * cols = new int[nz];
	double * vals = new double[nz];	
	nz = 0;
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
		{
			rows[nz] = iter->first;
			cols[nz] = i;
			vals[nz] = iter->second;			
			nz++;
		}
	}
	assert(sizeof(int)==4);
	fwrite(&m, 1, sizeof(int), file);
	fwrite(&n, 1, sizeof(int), file);
	fwrite(&nz, 1, sizeof(int), file);
	fwrite(rows, nz, sizeof(int), file);
	fwrite(cols, nz, sizeof(int), file);
	fwrite(vals, nz, sizeof(double), file);	
	delete [] rows;
	delete [] cols;
	delete [] vals;
}

void LinAlgMatrixSparseRealGeneral::ReadMatrixBin(FILE * file)
{
	int mread, nread;
	fread(&mread, 1, sizeof(int), file);
	fread(&nread, 1, sizeof(int), file);
	
	int nz;
	fread(&nz, 1, sizeof(int), file);
	assert(mread==m && nread==n);
	int * rows = new int[nz];
	int * cols = new int[nz];
	double * vals = new double[nz];	
	fread(rows, nz, sizeof(int), file);
	fread(cols, nz, sizeof(int), file);
	fread(vals, nz, sizeof(double), file);	
	
	for (int i=0; i<nz; i++)
		AddVal(rows[i], cols[i], vals[i]);
	
	delete [] rows;
	delete [] cols;
	delete [] vals;
}


//////////    LinAlgMatrixSparseRealARPACK - arpack /////////////
#ifdef WITH_ARPACK
#endif

