/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __CorrsCollection_H
#define __CorrsCollection_H

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "VkFunctions.h"
#include "VKString.h"

#include "Model3D.h"
#include "ModelDatabase.h"
#include "GeomShape.h"
#include "ShapePoint.h"
#include "ShapeCorrs.h"

#include "CollectionPairwise.h"


/** 
 * CorrsCollection
 */
class CorrsCollection : public ShapeCorrs
{
	public:
		static void FillShapes(ModelDatabase * db, const VKString & geomName,
							   std::vector<GeomShape*> & shapes);
	
		CorrsCollection(const VKString & corrsName,
						const std::vector<GeomShape*> & shapes, DBCorrs * corrs=NULL);
		virtual ~CorrsCollection();
		virtual VKString GetCorrsClassName() const;
	
		virtual int NumShapes();
		virtual GeomShape * GetShape(int localID);
		virtual int LocalShapeID(int modelID) const;
		virtual GeomShape * GetShapeByModelID(int modelID);
	
		virtual void Domain(std::vector<const GeomShape *> & domainShapes ) const;
		virtual void Range(std::vector<const GeomShape *> & rangeShapes ) const;

	// can map points if cached
		virtual int GetCorrs(const ShapePoint & pnt,
							 std::vector<ShapePoint> & corrs, 
							 const VKString & corrValName="none",
							 std::vector<double> *corrVals = NULL);
	
		// point enumeraction (assign id to every point on input shapes)
		virtual void EnumeratePoints();
		virtual bool Enumerated() const;
		virtual int NumGlobPnts() const;
		virtual int PointToId(const ShapePoint & pnt) const;
		virtual ShapePoint IdToPoint(int globalPntID) const;

		// NOTE: only 'default' corr value is allowed (todo other).
		void AutoFillCache(const VKString & corrValueName = "default");

		// also use InitCache to clear cache
		void InitCache(const VKString & corrValueName = "default");
	
		void SetCachedCorrs(int fromPntID, const std::vector<ShapePoint> & corrs,
							const std::vector<double> & vals);
		void GetCachedCorrs(int fromPntID, 
							std::vector<ShapePoint> & corrs,
							std::vector<double> * vals);
		bool IsCacheFilled(int fromPntID) const;
		int GetCachedCorrs(const ShapePoint & pnt,
						   std::vector<ShapePoint> & corrs, 
						   const VKString & corrValName,
						   std::vector<double> *corrVals);

	// read / write: only if all points are cached
		virtual void WriteCollectionHeader(FILE * file, 
										   const VKString & className="Native");

		static CorrsCollection * ReadCollection(ModelDatabase * db, 
												const VKString & filename, DBCorrs * corrs);

		static FILE * ReadCollectionHeader(const VKString & filename, 
										   int & nGeoms,
										   VKStringList & modelNames, 
										   VKStringList & geomNames,
										   VKString & className,
										   VKString & corrName);
	
		virtual void WriteShapeCorrs(FILE * file);
		virtual void ReadShapeCorrs(FILE * file);
		void WriteShapeCorrsAsCollection(FILE * file);
		void ReadShapeCorrsAsCollection(FILE * file);

	
	protected:
		std::vector<GeomShape*> m_Shapes;

	private: // point indexing
		int m_iMaxModelID;
		int * m_ModelIDToOffset;
		int * m_aModelIDToShapeID;
	
		int m_iMaxGlobPntID;
		int * m_GlobPntIDToShapeID;
		int * m_GlobPntIDToLocPntID;

	private: // corrs caching
		double ** m_aCachedCorrVals;
		int ** m_aCachedCorrs;
		int * m_aNumCachedCorrs;
	

};

#endif

