/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __CorrsCreator_H
#define __CorrsCreator_H

#include <stdio.h>
#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "GeomShape.h"
#include "ShapeCorrs.h"
#include "CollectionPairwise.h"

#include "CorrsPair.h"
#include "CorrsPairMap.h"
#include "CorrsCollection.h"
#include "CorrsCollectionMap.h"

/**
 * ShapeCorrs can optionally implement this params interface
 */
class CorrParams
{
	public:
		CorrParams(const VKString & corrsParamsName, 
				   const VKString & className, const VKString & paramClassName);
	
		VKString m_sParamsName;
		VKString m_sClassName;
		VKString m_sParamClassName;
};

/** 
 * Note: since corrs classes might be implemented in other libs, 
 *	this class is an abstract class for initializing one of the 
 *  'core' correspondence classes (pairwise or general collection). 
 */
class CorrsCreator
{
	public:
		static void InitCoreCorrsReaders();
	
		static CorrsCreator * GetCreator(const VKString & className);
		static void AddCreator(CorrsCreator * creator);	
		static void AddCreateParams(CorrParams * algParams);

		CorrsCreator(ShapeCorrs::CoreCorrsType tp, 
					 const VKString & className);
		virtual ~CorrsCreator();
	
		CorrParams * GetAlgParams(const VKString & algName);
		const VKString & ClassName() const;
		ShapeCorrs::CoreCorrsType ClassType() const;
	
		virtual CorrsPair * ReadCorrsPair(const VKString & corrName, 
										  GeomShape * shapeFrom, 
										  GeomShape * shapeTo, 
										  FILE * file, DBCorrs * corrs);
	
		virtual CorrsPairMap * ReadCorrsPairMap(const VKString & corrName, 
												GeomShape * shapeFrom, 
												GeomShape * shapeTo, 
												FILE * file, DBCorrs * corrs);
	
		virtual CorrsCollection * ReadCorrsCollection(const VKString & corrsName,
													  const std::vector<GeomShape*> & shapes,
													  FILE * file, DBCorrs * corrs);
	
		virtual CorrsCollectionMap * ReadCorrsCollectionMap(const VKString & corrsName,
															const std::vector<GeomShape*> & shapes,
															FILE * file, DBCorrs * corrs);

		virtual void CreateCorrsPair(const VKString & corrName, GeomShape * shapeFrom, 
									 GeomShape * shapeTo, CorrParams * params, 
									 std::vector<CorrsPair *> & saveCorrs, DBCorrs * corrs);
		
		virtual void CreateCorrsPairMap(const VKString & corrName, GeomShape * shapeFrom, 
										GeomShape * shapeTo, CorrParams * params, 
										std::vector<CorrsPairMap *> & saveCorrs, DBCorrs * corrs);
		
		virtual void CreateCorrsCollection(const VKString & corrsName, const std::vector<GeomShape*> & shapes,
										   CorrParams * params, 
										   std::vector<CorrsCollection *> & saveCorrs, DBCorrs * corrs);
		
		virtual void CreateCorrsCollectionMap(const VKString & corrsName,
											  const std::vector<GeomShape*> & shapes,
											  CorrParams * params, 
											  std::vector<CorrsCollectionMap *> & saveCorrs,
											  DBCorrs * corrs);

	
	protected:
		VKString m_sClassName;
		ShapeCorrs::CoreCorrsType m_Type;
		
		std::vector<CorrParams * > m_Algs;
		std::map<VKString, int> m_AlgNameToID;
		
		
		static std::map<VKString, CorrsCreator*> s_AllCreators;
};


class CoreCorrsReader : public CorrsCreator
{
	public:
		static void InitCoreCorrsReaders();
		CoreCorrsReader(ShapeCorrs::CoreCorrsType tp, 
						const VKString & className);
		
		virtual CorrsPair * ReadCorrsPair(const VKString & corrName, 
										  GeomShape * shapeFrom, GeomShape * shapeTo, 
										  FILE * file, DBCorrs * corrs);
		
		virtual CorrsPairMap * ReadCorrsPairMap(const VKString & corrName, 
												GeomShape * shapeFrom, 
												GeomShape * shapeTo, 
												FILE * file, DBCorrs * corrs);
		
		virtual CorrsCollection * ReadCorrsCollection(const VKString & corrsName,
													  const std::vector<GeomShape*> & shapes,
													  FILE * file, DBCorrs * corrs);
		
		virtual CorrsCollectionMap * ReadCorrsCollectionMap(const VKString & corrsName,
															const std::vector<GeomShape*> & shapes,
															FILE * file, DBCorrs * corrs);

};



#endif

