/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "DBCorrs.h"

const unsigned int DBCorrs::s_iCorrsCollHeader = 0x64c12afb;
const unsigned int DBCorrs::s_iCorrsPairHeader = 0xfd0493fd;
const unsigned int DBCorrs::s_iCorrsHeader	   = 0x93f0d719;

DBCorrs::DBCorrs(ModelDatabase * db)
{
	m_pDB = db;
}

DBCorrs::~DBCorrs()
{
	
}

void DBCorrs::AddPairColl(const VKString & pairwiseCorrName,
						  CollectionPairwise * pairwise)
{
	int pairID = (int)m_PairwiseCollections.size();
	m_PairwiseNameToID[pairwiseCorrName] = pairID;
	m_PairwiseCollections.push_back(pairwise);
}

void DBCorrs::AddCorrs(const VKString & corrName, ShapeCorrs * corrs)
{
	int corrID = (int)m_Corrs.size();
	m_CorrsNameToID[corrName] = corrID;
	m_Corrs.push_back(corrs);
}

int DBCorrs::NumPairColls() const
{
	return (int)m_PairwiseCollections.size();
}

int DBCorrs::NumCorrs() const
{
	return (int)m_Corrs.size();
}

CollectionPairwise * DBCorrs::GetPairColl(int id)
{
	assert(id >=0 && id < (int)m_PairwiseCollections.size());
	return m_PairwiseCollections[id];
}

ShapeCorrs * DBCorrs::GetCorrs(int id)
{
	assert(id >=0 && id < (int)m_Corrs.size());
	return m_Corrs[id];
}

CollectionPairwise * DBCorrs::GetPairColl(const VKString & collName)
{
	std::map<VKString, int>::iterator iter = m_PairwiseNameToID.find(collName);
	assert(iter!=m_PairwiseNameToID.end());
	return GetPairColl(iter->second);
}

ShapeCorrs * DBCorrs::GetCorrs(const VKString & corrName)
{
	std::map<VKString, int>::iterator iter = m_CorrsNameToID.find(corrName);
	if (iter==m_CorrsNameToID.end())
		std::cout<<"[ERROR] Cannot find corrs: "<<corrName.c_str()<<std::endl;
	assert(iter!=m_CorrsNameToID.end());
	return GetCorrs(iter->second);
}

DBCorrs::EntryType DBCorrs::CorrsType(const VKString & corrName)
{
	std::map<VKString, int>::iterator iter = m_PairwiseNameToID.find(corrName);
	if (iter!=m_PairwiseNameToID.end())
		return DBCORRS_PairwiseColl;

	assert(m_CorrsNameToID.find(corrName)!=m_CorrsNameToID.end());
	return DBCORRS_Corrs;
}

CollectionPairwise * DBCorrs::ReadPairwiseCollection(const VKString & corrName,
													 ModelDatabase * db, 
													 const VKString & filename,
													 DBCorrs * dbCorrs)
{
	CollectionPairwise::ReadConcatenated(corrName, db, filename, dbCorrs);
}


