/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "DistEuclidean.h"
#include "ShapeMesh.h"
#include "ShapePointCloud.h"

DistEuclidean::DistEuclidean(GeomShape * shape, const VKString & metricName, double norm)
: ShapeDistance(shape, metricName)
{
	m_fNorm = norm;
}

DistEuclidean::~DistEuclidean()
{
	
}

DistEuclidean * DistEuclidean::SafeCastToEuclidean()
{
	return this;
}

double DistEuclidean::NormalizedDistance(double distance) const
{
	if (m_fNorm<0)
	{
		ShapeMesh * mesh = m_pShape->SafeCastToMesh();
		ShapePointCloud * pnts = m_pShape->SafeCastToPointCloud();
		if (mesh!=NULL)
			((double&)m_fNorm) = mesh->AverageRadius(mesh->Centroid());
		else if (pnts!=NULL)
			((double&)m_fNorm) = pnts->AverageRadius(pnts->Centroid());
		else
			assert(false);
	}
	
	return distance / m_fNorm;
}

double DistEuclidean::NormalizedDistance(const ShapePoint & p1, 
											 const ShapePoint & p2) const
{
	return NormalizedDistance(R3Distance(p1.GetPosition(), p2.GetPosition()));
}

//void DistEuclidean::NearestNeighborhood(const ShapePoint & p, 
//										double distanceThreshold,
//										std::vector<ShapePoint> * neighborhood,
//										std::vector<double> * normDists) const
//{
//	assert(false);
//}
//
//ShapePoint DistEuclidean::NearestNeighbor(const ShapePoint & p, 
//										  double* normDists ) const
//{
//	assert(false);
//}

