/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __ModelDatabase_H
#define __ModelDatabase_H

#include "VKString.h"
#include "VKStringList.h"

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

class Model3D;

class ModelDatabase
{
	public:
		enum InputDirFormat
		{
			DBINPUTDIR_Unknown,
			DBINPUTDIR_MyFormat,
			DBINPUTDIR_MeshFiles,
			DBINPUTDIR_ColladaDirs
		};
	
		enum WriteDBFlags
		{
			WRITE_MODEL3D_FILES			= 1,
			WRITE_MODEL3D_GEOMS			= 2,
			WRITE_ARG_SINGLE_GEOM		= 4,
		};
		
		enum ReadingFlags
		{
			READ_MESH_AS_SOUP		= 1,
			READ_MESH_AS_MANIFOLD	= 2,
			READ_PNTS				= 4,
			READ_SHAPEVALS			= 8,
			READ_SHAPEDISTS			= 16,
		};
		
		// NOTE: this only lists simple things that can be created
		//		DO NOT use it for general pipeline / app specific stuff
		enum CreatingFlags
		{
			CREATE_PNTS_ALLVTX		= 1,
			CREATE_PNTS_ALLFACES	= 2,
			CREATE_DIST_EUCLIDEAN	= 4,
			CREATE_DIST_GEODESIC	= 8
		};	
	
		static const int EVERYTHING = 0xFFFFFFFF;
		static const int NOTHING = 0;
	
		static VKStringList GetAllModelsInDir(const VKString & inputDir, 
											  const InputDirFormat dirFormat);
		static void GetModelFiles(const VKString & inputDir, const InputDirFormat dirFormat,
								  const VKStringList & listModelsToLoad,
								  VKStringList & modelNames, VKStringList & modelFiles);

	
		static ModelDatabase * CreateSingleModelDB(Model3D * m, const VKString & dbDir = "none",
												   const VKString & dbName = "none");
		static ModelDatabase * CreatePairModelsDB(Model3D * m1, Model3D * m2, const VKString & dbDir = "none",
												  const VKString & dbName = "none");
		static ModelDatabase * CreateDBFromDir(const VKString & inputDir, InputDirFormat dirFormat,
											   VKStringList modelsToLoad = VKStringList(), int readingFlags = EVERYTHING,
											   int createFlags = NOTHING, const VKString & dbDir = "none",
											   const VKString & dbName = "none");
		
		// databse: full DB, quick: single model, pairwise map
		ModelDatabase(const VKString & dbDir = "none", 
					  const VKString & dbName = "none");
	
		virtual ~ModelDatabase();
	
		virtual void WriteDB(int writingFlags = EVERYTHING, 
							 const VKString & writeArg = "none");
		virtual void ReadDB();
			
		virtual void ReadEverythingInADir(const VKString & inputDir, InputDirFormat dirFormat, VKStringList & modelsToLoad,
										  int readingFlags = EVERYTHING, int createFlags = NOTHING);
		
		virtual int NumModels();
		virtual Model3D * GetModel(const VKString & modelName);
		virtual Model3D * GetModel(int modelID);
//		virtual Model3D * GetModelByUniqeuID(int uniqueID);
		
		virtual void AddModel(Model3D * model);
		virtual void RemoveModel(int modelID);
	
//		virtual void WriteDB(const VKString & filename);
//		static void ReadDB(const VKString & filename,
//						   int & nModels, int & maximalUniqueID,
//						   VKStringList * modelNames = NULL,
//						   std::vector<int> * modelIDs = NULL);
	
		virtual VKString GetDBDescriptorFilename();
		virtual VKString GetDBDirname();
		virtual VKString GetModelDirname(const VKString & modelName);
		virtual VKString GetModelFilename(const VKString & modelName);
		virtual VKString GetMetaDataFilename(const VKString & modelName, 
											 const VKString & metaData);
		static VKString GetModelDirname(const VKString & dbDir, const VKString & modelName);
		static VKString GetModelFilename(const VKString & dbDir, const VKString & modelName);
		virtual VKString GetModelGeometryFilename(const VKString & modelName, const VKString & geometryFile);
		static VKString GetModelGeometryDir();
	
		static VKString ShapeValueFilename(const VKString & modelDir,
										   const VKString & geomName, const VKString & geomValueName);
		
	protected:
		VKString m_sDBName;
		VKString m_sDBDir;
	
		// models
		std::map<VKString, int> m_MapModelNameToModelID;
		std::map<int, int> m_MapUniqueIDToModelID;
		std::vector<Model3D *> m_Models;
		
		int m_iMaximalUniqueID;
};


#endif




