/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __ShapeLabels_H
#define __ShapeLabels_H

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "ShapeValue.h"

class SceneGraphLabel
{
	public:
		SceneGraphLabel();
		VKString m_sLabel;
		int m_iID;
	
		// scene graph node with a label
		std::vector<SceneGraphLabel *> m_apChildren;
		SceneGraphLabel * m_pParent;

		// geometry data (i.e. symmetry of parts)
		std::vector<int> m_aiGeometryID;
		std::vector<int> m_iFirstFaceInGeometry;
		std::vector<int> m_iNFacesInGeometry;
		
		// cached faces (this is inverse of per-face info)
		bool m_bCachedIDs;
		std::vector<int> m_CachedIDs;
};

class SceneGraphGeometry
{
	public:
		VKString m_sGeometryName;
		int m_iID;
};

/**
 * 
 */
class ShapeLabels : public ShapeValue
{
	public:
		static ShapeLabels * CopyLabels(ShapeLabels * labels, GeomShape * target);
		static ShapeLabels * SplitLabelsIntoCC(ShapeLabels * labels);
	
		ShapeLabels(GeomShape * shape, const VKString & valueName);
		virtual ~ShapeLabels();	
		virtual ShapeLabels * SafeCastToLabels();
		virtual const ShapeLabels * SafeCastToLabels() const;
	
		virtual VKString GetClassName() const;

		virtual void SetFaceSegments(const std::vector<int> & segIDs);
		virtual void SetPntsSegments(const std::vector<int> & segIDs);
		virtual void SetAllLabels(std::vector<SceneGraphLabel*> & labels);
		virtual void SetAllGeometries(std::vector<SceneGraphGeometry*> & geoms);
		virtual VKStringList GetAllLabels(int segID) const;
		virtual VKStringList GetAllLabels() const;
	
		virtual void WriteShapeValues(FILE * file);
		virtual void ReadShapeValues(FILE * file);

		virtual void ReadFromLabFile(const VKString & filename);
		virtual void ReadFromSegFile(const VKString & filename, const VKString & segNames = "none");
		virtual void ReadFromPlyFile(const VKString & filename, ShapeLabels * optionalGTLabels = NULL);

	// geometries
		virtual int NGeometries() const;
		virtual const VKString & GeometryLabel(int geomID) const;	
	
	// scene graph structure
		virtual void InitFlatGraph(int nChildren);
		virtual int NNodes() const;
		virtual int RootNode() const;
		virtual void SetNodeLabel(int nodeID, const VKString & nodeLabel);
		virtual const VKString & NodeLabel(int nodeID) const;
		virtual int NodeNChildren(int nodeID) const;
		virtual int NodeChild(int nodeID, int cID) const;
		virtual int NodeParent(int nodeID) const;
		virtual int NodeDepth(int nodeID) const;
		virtual void NodeAllFaces(int nodeID, std::vector<int> & faceIDs) const;
		virtual const std::vector<int> & NodeFacesAtNode(int nodeID) const;
		virtual void NodeAllPoints(int nodeID, std::vector<int> & pointIDs) const;
		virtual const std::vector<int> & NodePointsAtNode(int nodeID) const;
		virtual int MaxDepth() const;
	
	// modify scene graph
		virtual int AddNode(int parentID, const VKString & nodeLabel);
		
		virtual int NumNodeGeometries(int nodeID) const;
		virtual int GetGeometryID(int nodeID, int geomID, int * f0=NULL, int *nf=NULL) const;

		virtual Model3D * ExtractNode(int nodeID);
	
	protected:
		virtual void ClearCache();
		virtual void CacheAllLeafIDs();
		std::vector<SceneGraphLabel*> m_AllNodes;
		std::vector<SceneGraphGeometry*> m_AllGeometries;
};

#endif

