/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __ShapeManifoldMesh_H
#define __ShapeManifoldMesh_H

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "ShapeMesh.h"

// for 2d meshes represented by r3mesh
struct FlatSearchNode
{
	FlatSearchNode(int id, const R2Point & pnt);
	FlatSearchNode();
	int m_iID;
	R2Point m_Pnt;
};

R2Point global_FlatNodePosition(FlatSearchNode * currNode, void * data);



/**
 * Manifold mesh: basically wraps around r3mesh
 */
class ShapeManifoldMesh : public ShapeMesh
{
	public:
		ShapeManifoldMesh(GeomShapeTypes shapeTp = GeomShape::SHAPE_TYPE_MANIFOLD, const VKString & geomFile = "none",
						  const VKString & geomName = "none", Model3D * model = NULL);
		virtual ~ShapeManifoldMesh();
		virtual ShapeManifoldMesh * SafeCastToManifold();
	
		virtual void WriteGeometryFile(const VKString & geomFile);
		virtual void ReadGeometryFile(const VKString & geomFile);
		
		virtual int AddVertex(double x, double y, double z, double nx = 0, double ny = 0, double nz = 0);
		virtual int AddFace(int v1, int v2, int v3);
		virtual void RemoveFace(int f);
		virtual void RemoveVertex(int v);
		
	
		// vertices
		virtual int NVertices() const;
		virtual R3Point VertexPosition(int vID) const;
		virtual R3Vector VertexNormal(int vID) const;
		virtual int FaceOnVertex(int vID) const;
		virtual int NAdjacentFaces(int vID) const;
		virtual int FaceOnVertex(int vID, int nFace) const;
		virtual int VertexValence(int vID) const;
		virtual std::vector<int> AdjacentVertices(int vID) const;
	
		//edges
		virtual int NEdges() const;
	
		// faces
		virtual int NFaces() const;
		virtual R3Point FaceCentroid(int fID) const;
		virtual int VertexOnFace(int fID, int vID) const;
		virtual R3Point FacePoint(int fID, double b1, double b2) const;
		virtual R3Point FaceBarycentric(int fID, const R3Point & p) const;
		virtual R3Vector FaceNormal(int fID) const;
		virtual double FaceArea(int fID) const;
		virtual void AdjacentFaces(int fID, std::vector<int> & fids) const;
	
		virtual void RearrangeAxes(char newOrder[3], int newSigns[3]);
	
		// R3Mesh
		void SetMesh(R3Mesh * mesh);
		const R3Mesh * GetMesh() const;
		R3Mesh * GetMesh();
		
		// if flattened: todo, currently uses 3d datastructures
		ShapePoint NearestPointOnFlatMesh(const R2Point & p) const;
		
		
	protected:
		virtual R3MeshSearchTree * CreateSearchTree(R3Mesh ** mesh,
													std::vector<int> & triangleMap,
													std::vector<int> & vertexMap) const ;

		R3Mesh * m_pMesh;
	
};

#endif

