/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __ShapeValue_H
#define __ShapeValue_H

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "GeomShape.h"
class ShapeMaterials;
class ShapeLabels;

class ShapeValue
{
	public:
		enum ShapeValueCacheType
		{
			SVTYPE_Unknown,
			SVTYPE_PerFace,			// single value for the whole face
			SVTYPE_PerIndexedFaces,	// for each face store indices 
			SVTYPE_PerVertex,		// defined at every vertex of a mesh
			SVTYPE_PerPointSet		// defined at every point in a set
		};
		
		static VKString m_sDefaultMaterials;
		static VKString m_sDefaultSceneGraphLabels;
		static VKString s_sDefaultConnectedComponents;
	
		ShapeValue(GeomShape * shape, const VKString & valueName, int nDims = 1);
		ShapeValue(GeomShape * shape);
		virtual void InitPtrs();
		virtual ~ShapeValue();
	
		virtual ShapeMaterials * SafeCastToMaterials();
		virtual ShapeLabels * SafeCastToLabels();
		virtual const ShapeMaterials * SafeCastToMaterials() const;
		virtual const ShapeLabels * SafeCastToLabels() const;
		virtual VKString GetClassName() const;
		virtual VKString GetShapeValueName() const;
		virtual GeomShape * GetShape();
		virtual const GeomShape * GetShape() const;
	
		virtual int NDims() const;
		virtual int NCachedValues() const;
		virtual void Values(const ShapePoint & pnt, double * storeVals) const;
		virtual double Value(const ShapePoint & pnt, int dim) const;
	
	// points must be indexable within the same point set
		virtual void InitCache(ShapeValueCacheType cacheType, int cacheSize = -1);
		virtual bool IsValueSet(const ShapePoint & pnt);
		virtual void SetPntsValues(const double * values, const int nValues);
		virtual void SetFaceValues(const double * values, const int nValues);
		virtual void SetIndexedFaceValues(const int * faceIndices, const int nFaceIndices,
										  const double * values, const int nValues);

		virtual void GetCachedFaceIndexedValues(int fID, int vID, double * saveVals);
		virtual double GetCached(const ShapePoint & pnt, int dim) const;
	
	// write (by default just saves cached values)
		virtual void WriteShapeValues(const VKString & filename);
		virtual void ReadShapeValues(const VKString & filename);
		virtual void WriteShapeValues(FILE * file);
		virtual void ReadShapeValues(FILE * file);
		virtual void WriteCachedShapeValues(FILE * file);
		virtual void ReadCachedShapeValues(FILE * file);
	
	// update structure...
		virtual void RemoveFace(int f);
		virtual void RemoveFaces(int f1, int f2);
		virtual void RemoveVertex(int v);
	
	protected:
		void ShapeValueRemoveFace(int f); // non-virtual works at this level only
		void ShapeValueRemoveFaces(int f1, int f2); // non-virtual works at this level only
		void ShapeValueRemoveVertex(int v); // non-virtual works at this level only
	
		GeomShape * m_pShape;
		ShapeMesh * m_pMesh;
		ShapePointCloud * m_pPnts;
		
		VKString m_sValueName;
		int m_iNDims;

	// cache type
		ShapeValueCacheType m_CacheType;
	
	// values cached
		int m_iNValues;
		double * m_aValues;
	
	// optional: only for SVTYPE_DefinedPerFaceVertices
		int m_iNPerFaceIDs;
		int * m_aPerFaceIDs;
		
};

#endif

