#include "VKString.h"
#include "Sortable.h"
#include "FileManagement.h"
#include "BasicNumerical.h"

#include "STemplate.h"
#include "TMatch.h"
#include "TMatchEnergy.h"
#include "TMatchAlg.h"
#include "PartOfEdge.h"
#include "ShapeLocalPCA.h"

#ifndef __ConsistentPartsAlg_H
#define __ConsistentPartsAlg_H

struct PointsCluster
{
	PointsCluster();
	PointsCluster(int modelID, int clusterID, int localNodeID, const VKString & label, const ShapePointCloud * pnts);
	virtual ~PointsCluster();
	virtual void CreateAlignedKdTree(const std::vector<int> & subset, const ShapePointCloud * pnts, TMatch * match);

	int m_iModelID;
	int m_iClusterID;
	int m_iLocalNodeID;
	
	ShapePointCloud * m_pAlignedSubsetPoints;
	ShapeLocalPCA * m_pLocalOrient;
	
	R3Affine m_Pnts2Tmplt;
	const ShapePointCloud * m_pAllPnts;
	std::vector<int> m_aSubset;
	PartBox * m_pGeomFit;
	VKString m_sLabel;
};

class ConsistentPartsAlg
{
	public:
		enum ConsistentPartsMethods
		{
				// initial clusters: from points
			CPM_InitWhole						= 1<<0,
			CPM_InitFromGT 						= 1<<1,
			CPM_InitWithMRF_TakeFurthest		= 1<<2,
			CPM_InitWithMRF_AssignToNonNull		= 1<<3,
			CPM_InitGreedy_AssignToNonNull		= 1<<4,
			    // Clustering methods
			CPM_ClusterTakeAll 					= 1<<5,
			CPM_ClusterSpectralAgglomerative	= 1<<6,
				// final 'average' part for a cluster
			CPM_AveClusterTakeAny				= 1<<7,
			CPM_AveClusterTakeAll				= 1<<8,
		};
		ConsistentPartsAlg(STemplate * st, const std::vector<TMatchEnergy*> & matches, int methodFlags, bool verb);
		virtual ~ConsistentPartsAlg();
		virtual void ExtractConsistentParts(int nodeID, int partID, int nParts, std::vector<PointsCluster> & saveParts);
		
	protected:
		virtual void InitializeClusters(int nodeID, int partID, int nParts);
		virtual void MRFPointClustering(int matchID, int localNodeID, std::vector<int> & allpnts);
		virtual double ClusterDistance(const TMatchEnergy * energy, const PointsCluster & c1, const PointsCluster & c2);
		virtual double OneWayDistance(const TMatchEnergy * energy, const PointsCluster & c1, const PointsCluster & c2);
		virtual void FindConsistentParts(const LinAlgMatrixReal & pairwiseSim,
										 const LinAlgVectorReal & evals, const LinAlgMatrixReal & evecs,
										 std::vector<int> & partIDs);
		virtual void AverageParts(std::vector<int> & partIDs, std::vector<PointsCluster> & finalClusters);
		virtual void ComputeEigs(LinAlgVectorReal ** evals, LinAlgMatrixReal ** evecs,
								 LinAlgMatrixReal ** simmtx, bool row_normalize);
		virtual void EraseClusters(std::vector<int> & clusterIDs);
	
		// input data (lists null nodes)
		STemplate * m_pST;
		const std::vector<TMatchEnergy*> & m_Matches;
		int m_iMethodFlags;
		bool m_bVerbose;
	
		// clustering
		std::vector<std::vector<PointsCluster> > m_ClusteredPoints;
		std::vector<int> m_MatchIDsForParts;
		std::vector<int> m_ClusterIDsForParts;
};



#endif



