/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __PartTransforms_H
#define __PartTransforms_H

#include "Model3D.h"
#include "ShapeMesh.h"
#include "ShapeLabels.h"
#include "ShapeDescriptor.h"

#include "ProbDistribution.h"
#include "PartBox.h"
#include "STemplate.h"

class TemplateNodeDeform;


/**
 * Random variable representing part transformations
 * 
 * For N parts provides aligning transformations (could be a distribution)
 * for each partID=[1, .., N]
 */
class PartTransforms
{
	public:	
		enum TransformTypes
		{
			PARTXFORM_ScaleIsotropic		= 1,
			PARTXFORM_ScaleAnisotropic		= 2,
			PARTXFORM_Translation			= 4,
			PARTXFORM_RotationEuler			= 8,
			PARTXFORM_RotationQuaternion	= 16
		};

		PartTransforms(int types = PARTXFORM_Translation);
		virtual ~PartTransforms();
	
		virtual PartTransforms * CreateCopy() const;
	
		// st1 - parent, st2 - child
		virtual int NSamples() const;
		virtual int XFormFlags() const;
		virtual void UpdateTransformStats(int partID, const PartBox & parent, const PartBox & child);
		virtual void UpdateTransformStats(const TemplateNodeDeform & deform, bool include);
		virtual void Reset();
		virtual void Normalize(const R3Point & centroid, double aveRadius);
		
		virtual ContProbDistribution * GetPartTransform(int partID);
		virtual const ContProbDistribution * GetPartTransform(int partID) const;
	
		//virtual ContProbDistribution * TransformsForPart(int partID, int & xformflags);
		//virtual const ContProbDistribution * TransformsForPart(int partID, int & xformflags) const;
		
	// read / write
		virtual void WriteTransforms(FILE * file) const;
		static PartTransforms * ReadTransforms(FILE * file, STemplate::STemplateVersions version);

	//protected:
		int m_iXFormFlags;

		std::vector<ContProbDistribution *> m_apXForms;
		int m_iNTotalFields;		// just a cached value for the storage (used internally only)
		double * m_aTransform;		// just a temporary storage (used internally only)
	
		static std::vector<TransformTypes> m_Types;
		static std::vector<int> m_TypeFields;
	
	// some static helper functions map transform
	public:
		static R3Affine CreateAffineTransform(const double * transform, int typeFlags);
	
	protected:
		static void ComputeAligningTransform(const R3Point & c1, const R3Vector & axes1X, const R3Vector & axes1Y,
											 const R3Vector & axes1Z, const R3Point & c2, const R3Vector & axes2X,
											 const R3Vector & axes2Y, const R3Vector & axes2Z, 
											 int typeFlags, double * transform);		
		static int Offset(TransformTypes type, int typeFlags, int * nFields = NULL, int * nTotalFields = NULL);
};

#endif

