/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __PartsCollection_H
#define __PartsCollection_H

#include "Model3D.h"
#include "ShapeMesh.h"
#include "ShapeLabels.h"
#include "ShapePointCloud.h"
#include "ShapeLocalPCA.h"

/**
 * Describes (deformable?) geometry of a single part
 */
class PartGeometry
{
	public:
		PartGeometry();
		PartGeometry(const ShapePointCloud * pnts, const std::vector<int> & subset, const VKString & partLabel,
					 const R3Point & centroid, const R3Triad & orientation, const R3Vector & scale,
					 const std::vector<int> & subsetMesh);
		virtual ~PartGeometry();
	
		virtual PartGeometry * CreateCopy() const;
		virtual const ShapeMesh * GetNormalizedMesh() const;
		virtual const ShapePointCloud * GetNormalizedPart() const;
		virtual R3Point PointPosition(int pntid) const;
//		virtual R3Triad GetPoint(int pntID, R3Point & tPos, double tPntEigs[3]) const;
		virtual int NearestPointID(const R3Point & pointInPartSpace) const;
		virtual double Area() const;
	
		virtual void WritePart(FILE * file) const;
		static PartGeometry * ReadPart(FILE * file);

		virtual const R3Point & Centroid() const;
		virtual const R3Triad & Orientation() const;
		virtual const R3Vector & Scale() const;
	
	protected:
		R3Point m_Centroid;
		R3Triad m_Orientation;
		R3Vector m_Scale;

		VKString m_sPartLabel;
		int m_iNPoints;
		ShapeMesh * m_pNormalizedPartMesh;
		ShapePointCloud * m_pNormalizedPart;	
};

/**
 * Stores all parts, accessed in a PartBox
 */
class PartsCollection
{
	public:
		PartsCollection();
		virtual ~PartsCollection();
		virtual int NParts() const;
		virtual int AddPart(PartGeometry * part);
		virtual const PartGeometry * GetPart(int partID) const;
		virtual void WriteParts(FILE * file) const;
		static PartsCollection * ReadParts(FILE * file);
	
		
	protected:
		std::vector<PartGeometry*> m_AllParts;
};


#endif


