/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __VisButtonsMenu_H
#define __VisButtonsMenu_H

#include "OpenGLHeaders.h"
#include "Model3D.h"
#include "GeomShape.h"
#include "ShapePointCloud.h"
#include "ShapeManifoldMesh.h"

class VisWindow;
class VisButton;
class VisButtonsGroup;
class VisButtonsMenu;


class VisButton
{
	public:
		friend class VisButtonsGroup;
		friend class VisButtonsMenu;
	
		VisButton(const VKString & buttonName, 
				  int initState = 0, int nStates = 2);
		virtual int GetState() const;
	
		virtual void ChangeState(int change);
	
	protected:
		VKString m_sButtonName;
		int m_iButtonState;
		int m_iMaximalStateValue;
		
		double m_fMinX, m_fMinY, m_fWidth, m_fHeight;
};

class VisButtonsGroup
{
	public:
		friend class VisButtonsMenu;
		
		enum ButtonsGroupType
		{
			BUTGROUP_RadioButtons,
			BUTGROUP_CheckboxButtons
		};

		VisButtonsGroup(const VKString & groupname, 
						ButtonsGroupType groupType = BUTGROUP_CheckboxButtons);
		virtual ~VisButtonsGroup();
		virtual void AddOnOffButtons(const VKStringList & buttonNames);
		virtual void AddButton(const VKString & buttonName, int nStates);
	
		virtual void FreeButtons();
	
	protected:
		int m_iFirstButtonPage;
		ButtonsGroupType m_GroupType;
		VKString m_sGroupName;
		std::vector<VisButton*> m_Buttons;
	
		double m_fHeaderMinX, m_fHeaderMinY, m_fHeaderWidth, m_fHeaderHeight;
		double m_fFooterMinX, m_fFooterMinY, m_fFooterWidth, m_fFooterHeight;
	
};

class VisButtonsMenu
{
	public:
		VisButtonsMenu();
		virtual ~VisButtonsMenu();
		virtual void AddButtonGroup(VisButtonsGroup * groups);
		
		virtual void Draw(VisWindow * window);
		virtual bool ClickInWindow(double x, double y, 
								   int & groupID, int & buttonID, int & dir);
		
		virtual void SetRenderingParams(double textSpacing, 
										double textXOffset, 
										double textYOffset, 
										int nMaxButtons);
		virtual void FillRenderablePos();
		
		virtual void ClearGroup(int groupID);
	
		virtual void ChangePage(int groupID, int advanceID);
		virtual void ChangeState(int groupID, int buttonID, int change);
		virtual const VKString & ButtonName(int groupID, int buttonID);
		virtual int State(int groupID, int buttonID);
		virtual VisButtonsGroup * GetGroup(int groupID);
		virtual VisButton * GetButton(int groupID, int buttonID);
	
	protected:
		double m_fTextSpacing;
		double m_fTextXOffset;
		double m_fTextYOffset;
	
		int m_iRenderTextForGroup;
		int m_iMaxButtons;
		std::vector<VisButtonsGroup * > m_ButtonGroups;
};

#endif

