/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "VisShapeTemplate.h"
#include "PartOfEdge.h"
#include "TMatch.h"
#include "TMatchAlg.h"
#include "ConsistentPartsAlg.h"
#include "STemplateNode.h"

VisShapeTemplate::VisShapeTemplate(ModelDatabase * db)
: DBApp(db)
{
	m_bPntsIBufferInitialized = false;
	m_CurrSorting = SRT_ByEnergy;
	m_sGTLabelsName = ShapeValue::m_sDefaultSceneGraphLabels;
	m_iSelDeformedX = -1;
	m_bDeformedPntsBufferInitialized = false;
	m_bSelectingDeformedPnts = false;
	m_pCoAlignedParts = NULL;
	m_iSelectedCluster = 0;
	m_iSelectedClustering = -1;
	m_iTestingSomething = 0;
	m_TemplateCamera = R3identity_affine;
	m_iSelectedPart = -1;
	
	m_pCreatingTemplate = NULL;
	m_iInteractionFlags = INTERACT_VisualizeOnly;
	
	m_iVisFlags = VISTMPLT_DrawTemplateAsBoxes | VISTMPLT_DrawTemplateAsAxes;
	m_iVisFlags |= VISTMPLT_TemplateSegPnts | VISTMPLT_MatchedSegPnts;
	//m_iVisFlags |= VISTMPLT_SeedTemplate | VISTMPLT_DeformedOverMatched;
	m_iVisFlags |= VISTMPLT_DeformedOverMatched;
	
	m_iCurrMatched = -1;
	m_iNIterations = 0;
	
	m_bIBufferInitialized = false;
	m_bOnlyRigidAlignment = false;
}

void VisShapeTemplate::SetInteractionCreateTemplate(const VKString & topologyFile, const VKString & outTemplate)
{
	assert(false);
}

void VisShapeTemplate::SetTemplate(STemplate * st)
{
	m_pCreatingTemplate = st;
}

void VisShapeTemplate::SetInteractionCreateTemplate(const VKString & outFile, STemplate * stemplate)
{
	m_iVisFlags |= VISTMPLT_SeedTemplate;
	m_iInteractionFlags  = INTERACT_CreateTempalte;
	m_sOutTemplateFile = outFile;
	m_pCreatingTemplate = stemplate;
	if (m_pCreatingTemplate==NULL)
	{
		m_pCreatingTemplate = new STemplate(NULL);
		PartBox root(NULL);
		root.InitBox(R3Point(0,0,0), R3Triad(R3Vector(1,0,0), R3Vector(0,1,0), R3Vector(0,0,1)), R3Vector(1, 1, 1), 0);
		m_pCreatingTemplate->InitAndAddNewNode(-1, "root", root, 0);
		
		PartBox nullPart(NULL);
		nullPart.InitNull();
		m_pCreatingTemplate->InitAndAddNewNode(0, "NullPart", nullPart, 0);
	}
	m_apSeedConfigs.clear();
	for (int i=0; i<m_pCreatingTemplate->NSeedConfigs(); i++)
	{
		TMatch * tmatch = m_pCreatingTemplate->CreateSeedConfig(i, NULL, std::vector<int>(), R3identity_affine);
		m_apSeedConfigs.push_back(tmatch);
		if (i==0)
			m_sConfigNames.push_back("all_parts");
		else
			m_sConfigNames.push_back(m_pCreatingTemplate->ConfigName(i));
	}
	m_iCurrMatched = 0;
	
}

void VisShapeTemplate::SetReferenceModel(const VKString & mName, const VKString & gName)
{
	m_sRefModelName = mName;
	m_sRefGeometry = gName;
}

void VisShapeTemplate::WriteTemplate()
{
	assert(m_pCreatingTemplate!=NULL);
	m_pCreatingTemplate->WriteTemplate(m_sOutTemplateFile);
}

void VisShapeTemplate::SetGTEval(const VKString & gtLabels, const VKString & equivFilename)
{
	m_sGTLabelsName = gtLabels;
	
	if (equivFilename!="none")
	{
		std::ifstream textStream(equivFilename.c_str());
		assert(textStream.is_open());
		VKString tempStr;
		while (!textStream.eof())
		{
			tempStr.readLine(textStream);
			int currID = m_BasicLabels.count();
			VKStringList allLabels = tempStr.split(" ");
			if (allLabels.count()==0)
				continue;
			for (int i=0; i<allLabels.count(); i++)
				m_EquivalentLabels[allLabels[i]] = currID;
			m_BasicLabels.push_back(allLabels[0]);
		}
		textStream.close();
	}
}

double VisShapeTemplate::GetConfidence(int matchID)
{
	if (m_pCoAlignedParts==NULL)
		return 0;
	
	
	
//	m_pCoAlignedParts->GetClusterForPoint(clusteringID, i, p, &cid, &pid, valid);
//	double cv = valid ? coAligned->GetConfidenceValue(clusteringID, cid, pid, i, p, valid) : -1;
//	if (!valid)
//	{
//		clusteringID = coAligned->GetClusteringFromMatchesID(match->SeedID());
//		const TemplateNodeDeform & deform = match->Deformation(lnid);
//		int method = CoAlignedParts::CCF_Dist2All;
//		int clusterID, partInClusterID;
//		bool valid;
//		coAligned->GetClusterWithPart(clusteringID, i, lnid, &clusterID, &partInClusterID, valid);
//		assert(valid);
//		coAligned->ComputeConfidences(clusteringID, clusterID, method);
//		cv = coAligned->GetConfidenceValue(clusteringID, cid, pid, i, p, valid);
//		assert(valid);
//	} 
	 
	
	return 0;
}

double VisShapeTemplate::GetAccuracy(int matchID)
{
	assert(matchID>=0 && matchID < (int)m_apMatchedConfigs.size());
	TMatch * match = m_apMatchedConfigs[matchID]->GetMatch();
	const ShapePointCloud * pnts = match->GetMatchedGeomAsPnts();
	if (pnts==NULL || pnts->GetShapeVal(m_sGTLabelsName)==NULL)
	{
		std::cout<<"[WARNING] No ground truth "<<m_sGTLabelsName.c_str();
		std::cout<<" in match: "<<match->ModelName().c_str()<<std::endl;;
		return 0;
	}
	assert(pnts!=NULL && pnts->GetShapeVal(m_sGTLabelsName)!=NULL);
	const ShapeLabels * gtLabels = pnts->GetShapeVal(m_sGTLabelsName)->SafeCastToLabels();
	if (gtLabels==NULL)	// no ground truth labels...
		return 0;
	int nCorrect = 0;
	int nTotal = 0;
	
	for (int p = 0; p < match->NShapePrimitives(); p++)
	{
		int d = match->LocalNodeID(p);
		if (m_iSelectedPart != -1 && d!=m_iSelectedPart )
			continue;
			
		int nodeID = match->Deformation(d).NodeID();
		VKString labelName = match->GetSTemplate()->GetNode(nodeID)->NodeName();
		std::map<VKString, int>::iterator iter1 = m_EquivalentLabels.find(labelName);
		
		VKString gtLabelName = gtLabels->NodeLabel((int)gtLabels->Value(pnts->GetPoint(p), 0));
		std::map<VKString, int>::iterator iter2 = m_EquivalentLabels.find(gtLabelName);
		
		bool sameLabel = (gtLabelName == labelName);
		if (iter1 != m_EquivalentLabels.end() && iter2 != m_EquivalentLabels.end())
			sameLabel = (iter1->second == iter2->second);
					
		if (sameLabel)
			nCorrect++;
		nTotal++ ;
	}
	
	if (nTotal==0)
		return 0;
	else
		return (double)nCorrect / (double)nTotal;
}

void VisShapeTemplate::AddNewPart(const STemplateNode * cNode)
{
	//m_iCurrMatched = 0;
	assert(m_pCreatingTemplate!=NULL);
	VKString partName;
	if (cNode==NULL)
	{
		static char buffer[1024];
		std::cout<<"What is part's name? Write here: "<<std::flush;
		scanf("%s", buffer);
		partName = VKString(buffer);
		std::cout<<"   Adding |"<<partName.c_str()<<"|"<<std::endl;
	}
	else
		partName = cNode->NodeName();
	
	int configID = 0;
	PartBox box0(NULL);
	if (cNode==NULL || cNode->GetPart(configID)->IsType(PartBox::BOXTYPE_NullBox))
		box0.InitBox(R3Point(0,0,0), R3Triad(R3Vector(1,0,0), R3Vector(0,1,0), R3Vector(0,0,1)), R3Vector(.1, .1, .1), 0);
	else
	{
		box0.InitBox(R3Point(0,0,0), cNode->GetPart(configID)->UndeformedOrientation(), cNode->GetMeanScale(0), 0);
	}
	
	m_pCreatingTemplate->InitAndAddNewNode(configID, partName, box0, 0);
	TMatch * tmatch = m_pCreatingTemplate->CreateSeedConfig(configID, NULL, std::vector<int>(), R3identity_affine);
	if (m_apSeedConfigs.size()>0)
	{
		delete m_apSeedConfigs[0];
		m_apSeedConfigs[0] = tmatch;
	}
	else
		m_apSeedConfigs.push_back(tmatch);
	//m_iSelectedPart++;
}


void VisShapeTemplate::AddMatchedModel(TMatch * seed, TMatchEnergy * matched)
{
	//m_iCurrMatched = (int)m_apMatchedConfigs.size();
	m_iCurrMatched = 0;
	m_apMatchedConfigs.push_back(matched);
	m_apSeedConfigs.push_back(seed);
}

bool VisShapeTemplate::KeyPressed(DatabaseVisWindow * window, const char key, int mouseX, int mouseY)
{
	if (m_bDisable)
		return false;
	
	if (m_iInteractionFlags & INTERACT_CreateTempalte)
	{
		bool fSelected = (m_iCurrMatched>=0 && m_iCurrMatched < (int)m_apSeedConfigs.size());
		fSelected = fSelected && m_iSelectedPart>=0 && m_iSelectedPart < m_apSeedConfigs[m_iCurrMatched]->NTemplateNodes();
		if (fSelected)
			fSelected = !m_apSeedConfigs[m_iCurrMatched]->NodeEllipsoid(m_iSelectedPart).IsType(PartBox::BOXTYPE_NullBox);
		int config = key-'0';
		// setting up configuration for a particular shape style
		if (config>=0 && config <= 9)
		{
			m_iCurrMatched = vkMin(config, (int)m_apSeedConfigs.size());
			if (m_iCurrMatched == (int)m_apSeedConfigs.size())
			{
				m_sConfigNames.push_back("unknown");
				m_apSeedConfigs.push_back(NULL);
				CreateNewConfig(m_iCurrMatched);
			}
			else
			{
				std::cout<<"Picked config: "<<m_sConfigNames[m_iCurrMatched].c_str();
				std::cout<<" ("<<m_iCurrMatched<<" / "<<m_sConfigNames.count()<<")"<<std::endl;
			}
			return true;
		}
		switch (key)
		{
			case '`':
				std::cout<<"Template editing mode"<<std::endl;
				std::cout<<"    +       Add new part"<<std::endl;
				std::cout<<"    > <     Select parts (also right-click)"<<std::endl;
				std::cout<<"    / ?     Select part's face / axis (also right-click)"<<std::endl;
				std::cout<<"    - =     Rescale"<<std::endl;
				std::cout<<"    ] [     Rotate part (also right-mouse + ctrl)"<<std::endl;
				std::cout<<"    ; '     Translate part (also right-mouse + shift)"<<std::endl;
				std::cout<<"    w     	Write template"<<std::endl;
				std::cout<<"    o     	Make part optional"<<std::endl;
				std::cout<<"    d     	Delete part"<<std::endl;
				std::cout<<"    c     	Clone part"<<std::endl;
				std::cout<<"    r     	Rename part"<<std::endl;
				std::cout<<"    p     	Print info"<<std::endl;
				std::cout<<"    R     	Reset current configuration"<<std::endl;
				std::cout<<"    D     	Remove current configuration"<<std::endl;
				std::cout<<"    a     	Draw all template info"<<std::endl;
				std::cout<<"    m       Show reference model (if applicable)"<<std::endl;
				return true;
			case '+': // add part
				if (m_iCurrMatched==0)
					AddNewPart();
				else
					std::cout<<"[WARNING] Cannot add part (choose config 0)"<<std::endl;
				return true;
			case 'A':
				Toggle(VISTMPLT_DrawDistribution);
				return true;
			case 'x':
			case 'X':
			case 'y':
			case 'Y':
			case 'z':
			case 'Z':
			{
				R3Point offset(0,0,0);
				if (key=='x')
					offset[0] -= .01;
				else if (key=='X')
					offset[0] += .01;
				else if (key=='y')
					offset[1] -= .01;
				else if (key=='Y')
					offset[1] += .01;
				else if (key=='z')
					offset[2] -= .01;
				else if (key=='Z')
					offset[2] += .01;
				m_pCreatingTemplate->Normalize(offset, 1.);
				if (IsFlagSet(VISTMPLT_DrawReferenceModel))
					UpdateTemplateScale();
				return true;
			}
			case 's':
			case 'S':	// scale template
				m_pCreatingTemplate->Normalize(R3Point(0,0,0), key=='s' ? .95 : 1.05);
				if (IsFlagSet(VISTMPLT_DrawReferenceModel))
					UpdateTemplateScale();
				return true;
			case 'm':
			{
				Toggle(VISTMPLT_DrawReferenceModel);
				if (IsFlagSet(VISTMPLT_DrawReferenceModel))
					UpdateTemplateScale();
				return true;
			}
			case 'p':	// print info
			{
				std::cout<<"Current configuration ("<<m_apSeedConfigs[m_iCurrMatched]->SeedID()<<")";
				std::cout<<": "<<m_sConfigNames[m_iCurrMatched].c_str();
				std::cout<<"  "<<m_iCurrMatched<<" / "<<m_sConfigNames.count()<<std::endl;
				std::vector<int> sortedIDs;
				for (int i=0; i<m_apSeedConfigs[m_iCurrMatched]->NTemplateNodes(); i++)
				{
					const TemplateNodeDeform & deform = m_apSeedConfigs[m_iCurrMatched]->Deformation(i);
					if (deform.IsIncluded())
						sortedIDs.insert(sortedIDs.begin(), i);
					else
						sortedIDs.push_back(i);
				}
				for (int i=0; i<(int)sortedIDs.size(); i++)
				{
					const TemplateNodeDeform & deform = m_apSeedConfigs[m_iCurrMatched]->Deformation(sortedIDs[i]);
					const STemplateNode * node = m_pCreatingTemplate->GetNode(deform.NodeID());
					const DiscreteProbDistribution * prob = NULL;

					prob = node->PartOfParent()->GetProbNParts(m_apSeedConfigs[m_iCurrMatched]->SeedID());
					
					VKString nameStr = node->NodeName();
					VKString partsStr = VKString::number(deform.PartID())+" / "+VKString::number(deform.NParts());
					VKString probStr = VKString::number((int)(prob->Prob(deform.NParts()) * 100))+"%";
					VKString exclStr = deform.IsIncluded() ? "Included" : "Excluded";
					while (nameStr.length()<15)
						nameStr += " ";
					while (partsStr.length()<10)
						partsStr += " ";
					while (probStr.length()<15)
						probStr += " ";
					
					std::cout<<"    "<<nameStr.c_str()<<" "<<partsStr.c_str()<<" "<<probStr.c_str();
					std::cout<<"    "<<exclStr.c_str()<<std::endl;
				}
				return true;
			}
			case 'P':	// print node-specific (detailed) info
			{
				if (m_iCurrMatched<0 || m_iSelectedPart<0
					|| m_iSelectedPart >= m_apSeedConfigs[m_iCurrMatched]->NTemplateNodes())
				{
					std::cout<<"[WARNING] Config ("<<m_iCurrMatched<<") or part ("<<m_iSelectedPart<<") not picked"<<std::endl;
					return true;
				}
				const TemplateNodeDeform & deform = m_apSeedConfigs[m_iCurrMatched]->Deformation(m_iSelectedPart);
				const STemplateNode * node = m_pCreatingTemplate->GetNode(deform.NodeID());
				std::cout<<"Node: "<<node->NodeName().c_str()<<std::endl;
				int configID = m_apSeedConfigs[m_iCurrMatched]->SeedID();
				int ancestor;
				R3Vector s = node->GetMeanScale(configID);
				R3Vector sv = node->ScaleVariance(configID, ancestor);
				int flags, ndims;
				const PartOfEdge * edge = node->PartOfParent();
				const double * xf = edge->GetMeanTransform(configID, deform.NParts(), deform.PartID(), flags, ndims);
				R3Vector xfv = edge->PosVariance(configID, deform.NParts(), deform.PartID(), ancestor);
				assert(flags = PartTransforms::PARTXFORM_Translation);
				
				std::cout<<"  Scale: M=["<<s.X()<<" "<<s.Y()<<" "<<s.Z()<<"] "<<std::endl;;
				std::cout<<"         V=["<<sv.X()<<" "<<sv.Y()<<" "<<sv.Z()<<"] "<<std::endl;
				std::cout<<"  XForm: M=["<<xf[0]<<" "<<xf[1]<<" "<<xf[2]<<"] "<<std::endl;
				std::cout<<"         V=["<<xfv.X()<<" "<<xfv.Y()<<" "<<xfv.Z()<<"] "<<std::endl;
				
//				const ContProbDistribution * scaleDistr = node->ScaleDistribution();
//				const PartTransforms * transforms = node->PartOfParent()->GetPartTransforms(deform.NParts());
//				int xformstyle;
//				const ContProbDistribution * transDistr = transforms->TransformsForPart(deform.PartID(), xformstyle);
//				double minvar[3];
//				
//				std::cout<<" - Transformation:\n"<<transDistr->Print().c_str();
//				//TMatchEnergy(m_apSeedConfigs[m_iCurrMatched]).GetXFromVarianceThresh(deform.NodeID(), transDistr, minvar, 1);
//				std::cout<<"       MinVar = "<<minvar[0]<<std::endl;
//				std::cout<<" - Scale: "<<scaleDistr->Print().c_str();
//				//TMatchEnergy(m_apSeedConfigs[m_iCurrMatched]).GetScaleVarianceThresh(deform.NodeID(), scaleDistr, minvar, 3);
//				std::cout<<"       MinVar = [ "<<minvar[0]<<" "<<minvar[1]<<" "<<minvar[2]<<" ]"<<std::endl;
//				std::cout<<" - Geometry: Box"<<std::endl;
				
				return true;
			}
			case 'E':
				if (m_pCreatingTemplate!=NULL)
				{
					m_pCreatingTemplate->RemoveExcludedDeformsFromSeeds();
					UpdateTemplateScale();
				}
				return true;
			case 'R': // reset configuration
			{
				if (m_iCurrMatched>0)
					CreateNewConfig(m_iCurrMatched);
				return true;
			}
//			case 'N':
//			{
//				//m_pCreatingTemplate->Normalize();
//				return true;
//			}
			case 'd':	// delete part from configuration
				if (fSelected)
				{
					TMatch * match = m_apSeedConfigs[m_iCurrMatched];
					TemplateNodeDeform & deform = (TemplateNodeDeform&)match->Deformation(m_iSelectedPart);
					STemplateNode * node = (STemplateNode*)m_pCreatingTemplate->GetNode(deform.NodeID());
					
					if (m_iCurrMatched>0)
					{
						std::cout<<"Excluding "<<node->NodeName().c_str()<<std::endl;
						deform.SetIncluded(false);
						
						if (node->PartOfParent()->GetProbNParts(match->SeedID())->Prob(deform.NParts()) > 0)
						{
							node->PartOfParent()->UpdateNParts(match->SeedID(), 0, true);
							while (node->PartOfParent()->GetProbNParts(match->SeedID())->Prob(deform.NParts())>0)
								node->PartOfParent()->UpdateNParts(match->SeedID(), deform.NParts(), false);
							std::cout<<"Config "<<m_sConfigNames[match->SeedID()].c_str()<<", ";
							std::cout<<"setting probability of "<<node->NodeName().c_str()<<" ";
							std::cout<<"to "<<node->PartOfParent()->GetProbNParts(match->SeedID())->Prob(deform.NParts())<<std::endl;
						}
					}
					else	// deleting parts forever
					{
						std::cout<<"Removing: "<<deform.NodeID()<<std::endl;
						m_pCreatingTemplate->RemoveNode(deform.NodeID());
						UpdateTemplateScale();
					}
				}
				return true;
			case 'D': // remove configuration
				assert(m_pCreatingTemplate!=NULL);
				if (m_iCurrMatched>0)
				{
					m_pCreatingTemplate->RemoveSeed(m_iCurrMatched);
					m_iCurrMatched = 0;
					UpdateTemplateScale();
				}
				return true;
			case 'c':	// clone part
				if (fSelected)
				{
					TMatch * match = m_apSeedConfigs[m_iCurrMatched];
					TemplateNodeDeform & deform = (TemplateNodeDeform&)match->Deformation(m_iSelectedPart);
					const STemplateNode * cNode = m_pCreatingTemplate->GetNode(deform.NodeID());
					std::cout<<"Cloning: "<<cNode->NodeName().c_str()<<std::endl;
					AddNewPart(cNode);
				}
				return true;
			case 'r':	// rename
				if (fSelected)
				{
					TMatch * match = m_apSeedConfigs[m_iCurrMatched];
					TemplateNodeDeform & deform = (TemplateNodeDeform&)match->Deformation(m_iSelectedPart);
					STemplateNode * cNode = (STemplateNode*)m_pCreatingTemplate->GetNode(deform.NodeID());
					static char buffer[1024];
					std::cout<<"What is part's name? Write here: "<<std::flush;
					scanf("%s", buffer);
					VKString partName(buffer);
					std::cout<<"   Updating Name |"<<cNode->NodeName().c_str()<<"| to |"<<partName.c_str()<<"|"<<std::endl;
					cNode->SetName(partName);
				}
				return true;
			case 'o':
				if (fSelected)
				{
					TMatch * match = m_apSeedConfigs[m_iCurrMatched];
					TemplateNodeDeform & deform = (TemplateNodeDeform&)match->Deformation(m_iSelectedPart);
					STemplateNode * node = (STemplateNode*)m_pCreatingTemplate->GetNode(deform.NodeID());
					double prob = node->PartOfParent()->GetProbNParts(match->SeedID())->Prob(1);
					if (prob == 1.)
						node->PartOfParent()->UpdateNParts(match->SeedID(), 0, true);
					else if (vkAbs(prob-0.5)<10E-10)
						node->PartOfParent()->UpdateNParts(match->SeedID(), 0, false);
					std::cout<<"Config "<<m_sConfigNames[match->SeedID()].c_str()<<", ";
					std::cout<<"setting probability of "<<node->NodeName().c_str()<<" ";
					std::cout<<"to "<<node->PartOfParent()->GetProbNParts(match->SeedID())->Prob(1)<<std::endl;
				}
				return true;
			case '>':
			case '<': // select part
				if (m_iCurrMatched>=0 && m_iCurrMatched < (int)m_apSeedConfigs.size())
				{
					m_iSelectedFace = 0;
					m_iSelectedPart+= (key=='>' ? 1 : -1);
					if (m_iSelectedPart >= m_apSeedConfigs[m_iCurrMatched]->NTemplateNodes())
						m_iSelectedPart = 0;
					else if (m_iSelectedPart < 0)
						m_iSelectedPart = m_apSeedConfigs[m_iCurrMatched]->NTemplateNodes()-1;
					int nodeID = m_apSeedConfigs[m_iCurrMatched]->Deformation(m_iSelectedPart).NodeID();
					VKString nname = m_pCreatingTemplate->GetNode(nodeID)->NodeName();
					std::cout<<"Selected "<<nname.c_str()<<" (id="<<m_iSelectedPart<<")"<<std::endl;
				}
				return true;
			case '/':
			case '?': // select face
				if (m_iCurrMatched>=0 && m_iCurrMatched < (int)m_apSeedConfigs.size())
				{
					m_iSelectedFace += (key=='/' ? 2 : -2);
					if (m_iSelectedFace>5)
						m_iSelectedFace = 0;
					else if (m_iSelectedFace < 0)
						m_iSelectedFace = 4;
				}
				return true;
			case '-':
			case '=': // rescale part
				RescaleTemplateBox((key=='-' ? -.05 : .05));
				return true;
			case '[':
			case ']': // rotate part
				RotateTemplateBox((key=='['?-1 : 1) * 3.14/40.);
				return true;
			case ';':
			case '\'': // translate part
				TranslateTemplateBox((key==';' ? -.05 : .05));
				return true;
			case 'w':
				if (m_sOutTemplateFile=="none")
					return true;
				std::cout<<"Writing: "<<m_sOutTemplateFile.c_str()<<std::endl;
				m_pCreatingTemplate->WriteTemplate(m_sOutTemplateFile);
				return true;
			case 'i':
				m_iInteractionFlags = INTERACT_VisualizeOnly;
				return true;
		}
	}
	else
	{
		switch(key)
		{
			case '`':
				std::cout<<"VisShapeTemplate keys:"<<std::endl;
				std::cout<<"Visualization:"<<std::endl;
				std::cout<<"    0 - "<<"Clear all visualization flags"<<std::endl;
				std::cout<<"    1 - "<<"template mesh rendering (mesh, labeled, nothing)"<<std::endl;
				std::cout<<"    2 - "<<"matched mesh rendering (mesh, labeled, nothing)"<<std::endl;
				std::cout<<"    3 - "<<"template"<<std::endl;
				std::cout<<"    4 - "<<"seed deformation of a template (over matched points)"<<std::endl;
				std::cout<<"    5 - "<<"deformed template (over matched points)"<<std::endl;
				std::cout<<"  X 6 - "<<"deformed points (disabled)"<<std::endl;
				std::cout<<"    7 - "<<"Template sampling"<<std::endl;
				std::cout<<"    8 - "<<"Nearest surface match"<<std::endl;
				std::cout<<"    9 - "<<"Mis-labeled only"<<std::endl;
				std::cout<<"Other"<<std::endl;
				std::cout<<"    S - "<<"Selection mode"<<std::endl;
				std::cout<<"    X - "<<"only rigid transform for alignment"<<std::endl;
				std::cout<<"    i - "<<"switch to interactive mode"<<std::endl;
				std::cout<<"    ~ - "<<"sort models (confidence, accuracy, ...)"<<std::endl;
				return false;
			case '0':
				m_iVisFlags = VISTMPLT_DrawTemplateAsBoxes|VISTMPLT_DrawTemplateAsAxes;
				return true;
//			case '1':
//				if (IsFlagSet(VISTMPLT_TemplateSegPnts))
//				{
//					SetRenderingFlag(VISTMPLT_TemplatePnts, true);
//					SetRenderingFlag(VISTMPLT_TemplateSegPnts, false);
//				}
//				else if (IsFlagSet(VISTMPLT_TemplatePnts))
//				{
//					SetRenderingFlag(VISTMPLT_TemplatePnts, false);
//					SetRenderingFlag(VISTMPLT_TemplateSegPnts, false);				
//				}
//				else
//					SetRenderingFlag(VISTMPLT_TemplateSegPnts, true);				
//				return true;
			case '1':
				if (!IsFlagSet(VISTMPLT_DeformedPoints) && !IsFlagSet(VISTMPLT_AllDeformedPoints))
				{
					std::cout<<"Rendering deformation of current match"<<std::endl;
					SetRenderingFlag(VISTMPLT_DeformedPoints, true);
				}
				else if (IsFlagSet(VISTMPLT_DeformedPoints))
				{
					std::cout<<"Rendering deformation ALL matches"<<std::endl;
					SetRenderingFlag(VISTMPLT_DeformedPoints, false);
					SetRenderingFlag(VISTMPLT_AllDeformedPoints, true);
				}
				else
				{
					std::cout<<"Not rendering any deformations"<<std::endl;
					SetRenderingFlag(VISTMPLT_DeformedPoints, false);
					SetRenderingFlag(VISTMPLT_AllDeformedPoints, false);
				}
				return true;
			case '2':
				if (IsFlagSet(VISTMPLT_MatchedSegPnts))
				{
					SetRenderingFlag(VISTMPLT_MatchedPnts, true);
					SetRenderingFlag(VISTMPLT_MatchedSegPnts, false);
				}
				else if (IsFlagSet(VISTMPLT_MatchedPnts))
				{
					SetRenderingFlag(VISTMPLT_MatchedPnts, false);
					SetRenderingFlag(VISTMPLT_MatchedSegPnts, false);				
				}
				else
				{
					SetRenderingFlag(VISTMPLT_MatchedPnts, false);
					SetRenderingFlag(VISTMPLT_MatchedSegPnts, true);
				}
				return true;
			case '3':
				Toggle(VISTMPLT_SeedTemplate);
				return true;
			case '4':
				Toggle(VISTMPLT_SeedOverMatched);
				return true;
			case '5':
				Toggle(VISTMPLT_DeformedOverMatched);
				return true;
			case '6':
				Toggle(VISTMPLT_MatchedGTSegMesh);
				Toggle(VISTMPLT_TemplateGTSegMesh);
				return true;
			case '7':
				Toggle(VISTMPLT_DrawTemplateSampling);
				return true;
			case '8':
				if (!IsFlagSet(VISTMPLT_DebugNearestSurface1) && !IsFlagSet(VISTMPLT_DebugNearestSurface2))
				{
					SetRenderingFlag(VISTMPLT_DebugNearestSurface1, true);
					SetRenderingFlag(VISTMPLT_DebugNearestSurface2, false);
				}
				else if (!IsFlagSet(VISTMPLT_DebugNearestSurface2))
				{
					SetRenderingFlag(VISTMPLT_DebugNearestSurface1, false);
					SetRenderingFlag(VISTMPLT_DebugNearestSurface2, true);
				}
				else
				{
					SetRenderingFlag(VISTMPLT_DebugNearestSurface1, false);
					SetRenderingFlag(VISTMPLT_DebugNearestSurface2, false);
				}
				return true;
			case '9':
				Toggle(VISTMPLT_DrawMislabeledOnly);
				return true;
			case 'A':
				Toggle(VISTMPLT_DrawDistribution);
				return true;
			case 's':
				Toggle(VISTMPLT_DebugSmoothness);
				return true;
			case 'S':
				m_bSelectingDeformedPnts = !m_bSelectingDeformedPnts;
				return true;
			case 'X':
				m_bOnlyRigidAlignment = !m_bOnlyRigidAlignment;
				return true;
			case ' ':
				if (IsFlagSet(VISTMPLT_DebugSmoothness))
				{
#ifndef MESA_RENDERING
					glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_PntsBuffer.m_iFrameBufferObject);
					glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
					GLint viewport[4];
					glGetIntegerv(GL_VIEWPORT, viewport);
					assert(m_PntsBuffer.m_iWidth >= viewport[2]);
					assert(m_PntsBuffer.m_iHeight >= viewport[3]);
					
					glReadPixels(0, 0, m_PntsBuffer.m_iWidth, m_PntsBuffer.m_iHeight,
								 GL_RGBA, GL_UNSIGNED_BYTE, m_PntsBuffer.m_aColorBuffer);
					window->LoadCurrentBuffer();
					
					int id = mouseY * m_PntsBuffer.m_iWidth + mouseX;
					int r = m_PntsBuffer.m_aColorBuffer[4 * id + 0];
					int g = m_PntsBuffer.m_aColorBuffer[4 * id + 1];
					int b = m_PntsBuffer.m_aColorBuffer[4 * id + 2];
					int a = m_PntsBuffer.m_aColorBuffer[4 * id + 3];
					
					int pntid = (int)(r | (g<<8) | (b<<16)) - 1;
					std::cout<<"Selected point: "<<pntid<<std::endl;
					if (pntid>=0)
					{
						TMatch * match = ((TMatchEnergy*)m_apMatchedConfigs[m_iCurrMatched])->GetMatch();
						const ShapePointCloud * pnts = match->GetMatchedGeomAsPnts();
						assert(pnts!=NULL);
						//VisModel3D * mvis = window->GetEmptyWindowXForm(1);
						VisModel3D * mvis = NULL;
						
						int id = pnts->GetModel()->ModelID();
						assert(id>=0 && id < (int)window->GetDrawableModels().size());
						mvis = window->GetDrawableModels()[id];
						mvis->SelectPoint(pnts->GetPoint(pntid));
					}
#endif
				}
				return true;
			case '~': // change sorting
			{
				((int&)m_CurrSorting)++;
				if ((int)m_CurrSorting>(int)SRT_ByAccuracy)
					m_CurrSorting = SRT_ByEnergy;
				std::cout<<"Sorting method: ";
				if (m_CurrSorting==SRT_ByEnergy)
					std::cout<<"ByEnergy"<<std::endl;
				else if (m_CurrSorting==SRT_ByConfidence)
					std::cout<<"ByConfidence"<<std::endl;
				else if (m_CurrSorting==SRT_ByAccuracy)
					std::cout<<"ByAccuracy"<<std::endl;
				else
					std::cout<<"Unknown. Id="<<m_CurrSorting<<std::endl;
				
				std::vector<Sortable> sortedMatches;
				TimeProfiler timer;
				for (int i=0; i<(int)m_apMatchedConfigs.size(); i++)
				{
					timer.WriteProgress("SRT", i, (int)m_apMatchedConfigs.size());
					double val = 0;
					if (m_CurrSorting==SRT_ByEnergy)
					{
						if (m_iSelectedPart==-1)
							val = m_apMatchedConfigs[i]->Energy();
						else
							val = m_apMatchedConfigs[i]->NodeEnergy(m_iSelectedPart);
					}
					else if (m_CurrSorting==SRT_ByConfidence)
						val = GetConfidence(i);
					else if (m_CurrSorting==SRT_ByAccuracy)
						val = GetAccuracy(i);
//					else
//						assert(false);
					sortedMatches.push_back(Sortable(val, m_apMatchedConfigs[i], i));
				}
				std::cout<<std::endl;
				
				std::sort(sortedMatches.begin(), sortedMatches.end());
				m_apMatchedConfigs.clear();
				m_SortingValues.clear();
				for (int i=0; i<(int)sortedMatches.size(); i++)
				{
					m_apMatchedConfigs.push_back((TMatchEnergy * )sortedMatches[i].ptr);
					m_SortingValues.push_back(sortedMatches[i].value);
				}
				return true;
			}
			case '>':
			case '<':
			{
				if (IsFlagSet(VISTMPLT_AllDeformedPoints))
				{
					int prevID = m_iCurrMatched;
					int currSeedID = m_apMatchedConfigs[m_iCurrMatched]->GetMatch()->SeedID();
					int goalID = currSeedID+1;
					while (m_apMatchedConfigs[m_iCurrMatched]->GetMatch()->SeedID() != goalID)
					{
						if (goalID >= m_apMatchedConfigs[m_iCurrMatched]->GetMatch()->GetSTemplate()->NSeedConfigs())
							goalID = 0;
						m_iCurrMatched++;
						if (m_iCurrMatched==prevID)
						{
							std::cout<<"[WARNING] Skipping seed: "<<goalID<<std::endl;
							goalID++;
						}
						else if (m_iCurrMatched>=(int)m_apMatchedConfigs.size())
							m_iCurrMatched = 0;
					}
					std::cout<<"Deformed all pnts: "<<"Next Config = "<<m_iCurrMatched<<std::endl;
				}
				else
				{
					m_iCurrMatched += (key=='>' ? 1 : -1);
					if (m_iCurrMatched<0)
						m_iCurrMatched = 0;
					else if (m_iCurrMatched>=(int)m_apMatchedConfigs.size())
						m_iCurrMatched = (int)m_apMatchedConfigs.size() - 1;
//					std::cout<<"Curr Matched = "<<m_iCurrMatched<<" / "<<m_apMatchedConfigs.size()<<". M=";
//					std::cout<<m_apMatchedConfigs[m_iCurrMatched]->GetMatch()->ModelName().c_str()<<std::endl;
//					std::cout<<"      "<<m_apMatchedConfigs[m_iCurrMatched]->EnergyStr().c_str()<<std::endl;
					std::cout<<"Match: "<<m_iCurrMatched<<" / "<<m_apMatchedConfigs.size()<<". ";
					std::cout<<"Model: "<<m_apMatchedConfigs[m_iCurrMatched]->GetMatch()->ModelName().c_str()<<". ";
					if (m_iCurrMatched < (int)m_SortingValues.size())
					{
						if (m_CurrSorting == SRT_ByEnergy)
							std::cout<<"Energy = "<<m_SortingValues[m_iCurrMatched]<<" ";
						else if (m_CurrSorting == SRT_ByConfidence)
							std::cout<<"Confidence = "<<m_SortingValues[m_iCurrMatched]<<"% ";
						else if (m_CurrSorting == SRT_ByAccuracy)
							std::cout<<"Accuracy = "<<(m_SortingValues[m_iCurrMatched]*100)<<"% ";
					}
					std::cout<<std::endl;
						
				}
				return true;
			}
			case '\'':
			case '\"':
				if (m_iCurrMatched>=0 && m_iCurrMatched < (int)m_apMatchedConfigs.size())
				{
					m_iSelectedPart += (key == '\'' ? 1 : -1);
					std::cout<<"Selected: "<<m_iSelectedPart<<": "<<std::flush;
					TMatch * match  = m_apMatchedConfigs[m_iCurrMatched]->GetMatch();
					const STemplate *st=match->GetSTemplate();
					if (m_iSelectedPart>=0 && m_iSelectedPart < match->NTemplateNodes())
						std::cout<<st->GetNode(match->Deformation(m_iSelectedPart).NodeID())->NodeName().c_str()<<std::endl;
					else
						std::cout<<std::endl;
				}
				return true;
			case 'p':
			{
				if (m_iCurrMatched<0 || m_iCurrMatched>=(int)m_apMatchedConfigs.size())
					return true;

				TMatchEnergy * energy = m_apMatchedConfigs[m_iCurrMatched];
				TMatch * match = energy->GetMatch();
				double acc = GetAccuracy(m_iCurrMatched);
				std::cout<<"Model: "<<match->ModelName().c_str();
				std::cout<<"  Acc: "<<((int)(acc*100))<<"."<<(((int)(acc*10000))%100)<<"%";
				std::cout<<"  Node Energies:"<<std::endl;
				
				std::vector<int> nPerPartPnts;
				for (int j=0; j<match->NTemplateNodes(); j++)
					nPerPartPnts.push_back(0);				
				for (int p=0; p<match->NShapePrimitives(); p++)
					nPerPartPnts[match->LocalNodeID(match->Loc2GlobPid(p))]++;

				std::vector<double> fullEnergies;
				for (int j=0; j<energy->GetMatch()->NTemplateNodes(); j++)
					fullEnergies.push_back(energy->NodeEnergy(j));
									
				int currTerms = energy->GetTerms();
				int terms = TMatchEnergy::ET_Pnts | TMatchEnergy::ET_NodeFit2Pnts;
				terms = terms | (currTerms & (TMatchEnergy::ET_PntsDistance | TMatchEnergy::ET_PntsOrientation));
				energy->SetTerms(terms);
				energy->ClearCache();
								
				std::vector<double> dataEnergies;
				for (int j=0; j<energy->GetMatch()->NTemplateNodes(); j++)
					dataEnergies.push_back(energy->NodeEnergy(j));
				
				std::cout<<"    "<<VKString("Part").CreateStrWithPadding(12, " ").c_str();
				std::cout<<" "<<VKString("DataEnergy").CreateStrWithPadding(12, " ").c_str();
				std::cout<<" "<<VKString("FullEnergy").CreateStrWithPadding(12, " ").c_str();
				std::cout<<" "<<VKString("NPnts").CreateStrWithPadding(12, " ").c_str()<<std::endl;;
				for (int j=0; j<energy->GetMatch()->NTemplateNodes(); j++)
				{
					int nodeID = match->Deformation(j).NodeID();
					VKString nodename = VKString::number(nodeID)+". "+match->GetSTemplate()->GetNode(nodeID)->NodeName();
					if (nPerPartPnts[j]>0)
					{
						nodename = nodename.CreateStrWithPadding(12, " ");
						VKString dataEStr = VKString::number(dataEnergies[j]/nPerPartPnts[j]).CreateStrWithPadding(12, " ");
						VKString fullEStr = VKString::number(fullEnergies[j]/nPerPartPnts[j]).CreateStrWithPadding(12, " ");
						VKString npntsStr = VKString::number(nPerPartPnts[j]).CreateStrWithPadding(12, " ");
						std::cout<<"    "<<nodename.c_str()<<" "<<dataEStr.c_str()<<" "<<fullEStr.c_str();
						std::cout<<" "<<npntsStr.c_str()<<std::endl;
					}
				}

				return true;
			}
			case 'i':
				assert(m_apMatchedConfigs.size()>0);
				SetInteractionCreateTemplate("none", (STemplate*)m_apMatchedConfigs[0]->GetMatch()->GetSTemplate());
				return true;
			case 't':
			case 'T':
			{
				// test arbitrary stuff
				m_iTestingSomething += (key == 't' ? 1 : -1);
				std::cout<<"Testing: "<<m_iTestingSomething<<std::endl;
				int iter = 0;
				TMatchAlgParams params;
				params.m_iICPIterations = 1;
				if (m_iSelectedPart>=0)
				{
					int pointsTIds[1000];
					int pointsMIds[1000];
					int nnew;
					for (int i=0; i<1000; i++)
						pointsTIds[i] = pointsMIds[i] = -1;
					TMatchAlg::FindNearestPointPairs(m_iSelectedPart, m_apMatchedConfigs[m_iCurrMatched],
													 params, pointsTIds, pointsMIds, 1000, nnew);
					
					TMatchAlg::OptimizeSingleNodeDeform(m_iSelectedPart, m_apMatchedConfigs[m_iCurrMatched],
														params, pointsTIds, pointsMIds, nnew);
				}
				else
					TMatchAlg::MStep_ICP(m_apMatchedConfigs[m_iCurrMatched], params,iter);
				std::cout<<"Energy = "<<m_apMatchedConfigs[m_iCurrMatched]->EnergyStr().c_str()<<std::endl;
				return true;
			}
			case '.':
			case ',':
			{
				if (m_pCoAlignedParts==NULL)
				{
					std::cout<<"[WARNING] Co-aligned parts are not defined - no clustering"<<std::endl;
					return true;
				}
				if (m_iCurrMatched<0 || m_iCurrMatched >= (int)m_apMatchedConfigs.size())
					return true;
				
				// check clustering id
				TMatch * match = m_apMatchedConfigs[m_iCurrMatched]->GetMatch();
				
				if (m_iSelectedPart<0 || m_iSelectedPart >= match->NTemplateNodes())
					return true;

				const TemplateNodeDeform & deform = match->Deformation(m_iSelectedPart);
				int clusteringID = m_pCoAlignedParts->GetClusteringID(match->SeedID(),deform.NodeID(),
																	  deform.PartID(), deform.NParts());
				
				if (clusteringID<0)
				{
					std::cout<<"[WARNING] Clustering not computed C="<<match->SeedID()<<" P="<<deform.NodeID()<<std::endl;
					return true;
				}
				
				if (key=='.')
					m_iSelectedCluster++;
				else
					m_iSelectedCluster--;
				
				if (m_iSelectedCluster>=m_pCoAlignedParts->NClusters(clusteringID))
					m_iSelectedCluster = -1;
				else if (m_iSelectedCluster < -1)
					m_iSelectedCluster = m_pCoAlignedParts->NClusters(clusteringID)-1;
				std::cout<<"Cluster: "<<m_iSelectedCluster;
				if (m_iSelectedCluster>=0)  
					std::cout<<" Size = "<<m_pCoAlignedParts->NPartsInCluster(clusteringID, m_iSelectedCluster);
				std::cout<<std::endl; 
				return true;
			}
			case 'c': // compute clustering
			{
				if (m_pCoAlignedParts==NULL)
				{
					std::cout<<"[WARNING] Co-aligned parts are not defined - no clustering"<<std::endl;
					return true;
				}

				std::cout<<"Clustering parts..."<<std::endl;
				int methodFlags = ConsistentPartsAlg::CPM_InitWhole | ConsistentPartsAlg::CPM_ClusterSpectralAgglomerative;
				methodFlags |= ConsistentPartsAlg::CPM_AveClusterTakeAll;
				TMatch * match = m_apMatchedConfigs[m_iCurrMatched]->GetMatch();
				STemplate * st = (STemplate*)match->GetSTemplate();
				
				if (m_iSelectedPart >= 0 && m_iSelectedPart < match->NTemplateNodes())
				{
					const TemplateNodeDeform & deform = match->Deformation(m_iSelectedPart);
					m_pCoAlignedParts->ClusterStyles(match->SeedID(), deform.NodeID(), deform.PartID(), deform.NParts());
					m_iSelectedClustering = m_pCoAlignedParts->GetClusteringID(match->SeedID(), deform.NodeID(),
																			   deform.PartID(), deform.NParts());
					m_iSelectedCluster = 0;
				} 
				return true;
			}
			case 'C':	// compute confidence values
			{
				if (m_pCoAlignedParts==NULL)
				{
					std::cout<<"[WARNING] Co-aligned parts are not defined - no clustering"<<std::endl;
					return true;
				}
				Toggle(VISTMPLT_DrawPointConfidence);
				int configID = m_apMatchedConfigs[m_iCurrMatched]->GetMatch()->SeedID();
				m_iSelectedClustering = m_pCoAlignedParts->GetClusteringFromMatchesID(configID);
				
//				std::cout<<"Compute confidence..."<<std::endl;
//				m_iSelectedClustering = m_pCoAlignedParts->GetClusteringFromMatchesID(configID);
//				int method = CoAlignedParts::CCF_Dist2All;
//				int clusterID, partInClusterID;
//				bool valid;
//				m_pCoAlignedParts->GetClusterWithPart(m_iSelectedClustering, m_iCurrMatched, m_iSelectedPart,
//													  &clusterID, &partInClusterID, valid);
//				m_pCoAlignedParts->ComputeConfidences(m_iSelectedClustering, clusterID, method);
//				
				return true;
			}
			case '=':
			case '-':	// more iterations
			case 'N':	
			case 'n':	// rescale null weight
			case 'm':	// m-step
			case 'e':	// e-step
			{			
				if (m_iCurrMatched<0 || m_iCurrMatched>=(int)m_apMatchedConfigs.size())
					return true;
				TMatchEnergy * matchEnergy = m_apMatchedConfigs[m_iCurrMatched];
				const STemplate * st = matchEnergy->GetMatch()->GetSTemplate();
				TMatch * match = m_apMatchedConfigs[m_iCurrMatched]->GetMatch();
				std::vector<int> allpnts;
				GeomShape * shape = (GeomShape *)match->GetMatchedGeomAsPnts();			
				TMatch * newSeed = st->CreateSeedConfig(match->SeedID(), shape, allpnts, match->SeedAlignment());
				TMatchAlgParams params;
				
				bool rescaleNull = (key=='N' || key=='n');
				bool changeNIter = (key=='=' || key=='-');
				int niter;
				if (key=='e')
					TMatchAlg::EStep_LabelPoints(matchEnergy, params, niter);
				else if (key=='m')
				{
					params.m_iNGreedyIterations = 1;
					TMatchAlg::MStep_DeformTemplate(matchEnergy, params, niter);
				}
				else if (changeNIter)
				{
					if (key=='=')
						m_iNIterations++;
					else if (key=='-')
						m_iNIterations = 1;
					std::cout<<"Iter="<<(m_iNIterations/2)<<std::flush;
					if (m_iNIterations%2==1)
						std::cout<<" - MRF Labeling"<<std::endl;
					else
						std::cout<<" - Deform 2 Fit"<<std::endl;				

					params.m_iIterations = m_iNIterations;
					//params.m_iMRFIterations = m_iNIterations;
					//params.m_iNGreedyIterations = m_iNIterations;
					m_apMatchedConfigs[m_iCurrMatched] = TMatchAlg::OptimizeConfigured(st, newSeed,  params);
				}
				else if (rescaleNull)
				{
					double nullPenalty = matchEnergy->GetNullPenalty();
					if (key=='n')
						nullPenalty += .05;
					else if (key=='N')
						nullPenalty -= .05;
					if (nullPenalty < 0)
						nullPenalty = 0;
					std::cout<<"New NULL Penalty: "<<nullPenalty<<std::endl;
					matchEnergy->SetNullPenalty(nullPenalty);
					int nChanged;
					TMatchAlg::EStep_LabelPoints(m_apMatchedConfigs[m_iCurrMatched], TMatchAlgParams(), nChanged);
				}

				std::cout<<"Best Energy: "<<m_apMatchedConfigs[m_iCurrMatched]->EnergyStr().c_str()<<std::endl;

				// not relevant for visualization...
	//			delete matchEnergy->GetMatch();
	//			delete matchEnergy;
				return true;
			}
		}
	}
	return false;
}

void VisShapeTemplate::CreateNewConfig(int configID)
{
	assert(configID>=0 && configID < m_sConfigNames.count() && configID < (int)m_apSeedConfigs.size());
	std::cout<<"Creating new config: "<<configID<<std::endl;
	static char buffer[1024];
	std::cout<<"     What is config name? Write here: "<<std::flush;
	scanf("%s", buffer);
	m_sConfigNames[configID] = VKString(buffer);
	std::cout<<"     Adding config: |"<<m_sConfigNames[configID].c_str()<<"| seed="<<configID<<std::endl;
	TMatch * match = m_pCreatingTemplate->CreateSeedConfig(0, NULL, std::vector<int>(), R3identity_affine);
	match->SetSeedID(configID);
	m_apSeedConfigs[configID] = match;
	
	m_pCreatingTemplate->AddOrReplaceSeed(configID, m_sConfigNames[configID], m_apSeedConfigs[configID], 0);

}

void VisShapeTemplate::PickPartFromDeformedPnts(DatabaseVisWindow * window, int x, int y)
{
#ifndef MESA_RENDERING
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_DeformedPntsBuffer.m_iFrameBufferObject);
	glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
	GLint viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);
	assert(m_DeformedPntsBuffer.m_iWidth >= viewport[2]);
	assert(m_DeformedPntsBuffer.m_iHeight >= viewport[3]);
	
	glReadPixels(0, 0, m_DeformedPntsBuffer.m_iWidth, m_DeformedPntsBuffer.m_iHeight,
				 GL_RGBA, GL_UNSIGNED_BYTE, m_DeformedPntsBuffer.m_aColorBuffer);
	window->LoadCurrentBuffer();
	
//	std::cout<<"Picking point"<<std::endl;
	// find triangle from color: r, g, b (if a = 0) at x, y
	if (IsFlagSet(VISTMPLT_DebugSmoothness))
	{
		m_iSelDeformedX = x;
		m_iSelDeformedY = y;
	}
	for (int i=vkMin(m_iSelDeformedX, x); i<=vkMax(m_iSelDeformedX, x); i++)
	for (int j=vkMin(m_iSelDeformedY, y); j<=vkMax(m_iSelDeformedY, y); j++)
	{
		if (j<0 || j >= m_DeformedPntsBuffer.m_iHeight || i<0 || i>=m_DeformedPntsBuffer.m_iWidth)
			continue;
		int id = j * m_DeformedPntsBuffer.m_iWidth + i;
		int r = m_DeformedPntsBuffer.m_aColorBuffer[4 * id + 0];
		int g = m_DeformedPntsBuffer.m_aColorBuffer[4 * id + 1];
		int b = m_DeformedPntsBuffer.m_aColorBuffer[4 * id + 2];
		int a = m_DeformedPntsBuffer.m_aColorBuffer[4 * id + 3];
		
		int modelID = (int)(r | (g<<8)) - 1;
		int pntID = (int)(b | (a<<8));
		if (modelID<0)
			continue;
		assert(modelID>=0 && modelID<(int)m_apMatchedConfigs.size());
		assert(pntID>=0 && pntID < (int)m_apMatchedConfigs[modelID]->GetMatch()->NShapePrimitives());
		
//		
//		if (IsFlagSet(VISTMPLT_DebugSmoothness))
//		{
//			TMatch * match = ((TMatchEnergy*)m_apMatchedConfigs[m_iCurrMatched])->GetMatch();
//			const ShapePointCloud * pnts = match->GetMatchedGeomAsPnts();
//			assert(pnts!=NULL);
//			//VisModel3D * mvis = window->GetEmptyWindowXForm(1);
//			VisModel3D * mvis = NULL;
//			
//			int id = pnts->GetModel()->ModelID();
//			
//			if (id<0 || id >= (int)window->GetDrawableModels().size())
//			{
//				std::cout<<"[WARNING] Cannot find selected point. NDraw="<<window->GetDrawableModels().size()<<std::endl;
//				return;
//			}
//			mvis = window->GetDrawableModels()[id];
//		}
//		else
		{
			while (modelID >= (int)m_SelectedPointsFromDeformed.size())
				m_SelectedPointsFromDeformed.push_back(std::vector<bool>());
			int npnts =(int)m_apMatchedConfigs[modelID]->GetMatch()->NShapePrimitives();
			if (m_SelectedPointsFromDeformed[modelID].size()==0)
			{
				for (int p=0; p<npnts; p++)
					m_SelectedPointsFromDeformed[modelID].push_back(false);
			}
			assert(npnts == (int)m_SelectedPointsFromDeformed[modelID].size());
			m_SelectedPointsFromDeformed[modelID][pntID] = true;
		}
	}
//	std::cout<<"Picking point - done!"<<std::endl;
//
//	int id = y * m_DeformedPntsBuffer.m_iWidth + x;
//	int r = m_DeformedPntsBuffer.m_aColorBuffer[4 * id + 0];
//	int g = m_DeformedPntsBuffer.m_aColorBuffer[4 * id + 1];
//	int b = m_DeformedPntsBuffer.m_aColorBuffer[4 * id + 2];
//	int a = m_DeformedPntsBuffer.m_aColorBuffer[4 * id + 3];
//	
//	if (a!=0)
//	{
//		m_iSelectedPart = (int)(r | (g<<8) | (b<<16)) - 1;
//		m_iSelectedFace = ((int)a)-1;
//	}
//	else
//		m_iSelectedPart = -1;
//	
//	std::cout<<"Selected Template Part ";
//	if (m_iCurrMatched>=0 && m_iCurrMatched < (int)m_apSeedConfigs.size() && m_iSelectedPart>=0)
//	{
//		int nid = m_apSeedConfigs[m_iCurrMatched]->Deformation(m_iSelectedPart).NodeID();
//		std::cout<<m_apSeedConfigs[m_iCurrMatched]->GetSTemplate()->GetNode(nid)->NodeName().c_str();
//	}
//	std::cout<<" P="<<(m_iSelectedPart+1)<<" F="<<m_iSelectedFace<<std::endl;
#else
	std::cout<<"[WARNING] Cannot do interaction w/ MESA renderer"<<std::endl;
#endif

}

void VisShapeTemplate::PickFace(DatabaseVisWindow * window, int x, int y)
{
#ifndef MESA_RENDERING
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_IntersectionBuffer.m_iFrameBufferObject);
	glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
	GLint viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);
	assert(m_IntersectionBuffer.m_iWidth >= viewport[2]);
	assert(m_IntersectionBuffer.m_iHeight >= viewport[3]);

	glReadPixels(0, 0, m_IntersectionBuffer.m_iWidth, m_IntersectionBuffer.m_iHeight,
				 GL_RGBA, GL_UNSIGNED_BYTE, m_IntersectionBuffer.m_aColorBuffer);
	window->LoadCurrentBuffer();
		
	// find triangle from color: r, g, b (if a = 0) at x, y
	int id = y * m_IntersectionBuffer.m_iWidth + x;
	int r = m_IntersectionBuffer.m_aColorBuffer[4 * id + 0];
	int g = m_IntersectionBuffer.m_aColorBuffer[4 * id + 1];
	int b = m_IntersectionBuffer.m_aColorBuffer[4 * id + 2];
	int a = m_IntersectionBuffer.m_aColorBuffer[4 * id + 3];

	if (a!=0)
	{
		m_iSelectedPart = (int)(r | (g<<8) | (b<<16)) - 1;
		m_iSelectedFace = ((int)a)-1;
	}
	else
		m_iSelectedPart = -1;
	
	std::cout<<"Selected Template Part ";
	if (m_iCurrMatched>=0 && m_iCurrMatched < (int)m_apSeedConfigs.size() && m_iSelectedPart>=0)
	{
		int nid = m_apSeedConfigs[m_iCurrMatched]->Deformation(m_iSelectedPart).NodeID();
		std::cout<<m_apSeedConfigs[m_iCurrMatched]->GetSTemplate()->GetNode(nid)->NodeName().c_str();
	}
	std::cout<<" P="<<(m_iSelectedPart+1)<<" F="<<m_iSelectedFace<<std::endl;
#else
	std::cout<<"[WARNING] Cannot do interaction w/ MESA renderer"<<std::endl;
#endif
}

void VisShapeTemplate::RescaleTemplateBox(double step)
{
	if (!(m_iCurrMatched>=0 && m_iCurrMatched < (int)m_apSeedConfigs.size()))
		return;
	if (!(m_iSelectedPart>=0 && m_iSelectedPart < m_apSeedConfigs[m_iCurrMatched]->NTemplateNodes()))
		return;
	if (m_apSeedConfigs[m_iCurrMatched]->NodeEllipsoid(m_iSelectedPart).IsType(PartBox::BOXTYPE_NullBox))
		return;

	TMatch * match = m_apSeedConfigs[m_iCurrMatched];
	TemplateNodeDeform & deform = (TemplateNodeDeform&)match->Deformation(m_iSelectedPart);
	R3Vector scale = deform.GetScale();
	scale[m_iSelectedFace/2] += step;
	deform.SetScale(scale);
	match->UpdateEllipsoid(m_iSelectedPart);
	if (m_iCurrMatched==0)
		m_pCreatingTemplate->SetDeformAsRestPose(match);
	
}
void VisShapeTemplate::RotateTemplateBox(double step)
{
	if (!(m_iCurrMatched>=0 && m_iCurrMatched < (int)m_apSeedConfigs.size()))
		return;
	if (!(m_iSelectedPart>=0 && m_iSelectedPart < m_apSeedConfigs[m_iCurrMatched]->NTemplateNodes()))
		return;
	if (m_apSeedConfigs[m_iCurrMatched]->NodeEllipsoid(m_iSelectedPart).IsType(PartBox::BOXTYPE_NullBox))
		return;

	if (m_iCurrMatched!=0)
	{
		std::cout<<"[WARNING] Cannot rotate in seed"<<std::endl;
		return;
	}
	TMatch * match = m_apSeedConfigs[m_iCurrMatched];
	TemplateNodeDeform & deform = (TemplateNodeDeform&)match->Deformation(m_iSelectedPart);
	const STemplateNode * node = m_pCreatingTemplate->GetNode(deform.NodeID());
	R3Triad & o1 = (R3Triad&)node->GetPart(match->SeedID())->UndeformedOrientation();
	double a = 3.14/10.;
	o1.Rotate(m_iSelectedFace/2, step);
	R3Triad & o2 = (R3Triad&)(match->NodeEllipsoid(m_iSelectedPart).UndeformedOrientation());
	o2.Rotate(m_iSelectedFace/2, step);
	deform.SetScale(deform.GetScale());
	match->UpdateEllipsoid(m_iSelectedPart);
}
void VisShapeTemplate::TranslateTemplateBox(double step)
{
	if (!(m_iCurrMatched>=0 && m_iCurrMatched < (int)m_apSeedConfigs.size()))
		return;
	if (!(m_iSelectedPart>=0 && m_iSelectedPart < m_apSeedConfigs[m_iCurrMatched]->NTemplateNodes()))
		return;
	if (m_apSeedConfigs[m_iCurrMatched]->NodeEllipsoid(m_iSelectedPart).IsType(PartBox::BOXTYPE_NullBox))
		return;

	TMatch * match = m_apSeedConfigs[m_iCurrMatched];
	TemplateNodeDeform & deform = (TemplateNodeDeform&)match->Deformation(m_iSelectedPart);
	int xformndims, xformflags;
	double newxform[3];
	const double * xform = deform.GetTransform(xformndims, xformflags);
	const STemplateNode * node = m_pCreatingTemplate->GetNode(deform.NodeID());
	const R3Triad & o = node->GetPart(match->SeedID())->UndeformedOrientation();
	assert(xformndims==3);
	R3Vector offset = o[m_iSelectedFace / 2] * step;
	for (int i=0; i<3; i++)
		newxform[i] = xform[i] + offset[i];
	deform.SetTransform(newxform);
	match->UpdateEllipsoid(m_iSelectedPart);
	if (m_iCurrMatched==0)
		m_pCreatingTemplate->SetDeformAsRestPose(match);
}

void VisShapeTemplate::SetCoAlignedPartsHolder(CoAlignedParts * coAligned)
{
	m_pCoAlignedParts = coAligned;
}

bool VisShapeTemplate::MouseButtonEvent(DatabaseVisWindow * window, int x, int y,
										VisWindow::MouseButton mouseButtonID,
										VisWindow::MouseButtonState newState)
{
	if (mouseButtonID == VisWindow::MOUSE_RIGHT && newState == VisWindow::MOUSE_BT_DOWN)
	{
		if (m_bSelectingDeformedPnts)
		{
			m_iSelDeformedX = x;
			m_iSelDeformedY = y;
			m_iSelDeformedNewX = x;
			m_iSelDeformedNewY = y;
			m_SelectedPointsFromDeformed.clear();
		}
		else
			PickFace(window, x, y);
		return true;
	}
	else if (mouseButtonID == VisWindow::MOUSE_RIGHT && newState == VisWindow::MOUSE_BT_UP)
	{
		if (m_bSelectingDeformedPnts)
			m_iSelDeformedX = -1;
		return true;
	}
	return false;
}

bool VisShapeTemplate::MouseMotionEvent(DatabaseVisWindow * window, int x, int y, int dx, int dy)
{
	if (m_bDisable)
		return false;
	
	if (window->ClickInWindow(x, y) > window->NumMainWindows())
		return false;

	if (window->IsMouseButtonDown(VisWindow::MOUSE_RIGHT))
	{
		if (m_bSelectingDeformedPnts)
		{
			// lookup all pnts in a square
			m_iSelDeformedNewX = x;
			m_iSelDeformedNewY = y;
			PickPartFromDeformedPnts(window, x, y);
			return true;
		}
		
		if (m_iCurrMatched<0 || m_iCurrMatched >= (int)m_apSeedConfigs.size())
			return false;
		if (m_iSelectedPart<0 || m_iSelectedPart >= m_apSeedConfigs[m_iCurrMatched]->NTemplateNodes())
			return false;
		if (m_apSeedConfigs[m_iCurrMatched]->NodeEllipsoid(m_iSelectedPart).IsType(PartBox::BOXTYPE_NullBox))
			return false;
		if (m_pCreatingTemplate==NULL)
			return false;
		double ddx = -(double)dx/(double)window->Width();
		double ddy = -(double)dy/(double)window->Height();

		TMatch * match = m_apSeedConfigs[m_iCurrMatched];
		TemplateNodeDeform & deform = (TemplateNodeDeform&)match->Deformation(m_iSelectedPart);
		const STemplateNode * node = m_pCreatingTemplate->GetNode(deform.NodeID());
		R3Triad & o1 = (R3Triad&)node->GetPart(match->SeedID())->UndeformedOrientation();
		
		if (window->Ctrl())	// rotate
		{
			R3Vector dispx(ddx, 0, 0);
			m_TemplateCamera.ApplyInverse(dispx);
			R3Vector dispy(0, ddy, 0);
			m_TemplateCamera.ApplyInverse(dispy);
			
			int a1 = m_iSelectedFace/2, a2 = 1, a3 = 2;
			if (a1==1)
				a2 = 0;
			else if (a1==2)
				a3 = 0;

			TMatch * match = m_apSeedConfigs[m_iCurrMatched];
			TemplateNodeDeform & deform = (TemplateNodeDeform&)match->Deformation(m_iSelectedPart);
			const STemplateNode * node = m_pCreatingTemplate->GetNode(deform.NodeID());
			R3Triad & o1 = (R3Triad&)node->GetPart(match->SeedID())->UndeformedOrientation();
			R3Triad & o2 = (R3Triad&)(match->NodeEllipsoid(m_iSelectedPart).UndeformedOrientation());
			double a = 3.14 / 30.;
			
			R3Vector right(1., 0, 0);
			R3Vector up(0, 1., 0);
			
			RNAngle theta = 4.0 * (fabs(ddx) + fabs(ddy));
			R3Vector vector = (right * -ddx) + (up * -ddy);
			R3Vector axis = (right % up) % vector;
			axis.Normalize();
			
			o1.Rotate(axis, theta);
			
			o2 = o1;
			deform.SetScale(deform.GetScale());
			match->UpdateEllipsoid(m_iSelectedPart);
			return true;
		}
		else if (window->Shift())	// translate
		{
			R3Point p1(0, 0, -1);
			R3Point p2(-ddx, -ddy, -1);
			m_TemplateCamera.ApplyInverse(p1);
			m_TemplateCamera.ApplyInverse(p2);
			R3Vector disp = p2 - p1;

			TMatch * match = m_apSeedConfigs[m_iCurrMatched];
			TemplateNodeDeform & deform = (TemplateNodeDeform&)match->Deformation(m_iSelectedPart);
			int xformndims, xformflags;
			double newxform[3];
			const double * xform = deform.GetTransform(xformndims, xformflags);
			const STemplateNode * node = m_pCreatingTemplate->GetNode(deform.NodeID());
			const R3Triad & o = node->GetPart(match->SeedID())->UndeformedOrientation();
			assert(xformndims==3);
			for (int i=0; i<3; i++)
				newxform[i] = xform[i] + disp[i];
			deform.SetTransform(newxform);
			match->UpdateEllipsoid(m_iSelectedPart);
			if (m_iCurrMatched==0)
				m_pCreatingTemplate->SetDeformAsRestPose(match);
			return true;
		}
	}
	else if (window->IsMouseButtonDown(VisWindow::MOUSE_LEFT))
	{
		int subwindow = window->SelectedWindow();
		//std::cout<<"selected window = "<<subwindow<<std::endl;
		if (subwindow>=0 && subwindow < window->NumWindows())
		{
			for (int i=0; i<2; i++)				
			{	
				VisModel3D * mvis = window->GetEmptyWindowXForm(subwindow);
				if (i==1)
				{
					int mid = -1;
					std::vector<VisModel3D *> & allModels = window->GetDrawableModels();
					if (m_iCurrMatched>=0 && m_iCurrMatched < (int)m_apMatchedConfigs.size())
					{
						TMatch * match = m_apMatchedConfigs[m_iCurrMatched]->GetMatch();
						if (subwindow==0)
						{
							const ShapeLabels * trainLabels = match->GetSTemplate()->TrainingLabels();
							if (trainLabels!=NULL)
								mid = trainLabels->GetShape()->GetModel()->ModelID();
						}
						else if (match->GetMatchedGeomAsPnts()!=NULL)
							mid = match->GetMatchedGeomAsPnts()->GetModel()->ModelID();
					}
					assert(mid==-1 || (mid>=0 && mid < allModels.size()));
					mvis = (mid==-1 ? NULL : allModels[mid]);
				}
				
				if (mvis==NULL)
					continue;
							 
				int cameraID = mvis->SubwindowCameraID(subwindow, window);
				if (!window->Ctrl() && !window->Alt() && !window->Shift())	// rotate
					mvis->Rotate(cameraID, x, y, dx, dy, window);
				else if (window->Ctrl())
					mvis->Scale(cameraID, x, y, dx, dy, window);
				else if (window->Shift())
					mvis->Translate(cameraID, x, y, dx, dy, window);
			}
		}
		return true;
	}
	return false;
}

void VisShapeTemplate::UpdateTemplateScale()
{
	for (int i=0; i<(int)m_apMatchedConfigs.size(); i++)
		delete m_apMatchedConfigs[i];
//	for (int i=0; i<(int)m_apSeedConfigs.size(); i++)
//		delete m_apSeedConfigs[i];
	m_apMatchedConfigs.clear();
	m_apSeedConfigs.clear();
	
	GeomShape * g = NULL;
	if (m_pDB!=NULL && m_pDB->NumModels()>0)
	{
		Model3D * m = m_pDB->GetModel(m_sRefModelName);
		assert(m!=NULL);
		g = m->GetGeomShape(m_sRefGeometry);
	}
	
	for (int i=0; i<m_pCreatingTemplate->NSeedConfigs(); i++)
	{
		TMatch * nmatch = m_pCreatingTemplate->CreateSeedConfig(i, g, std::vector<int>(), R3identity_affine);
		m_apMatchedConfigs.push_back(new TMatchEnergy(nmatch));
		m_apSeedConfigs.push_back(nmatch);
	}
}

void VisShapeTemplate::SetScreenshotType(const VKString & sName, int value)
{
	if (sName=="Template" || sName=="CoAligned")
	{
		int flags = VISTMPLT_TemplatePnts | VISTMPLT_SeedTemplate | VISTMPLT_DrawDistribution;
		flags |=  VISTMPLT_DrawTemplateAsBoxes | VISTMPLT_DrawTemplateAsAxes;
		if (sName=="CoAligned")
			flags = 0;

		SetRenderingFlags(flags);
		m_iCurrMatched = value;
		if (m_apSeedConfigs.size()==0)
		{
			assert(m_pCreatingTemplate!=NULL);
			std::vector<int> emptySet;
			
			for (int i=0; i<m_pCreatingTemplate->NSeedConfigs(); i++)
			{
				TMatch * m = m_pCreatingTemplate->CreateSeedConfig(i, NULL, emptySet, R3identity_affine);
				m_apSeedConfigs.push_back(m);
			}
		}
		
		//std::cout<<"Curr Matched: "<<m_iCurrMatched<<" / "<<m_apSeedConfigs.size()<<std::endl;
		assert(m_iCurrMatched < (int)m_apSeedConfigs.size());
	}
	else
		SetRenderingFlags(value);
}

void VisShapeTemplate::SetCoAlignedPoints(const VKString & coAlignedPntsFile,
										  const std::vector<double> & scores, double threshold)
{
	FILE * file = fopen(coAlignedPntsFile.c_str(), "rb");
	if (file==NULL)
		std::cout<<"[ERROR] Cannot read file: "<<coAlignedPntsFile.c_str()<<std::endl;
	assert(file!=NULL);
	int mid, nseeds, npnts;
	
	int counter = 0;
	while (!feof(file))
	{
		fread(&mid, sizeof(int), 1, file);
		mid--;
		//std::cout<<"mid="<<mid<<std::endl;
		fread(&nseeds, sizeof(int), 1, file);
		fread(&npnts, sizeof(int), 1, file);
		if (feof(file))
			break;
		bool include = true;
		if (mid>=(int)scores.size()||mid<0)
		{
			std::cout<<"[WARNING] CoAlignedPnts: "<<coAlignedPntsFile.c_str()<<" mid does not match scores: ";
			std::cout<<mid<<" / "<<scores.size()<<std::endl;
			include = false;
		}
		else
		{
			include = scores[mid] < threshold;
			//std::cout<<"Including: "<<mid<<std::endl;
		}
		///assert(mid < (int)scores.size());

		if (include)
		{
			m_CoAlignedPntSeeds.push_back(nseeds);
			m_CoAlignedPoints.push_back(std::vector<R3Point>());
		}
		assert(npnts<1000000000);
		for (int i=0; i<npnts; i++)
		{
			double xyz[3];
			fread(xyz, sizeof(double), 3, file);
			if (include)
				m_CoAlignedPoints[counter].push_back(R3Point(xyz[0], xyz[1], xyz[2]));
		}
		if (include)
			counter++;
	}
	
	fclose(file);
}




