/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __VisShapeTemplate_H
#define __VisShapeTemplate_H

#include <vector>
#include <assert.h>

#include "DBApp.h"
#include "STemplate.h"
#include "TMatchEnergy.h"
#include "CoAlignedParts.h"

class VisShapeTemplate : public DBApp
{
	public:
		enum VisTemplateFlag
		{
			VISTMPLT_TemplatePnts			= 1<<0,
			VISTMPLT_TemplateSegPnts		= 1<<1,
			VISTMPLT_TemplateGTSegMesh		= 1<<2,
			VISTMPLT_MatchedPnts			= 1<<3,
			VISTMPLT_MatchedSegPnts			= 1<<4,
			VISTMPLT_MatchedGTSegMesh		= 1<<5,
			VISTMPLT_SeedTemplate			= 1<<6,
			VISTMPLT_SeedOverMatched		= 1<<7,		// undeformed template
			VISTMPLT_DeformedOverMatched	= 1<<8,		// deformed template (least-energy deformation)
			VISTMPLT_DeformedPoints			= 1<<9,		// deform points to template (disabled)
			VISTMPLT_AllDeformedPoints		= 1<<10,		// deform points to template (disabled)
			VISTMPLT_ShowAllMatchesStats	= 1<<11,
			VISTMPLT_DebugNearestSurface1	= 1<<12,
			VISTMPLT_DebugNearestSurface2	= 1<<13,

			VISTMPLT_DrawTemplateAsBoxes	= 1<<14,
			VISTMPLT_DrawTemplateAsAxes		= 1<<15,
			VISTMPLT_DrawTemplateSampling	= 1<<16,
			VISTMPLT_DrawMislabeledOnly		= 1<<17,
			VISTMPLT_DrawDistribution		= 1<<18,
			VISTMPLT_DrawReferenceModel		= 1<<19,
			
			VISTMPLT_DebugSmoothness		= 1<<20,
			VISTMPLT_DrawPointConfidence	= 1<<21,
			VISTMPLT_DrawMeshTextures		= 1<<22,
			VISTMPLT_DrawMeshPartMask		= 1<<23,
			
			VISTMPLT_DrawSegmentedMesh		= 1<<24
			
		};
	
		enum InteractionMode
		{
			INTERACT_VisualizeOnly 			= 0,
			INTERACT_CreateTempalte			= 1<<0
		};
	
		enum SortMatches
		{
			SRT_ByEnergy 	 = 1,
			SRT_ByConfidence = 2,
			SRT_ByAccuracy   = 3
			
		};
	
		// static visualiaztion functions	
		static void DrawTemplateConfig(const TMatch * tConfig, const DatabaseVisWindow * window, 
									   int flags, std::vector<RenderableFace> & faces, const R3Affine & xform);

		static void DrawTemplateAsBoxes(const PartBox & ellipsoid, int nodeID, const DatabaseVisWindow * window,
										std::vector<RenderableFace> & faces, const R3Affine & xform, bool outline,
										int highlightFace);
		static void DrawTemplateMesh(const PartBox & ellipsoid, int nodeID, const DatabaseVisWindow * window,
									 std::vector<RenderableFace> & faces, const R3Affine & xform, bool outline);
		static void DrawTemplateAsAxes(const PartBox & ellipsoid, int nodeID, const DatabaseVisWindow * window,
									   const R3Affine & xform, const R3Vector * sVar = NULL,
									   const R3Vector  * pVar = NULL, double prob = -1);
		static void DrawDeformedGeometry(DatabaseVisWindow * window, const R3Affine & camera, TMatch * currMatch,
										 std::vector<TMatchEnergy*> & allMatches, CoAlignedParts * coAligned,
										 int clusteringID, int clusterID, bool drawCurrent, bool drawAll,
										 bool drawConfidence, int selectedPart, bool onlyRigid,
										 BufferID * isectBuffer = NULL,
										 std::vector<std::vector<bool> > * matchPart = NULL);
		static void DrawTemplateMesh(bool drawMesh, bool drawLabels, bool drawMislabeled, bool drawGTMesh,
									 DatabaseVisWindow * window, int subwindow, 
									 const R3Affine & camera, const ShapeLabels * labels);
		static void DrawDebugNearestSurface(bool drawMislabeled, bool surf1, bool surf2, const TMatch * match,
											int partID, const DatabaseVisWindow * window, const R3Affine & xform);
		static void DrawPart(TMatchEnergy * match, std::vector<bool> & selectedPnt,
							 DatabaseVisWindow * window, const R3Affine & camera);
	
		VisShapeTemplate(ModelDatabase * db);
		virtual void DrawDebugEnergy(TMatchEnergy * energy, DatabaseVisWindow * window, TMatch * seedConfig,
									 const R3Affine & xform, int selectedPart, int selectedFace, int testValue);
	
		virtual void SetTemplate(STemplate * st);
		virtual void SetInteractionCreateTemplate(const VKString & topologyFile, const VKString & outFile);
		virtual void SetInteractionCreateTemplate(const VKString & outFile, STemplate * initTemplate);
		virtual void SetReferenceModel(const VKString & mName, const VKString & gName);
		virtual void WriteTemplate();
		virtual void AddNewPart(const STemplateNode * cNode = NULL);
	
		virtual void AddMatchedModel(TMatch * seed, TMatchEnergy * match);

		virtual bool KeyPressed(DatabaseVisWindow * window, const char key, int mouseX, int mouseY);
		virtual bool MouseButtonEvent(DatabaseVisWindow * window, int x, int y, 
									  VisWindow::MouseButton mouseButtonID, 
									  VisWindow::MouseButtonState newState);
		virtual bool MouseMotionEvent(DatabaseVisWindow * window, int x, int y, int dx, int dy);
		
		virtual bool Draw(DatabaseVisWindow * window);
		virtual void DrawIntersectionBuffers(const TMatch * tConfig, DatabaseVisWindow * window, const R3Affine & cam);
		virtual void DrawPntsIntersectionBuffer(const TMatch * tConfig, DatabaseVisWindow * window, const R3Affine & cam);
		virtual void DrawSelectedParts(DatabaseVisWindow * window);


		virtual GeomShape * IdentityToModelXForm(DatabaseVisWindow * window, int wndID) const;
        virtual R3Affine IdentityToModelXForm(DatabaseVisWindow * window, int wndID, const GeomShape * shape) const; 
	
		virtual void PickPartFromDeformedPnts(DatabaseVisWindow * window, int x, int y);
		virtual void PickFace(DatabaseVisWindow * window, int x, int y);
		virtual void RescaleTemplateBox(double step);
		virtual void RotateTemplateBox(double step);
		virtual void TranslateTemplateBox(double step);
	
		virtual void SetCoAlignedPartsHolder(CoAlignedParts * coAligned);
	
		virtual void SetGTEval(const VKString & gtLabels, const VKString & equivFilename = "none");
		virtual double GetAccuracy(int matchID);
		virtual double GetConfidence(int matchID);
	
		virtual void SetScreenshotType(const VKString & sName, int value);
	
		virtual void SetCoAlignedPoints(const VKString & coAlignedPntsFile,
										const std::vector<double> & scores,
										double threshold);
	
	protected:
		BufferID m_IntersectionBuffer;
		bool m_bIBufferInitialized;
		R3Affine m_TemplateCamera;
	
		BufferID m_PntsBuffer;
		bool m_bPntsIBufferInitialized;
	
		virtual void CreateNewConfig(int configID);
		int m_iCurrMatched;
		std::vector<TMatchEnergy *> m_apMatchedConfigs;
		std::vector<TMatch *> m_apSeedConfigs;
		int m_iNIterations;
	
		int m_iSaveFlags;

		int m_iInteractionFlags;
	
	// creating template
		int m_iSelectedPart;
		int m_iSelectedFace;
	
		VKString m_sOutTemplateFile;
		STemplate * m_pCreatingTemplate;
		VKStringList m_sConfigNames;
	
		virtual void UpdateTemplateScale();
		VKString m_sRefModelName;
		VKString m_sRefGeometry;
	
		int m_iTestingSomething;	// special flag for debugging
	
		CoAlignedParts * m_pCoAlignedParts;
		int m_iSelectedClustering;
		int m_iSelectedCluster;
		bool m_bSelectingDeformedPnts;
		bool m_bDeformedPntsBufferInitialized;
		BufferID m_DeformedPntsBuffer;
	
		int m_iSelDeformedX, m_iSelDeformedY;
		int m_iSelDeformedNewX, m_iSelDeformedNewY;
		//std::vector<std::map<int, int> > m_SelectedPartsFromDeformed;
		std::vector<std::vector<bool> > m_SelectedPointsFromDeformed;
	
		SortMatches m_CurrSorting;
		VKString m_sGTLabelsName;
		std::map<VKString, int> m_EquivalentLabels;
		VKStringList m_BasicLabels;
		std::vector<double> m_SortingValues;
	
		bool m_bOnlyRigidAlignment;
	
		std::vector<std::vector<R3Point> > m_CoAlignedPoints;
		std::vector<int> m_CoAlignedPntSeeds;
};

#endif



