#!/usr/bin/python

##########
# Current evolution algorithm
##########
import sys, os;
sys.path.append(os.path.realpath("./"+os.path.dirname(sys.argv[0])+"/../largescale/"))
sys.path.append(os.path.realpath("./"+os.path.dirname(sys.argv[0])+"/../experiment/"))
sys.path.append(os.path.realpath("./"+os.path.dirname(sys.argv[0])+"/../evolution/"))
import string, decimal, glob, re, subprocess, shlex, datetime, mydb, expdb

evolPath = "./scripts/evolution/"


evalCorrs = "none"
eval = "none";
equivLabels = "none";
genID = 0;
maxGen = 10;
readManifold = False;
singleStep = "none";
alignments = "none";
writeScheduledIDs = "none";

evolveSubset = "all";
prefixIsAbosolutePath = False;

if len(sys.argv)>=4:
	keyword = sys.argv[1];
	template = sys.argv[2];	# currently just a model name
	outDirPfx = sys.argv[3];

	i=4;
	while i<len(sys.argv):
		if (sys.argv[i]=="-eval"):
			i=i+1;
			eval = sys.argv[i];
		elif (sys.argv[i]=="-evalCorrs"):
			i=i+1;
			evalCorrs = sys.argv[i];
		elif (sys.argv[i]=="-equivLabels"):
			i=i+1;
			equivLabels = sys.argv[i];
		elif (sys.argv[i]=="-startGen"):
			i=i+1;
			genID = int(sys.argv[i]);
		elif (sys.argv[i]=="-endGen"):
			i=i+1;
			maxGen = int(sys.argv[i]);
		elif (sys.argv[i]=="-readManifold"):
			readManifold = True;
		elif (sys.argv[i]=="-alignments"):
			i = i + 1;
			alignments = sys.argv[i];
		elif (sys.argv[i]=="-singleStep"):
			i=i+1;
			genID = int(sys.argv[i]);
			maxGen = genID + 1;
			i=i+1;
			singleStep = sys.argv[i];
		elif (sys.argv[i]=="-subset"):
			i=i+1;
			evolveSubset = sys.argv[i];
		elif (sys.argv[i]=="-writeScheduledIDs"):
			i=i+1;
			writeScheduledIDs = sys.argv[i];
		elif (sys.argv[i]=="-prefixIsAbosolutePath"):
			prefixIsAbosolutePath = True;
		else:
			print("[WARNING] Cannot understand command: "+sys.argv[i]+" in evolve.py");
		i=i+1;
else:
	print("Usage:");
	print("   python "+sys.argv[0]+" keyword seedTemplate outWorkDir [options]");
	print("   OPTIONS:");
	print("       -eval fileprefix - perform evaluation");
	print("       -evalCorrs fileprefix - perform evaluation for correspondences");
	print("       -equivLabels filename - label equivalence");
	print("       -startGen iter0 - prescribe evolutions");
	print("       -endGen iter1 - prescribe evolutions");
	print("       -readManifold - shapes are manifold");
	print("       -alignments [IdentityOnly AllXYZ AllXY AllPCA] - alignments to try");	
	print("       -singleStep prevGen [SelectToLabel, Label, AggLabeled, DeformOnly, AggDeformed, Update, PrepNextLvl] ");
	print("       -singleStep prevGen [LastStepLabel, LastStepAgg] ");
	print("       -subset subsetFile - evolve only on a subset of models ");
	print("       -writeScheduledIDs jobsfile ");
	print("		  -prefixIsAbosolutePath - set if prefix has absolute path")
	print("   EXAMPLE:");
	exmplStr = "./scripts/evolution/evolve.py PSB_dining_chair"
	exmplStr += "./Processing/SharedTemplates/ManualTemplates/v1_0/seating_4legschair.tmplt evolution ";
	exmplStr += "./Processing/SharedTemplates/Evolved/seating_4legschair.tmplt -eval eval ";
	exmplStr += "-equivLabels ./scripts/labels_equivalence/Chair.txt";
	print(exmplStr);
	exit();

workDir =  mydb.Webpage(keyword)+"/Templates/";
if (not os.path.isdir(mydb.Webpage(keyword))):
	os.system("mkdir "+mydb.Webpage(keyword)+";");

if (not os.path.isdir(workDir)):
	os.system("mkdir "+workDir+";");
workDir = workDir+outDirPfx;

if prefixIsAbosolutePath:
	workDir = outDirPfx;

#os.system("if [ -d "+workDir+" ]; then rm -r -f "+workDir+"*; fi")
#os.system("if [ ! -d "+workDir+" ]; then mkdir "+workDir+"; fi");

labelPointsEval ="";
labelPointsFlags ="";
if eval != "none":
	labelPointsEval += "-eval "+eval+" ";
if equivLabels != "none":
	labelPointsEval += "-equivLabels "+equivLabels+" ";
if readManifold:
	labelPointsFlags += "-readManifold ";
if (alignments!="none"):
	labelPointsFlags += "-alignments "+alignments+" ";


# iterate until no new nodes
nTotalShapes = len(mydb.ListModels(keyword));
while (genID < maxGen):
	# STEP 0: Setup generation names, etc..
	prevGen = workDir+"_Gen"+str(genID)+"/";
	if genID==0:
		# STEP 0.1 initialize template
		os.system("if [ ! -d "+prevGen+" ]; then mkdir "+prevGen+"; fi")
		os.system("cp "+template+" "+prevGen+"/main.tmplt");

		# STEP 0.2 initialize lists: SucceededModels.txt, RemainingModels.txt
		if evolveSubset=="all":
			mListFile = mydb.MyDBDir(keyword)+"/../ModelList.txt";
		else:
			mListFile = evolveSubset;
		if not os.path.isfile(mListFile):
			print("[ERROR] Cannot find mlist: "+mListFile);
			exit();
		os.system("cp "+mListFile+" "+prevGen+"/RemainingModels.txt; ");
	
	genID = genID + 1;
	nextGen = workDir+"_Gen"+str(genID)+"/";
	
	# check if previous template exists
	if not os.path.isfile(prevGen+"/main.tmplt"):
		if (singleStep!="none"):	# probably converged earlier
			break;
		else:
			print("[ERROR] Invalid prev gen: "+prevGen);
			break;
	commandStr = "";
	
	# STEP 1. Choose models to label
	# 	INPUT: 	(i) poorly matched models sorted by energy, (ii) sucessfully matched models (not used?)
	# 	OUTPUT:	a subset of models in list (i)
	if singleStep=="none" or singleStep=="SelectToLabel":
		cmd = evolPath+"selectSubset.py ";
		cmd +="-inListSuccess "+prevGen+"/SucceededModels.txt ";
		cmd +="-inListRemaining "+prevGen+"/RemainingModels.txt ";
		cmd +="-outListToLabel "+prevGen+"/ToLabelModels.txt ";
		cmd +="-iter "+str(genID-1)+" ";
		cmd +="-mode SelectToLabel ";
		id = mydb.ExecuteNoWait(cmd, "SelectToLabel"+str(genID));
		commandStr += cmd+"\n\n";
		if (singleStep=="none"):
			mydb.WaitForJobsInArray([id]);
		else:
			mydb.WriteJobsFile([id], writeScheduledIDs);

	# STEP 2: Label all chosen models from the collection
	toLabelFile = prevGen+"/ToLabelModels.txt";
	if (singleStep=="none" or singleStep=="Label") and os.path.isfile(toLabelFile):
		# label models in a collection
		cmd = evolPath+"labelPoints.py "+keyword+" "+prevGen+"/main.tmplt "+prevGen+" -outDirIsAbsolute ";
		cmd += "-inModelList "+prevGen+"/ToLabelModels.txt ";
		cmd +="-writeScheduledIDs "+writeScheduledIDs+" ";
		cmd += labelPointsEval + labelPointsFlags;
		os.system(cmd);
		commandStr += cmd+"\n\n";

		if (singleStep=="none"):
			mydb.WaitForJobsInFile(writeScheduledIDs);

	# STEP 3: Choose Working Set
	#   INPUT:  (i) list of models with fitting energies
	#   OUTPUT: A list of models that were 'matched reasonably well' - to learn
	if (singleStep=="none" or singleStep=="AggLabeled") and os.path.isfile(toLabelFile):
		# STEP 3.1: Aggregate errors / info from all matches
		cmd = evolPath+"aggLabelStats.py "+keyword+" "+prevGen+" LabeledModels.txt -outDirIsAbsolute ";
		cmd += "-filepfx match -inModelList "+prevGen+"ToLabelModels.txt ";
		#-processMetaInfo
		if (eval!="none"):
			cmd += "-processEval ";
		cmd += ";";
		
		# STEP 3.2 (optional): Evaluate correspondences
		if (evalCorrs!="none"):
			cmd += "./scripts/evolution/evalTemplate.py "+keyword+" "+prevGen+" "+evalCorrs+" -metric Corrs ";
			cmd += "-prefixIsAbosolutePath; ";

		# STEP 3.3: Choose a subset of models for full deformation
		cmd +=evolPath+"selectSubset.py ";
		cmd +="-inListSuccess "+prevGen+"/SucceededModels.txt ";
		cmd +="-inListLabeled "+prevGen+"/LabeledModels.txt ";
		cmd +="-outListToDeform "+prevGen+"/ToDeformModels.txt ";
		cmd +="-iter "+str(genID-1)+" ";
		cmd +="-mode SelectToDeform; ";
		#print(cmd);
		id = mydb.ExecuteNoWait(cmd, "ChooseWS"+str(genID));
		commandStr += cmd+"\n\n";

		if singleStep=="none":
			mydb.WaitForJobsInArray([id]);
		else:
			mydb.WriteJobsFile([id], writeScheduledIDs);
 
	# STEP 4: Find optimal deformations (i) no deformation penalty, (ii) deform to every model
	toDeformFile = prevGen+"/ToDeformModels.txt";
	if (singleStep=="none" or singleStep=="DeformOnly") and os.path.isfile(toDeformFile):
		cmd = evolPath+"labelPoints.py "+keyword+" "+prevGen+"/main.tmplt "+prevGen+" -outDirIsAbsolute ";
		cmd += "-deformOnly -inModelList " + prevGen + "/ToDeformModels.txt ";
		cmd += "-readMatches "+prevGen+" match.match.bin ";
		cmd += "-writeScheduledIDs "+writeScheduledIDs+" ";
		cmd += labelPointsFlags+";";
		os.system(cmd);
		commandStr += cmd+"\n\n";
		jobIDs1 = mydb.ReadJobsFile(writeScheduledIDs);
		
		cmd = "./scripts/evolution/coAlignAllPnts.py "+keyword+" "+prevGen+"/ -outDirIsAbsolute -render ";
		#cmd += "-writeScheduledIDs "+writeScheduledIDs+"; ";
		os.system(cmd);
		jobIDs2 = [];
		#jobIDs2 = mydb.ReadJobsFile(writeScheduledIDs);

		if (singleStep=="none"):
			mydb.WaitForJobsInArray(jobIDs1+jobIDs2);
		else:
			mydb.WriteJobsFile(jobIDs1 + jobIDs2, writeScheduledIDs);

	# STEP 5: Aggregate info on working set
	if (singleStep=="none" or singleStep=="AggDeformed") and os.path.isfile(toDeformFile):
		cmd = evolPath+"aggLabelStats.py "+keyword+" "+prevGen+" DeformedModels.txt -outDirIsAbsolute ";
		cmd += " -filepfx deformOnly -inModelList "+prevGen+"/ToDeformModels.txt ";
		id1 = [mydb.ExecuteNoWait(cmd, "ChooseWS"+str(genID))];
		commandStr += cmd+"\n\n";
		
		#cmd = "./scripts/evolution/renderCoAlignPnts.py "+keyword+" "+prevGen+"/ -outDirIsAbsolute";
		#id2 = [mydb.ExecuteNoWait(cmd, "renderCoAligned"+str(genID))];
		#commandStr += cmd+"\n\n";
		id2 = [];
		
		if (singleStep=="none"):
			mydb.WaitForJobsInArray(id1 + id2);
		else:
			mydb.WriteJobsFile(id1 + id2, writeScheduledIDs);

	# STEP 6: Clustering to detect new modes of variations
	#   INPUT:  (i) deformations, (ii) deformed point clouds, (iii) seed templates
	#   OUTPUT: (i) new seed gaussians, (ii) new geometries
	if singleStep=="none" or singleStep=="Update":
		cmd = "";
		if (os.path.isfile(toDeformFile)):	 # create updated template
			cmd += "./UpdateTemplate ";
			cmd += "-dbDir "+mydb.MyDBDir(keyword)+" -readGeometries ";
			cmd += "-readNMatches 100%_0_100 -updateFromAll -totalShapes "+str(nTotalShapes)+" ";
			cmd += "-readTemplate "+prevGen+"/main.tmplt -readMatches "+prevGen+"/DeformedModels.txt ";
			cmd += "-writeTemplate "+nextGen+"/main.tmplt -writeUpdateLog "+nextGen+"/update.log.txt ";
			cmd += "-writeClustered "+prevGen+"/ClusteredModels.txt ";
			if readManifold:
				cmd += "-readManifold ";
			cmd += "; "
		else:		# copy template
			cmd += "if [ ! -d "+nextGen+" ]; then mkdir "+nextGen+"; fi; ";
			cmd += "cp "+prevGen+"/main.tmplt "+nextGen+"/main.tmplt; ";
			cmd += "echo No update, just copying > "+nextGen+"update.log.txt; ";
		# render after it's been created
		cmd += "./CreateTemplate -readTemplate "+nextGen+"/main.tmplt ";
		cmd += "-shotTemplate "+nextGen+"/template.jpg ";
		cmd += "-writeSeedHierarchy "+nextGen+"/hierarchy.bin "
		id = mydb.ExecuteNoWait(cmd, "ClusterG"+str(genID));
		
		commandStr += cmd+"\n\n";
		if singleStep=="none":
			mydb.WaitForJobsInArray([id]);
		else:
			mydb.WriteJobsFile([id], writeScheduledIDs);


	# STEP 7: Update lists (succeeded / remaining models)
	if singleStep=="none" or singleStep=="PrepNextLvl":
		cmd =evolPath+"selectSubset.py ";
		cmd +="-inListSuccess "+prevGen+"/SucceededModels.txt ";
		cmd +="-inListRemaining "+prevGen+"/RemainingModels.txt ";
		cmd +="-inListExcluded "+prevGen+"/ExcludedModels.txt ";
		cmd +="-inListDeformed "+prevGen+"/DeformedModels.txt ";
		cmd +="-inListClustered "+prevGen+"/ClusteredModels.txt ";
		cmd +="-inListLabeled "+prevGen+"/LabeledModels.txt ";

		cmd +="-outListSuccess "+nextGen+"/SucceededModels.txt ";
		cmd +="-outListRemaining "+nextGen+"/RemainingModels.txt ";
		cmd +="-outListExcluded "+nextGen+"/ExcludedModels.txt ";
		cmd +="-iter "+str(genID-1)+" ";

		cmd +="-mode UpdNextLevel; ";
		id = mydb.ExecuteNoWait(cmd, "UpdLists"+str(genID));
		commandStr += cmd+"\n";
		if singleStep=="none":
			mydb.WaitForJobsInArray([id]);
		else:
			mydb.WriteJobsFile([id], writeScheduledIDs);


	lastIter = False;	# is this even relevant? NOTE: when running allBenchmarsk - does not necessarily wait...
#	print("Remaining: "+nextGen+"/RemainingModels.txt");
#	lastIter = genID<maxGen and (not os.path.isfile(nextGen+"/RemainingModels.txt"));
#	if not lastIter:
#		lastIter = (len(expdb.readModelList(nextGen+"/RemainingModels.txt"))==0);
#	print("     Last Iteration: "+str(lastIter));

	# Last iteration: re-label everything
	if ((lastIter or genID==maxGen) and singleStep=="none") or (singleStep=="LastStepLabel"):
		if not os.path.isfile(nextGen+"/main.tmplt"):
			print("[ERROR] Cannot find template: "+nextGen+"/main.tmplt"+" (see LastStepLabel step in evolve.py)");
		cmd = evolPath+"labelPoints.py "+keyword+" "+nextGen+"/main.tmplt "+nextGen+" -outDirIsAbsolute ";
		cmd +="-writeScheduledIDs "+writeScheduledIDs+" -inModelList "+evolveSubset+" ";
		# TODO: read previous matches
		cmd += labelPointsEval + labelPointsFlags;
		os.system(cmd);
		commandStr += cmd+"\n\n";
		if singleStep=="none":
			mydb.WaitForJobsInFile(writeScheduledIDs);
 
	if ((lastIter or genID==maxGen) and singleStep=="none") or (singleStep=="LastStepAgg"):
		cmd = evolPath+"aggLabelStats.py "+keyword+" "+nextGen+" LabeledModels.txt -outDirIsAbsolute ";
		cmd += "-filepfx match ";
		if (eval!="none"):
			cmd += "-processEval ";
		cmd += ";";

		if (evalCorrs!="none"):	 #(optional) evaluate correspondences
			#cmd += "cp "+prevGen+"/main.tmplt "+nextGen+"; ";
			cmd += "./scripts/evolution/evalTemplate.py "+keyword+" "+nextGen+" "+evalCorrs+" -metric Corrs ";
			cmd += "-prefixIsAbosolutePath; ";

 		id = mydb.ExecuteNoWait(cmd, "UpdLists"+str(genID));
		commandStr += cmd+"\n\n";

		if singleStep=="none":
			mydb.WaitForJobsInArray([id]);
		else:
			mydb.WriteJobsFile([id], writeScheduledIDs);

	# Write Log
	cmdLogFile = "cmd.txt";
	if singleStep!="none":
		cmdLogFile = "cmd_"+singleStep+".txt";
	f = open(prevGen+"/"+cmdLogFile, 'w');
	f.write(commandStr);
	f.close();

	if (lastIter):
		print("Converged at "+str(genID)+" / "+str(maxGen));
		break;



if (singleStep=="none"):
	if (genID==maxGen):
		print("Reached maximal generation "+str(genID));
	# Copy to GenFinal
	os.system("mv "+nextGen+" "+workDir+"_GenFinal");





