/***************************************************************************
  **************************************************************************
  
                SOFT: SO(3) Fourier transform code
  
   Peter Kostelec, Dan Rockmore
   {geelong,rockmore}@cs.dartmouth.edu
  
   Contact: Peter Kostelec
            geelong@cs.dartmouth.edu
  
  
   Copyright 2003 Peter Kostelec, Dan Rockmore
  
  
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
  
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
  
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  
   Commercial use is absolutely prohibited.
  
   See the accompanying LICENSE file for details.
  
  ************************************************************************
  ************************************************************************/



/*

  test program to see if the inverse full SO3 transform
  actually works.

  I'll generate in Mma my sample values, and will see what
  coefficients I get at the end

  input: - bandwidth bw 
         - input file names of real and imaginary parts of
	   coefficients ... should be 1/3 bw (bw^2-1) values
	   each
	 - output files for real and imaginary parts of samples,
	   each of size (2 bw)^3

  example: test_FST_SO3_Inv 16 rCoeff.dat iCoeff.dat rSam.dat iSam.dat

*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "utils.h"
#include "FST_SO3.h"
#include "csecond.h"

int main( int argc,
	  char **argv )

{
  int j, i, bw, n, n3 ;
  double *rsignal, *isignal ;
  double *rcoeffs, *icoeffs ;
  double *workspace1, *workspace2 ;
  double tstart, tstop, runtime ;
  double max_error, max_relerror, realtmp, imagtmp ;
  double origmag, tmpmag ;
  FILE *fp ;
  int whichJ ;
  double *tmpRealSam, *tmpImagSam ;
  
  if (argc < 6)
    {
      fprintf(stdout,"Usage: test_FST_SO3_Inv bw realCoeff imagCoeff ");
      fprintf(stdout,"realSam imagSam ");
      fprintf(stdout,"[trueRealSam trueImagSam]\n");
      exit(0);
    }

  bw = atoi( argv[1] );
  n = 2 * bw ;
  n3 = n * n * n ;

  /* real and imaginary parts of signal each need n^3 space */
  rsignal = ( double * ) malloc( sizeof( double ) * n3 ) ;
  isignal = ( double * ) malloc( sizeof( double ) * n3 ) ;

  /* real and imaginary parts of coeffs each need
     totalCoeffs( bw) amount of space */
  rcoeffs = ( double * ) malloc(sizeof( double ) * totalCoeffs( bw ) ) ;
  icoeffs = ( double * ) malloc(sizeof( double ) * totalCoeffs( bw ) ) ;

  /* now for LOTS OF workspace */
  workspace1 = ( double * ) malloc(sizeof( double ) * 4 * n3 ) ;
  workspace2 = ( double * ) malloc(sizeof( double ) * ( 24 * bw + 2 * bw * bw) );


  /* read in coefficient files */
  fp = fopen( argv[2] , "r" );
  for ( i = 0 ; i < totalCoeffs( bw ) ; i ++ )
    fscanf(fp, "%lf", rcoeffs+i) ;
  fclose ( fp ) ;

  fp = fopen( argv[3] , "r" );
  for ( i = 0 ; i < totalCoeffs( bw ) ; i ++ )
    fscanf(fp, "%lf", icoeffs+i) ;
  fclose ( fp ) ;


  /* turn on stopwatch */
  tstart = csecond( ) ;

  /* now do the forward transform */
  Inverse_SO3_Naive( bw,
		     rcoeffs, icoeffs,
		     rsignal, isignal,
		     workspace1, workspace2 ) ;

  /* turn off stopwatch */
  tstop = csecond( ) ;
  runtime = tstop - tstart ;

  fprintf(stdout,"runtime: %.5f seconds\n", runtime);

  /* write out samples to disk */
  fp = fopen( argv[ 4 ], "w" );
  for ( i = 0 ; i < n3 ; i ++ )
    fprintf(fp, "%.16f\n", rsignal[ i ] ) ;
  fclose( fp ) ;

  fp = fopen( argv[ 5 ], "w" );
  for ( i = 0 ; i < n3 ; i ++ )
    fprintf(fp, "%.16f\n", isignal[ i ] ) ;
  fclose( fp ) ;


  /*
    HOLD OFF THIS COMPARISON CHECKING FOR NOW ...
  */

  /* any thing to compare it with ??? */
  if ( argc > 6 )
    {

      tmpRealSam = workspace1 ;
      tmpImagSam = tmpRealSam + n3 ;

      /* read in "truth", i.e. Mma samples */

      fp = fopen( argv[ 6 ], "r") ;
      for ( j = 0 ; j < n3 ; j ++ )
	fscanf(fp, "%lf", tmpRealSam + j );
      fclose( fp ) ;
      fp = fopen( argv[ 7 ], "r") ;
      for ( j = 0 ; j < n3 ; j ++ )
	fscanf(fp, "%lf", tmpImagSam + j );
      fclose( fp ) ;

      max_error = 0.0 ;
      max_relerror = 0.0 ;
      whichJ = 0 ;
      /* now figure out errors */
      for( j = 0 ; j < totalCoeffs( bw ) ; j ++ )
	{
	  realtmp = rsignal[j] - tmpRealSam[j] ;
	  imagtmp = isignal[j] - tmpImagSam[j] ;
	  origmag = sqrt((tmpRealSam[j]*tmpRealSam[j]) +
			 (tmpImagSam[j]*tmpImagSam[j]));
	  tmpmag = sqrt((realtmp*realtmp) +
			(imagtmp*imagtmp));
	  if ( max_error < tmpmag )
	    whichJ = j ;
	  max_error = MAX( max_error , tmpmag ) ;
	  max_relerror = MAX( max_relerror , tmpmag/(origmag+pow(10.0,-50.0)));
	}
      fprintf(stderr,"whichJ = %d\n", whichJ);
      fprintf(stderr,"r-o error\t = %.12e\n", max_error);
      fprintf(stderr,"(r-o)/o error\t = %.12e\n\n", max_relerror);
    }


  /* free up memory (and there's lots of it) */
  free( workspace2 );
  free( workspace1 );
  free( icoeffs );
  free( rcoeffs );
  free( isignal );
  free( rsignal );

  return 0 ;
}
