// Header file for WignerD transform class



// Class definition

class R3WignerD {
public:
  // Constructors
  R3WignerD(int bandwidth = 0);
  R3WignerD(const R3WignerD& w);
  ~R3WignerD(void);

  // WignerD property functions
  int Bandwidth(void) const;
  int NCoefficients(void) const;

  // Coefficient access functions
  const double *Coefficients(void) const;
  double RealCoefficient(int L, int M1, int M2) const;
  double ImaginaryCoefficient(int L, int M1, int M2) const;

  // WignerD manipulation functions
  void Clear(void);
  void AddSphericalHarmonics(const R2SphericalHarmonics& harmonics1, const R2SphericalHarmonics& harmonics2);
  void SetRealCoefficient(int L, int M1, int M2, RNScalar value);
  void SetImaginaryCoefficient(int L, int M1, int M2, RNScalar value);

  // I/O functions
  int ReadFile(const char *filename);
  int WriteFile(const char *filename) const;
  int Read(FILE *fp = NULL);
  int Write(FILE *fp = NULL) const;
  int Print(FILE *fp = NULL) const;

  // WignerD transform
  int InverseWignerDTransform(double *euler_grid) const;

  // Utility functions
  int Index(int L, int M1, int M2) const;

private:
  int bandwidth;
  int ncoefficients;
  double *coefficients;
};



// Inline functions

inline int R3WignerD::
Bandwidth(void) const
{
  // Return bandwidth
  return bandwidth;
}



inline int R3WignerD::
NCoefficients(void) const
{
  // Return ncoefficients
  return ncoefficients;
}



inline const double *R3WignerD::
Coefficients(void) const
{
  // Return coefficients
  return coefficients;
}



inline RNScalar R3WignerD::
RealCoefficient(int L, int M1, int M2) const
{
  // Return real coefficient
  return coefficients[Index(L, M1, M2)];
}



inline RNScalar R3WignerD::
ImaginaryCoefficient(int L, int M1, int M2) const
{
  // Return imaginary coefficient
  return coefficients[ncoefficients/2 + Index(L, M1, M2)];
}



