//
// $Id: BindableNodes.h,v 1.1 2000/01/17 18:47:59 min Exp min $
//

#ifndef __BINDABLENOTES_H
#define __BINDABLENOTES_H

#include "VrmlNode.h"
#include "Fields.h"





class Background : public VrmlNode
{

public:

  Background() { 
    set_node_name("Background");
    set_node_type(BackgroundType);

    add_field("set_bind", &set_bind, eventIn);
    add_field("groundAngle", &groundAngle, exposedField);
    add_field("groundColor", &groundColor, exposedField);
    add_field("backUrl", &backUrl, exposedField);
    add_field("bottomUrl", &bottomUrl, exposedField);
    add_field("frontUrl", &frontUrl, exposedField);
    add_field("leftUrl", &leftUrl, exposedField);
    add_field("rightUrl", &rightUrl, exposedField);
    add_field("topUrl", &topUrl, exposedField);
    add_field("skyAngle", &skyAngle, exposedField);
    add_field("skyColor", &skyColor, exposedField);
    add_field("isBound", &isBound, eventOut);

    SFColor *my_colour = new SFColor;
    my_colour->set_colour(0, 0, 0);
    skyColor.add(my_colour);
  }  // constructor
  ~Background() {}


  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return ReturnVrmlNode; }


private:

  SFBool set_bind;
  MFFloat groundAngle;
  MFColor groundColor;
  MFString backUrl;
  MFString bottomUrl;
  MFString frontUrl;
  MFString leftUrl;
  MFString rightUrl;
  MFString topUrl;
  MFFloat skyAngle;
  MFColor skyColor;
  SFBool isBound;


};  // Background class





class Fog : public VrmlNode
{

public:

  Fog() { 
    set_node_name("Fog");
    set_node_type(FogType);

    add_field("color", &color, exposedField);
    add_field("fogType", &fogType, exposedField);
    add_field("visibilityRange", &visibilityRange, exposedField);
    add_field("set_bind", &set_bind, eventIn);
    add_field("isBound", &isBound, eventOut);

    color.set_colour(1, 1, 1);
    fogType.set_string("LINEAR");
    visibilityRange.set_float(0);

  }  // constructor

  ~Fog() {}

  void sendEventIn(int field_id, void *new_value);


private:

  SFColor color;
  SFString fogType;
  SFFloat visibilityRange;
  SFBool set_bind;
  SFBool isBound;


};  // Fog class





class NavigationInfo : public VrmlNode
{

public:

  NavigationInfo() { 
    set_node_name("NavigationInfo");
    set_node_type(NavigationInfoType);

    add_field("set_bind", &set_bind, eventIn);
    add_field("avatarSize", &avatarSize, exposedField);
    add_field("headlight", &headlight, exposedField);
    add_field("speed", &speed, exposedField);
    add_field("type", &type, exposedField);
    add_field("visibilityLimit", &visibilityLimit, exposedField);
    add_field("isBound", &isBound, eventOut);

    set_bind.set_boolean(FALSE);
    SFFloat *collision_distance = new SFFloat(0.25);
    avatarSize.add(collision_distance);
    SFFloat *viewer_height = new SFFloat(1.6);
    avatarSize.add(viewer_height);
    SFFloat *step_height = new SFFloat(0.75);
    avatarSize.add(step_height);
    headlight.set_boolean(TRUE);
    speed.set_float(1.0);
    SFString *navigation_type = new SFString("WALK");
    type.add(navigation_type);
    visibilityLimit.set_float(0.0);
    isBound.set_boolean(FALSE);

  }  // constructor

  ~NavigationInfo() {}


  USE_ActionType get_USE_action() { return ReturnVrmlNode; }

  void sendEventIn(int field_id, void *new_value);


private:

  SFBool set_bind;
  MFFloat avatarSize;
  SFBool headlight;
  SFFloat speed;
  MFString type;
  SFFloat visibilityLimit;
  SFBool isBound;


};  // NavigationInfo class





class Viewpoint : public VrmlNode
{

public:

  Viewpoint() { 
    set_node_name("Viewpoint");
    set_node_type(ViewpointType);

    add_field("set_bind", &set_bind, eventIn);
    add_field("fieldOfView", &fieldOfView, exposedField);
    add_field("jump", &jump, exposedField);
    add_field("orientation", &orientation, exposedField);
    add_field("position", &position, exposedField);
    add_field("description", &description, field);
    add_field("bindTime", &bindTime, eventOut);
    add_field("isBound", &isBound, eventOut);

    set_bind.set_boolean(FALSE);
    fieldOfView.set_float(0.785398);
    jump.set_boolean(TRUE);
    orientation.set_axis(0, 0, 1);
    orientation.set_angle(0);
    position.set_vector(0, 0, 10);
    description.set_string("(no name)");

  }  // constructor

  ~Viewpoint() {}


  void add_Ring();

  int move_to(R3Session *session_ptr);

  USE_ActionType get_USE_action() { return ReturnVrmlNode; }

  void sendEventIn(int field_id, void *new_value);


private:

  SFBool set_bind;
  SFFloat fieldOfView;
  SFBool jump;
  SFRotation orientation;
  SFVec3f position;
  SFString description;
  SFTime bindTime;
  SFBool isBound;


};  // Viewpoint class





class ViewpointStack;
extern ViewpointStack viewpoint_stack;

class ViewpointStack {

public:

  ViewpointStack() {
    current_index = -1;
  }  // constructor
  ~ViewpointStack() {}


  void Push(Viewpoint *viewpoint_ptr) { 
    my_stack.Insert(viewpoint_ptr);
    current_index = 0;
  }  // Push

  int home_viewpoint(R3Session *session_ptr);
  int prev_viewpoint(R3Session *session_ptr);
  int next_viewpoint(R3Session *session_ptr);


private:

  RNArray<Viewpoint *> my_stack;
  int current_index;


};  // ViewpointStack class


#endif


