//
// $Id: Fields.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include "Fields.h"
#include "Runtime.h"
#include "VrmlNode.h"
#include "PROTONodes.h"
#include "parse_nodes.h"
#include "GroupingNodes.h"





Field::Field()
{
  outgoing_routes = NULL;
}  // constructor



Field::~Field()
{
}  // destructor



void
Field::parse()
{
  cerr << "Error: called Field's parse function" << endl;
}  // Field::parse



void
Field::add_route(Event *new_route)
{
  if (!outgoing_routes) outgoing_routes = new RNArray<Event *>;

  outgoing_routes->Insert(new_route);

}  // Field::add_route



int
Field::sendEventOut()
{
  if (!outgoing_routes) return 0;

  int nr_routes = outgoing_routes->NEntries();
  for(int i=0; i<nr_routes; i++) {
    event_queue.add_event(outgoing_routes->Kth(i));
  }  // for
  return 1;

} // Field::sendEventOut
				    




//
// MFField
//

MFField::~MFField()
{
#ifdef __DEBUG_DESTRUCTORS__
  cout << "MFField destructor: " << get_field_var_type_string() << endl;
#endif
  int nr_entries = array->NEntries();
  for(int i=0; i<nr_entries; i++) delete array->Kth(i);
  delete array;
}  // destructor


      
MFField::MFField(MFField& other)
{
  int nr_elements = other.get_size();
  array = new RNArray<Field *>;

  for(int i=0; i<nr_elements; i++) {
    Field *field_ptr = other.get_field(i);
    Field *new_ptr = field_ptr->create_copy();
    add(new_ptr);
  }  // for
}  // MFField copy constructor



char *
MFField::to_string(char *buffer)
{
  int nr_entries = get_size();

  buffer += sprintf(buffer, " [ ");
  for(int i=0; i<nr_entries; i++) {
    buffer = array->Kth(i)->to_string(buffer);
  }
  buffer += sprintf(buffer, " ] ");
  return buffer;
}  // MFField::to_string



void
MFField::copy_from(MFField *other)
{
  clear();
  int nr_fields = other->get_size();
  for(int i=0; i<nr_fields; i++) {
    add(other->get_field(i)->create_copy());
    // also do some Ring stuff here?
  }  // for

}  // MFField::copy_from



void
MFField::clear()
{
  int nr_fields = get_size();
  for(int i=0; i<nr_fields; i++)
    delete array->Kth(i);
  array->Empty();
}  // MFField::clear




//
// SFBool
//

void
SFBool::parse()
{
  char token[MAX_TOKEN_LENGTH];
  cur_vrml_file->get_token(token);
  if (strcmp(token, "IS") == 0)
    my_value = ((SFBool *) PROTOVarTable.handle_IS())->get_boolean();
  else if (strcmp(token, "TRUE") == 0) my_value = TRUE;
  else if (strcmp(token, "FALSE") == 0) my_value = FALSE;
  else {
    cerr << "Error: wrong value " << token << " for SFBool" << endl;
  }
}  // SFBool::parse



char *
SFBool::to_string(char *buffer)
{
  if (my_value) return buffer + sprintf(buffer, " TRUE ");
  else return buffer + sprintf(buffer, " FALSE ");
}  // SFBool::to_string



void
SFBool::sendEventIn(void *new_value)
{ 
  my_value = ((SFBool *) new_value)->get_boolean();
}  // SFBool::sendEventIn



Field *
SFBool::operator=(Field *other)
{
  switch (other->get_field_var_type()) {
    case SFBoolType: 
      my_value = ((SFBool *) other)->get_boolean(); break;
    case SFInt32Type:
      if (((SFInt32 *) other)->get_integer()) my_value = TRUE;
      else my_value = FALSE; break;
    case SFFloatType: 
      if (((SFFloat *) other)->get_float()) my_value = TRUE;
      else my_value = FALSE; break;
    case SFTimeType: 
      if (((SFTime *) other)->get_time()) my_value = TRUE;
      else my_value = FALSE; break;
    default:
      binary_operator_error("=", other);
  }  // switch
  return this;
}  // SFBool::operator=



Field *
SFBool::operator==(Field *other)
{
  if (other->get_field_var_type() == SFBoolType)
    return new SFBool(my_value == ((SFBool *) other)->get_boolean());
  
  binary_operator_error("==", other);
  return NULL;
}  // SFBool::operator==



Field *
SFBool::operator!=(Field *other)
{
  if (other->get_field_var_type() == SFBoolType)
    return new SFBool(my_value != ((SFBool *) other)->get_boolean());

  binary_operator_error("!=", other);
  return NULL;
}  // SFBool::operator!=



Field *
SFBool::operator&&(Field *other)
{
  if (other->get_field_var_type() == SFBoolType)
    return new SFBool(my_value && ((SFBool *) other)->get_boolean());

  binary_operator_error("&&", other);
  return NULL;
}  // SFBool::operator&&



Field *
SFBool::operator||(Field *other)
{
  if (other->get_field_var_type() == SFBoolType)
    return new SFBool(my_value || ((SFBool *) other)->get_boolean());

  binary_operator_error("||", other);
  return NULL;
}  // SFBool::operator||



Field *
SFBool::operator&(Field *other)
{
  if (other->get_field_var_type() == SFBoolType)
    return new SFBool(my_value & ((SFBool *) other)->get_boolean());

  binary_operator_error("&", other);
  return NULL;
}  // SFBool::operator&



Field *
SFBool::operator|(Field *other)
{
  if (other->get_field_var_type() == SFBoolType)
    return new SFBool(my_value | ((SFBool *) other)->get_boolean());

  binary_operator_error("|", other);
  return NULL;
}  // SFBool::operator|



Field *
SFBool::operator^(Field *other)
{
  if (other->get_field_var_type() == SFBoolType)
    return new SFBool(my_value ^ ((SFBool *) other)->get_boolean());

  binary_operator_error("^", other);
  return NULL;
}  // SFBool::operator^





//
// SFColor
//

void
SFColor::parse()
{
  char token[MAX_TOKEN_LENGTH];

  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);
    my_colour = ((SFColor *) PROTOVarTable.handle_IS())->get_colour();
  }
  else {
    SFFloat red, green, blue;
    red.parse();
    green.parse();
    blue.parse();
    my_colour.R(red.get_float());
    my_colour.G(green.get_float());
    my_colour.B(blue.get_float());
  }
}  // SFColor::parse



void
SFColor::sendEventIn(void *new_value)
{
  set_colour(((SFColor *) new_value)->get_colour());
}  // SFColor::set_value



Field *
SFColor::operator=(Field *other)
{
  my_colour = ((SFColor *) other)->get_colour();
  //  cout << "Set my_colour to: " << my_colour.R() << ", " << my_colour.G()
  //    << ", " << my_colour.B() << endl;
  //  cout << "Got it from: " << ((SFColor *) other)->get_colour().R() << ", "
  //    << ((SFColor *) other)->get_colour().G() << ", "
  //    << ((SFColor *) other)->get_colour().B() << endl;
  return this;
}  // SFColor::operator=





//
// SFFloat
//

void
SFFloat::parse()
{
  char token[MAX_TOKEN_LENGTH];

  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);
    my_float = ((SFFloat *) PROTOVarTable.handle_IS())->get_float();
  }
  else {
    cur_vrml_file->get_number(token);
    my_float = atof(token);
  }

}  // SFFloat::parse



char *
SFFloat::to_string(char *buffer)
{
  return buffer + sprintf(buffer, " %.12g ", my_float);
}  // SFFloat::to_string



double
SFFloat::to_float(Field *other)
{
  switch (other->get_field_var_type()) {
    case SFBoolType:
      return ((SFBool *) other)->get_boolean();
    case SFInt32Type:
      return ((SFInt32 *) other)->get_integer();
    case SFFloatType:
      return ((SFFloat *) other)->get_float();
    case SFTimeType:
      return ((SFTime *) other)->get_time();
    default:
      conversion_error(other);
  }  // switch
  return -1;
}  // SFFloat::to_float


//
// operators
//

//
// unary
//

Field *
SFFloat::operator-()
{
  return new SFFloat(-my_float);
}  // SFFloat::operator- (unary)


    
Field *
SFFloat::operator!()
{
  return new SFFloat(!my_float);
}  // SFFloat::operator!



Field *
SFFloat::operator=(Field *other)
{
  my_float = to_float(other);
  //  cout << "SFFloat::operator= : " << my_float << " is new value" << endl;
  return this;
}  // SFBool::operator=



Field *
SFFloat::operator==(Field *other)
{
  double other_float = to_float(other);
  return new SFBool(my_float == other_float);
}  // SFFloat::operator==



Field *
SFFloat::operator!=(Field *other)
{
  double other_float = to_float(other);
  return new SFBool(my_float != other_float);
}  // SFFloat::operator!=



Field *
SFFloat::operator<(Field *other)
{
  double other_float = to_float(other);
  return new SFBool(my_float < other_float);
}  // SFFloat::operator<



Field *
SFFloat::operator<=(Field *other)
{
  double other_float = to_float(other);
  return new SFBool(my_float <= other_float);
}  // SFFloat::operator<=



Field *
SFFloat::operator>=(Field *other)
{
  double other_float = to_float(other);
  return new SFBool(my_float >= other_float);
}  // SFFloat::operator>=



Field *
SFFloat::operator>(Field *other)
{
  double other_float = to_float(other);
  return new SFBool(my_float > other_float);
}  // SFFloat::operator>



Field *
SFFloat::operator+(Field *other)
{
  double other_float = to_float(other);
  return new SFFloat(my_float + other_float);
}  // SFFloat::operator+



Field *
SFFloat::operator-(Field *other)
{
  double other_float = to_float(other);
  //  cout << "SFFloat::operator- : " << my_float << " - " << other_float
  //       << " = " << my_float - other_float << endl;
  return new SFFloat(my_float - other_float);
}  // SFFloat::operator-



Field *
SFFloat::operator+=(Field *other)
{
  my_float += to_float(other);
  return this;
}  // SFFloat::operator +=



Field *
SFFloat::operator-=(Field *other)
{
  my_float -= to_float(other);
  return this;
}  // SFFloat::operator -=



Field *
SFFloat::operator*(Field *other)
{
  double other_float = to_float(other);
  return new SFFloat(my_float * other_float);
}  // SFFloat::operator*



Field *
SFFloat::operator/(Field *other)
{
  double other_float = to_float(other);
  return new SFFloat(my_float / other_float);
}  // SFFloat::operator/



Field *
SFFloat::operator++()
{
  my_float++;
  return this;
}  // SFFloat::operator++ (prefix)



Field *
SFFloat::operator--()
{
  my_float--;
  return this;
}  // SFFloat::operator-- (prefix)



Field *
SFFloat::operator++(int dummy)
{
  my_float++;
  return new SFFloat(my_float - 1);
}  // SFFloat::operator++ (postfix)



Field *
SFFloat::operator--(int dummy)
{
  my_float--;
  return new SFFloat(my_float + 1);
}  // SFFloat::operator-- (postfix)





//
// SFImage
//

void
SFImage::parse()
{
  int width, height, depth;

  char token[MAX_TOKEN_LENGTH];
  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);
    // create a copy of the image
    R2Image *image_ptr = 
      ((SFImage *) PROTOVarTable.handle_IS())->get_image();
    my_image = copy_image(image_ptr);
  }
  else {
    SFInt32 vrml_int;

    vrml_int.parse();
    width = vrml_int.get_integer();
    vrml_int.parse();
    height = vrml_int.get_integer();
    vrml_int.parse();
    depth = vrml_int.get_integer();

    int num_values = width * height * depth;
    byte *data;

    if (num_values == 0) {  // shouldn't happen
      data = new byte[1];
      data[0] = 0;
      width = height = depth = 1;
    }

    data = new byte[num_values];
    my_image = new R2ColorImage(width, height, depth, data);

    for(int i=0; i<num_values; i++) {
      vrml_int.parse();
      data[i] = vrml_int.get_integer();
    }  // for
  }
}  // SFImage::parse



char *
SFImage::to_string(char *buffer)
{
  // get width, height, numcomponents
  int width = my_image->Width();
  int height = my_image->Height();
  int depth = my_image->Depth();

  buffer += sprintf(buffer, " %d %d %d\n", width, height, depth);

  long size = width * height * depth;
  unsigned char *data = (unsigned char *) my_image->Data();

  for(long i=0; i<size; i++) {
    buffer += sprintf(buffer, "%d ", data[i]);
  }
  return buffer;
}  // SFImage::to_string


    

R2Image*
SFImage::copy_image(R2Image *source_image)
{
  int width = source_image->Width();
  int height = source_image->Height();
  int depth = source_image->Depth();

  // this constructor will allocate memory for the new image
  R2Image *new_image = new R2ColorImage(width, height, depth);
  if (new_image)
    memcpy((void *) new_image->Data(), source_image->Data(), 
	   width * height * depth);
  else
    cerr << "Error: could not allocate memory for SFImage" << endl;

  return new_image;
}  // SFImage::copy_image





//
// SFInt32
//

void
SFInt32::parse()
{
  char token[MAX_TOKEN_LENGTH];
  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);
    my_integer = ((SFInt32 *) PROTOVarTable.handle_IS())->get_integer();
  }
  else {
    cur_vrml_file->get_number(token);
    my_integer = atoi(token);
  }

}  // SFInt32::parse



int
SFInt32::to_int(Field *other)
{
  switch (other->get_field_var_type()) {
    case SFBoolType:
      return ((SFBool *) other)->get_boolean();
    case SFInt32Type:
      return ((SFInt32 *) other)->get_integer();
    case SFFloatType:
      return (int)((SFFloat *) other)->get_float();
    case SFTimeType:
      return (int)((SFTime *) other)->get_time();
    default:
      conversion_error(other);
  }  // switch
  return -1;
}  // SFInt32::to_int



Field *
SFInt32::operator=(Field *other)
{
  my_integer = to_int(other);
  return this;
}  // SFInt32::operator=



Field *
SFInt32::operator==(Field *other)
{
  int other_int = to_int(other);
  return new SFBool(my_integer == other_int);
}  // SFInt32::operator==



Field *
SFInt32::operator!=(Field *other)
{
  int other_int = to_int(other);
  return new SFBool(my_integer != other_int);
}  // SFInt32::operator!=



Field *
SFInt32::operator<(Field *other)
{
  int other_int = to_int(other);
  return new SFBool(my_integer < other_int);
}  // SFInt32::operator<



Field *
SFInt32::operator<=(Field *other)
{
  int other_int = to_int(other);
  return new SFBool(my_integer <= other_int);
}  // SFInt32::operator<=



Field *
SFInt32::operator>=(Field *other)
{
  int other_int = to_int(other);
  return new SFBool(my_integer >= other_int);
}  // SFInt32::operator>=



Field *
SFInt32::operator>(Field *other)
{
  int other_int = to_int(other);
  return new SFBool(my_integer > other_int);
}  // SFInt32::operator>



Field *
SFInt32::operator+(Field *other)
{
  int other_int = to_int(other);
  return new SFInt32(my_integer + other_int);
}  // SFInt32::operator+



Field *
SFInt32::operator-(Field *other)
{
  int other_int = to_int(other);
  return new SFInt32(my_integer - other_int);
}  // SFInt32::operator-



Field *
SFInt32::operator*(Field *other)
{
  int other_int = to_int(other);
  return new SFInt32(my_integer * other_int);
}  // SFInt32::operator*



Field *
SFInt32::operator/(Field *other)
{
  int other_int = to_int(other);
  return new SFInt32(my_integer / other_int);
}  // SFInt32::operator/





//
// SFNode
//

SFNode::~SFNode()
{
  if (my_node) {
#ifdef __DEBUG_DESTRUCTORS__
    cout << "SFNode destructor for a " << my_node->get_node_name() << endl;
#endif
    // check if it has a name
    R3Node *ring_ptr = my_node->get_ring_node_ptr();
    if (ring_ptr) {
      char *name = (char *) ring_ptr->Name();
      if (name) {
	//	cout << "SFNode: my VrmlNode has a name: " << name << endl;
	// if so, find the name in my_DEF_table
	VrmlNode *vrmlnode_ptr = my_DEF_table.find_DEF(name);
	// if found, delete it 
	if (vrmlnode_ptr) {
	  //	  cout << "  found it in the DEFTable" << endl;
	  delete my_node;
	  // and remove name from my_DEF_table
	  my_DEF_table.remove(name);
	}
      }
    }
    else {
      delete my_node;
    }
  }
  if (my_string) delete my_string;
}  // destructor



void
SFNode::parse()
{
#ifdef __DEBUG__
  cout << "Entering SFNode::parse" << endl;
#endif

  char token[MAX_TOKEN_LENGTH];
  cur_vrml_file->get_token(token);

  if (strcmp(token, "IS") == 0) {
    SFNode *SFNode_ptr = (SFNode *) PROTOVarTable.handle_IS();
    my_string = SFNode_ptr->get_my_string();
    parse_my_string();
  }  
  else
    my_node = parse_node(token);

}  // SFNode::parse



void
SFNode::parse_my_string()
{
#ifdef __DEBUG_PARSE__
  cout << "Entering SFNode::parse_my_string" << endl;
#endif

  char *save_parse_source = cur_vrml_file->get_parse_source();
  cur_vrml_file->set_parse_source(my_string);

  char token[MAX_TOKEN_LENGTH];
  cur_vrml_file->get_token(token);
  my_node = parse_node(token);

  // parse source might have been something other than a file
  cur_vrml_file->set_parse_source(save_parse_source);  

}  // SFNode::parse_my_string



void
SFNode::parse_and_store_string()
{
  cur_vrml_file->set_just_parsing(TRUE);
  my_string = new char[MAX_NODE_TEXT_LENGTH];
  my_string[0] = ' ';  // trick to help MFNode parse_and_store_string
  my_string++;
  cur_vrml_file->start_storage(my_string);
  my_string--;

  char token[MAX_TOKEN_LENGTH];
  cur_vrml_file->get_token(token);
  parse_node(token);

  cur_vrml_file->stop_storage();
  cur_vrml_file->set_just_parsing(FALSE);

}  // SFNode::parse_and_store_string





//
// SFRotation
//

void
SFRotation::parse()
{
  char token[MAX_TOKEN_LENGTH];
  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);
    SFRotation *rotation_ptr = (SFRotation *) PROTOVarTable.handle_IS();
    axis = rotation_ptr->get_axis();
    angle = rotation_ptr->get_angle();
  }
  else {
    SFFloat x, y, z, new_angle;
    x.parse(), y.parse(), z.parse(), new_angle.parse();
    axis.Reset(x.get_float(), y.get_float(), z.get_float());
    angle = new_angle.get_float();
  }
}  // SFRotation::parse




//
// just for value_changed =
// in the OrientationInterpolator
//
Field * 
SFRotation::operator=(Field* other)
{
  axis = ((SFRotation *) other)->get_axis();
  angle = ((SFRotation *) other)->get_angle();

  return this;
  
}  // SFRotation::operator=





//
// SFString
//

SFString::SFString(char *new_string)
{
  set_string(new_string);
}  // SFString constructor with char *



SFString::SFString(SFString& other)
{
  set_string(other.get_string());
}  // SFString copy constructor



SFString::~SFString()
{
#ifdef __DEBUG_DESTRUCTORS__
  cout << "SFString destructor: " << my_string << endl;
#endif
  if (my_string) delete my_string;
}  // destructor



void
SFString::set_string(char *new_string)
{
  my_string = new char[strlen(new_string) + 1];
  strcpy(my_string, new_string);
  current_length = strlen(my_string);
}  // SFString::set_string



void
SFString::parse()
{
  parse(DOUBLE_QUOTE);
}  // SFString::parse



void
SFString::parse(char quote_char)
{
  char token[MAX_TOKEN_LENGTH];

  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);
    char *string_ptr = 
      ((SFString *) PROTOVarTable.handle_IS())->get_string();
    delete my_string;
    my_string = new char[strlen(string_ptr) + 1];
    strcpy(my_string, string_ptr);
    return;
  }
  else if (token[0] == quote_char) {
    cur_vrml_file->get_token(token);

    int done = 0;
    int length = 0;
    char k;

    while (!done) {
      k = cur_vrml_file->get_buffered_char();
      if ((k == quote_char) || (k == MY_EOF)) done = 1;
      else {
	if (k == BACKSLASH) k = cur_vrml_file->get_buffered_char();
	else {
	  my_string[length] = k;
	  length++;
	  if (length == (current_length - 1)) {  // reallocate twice the space
	    my_string[length] = '\0';
	    current_length *= 2;
	    //	  cout << "Current length is " << current_length << endl;
	    char *new_ptr = new char[current_length];
	    strcpy(new_ptr, my_string);
	    delete my_string;
	    my_string = new_ptr;
	  }
	}
      }
    }  // while

    my_string[length] = '\0';
#ifdef __DEBUG_TOKENS__
    cout << ">> STRING: [" << my_string << "]" << endl;
#endif
  }
  else {
    cur_vrml_file->print_error_line();
    cerr << "Expected a string, got a [" << token << "]" << endl;
  }

}  // SFString::parse(char quote_char)





//
// SFTime
//

//
// SFVec2f
//

void
SFVec2f::parse()
{
  char token[MAX_TOKEN_LENGTH];
  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);
    my_vector = ((SFVec2f *) PROTOVarTable.handle_IS())->get_point();
  }
  else {
    SFFloat x, y;
    x.parse(), y.parse();
    my_vector.X(x.get_float());
    my_vector.Y(y.get_float());
  }

}  // SFVec2f::parse





//
// SFVec3f
//

void
SFVec3f::parse()
{
  char token[MAX_TOKEN_LENGTH];
  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);
    my_vector = ((SFVec3f *) PROTOVarTable.handle_IS())->get_point();
  }
  else {
    SFFloat x, y, z;
    x.parse(), y.parse(), z.parse();
    my_vector.Reset(x.get_float(), y.get_float(), z.get_float());
  }

}  // SFVec3f::parse





//
// MFColor
//

void
MFColor::parse()
{
  char token[MAX_TOKEN_LENGTH];

  SFColor *my_color_ptr;

  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);
    MFColor *mfcolor_ptr = (MFColor *) PROTOVarTable.handle_IS();
    copy_from(mfcolor_ptr);
    return;
  }

  if (token[0] != OPEN_BRACKET) {
    my_color_ptr = new SFColor;
    my_color_ptr->parse();
    add(my_color_ptr);
  }
  else {
    int done = 0;
    cur_vrml_file->get_token(token);

    while (!done) {
      cur_vrml_file->lookahead_token(token);
      if (token[0] == CLOSE_BRACKET) done = 1;
      else {
	my_color_ptr = new SFColor;
	my_color_ptr->parse();
	add(my_color_ptr);
      }
    }  // while
    cur_vrml_file->get_token(token);
  }

}  // MFColor::parse





//
// MFFloat
//

void
MFFloat::parse()
{
  char token[MAX_TOKEN_LENGTH];

  SFFloat *my_float_ptr;

  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);
    MFFloat *mffloat_ptr = (MFFloat *) PROTOVarTable.handle_IS();
    copy_from(mffloat_ptr);
    return;
  }

  if (token[0] != OPEN_BRACKET) {  // MFFloat, but just one Float...
    my_float_ptr = new SFFloat;
    my_float_ptr->parse();
    add(my_float_ptr);
  }
  else {
    int done = 0;
    cur_vrml_file->get_token(token);

    while (!done) {
      cur_vrml_file->lookahead_token(token);
      if (token[0] == CLOSE_BRACKET) done = 1;
      else {
	my_float_ptr = new SFFloat;
	my_float_ptr->parse();
	add(my_float_ptr);
      }
    }  // while
    cur_vrml_file->get_token(token);  // get the close bracket
  }
}  // MFFloat::parse





//
// MFInt32
//

void
MFInt32::parse()
{
  char token[MAX_TOKEN_LENGTH];

  SFInt32 *my_int32_ptr;

  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);
    MFInt32 *mfint32_ptr = (MFInt32 *) PROTOVarTable.handle_IS();
    copy_from(mfint32_ptr);
    return;
  }

  if (token[0] != OPEN_BRACKET) {
    my_int32_ptr = new SFInt32;
    my_int32_ptr->parse();
    add(my_int32_ptr);
  }
  else {
    int done = 0;
    cur_vrml_file->get_token(token);

    while (!done) {
      cur_vrml_file->lookahead_token(token);
      if (token[0] == CLOSE_BRACKET) done = 1;
      else {
	my_int32_ptr = new SFInt32;
	my_int32_ptr->parse();
	add(my_int32_ptr);
      }
    }  // while
    cur_vrml_file->get_token(token);
  }
}  // MFInt32::parse





//
// MFNode
//

void
MFNode::parse()
{
  char token[MAX_TOKEN_LENGTH];

  SFNode *my_node_ptr;

  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {  // could be an IS
    cur_vrml_file->get_token(token);
    MFNode *mfnode_ptr = (MFNode *) PROTOVarTable.handle_IS();
    copy_from(mfnode_ptr);
    // now we have a list of SFNode pointers, parse their strings
    int nr_nodes = get_size();
    for(int i=0; i<nr_nodes; i++) {
      ((SFNode *) get_field(i))->parse_my_string();
    }  // for
    return;
  }
  else if (strcmp(token, "Inline") == 0) {  // could be an Inline
    cur_vrml_file->get_token(token);
    Inline *inline_ptr = new Inline;
    inline_ptr->init_Ring();
    inline_ptr->parse();
    inline_ptr->add_Ring();
    copy_from(inline_ptr->get_children());
    //    delete inline_ptr;  // don't want to delete grouping_node_ptr
    // which is done in GroupingNode destructor...
    return;
  }
  
  if (token[0] != OPEN_BRACKET) {
    cur_vrml_file->get_token(token);
    my_node_ptr = new SFNode();
    my_node_ptr->set_node_ptr(parse_node(token));
    add(my_node_ptr);
  }
  else {
    int done = 0;
    cur_vrml_file->get_token(token);

    while (!done) {
      cur_vrml_file->get_token(token);
      if (token[0] == CLOSE_BRACKET) done = 1;
      else {
	my_node_ptr = new SFNode();
	my_node_ptr->set_node_ptr(parse_node(token));
	add(my_node_ptr);
      }
    }  // while
  }  // else, we got an open bracket

}  // MFNode::parse



void
MFNode::parse_and_store_string()
{
  int k = cur_vrml_file->skip_whitespace();

  if (k != OPEN_BRACKET) {
    SFNode *node_ptr = new SFNode();
    node_ptr->parse_and_store_string();
    // SFNode reserved one space for a possible skipped first character...
    char *node_string = node_ptr->get_my_string();
    node_string[0] = k;
    add(node_ptr);
  }
  else {
    int done = 0;
    char token[MAX_TOKEN_LENGTH];
    cur_vrml_file->set_current_char(' ');  // get rid of the open bracket

    while (!done) {
      cur_vrml_file->get_token(token);
      if (token[0] == CLOSE_BRACKET) done = 1;
      else {
	SFNode *node_ptr = new SFNode();
	node_ptr->parse_and_store_string();
	add(node_ptr);
      }
    }  // while
  }  // else, we got an open bracket
}  // MFNode::parse_and_store_string





//
// MFRotation
//

void
MFRotation::parse()
{
  char token[MAX_TOKEN_LENGTH];

  SFRotation *my_rotation_ptr;

  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);
    MFRotation *mfrotation_ptr = (MFRotation *) PROTOVarTable.handle_IS();
    copy_from(mfrotation_ptr);
    return;
  }

  if (token[0] != OPEN_BRACKET) {
    my_rotation_ptr = new SFRotation;
    my_rotation_ptr->parse();
    add(my_rotation_ptr);
  }
  else {
    int done = 0;
    cur_vrml_file->get_token(token);

    while (!done) {
      cur_vrml_file->lookahead_token(token);
      if (token[0] == CLOSE_BRACKET) done = 1;
      else {
	my_rotation_ptr = new SFRotation;
	my_rotation_ptr->parse();
	add(my_rotation_ptr);
      }
    }  // while
    cur_vrml_file->get_token(token);
  }

}  // MFRotation::parse





//
// MFString
//

void
MFString::parse()
{
  char token[MAX_TOKEN_LENGTH];

  SFString *my_string_ptr;

  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);
    MFString *mfstring_ptr = (MFString *) PROTOVarTable.handle_IS();
    copy_from(mfstring_ptr);
    return;
  }

  if (token[0] != OPEN_BRACKET) {
    my_string_ptr = new SFString;
    my_string_ptr->parse();
    add(my_string_ptr);
  }
  else {
    int done = 0;
    cur_vrml_file->get_token(token);

    while (!done) {
      cur_vrml_file->lookahead_token(token);
      if (token[0] == CLOSE_BRACKET) done = 1;
      else {
	my_string_ptr = new SFString;
	my_string_ptr->parse();
	add(my_string_ptr);
      }
    }  // while
    cur_vrml_file->get_token(token);
  }

}  // MFString::parse





//
// MFTime
//

void
MFTime::parse()
{
  char token[MAX_TOKEN_LENGTH];

  SFTime *my_time_ptr;

  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);
    MFTime *mftime_ptr = (MFTime *) PROTOVarTable.handle_IS();
    copy_from(mftime_ptr);
    return;
  }

  if (token[0] != OPEN_BRACKET) {
    my_time_ptr = new SFTime;
    my_time_ptr->parse();
    add(my_time_ptr);
  }
  else {
    int done = 0;
    cur_vrml_file->get_token(token);

    while (!done) {
      cur_vrml_file->lookahead_token(token);
      if (token[0] == CLOSE_BRACKET) done = 1;
      else {
	my_time_ptr = new SFTime;
	my_time_ptr->parse();
	add(my_time_ptr);
      }
    }  // while
    cur_vrml_file->get_token(token);
  }

}  // MFTime::parse





//
// MFVec2f
//

void
MFVec2f::parse()
{
  char token[MAX_TOKEN_LENGTH];

  SFVec2f *my_vec2f_ptr;

  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);
    MFVec2f *mfvec2f_ptr = (MFVec2f *) PROTOVarTable.handle_IS();
    copy_from(mfvec2f_ptr);
    return;
  }

  if (token[0] != OPEN_BRACKET) {
    my_vec2f_ptr = new SFVec2f;
    my_vec2f_ptr->parse();
    add(my_vec2f_ptr);
  }
  else {
    int done = 0;
    cur_vrml_file->get_token(token);

    while (!done) {
      cur_vrml_file->lookahead_token(token);
      if (token[0] == CLOSE_BRACKET) done = 1;
      else {
	my_vec2f_ptr = new SFVec2f;
	my_vec2f_ptr->parse();
	add(my_vec2f_ptr);
      }
    }  // while
    cur_vrml_file->get_token(token);
  }

}  // MFVec2f::parse





//
// MFVec3f
//

void
MFVec3f::parse()
{
  char token[MAX_TOKEN_LENGTH];

  SFVec3f *my_vec3f_ptr;

  cur_vrml_file->lookahead_token(token);

  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);
    MFVec3f *mfvec3f_ptr = (MFVec3f *) PROTOVarTable.handle_IS();
    copy_from(mfvec3f_ptr);
    return;
  }

  if (token[0] != OPEN_BRACKET) {
    my_vec3f_ptr = new SFVec3f;
    my_vec3f_ptr->parse();
    add(my_vec3f_ptr);
  }
  else {
    int done = 0;
    cur_vrml_file->get_token(token);

    while (!done) {
      cur_vrml_file->lookahead_token(token);
      if (token[0] == CLOSE_BRACKET) done = 1;
      else {
	my_vec3f_ptr = new SFVec3f;
	my_vec3f_ptr->parse();
	add(my_vec3f_ptr);
      }
    }  // while
    cur_vrml_file->get_token(token);
  }

}  // MFVec3f::parse



    
