//
// $Id: Fields.h,v 1.1 2000/01/17 18:47:59 min Exp min $
//

#ifndef __FIELDS_H
#define __FIELDS_H

#include "Field.h"

class VrmlNode;





class MFField : public Field
{

public:

  MFField() {
    array = new RNArray<Field *>;
  }
  MFField(MFField& other);

  ~MFField();


  void add(Field *new_SFField) { array->Insert(new_SFField); }
  void clear();
  int get_size() { return array->NEntries(); }
  Field *get_field(int index) { return array->Kth(index); }

  char *to_string(char *buffer);

  void copy_from(MFField *other);


protected:

  RNArray<Field *> *array;

};  // MFField class





//
// SF fields
//

class SFBool : public Field
{

public:

  SFBool() {}
  SFBool(SFBool& other) { my_value = other.get_boolean(); }
  SFBool(RNBoolean new_value) { my_value = new_value; }
  Field *create_copy() { return new SFBool(my_value); }

  ~SFBool() {}


  FieldVarType get_field_var_type() { return SFBoolType; }
  char *get_field_var_type_string() { return (char*)"SFBool"; }

  void set_boolean(RNBoolean new_value) { my_value = new_value; }
  RNBoolean get_boolean() { return my_value; }

  void parse();
  char *to_string(char *buffer);

  void sendEventIn(void *new_value);


  //
  // operators
  //

  //
  // unary
  //

  // virtual Field *operator-()
    
  Field *operator!() {
    return new SFBool(!my_value);
  }
    
  Field *operator~() {
    return new SFBool(!my_value);  // one's complement of an SFBool...
  }

  //  Field *operator++();

  //  Field *operator--();

  //  Field *operator++(int dummy);

  //  Field *operator--(int dummy);

  //
  // assignment
  //

  Field *operator=(Field *other);

  //  Field *operator+=(Field *other);

  //  Field *operator-=(Field *other);

  //  Field *operator*=(Field *other);

  //  Field *operator/=(Field *other);

  //  Field *operator%=(Field *other);

  //  Field *operator&=(Field *other);

  //  Field *operator|=(Field *other);

  //  Field *operator^=(Field *other);

  //  Field *operator<<=(Field *other);

  //  Field *operator>>=(Field *other);

  //  Field *operator>>>=(Field *other);

  //
  // binary
  //

  Field *operator==(Field *other);

  Field *operator!=(Field *other);

  //  Field *operator<(Field *other);

  //  Field *operator<=(Field *other);

  //  Field *operator>=(Field *other);

  //  Field *operator>(Field *other);

  //  Field *operator+(Field *other);

  //  Field *operator-(Field *other);

  //  Field *operator*(Field *other);

  //  Field *operator/(Field *other);

  //  Field *operator%(Field *other);

  Field *operator&&(Field *other);

  Field *operator||(Field *other);

  Field *operator&(Field *other);

  Field *operator|(Field *other);

  Field *operator^(Field *other);

  //  Field *operator<<(Field *other);

  //  Field *operator>>(Field *other);

  //  Field *operator>>>(Field *other);

  //
  // tertiary: a ? b : c
  //


private:

  RNBoolean my_value;


};  // SFBool class





class SFColor : public Field
{

public:

  SFColor() {}
  SFColor(RNRgb new_colour) { my_colour = new_colour; }
  SFColor(SFColor& other) { my_colour = other.get_colour(); }
  Field *create_copy() { return new SFColor(my_colour); }

  ~SFColor() {}


  FieldVarType get_field_var_type() { return SFColorType; }
  char *get_field_var_type_string() { return (char*)"SFColor"; }

  void set_colour(RNRgb new_colour) { my_colour = new_colour; }
  void set_colour(RNScalar r, RNScalar g, RNScalar b) {
    my_colour.SetRed(r), my_colour.SetGreen(g), my_colour.SetBlue(b);
  }
  RNRgb get_colour() { return my_colour; }

  void parse();
  char *to_string(char *buffer) {
    return buffer + sprintf(buffer, " %g %g %g ", 
			    my_colour.R(), my_colour.G(), my_colour.B());
  }

  void sendEventIn(void *new_value);


  //
  // operators
  //

  Field *operator=(Field *other);


private:

  RNRgb my_colour;


};  // SFColor class





class SFFloat : public Field
{

public:

  SFFloat() {}
  SFFloat(double new_float) { my_float = new_float; }
  SFFloat(SFFloat& other) { my_float = other.get_float(); }
  Field *create_copy() { return new SFFloat(my_float); }

  ~SFFloat() {}


  FieldVarType get_field_var_type() { return SFFloatType; }
  char *get_field_var_type_string() { return (char*)"SFFloat"; }

  void set_float(double new_float) { my_float = new_float; }
  double get_float() { return my_float; }
  void *get_value() { return &my_float; }

  void parse();
  char *to_string(char *buffer);

  void sendEventIn(void *new_value) { 
    my_float = ((SFFloat *) new_value)->get_float();
  }


  //
  // operators
  //

  //
  // unary
  //

  Field *operator-();
  Field *operator!();
    
  //  Field *operator~();

  // prefix
  Field *operator++();
  Field *operator--();

  // postfix
  Field *operator++(int dummy);
  Field *operator--(int dummy);


  //
  // assignment
  //

  Field *operator=(Field *other);

  Field *operator+=(Field *other);

  Field *operator-=(Field *other);

  //  Field *operator*=(Field *other);

  //  Field *operator/=(Field *other);

  //  Field *operator%=(Field *other);

  //  Field *operator&=(Field *other);

  //  Field *operator|=(Field *other);

  //  Field *operator^=(Field *other);

  //  Field *operator<<=(Field *other);

  //  Field *operator>>=(Field *other);

  //  Field *operator>>>=(Field *other);

  //
  // binary
  //

  Field *operator==(Field *other);

  Field *operator!=(Field *other);

  Field *operator<(Field *other);

  Field *operator<=(Field *other);

  Field *operator>=(Field *other);

  Field *operator>(Field *other);

  Field *operator+(Field *other);

  Field *operator-(Field *other);

  Field *operator*(Field *other);

  Field *operator/(Field *other);

  //  Field *operator%(Field *other);

  //  Field *operator&&(Field *other);

  //  Field *operator||(Field *other);

  //  Field *operator&(Field *other);

  //  Field *operator|(Field *other);

  //  Field *operator^(Field *other);

  //  Field *operator<<(Field *other);

  //  Field *operator>>(Field *other);

  //  Field *operator>>>(Field *other);

  //
  // tertiary: a ? b : c
  //


protected:

  double to_float(Field *other);

  double my_float;


};  // SFFloat class





class SFImage : public Field
{

public:

  SFImage() {}
  SFImage(SFImage& other) { my_image = other.copy_image(other.get_image()); }
  SFImage(R2Image *new_image) { my_image = copy_image(new_image); }
  Field *create_copy() { return new SFImage(my_image); }
  R2Image *copy_image(R2Image *source_image);

  ~SFImage() {}


  FieldVarType get_field_var_type() { return SFImageType; }
  char *get_field_var_type_string() { return (char*)"SFImage"; }

  void set_image(R2Image *new_image) { my_image = new_image; }
  R2Image *get_image() { return my_image; }

  void parse();
  char *to_string(char *buffer);

  void sendEventIn(void *new_value) {
    // copy the image
    cout << "eventIn on an SFImage..." << endl;
  }

  int get_size_in_bytes() { 
    return my_image->Width() * my_image->Height() * my_image->Depth();
  }  // get_size_in_bytes


private:

  R2Image *my_image;


};  // SFImage class





class SFInt32 : public Field
{

public:

  SFInt32() {}
  SFInt32(SFInt32& other) { my_integer = other.get_integer(); }
  SFInt32(int new_integer) { my_integer = new_integer; }
  Field *create_copy() { return new SFInt32(my_integer); }

  ~SFInt32() {}

  FieldVarType get_field_var_type() { return SFInt32Type; }
  char *get_field_var_type_string() { return (char*)"SFInt32"; }

  void set_integer(int new_integer) { my_integer = new_integer; }
  int get_integer() { return my_integer; }

  void parse();
  char *to_string(char *buffer) {
    return buffer + sprintf(buffer, " %d ", my_integer);
  }

  void sendEventIn(void *new_value) {
    my_integer = ((SFInt32 *) new_value)->get_integer();
  }


  //
  // operators
  //

  //
  // unary
  //

  Field *operator-() {
    return new SFInt32(-my_integer);
  }
    
  Field *operator!() {
    return new SFBool(!my_integer);
  }
    
  Field *operator~() {
    return new SFInt32(~my_integer);
  }

  Field *operator++() {
    my_integer++;
    return new SFInt32(my_integer);
  }

  Field *operator--() {
    my_integer--;
    return new SFInt32(my_integer);
  }

  Field *operator++(int dummy) {
    my_integer++;
    return new SFInt32(my_integer - 1);
  }

  Field *operator--(int dummy) {
    my_integer--;
    return new SFInt32(my_integer + 1);
  }

  //
  // assignment
  //

  Field *operator=(Field *other);

  //  Field *operator+=(Field *other);

  //  Field *operator-=(Field *other);

  //  Field *operator*=(Field *other);

  //  Field *operator/=(Field *other);

  //  Field *operator%=(Field *other);

  //  Field *operator&=(Field *other);

  //  Field *operator|=(Field *other);

  //  Field *operator^=(Field *other);

  //  Field *operator<<=(Field *other);

  //  Field *operator>>=(Field *other);

  //  Field *operator>>>=(Field *other);

  //
  // binary
  //

  Field *operator==(Field *other);

  Field *operator!=(Field *other);

  Field *operator<(Field *other);

  Field *operator<=(Field *other);

  Field *operator>=(Field *other);

  Field *operator>(Field *other);

  Field *operator+(Field *other);

  Field *operator-(Field *other);

  Field *operator*(Field *other);

  Field *operator/(Field *other);

  //  Field *operator%(Field *other);

  //  Field *operator&&(Field *other);

  //  Field *operator||(Field *other);

  //  Field *operator&(Field *other);

  //  Field *operator|(Field *other);

  //  Field *operator^(Field *other);

  //  Field *operator<<(Field *other);

  //  Field *operator>>(Field *other);

  //  Field *operator>>>(Field *other);

  //
  // tertiary: a ? b : c
  //


private:

  int to_int(Field *other);

  int my_integer;


};  // SFInt32 class





// as an actual parameter in a PROTO instantiation
static const int MAX_NODE_TEXT_LENGTH = 2048;

class SFNode : public Field
{

public:

  SFNode() {
    my_string = NULL;
  }
  SFNode(SFNode& other) {
    my_string = NULL;
    cerr << "Error: no copy constructor for SFNode" << endl;
  }
  SFNode(VrmlNode *new_node) { my_node = new_node; }
  Field *create_copy() { return this; }  // Note: no copy

  ~SFNode();


  FieldVarType get_field_var_type() { return SFNodeType; }
  char *get_field_var_type_string() { return (char*)"SFNode"; }

  void set_node_ptr(VrmlNode *new_node) { my_node = new_node; }
  VrmlNode *get_node_ptr() { return my_node; }

  void parse();

  void parse_and_store_string();
  void parse_my_string();
  char *get_my_string() { return my_string; }
  void set_my_string(char *new_string) { my_string = new_string; }

  char *to_string(char *buffer) {
    cerr << "Error: no to_string implementation for SFNode" << endl;
    return NULL;
  }

  void sendEventIn(void *new_value) {
    my_node = ((SFNode *) new_value)->get_node_ptr();
  }


private:

  VrmlNode *my_node;
  char *my_string;


};  // SFNode class





class SFRotation : public Field
{

public:

  SFRotation() {
    axis.Reset(0, 0, 1);
    angle = 0;
  }
  SFRotation(SFRotation& other) {
    axis = other.get_axis();
    angle = other.get_angle();
  }
  SFRotation(R3Vector new_axis, RNScalar new_angle) {
    axis = new_axis;
    angle = new_angle;
  }
  Field *create_copy() { return new SFRotation(axis, angle); }

  ~SFRotation() {}


  FieldVarType get_field_var_type() { return SFRotationType; }
  char *get_field_var_type_string() { return (char*)"SFRotation"; }

  void set_axis(RNScalar x, RNScalar y, RNScalar z) { axis.Reset(x, y, z); }
  void set_axis(R3Vector new_axis) { axis = new_axis; }
  void set_angle(RNScalar new_angle) { angle = new_angle; }
  R3Vector& get_axis() { return axis; }
  RNScalar get_angle() { return angle; }

  void parse();
  char  *to_string(char *buffer) {
    return buffer + sprintf(buffer, " %g %g %g %g ", 
			    axis.X(), axis.Y(), axis.Z(), angle);
  }

  void sendEventIn(void *new_value) {
    axis = ((SFRotation *) new_value)->get_axis();
    angle = ((SFRotation *) new_value)->get_angle();
  }


  //
  // operators
  //

  Field *operator=(Field* other);


private:

  R3Vector axis;
  RNScalar angle;


};  // SFRotation class





static const int DEFAULT_SFSTRING_LENGTH = 64;

class SFString : public Field
{

public:

  SFString() {
    my_string = new char[DEFAULT_SFSTRING_LENGTH];
    current_length = DEFAULT_SFSTRING_LENGTH;
  }
  SFString(char *new_string);
  SFString(SFString& other);
  Field *create_copy() { return new SFString(my_string); }

  ~SFString();


  FieldVarType get_field_var_type() { return SFStringType; }
  char *get_field_var_type_string() { return (char*)"SFString"; }

  void set_string(char *new_string);
  char *get_string() { return my_string; }

  void parse();
  void parse(char quote_char);
  char *to_string(char *buffer) {
    return buffer + sprintf(buffer, " \"%s\" ", my_string);
  }  // to_string

  void sendEventIn(void *new_value) {
    set_string(((SFString *) new_value)->get_string());
  }

  //
  // operators
  //


private:

  char *my_string;
  int current_length;


};  // SFString class





class SFTime : public SFFloat
{

public:

  SFTime() {}
  SFTime(SFTime& other) { my_float = other.get_time(); }
  SFTime(double new_time) { my_float = new_time; }
  Field *create_copy() { return new SFTime(my_float); }

  ~SFTime() {}


  FieldVarType get_field_var_type() { return SFTimeType; }
  char *get_field_var_type_string() { return (char*)"SFTime"; }

  void set_time(double new_time) { my_float = new_time; }
  double get_time() { return my_float; }

  void sendEventIn(void *new_value) {
    my_float = ((SFTime *) new_value)->get_time();
  }


private:


};  // SFTime class





class SFVec2f : public Field
{

public:

  SFVec2f() {}
  SFVec2f(R2Point new_vector) { my_vector = new_vector; }
  SFVec2f(SFVec2f& other) { my_vector = other.get_point(); }
  SFVec2f(RNScalar x, RNScalar y) { set_vector(x, y); }
  Field *create_copy() { return new SFVec2f(my_vector) ; }

  ~SFVec2f() {}


  FieldVarType get_field_var_type() { return SFVec2fType; }
  char *get_field_var_type_string() { return (char*)"SFVec2f"; }

  void set_vector(R2Point new_point) { my_vector = new_point; }
  void set_vector(R2Vector new_vector) { 
    my_vector.X(new_vector.X());
    my_vector.Y(new_vector.Y());
  }
  void set_vector(RNScalar x, RNScalar y) {
    my_vector.X(x), my_vector.Y(y);
  }
  R2Point get_point() { return my_vector; }
  R2Vector get_vector() {
    return R2Vector(my_vector.X(), my_vector.Y());
  }

  void parse();
  char *to_string(char *buffer) {
    return buffer + sprintf(buffer, " %g %g ", my_vector.X(), my_vector.Y());
  }

  void sendEventIn(void *new_value) {
    my_vector = ((SFVec2f *) new_value)->get_point();
  }


private:

  R2Point my_vector;


};  // SFVec2f class





class SFVec3f : public Field
{

public:

  SFVec3f() {}
  SFVec3f(SFVec3f& other) { my_vector = other.get_point(); }
  SFVec3f(R3Point new_vector) { my_vector = new_vector; }
  SFVec3f(RNScalar x, RNScalar y, RNScalar z) { set_vector(x, y, z); }
  Field *create_copy() { return new SFVec3f(my_vector); }


  FieldVarType get_field_var_type() { return SFVec3fType; }
  char *get_field_var_type_string() { return (char*)"SFVec3f"; }

  void set_vector(R3Point new_point) { my_vector = new_point; }
  void set_vector(R3Vector new_vector) { 
    my_vector.Reset(new_vector.X(), new_vector.Y(), new_vector.Z());
  }
  void set_vector(RNScalar x, RNScalar y, RNScalar z) {
    my_vector.Reset(x, y, z);
  }
  R3Point get_point() { return my_vector; }
  R3Vector get_vector() { 
    return R3Vector(my_vector.X(), my_vector.Y(), my_vector.Z());
  }

  void parse();
  char *to_string(char *buffer) {
    return buffer + sprintf(buffer, " %g %g %g ", 
			    my_vector.X(), my_vector.Y(), my_vector.Z());
  }

  void sendEventIn(void *new_value) { 
    my_vector = ((SFVec3f *) new_value)->get_point();
  }

  //
  // operators
  //


private:

  R3Point my_vector;


};  // SFVec3f class





//
// MF fields
//

class MFColor : public MFField
{

public:

  MFColor() {}
  Field *create_copy() { return new MFColor(*this); }

  ~MFColor() {}


  FieldVarType get_field_var_type() { return MFColorType; }
  char *get_field_var_type_string() { return (char*)"MFColor"; }

  void parse();

  SFColor *operator[](int index) { return (SFColor *) array->Kth(index); }


private:


};  // MFColor class





class MFFloat : public MFField
{

public:

  MFFloat() {}
  Field *create_copy() { return new MFFloat(*this); }

  ~MFFloat() {}


  FieldVarType get_field_var_type() { return MFFloatType; }
  char *get_field_var_type_string() { return (char*)"MFFloat"; }

  void parse();

  SFFloat *operator[](int index) { return (SFFloat *) array->Kth(index); }


private:


};  // MFFloat class





class MFInt32 : public MFField
{

public:

  MFInt32() {}
  Field *create_copy() { return new MFInt32(*this); }

  ~MFInt32() {}


  FieldVarType get_field_var_type() { return MFInt32Type; }
  char *get_field_var_type_string() { return (char*)"MFInt32"; }

  void parse();

  SFInt32 *operator[](int index) { return (SFInt32 *) array->Kth(index); }


private:


};  // MFInt32 class





class MFNode : public MFField
{

public:

  MFNode() {}
  Field *create_copy() { return new MFNode(*this); }

  ~MFNode() {}


  FieldVarType get_field_var_type() { return MFNodeType; }
  char *get_field_var_type_string() { return (char*)"MFNode"; }

  void parse();
  void parse_and_store_string();

  SFNode *operator[](int index) { return (SFNode *) array->Kth(index); }


private:


};  // MFNode class





class MFRotation : public MFField
{

public:

  MFRotation() {}
  Field *create_copy() { return new MFRotation(*this); }

  ~MFRotation() {}


  FieldVarType get_field_var_type() { return MFRotationType; }
  char *get_field_var_type_string() { return (char*)"MFRotation"; }

  void parse();

  SFRotation *operator[](int index) { 
    return (SFRotation *) array->Kth(index);
  }


private:


};  // MFRotation class





class MFString : public MFField
{

public:

  MFString() {}
  Field *create_copy() { return new MFString(*this); }

  ~MFString() {}


  FieldVarType get_field_var_type() { return MFStringType; }
  char *get_field_var_type_string() { return (char*)"MFString"; }

  void parse();

  SFString *operator[](int index) { return (SFString *) array->Kth(index); }


private:


};  // MFString class





class MFTime : public MFField
{

public:

  MFTime() {}
  Field *create_copy() { return new MFTime(*this); }

  ~MFTime() {}

  void parse();

  FieldVarType get_field_var_type() { return MFTimeType; }
  char *get_field_var_type_string() { return (char*)"MFTime"; }

  SFTime *operator[](int index) { return (SFTime *) array->Kth(index); }


private:


};  // MFTime class





class MFVec2f : public MFField
{

public:

  MFVec2f() {}
  Field *create_copy() { return new MFVec2f(*this); }

  ~MFVec2f() {}


  FieldVarType get_field_var_type() { return MFVec2fType; }
  char *get_field_var_type_string() { return (char*)"MFVec2f"; }

  void parse();

  SFVec2f *operator[](int index) { return (SFVec2f *) array->Kth(index); }


private:


};  // MFVec2f class





class MFVec3f : public MFField
{

public:

  MFVec3f() {}
  Field *create_copy() { return new MFVec3f(*this); }

  ~MFVec3f() {}


  FieldVarType get_field_var_type() { return MFVec3fType; }
  char *get_field_var_type_string() { return (char*)"MFVec3f"; }

  void parse();

  SFVec3f *operator[](int index) { return (SFVec3f *) array->Kth(index); }


private:


};  // MFVec3f class




#endif


