//
// $Id: Files.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include <ctype.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <iostream.h>
#include <stdio.h>
#include "Files.h"





//
// Datafile
//

Datafile::Datafile() {
  my_filespec = NULL;
  line_nr = 1;
  string_source = NULL;
  string_dest = NULL;
  current_char = 0x20;
  buffer_index = 0;
}  // constructor



int
Datafile::open(char *filespec)
{
  ifstream::open(filespec, ios::in);
  int open_good = good();
  my_filespec = filespec;
  precision(12);
  if (open_good) fill_buffer();  // file may be so small that EOF is reached...
  current_char = 0x20;
  return (open_good);  // that's why we now use open_good...
}  // Datafile::open



void
Datafile::fill_buffer()
{
  for(int i=0; i<MAX_BUFFERSIZE; i++) {
    char k = (char) get_source_char();
    buffer[i] = k;
    if (k == MY_EOF) return;
  }  // for
}  // Datafile::fill_buffer



void
Datafile::print_error_line(int warning)
{
  if (warning) cerr << "Warning";
  else cerr << "Error";

  cerr << " in " << my_filespec << " at line " << get_line_nr()
    << ":" << endl << "    ";
}  // Datafile::print_error_line



int
Datafile::get_source_char()
{
  // first check if there's a string source
  if (string_source != NULL) {
    current_char = string_source[0];
    string_source++;
  }
  // else get a character from the actual file
  else {
    current_char = get();
  }
  if (string_dest != NULL) {
    string_dest[0] = current_char;
    string_dest++;
  }
  if (current_char == EOF) current_char = MY_EOF;
  return current_char;
}  // Datafile::get_source_char



void
Datafile::refill_buffer()
{
  if (buffer_index > 1) {  // 1, not 0...
    int prev_char = current_char;
    int nr_to_move = MAX_BUFFERSIZE - buffer_index;
    memmove(buffer, buffer + buffer_index, nr_to_move);
    for(int i=0; i<buffer_index; i++)
      buffer[nr_to_move + i] = (char) get_source_char();
    buffer_index = 0;
    current_char = prev_char;
  }
}  // Datafile::refill_buffer



int
Datafile::get_buffered_char()
{
  if (buffer_index == MAX_BUFFERSIZE/2) refill_buffer();
  current_char = buffer[buffer_index];
  if (current_char == '\n') line_nr++;
  buffer_index++;
  return current_char;
}  // Datafile::get_buffered_char



void
Datafile::go_back_char()
{
  if (buffer_index == 0) {
    print_error_line();
    cerr << "attempt to move back in buffer at start of buffer" << endl;
  }
  else {
    buffer_index--;
    if (buffer[buffer_index] == '\n') line_nr--;
  }
}  // Datafile::go_back_char



void
Datafile::set_parse_source(char *new_string_source)
{
  string_source = new_string_source;
  fill_buffer();
  set_current_char(' ');
}  // Datafile::set_parse_source





//
// Tokenfile
//

void
Tokenfile::get_token(char *token)
{
  token[0] = '\0';

  current_char = skip_whitespace();
  if (current_char == MY_EOF) return;

  // first check for a standard token
  if (get_standard_token(token)) return;

  // now we have an id or a number
  //  token[0] = current_char;
  int index = 0;

  do {
    current_char = get_buffered_char();
    if (is_id_char(current_char))
    {
      token[index] = current_char;
      index++;
      if (index == MAX_TOKEN_LENGTH) {
	print_error_line();
	cerr << "Token too long (max length is " << MAX_TOKEN_LENGTH
	     << ")" << endl;
	return;
      }
    }
  } while (is_id_char(current_char));
  token[index] = '\0';
  go_back_char();

#ifdef __DEBUG_LINE_NR__
  printf("\rLine nr %d   ", get_line_nr());
  fflush(stdout);
#endif
#ifdef __DEBUG_TOKENS__
  cout << ">> TOKEN: [" << token << "]" << endl;
#endif

}  // Tokenfile::get_token



void
Tokenfile::lookahead_token(char *token)
{
  refill_buffer();
  int prev_index = buffer_index;
  int prev_char = current_char;
  int prev_line_nr = line_nr;
#ifdef __DEBUG_TOKENS__
  cout << "Lookahead: ";
#endif
  get_token(token);
  line_nr = prev_line_nr;
  current_char = prev_char;
  buffer_index = prev_index;
}  // Tokenfile::lookahead_token



int
Tokenfile::accept_token(const char *wanted)
{
  char token[MAX_TOKEN_LENGTH];
  get_token(token);

  if (strcmp(token, wanted) == 0) return 1;
  else {
    print_error_line();
    cerr << "Expected [" << wanted << "], read [" << token
	 << "]" << endl;
    return 0;
  }
}  // Tokenfile::accept_token with string parameter



int
Tokenfile::accept_token(const char wanted)
{
  char wanted_string[2];
  wanted_string[0] = wanted;
  wanted_string[1] = '\0';
  return accept_token(wanted_string);
}  // Tokenfile::accept_token with char parameter



void
Tokenfile::set_standard_tokens(char (*new_standard_tokens)[MAX_TOKEN_LENGTH])
{
  standard_tokens = new_standard_tokens;
  int index = 0;
  // find the longest token
  while (standard_tokens[index][0] != '\0') {
    int length = strlen(standard_tokens[index]);
    if (length > max_standard_token_length) max_standard_token_length = length;
    index++;
  }  // while
}  // Tokenfile::set_standard_tokens



int
Tokenfile::get_standard_token(char *token)
{
  int index = 0;
  char candidate[MAX_TOKEN_LENGTH];
  candidate[0] = '\0';
  int candidate_length = 0;
  while (standard_tokens[index][0] != '\0') {
    int length = strlen(standard_tokens[index]);
    if (length > candidate_length) {
      if (strncmp(standard_tokens[index], 
		  buffer + buffer_index, length) == 0) {
	strcpy(candidate, standard_tokens[index]);
	candidate_length = strlen(candidate);
      }
    }
    index++;
  }  // while
  if (candidate_length) {
    strcpy(token, candidate);
    for(int i=0; i<candidate_length; i++) get_buffered_char();
    current_char = ' ';
#ifdef __DEBUG_TOKENS__
    cout << ">> STANDARD TOKEN: [" << token << "]" << endl;
#endif
    return 1;
  }
  return 0;
}  // Tokenfile::get_standard_token



void
Tokenfile::read_until(char until_char)
{
  char k;
  do {
    k = get_buffered_char();
  } while (k != until_char);
}  // Tokenfile::read_until



int
Tokenfile::is_number_char(int k)
{
  return (strchr(number_chars, k) != NULL);
}  // Tokenfile::is_number_char



void
Tokenfile::get_number(char *number_string)
{
  current_char = skip_whitespace();
  if (current_char == MY_EOF) return;

  number_string[0] = '\0';
  
  int index = 0;
  do {
    current_char = get_buffered_char();
    if (is_number_char(current_char)) {
      number_string[index] = current_char;
      index++;
    }
  } while (is_number_char(current_char));
  number_string[index] = '\0';

  if (index == 0) {
    print_error_line();
    cerr << "expected number, but didn't find one" << endl;
  }
  go_back_char();
#ifdef __DEBUG_TOKENS__
    cout << ">> NUMBER: [" << number_string << "]" << endl;
#endif

}  // Tokenfile::get_number



int
Tokenfile::get_int()
{
  int i;

  *this >> i;
  return i;
}  // Tokenfile::get_int



Float
Tokenfile::get_float()
{
  Float f;

  *this >> f;
  return f;
}  // Tokenfile::get_float



int
Tokenfile::is_whitespace(int k)
{
  return (strchr(whitespace_chars, k) != NULL);
}  // Tokenfile::is_whitespace



int
Tokenfile::skip_whitespace()
{
  //  if (!is_whitespace(current_char)) return current_char;

  do {
    current_char = get_buffered_char();
    if (current_char == '#') skip_comments();  // needs to improve
    if (current_char == '/') {
      char next_char = get_buffered_char();
      if (next_char == '/') skip_comments();
      else {
	go_back_char();
	current_char = '/';
      }
    }
  } while (is_whitespace(current_char) && (current_char != MY_EOF));

  go_back_char();
  return current_char;
}  // Tokenfile::skip_whitespace


int
Tokenfile::is_id_char(int k)
{
  return (isalnum(k) || (k == '_') || (k == '%'));
}  // Tokenfile::is_id_char



int
Tokenfile::is_id_firstchar(int k)
{
  return (is_id_char(k) && !isdigit(k));
}  // Tokenfile::is_id_firstchar



void
Tokenfile::skip_comments()
{
  do {
    current_char = get_buffered_char();
  } while (current_char != '\n');
}  // Tokenfile::skip_comments





//
// VrmlFile
//

void
VrmlFile::skip_node()
{
  accept_token(OPEN_BRACE);

  int k;
  int done = 0;

  int tobe_closed = 1;

  while (!done) {
    k = get_buffered_char();
    if (k == MY_EOF) done = 1;
    else if (k == OPEN_BRACE) tobe_closed++;
    else if (k == CLOSE_BRACE) {
      tobe_closed--;
      done = (tobe_closed == 0);
    }
  }  // while
}  // VrmlFile::skip_node



