//
// $Id: GeometryNodes.h,v 1.1 2000/01/17 18:47:59 min Exp min $
//

#ifndef __GEOMETRYNODES_H
#define __GEOMETRYNODES_H

#include "VrmlNode.h"
#include "Fields.h"





class Box : public VrmlNode
{

public:

  Box() { 
    set_node_name("Box"); 
    set_node_type(BoxType);

    add_field("size", &size, field);

    size.set_vector(2, 2, 2);

  }  // constructor

  void add_Ring();

  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return HeadInsertNode; }


private:

  SFVec3f size;

};  // Box class





class Cone : public VrmlNode
{

public:

  Cone() { 
    set_node_name("Cone");
    set_node_type(ConeType);

    add_field("bottomRadius", &bottomRadius, field);
    add_field("height", &height, field);
    add_field("side", &side, field);
    add_field("bottom", &bottom, field);

    bottomRadius.set_float(1.0);
    height.set_float(2.0);
    side.set_boolean(TRUE);
    bottom.set_boolean(TRUE);

  }  // constructor

  ~Cone() {}


  void add_Ring();
  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return HeadInsertNode; }


private:

  SFFloat bottomRadius;
  SFFloat height;
  SFBool side;
  SFBool bottom;

};  // Cone class





class Cylinder : public VrmlNode
{

public:

  Cylinder() { 
    set_node_name("Cylinder"); 
    set_node_type(CylinderType); 

    add_field("bottom", &bottom, field);
    add_field("height", &height, field);
    add_field("radius", &radius, field);
    add_field("side", &side, field);
    add_field("top", &top, field);

    bottom.set_boolean(TRUE);
    height.set_float(2.0);
    radius.set_float(1.0);
    side.set_boolean(TRUE);
    top.set_boolean(TRUE);

  }  // constructor
  ~Cylinder() {}


  void add_Ring();
  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return HeadInsertNode; }


private:

  SFBool bottom;
  SFFloat height;
  SFFloat radius;
  SFBool side;
  SFBool top;


};  // Cylinder class





class ElevationGrid : public VrmlNode
{

public:

  ElevationGrid() { 
    set_node_name("ElevationGrid"); 
    set_node_type(ElevationGridType);

    add_field("set_height", &set_height, eventIn);
    add_field("color", &color, exposedField);
    add_field("normal", &normal, exposedField);
    add_field("texCoord", &texCoord, exposedField);
    add_field("height", &height, field);
    add_field("ccw", &ccw, field);
    add_field("colorPerVertex", &colorPerVertex, field);
    add_field("creaseAngle", &creaseAngle, field);
    add_field("normalPerVertex", &normalPerVertex, field);
    add_field("solid", &solid, field);
    add_field("xDimension", &xDimension, field);
    add_field("xSpacing", &xSpacing, field);
    add_field("zDimension", &zDimension, field);
    add_field("zSpacing", &zSpacing, field);

    ccw.set_boolean(TRUE);
    colorPerVertex.set_boolean(TRUE);
    creaseAngle.set_float(0);
    normalPerVertex.set_boolean(TRUE);
    solid.set_boolean(TRUE);
    xDimension.set_integer(0);
    xSpacing.set_float(0);
    zDimension.set_integer(0);
    zSpacing.set_float(0);

  }  // constructor

  ~ElevationGrid() {}


  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return HeadInsertNode; }


private:

  MFFloat set_height;
  SFNode color;
  SFNode normal;
  SFNode texCoord;
  MFFloat height;
  SFBool ccw;
  SFBool colorPerVertex;
  SFFloat creaseAngle;
  SFBool normalPerVertex;
  SFBool solid;
  SFInt32 xDimension;
  SFFloat xSpacing;
  SFInt32 zDimension;
  SFFloat zSpacing;


};  // ElevationGrid class





class Extrusion : public VrmlNode
{

public:

  Extrusion() { 
    set_node_name("Extrusion"); 
    set_node_type(ExtrusionType);

    add_field("crossSection", &set_crossSection, eventIn);
    add_field("orientation", &set_orientation, eventIn);
    add_field("scale", &set_scale, eventIn);
    add_field("spine", &set_spine, eventIn);
    add_field("beginCap", &beginCap, field);
    add_field("ccw", &ccw, field);
    add_field("convex", &convex, field);
    add_field("creaseAngle", &creaseAngle, field);
    add_field("crossSection", &crossSection, field);
    add_field("endCap", &endCap, field);
    add_field("orientation", &orientation, field);
    add_field("scale", &scale, field);
    add_field("solid", &solid, field);
    add_field("spine", &spine, field);

    beginCap.set_boolean(TRUE);
    ccw.set_boolean(TRUE);
    convex.set_boolean(TRUE);
    creaseAngle.set_float(0);
    SFVec2f *vec_ptr = new SFVec2f(1, 1);
    crossSection.add(vec_ptr);
    vec_ptr = new SFVec2f(1, -1);
    crossSection.add(vec_ptr);
    vec_ptr = new SFVec2f(-1, -1);
    crossSection.add(vec_ptr);
    vec_ptr = new SFVec2f(-1, 1);
    crossSection.add(vec_ptr);
    vec_ptr = new SFVec2f(1, 1);
    crossSection.add(vec_ptr);
    endCap.set_boolean(TRUE);
    SFRotation *initial_orientation = new SFRotation(R3Vector(0, 0, 1), 0);
    orientation.add(initial_orientation);
    vec_ptr = new SFVec2f(1, 1);
    scale.add(vec_ptr);
    solid.set_boolean(TRUE);
    SFVec3f *vec3_ptr = new SFVec3f(0, 0, 0);
    spine.add(vec3_ptr);
    vec3_ptr = new SFVec3f(0, 1, 0);
    spine.add(vec3_ptr);

  }  // constructor

  ~Extrusion() {}


  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return HeadInsertNode; }


private:

  MFVec2f set_crossSection;
  MFRotation set_orientation;
  MFVec2f set_scale;
  MFVec3f set_spine;
  SFBool beginCap;
  SFBool ccw;
  SFBool convex;
  SFFloat creaseAngle;
  MFVec2f crossSection;
  SFBool endCap;
  MFRotation orientation;
  MFVec2f scale;
  SFBool solid;
  MFVec3f spine;


};  // Extrusion class





class IndexedFaceSet : public VrmlNode
{

public:

  IndexedFaceSet() { 
    set_node_name("IndexedFaceSet"); 
    set_node_type(IndexedFaceSetType); 

    add_field("set_colorIndex", &set_colorIndex, eventIn);
    add_field("set_coordIndex", &set_coordIndex, eventIn);
    add_field("set_normalIndex", &set_normalIndex, eventIn);
    add_field("set_texCoordIndex", &set_texCoordIndex, eventIn);
    add_field("color", &color, exposedField);
    add_field("coord", &coord, exposedField);
    add_field("normal", &normal, exposedField);
    add_field("texCoord", &texCoord, exposedField);
    add_field("ccw", &ccw, field);
    add_field("colorIndex", &colorIndex, field);
    add_field("colorPerVertex", &colorPerVertex, field);
    add_field("convex", &convex, field);
    add_field("coordIndex", &coordIndex, field);
    add_field("creaseAngle", &creaseAngle, field);
    add_field("normalIndex", &normalIndex, field);
    add_field("normalPerVertex", &normalPerVertex, field);
    add_field("solid", &solid, field);
    add_field("texCoordIndex", &texCoordIndex, field);

    color.set_node_ptr(NULL);
    coord.set_node_ptr(NULL);
    normal.set_node_ptr(NULL);
    texCoord.set_node_ptr(NULL);

    ccw.set_boolean(TRUE);
    colorPerVertex.set_boolean(TRUE);
    convex.set_boolean(TRUE);
    creaseAngle.set_float(0);
    normalPerVertex.set_boolean(TRUE);
    solid.set_boolean(TRUE);

  }  // constructor

  ~IndexedFaceSet() {}


  void add_Ring();
  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return HeadInsertNode; }


private:

  MFInt32 set_colorIndex;
  MFInt32 set_coordIndex;
  MFInt32 set_normalIndex;
  MFInt32 set_texCoordIndex;

  SFNode color;
  SFNode coord;
  SFNode normal;
  SFNode texCoord;

  SFBool ccw;
  MFInt32 colorIndex;
  SFBool colorPerVertex;
  SFBool convex;
  MFInt32 coordIndex;
  SFFloat creaseAngle;
  MFInt32 normalIndex;
  SFBool normalPerVertex;
  SFBool solid;
  MFInt32 texCoordIndex;


};  // IndexedFaceSet class





class IndexedLineSet : public VrmlNode
{

public:

  IndexedLineSet() { 
    set_node_name("IndexedLineSet");
    set_node_type(IndexedLineSetType);

    add_field("set_colorIndex", &set_colorIndex, eventIn);
    add_field("set_coordIndex", &set_coordIndex, eventIn);
    add_field("color", &color, exposedField);
    add_field("coord", &coord, exposedField);
    add_field("colorIndex", &colorIndex, field);
    add_field("colorPerVertex", &colorPerVertex, field);
    add_field("coordIndex", &coordIndex, field);

    color.set_node_ptr(NULL);
    coord.set_node_ptr(NULL);
    colorPerVertex.set_boolean(TRUE);

  }  // constructor

  ~IndexedLineSet() {}


  void add_Ring();
  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return HeadInsertNode; }


private:

  MFInt32 set_colorIndex;
  MFInt32 set_coordIndex;
  SFNode color;
  SFNode coord;
  MFInt32 colorIndex;
  SFBool colorPerVertex;
  MFInt32 coordIndex;


};  // IndexedLineSet class





class PointSet : public VrmlNode
{

public:

  PointSet() { 
    set_node_name("PointSet"); 
    set_node_type(PointSetType);

    add_field("color", &color, exposedField);
    add_field("coord", &coord, exposedField);

    color.set_node_ptr(NULL);
    coord.set_node_ptr(NULL);

  }  // constructor

  ~PointSet() {}


  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return HeadInsertNode; }


private:

  SFNode color;
  SFNode coord;


};  // PointSet class





class Sphere : public VrmlNode
{

public:

  Sphere() { 
    set_node_name("Sphere");
    set_node_type(SphereType);

    add_field("radius", &radius, field);

    radius.set_float(1);

  }  // constructor

  ~Sphere() {}


  void add_Ring();
  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return HeadInsertNode; }


private:

  SFFloat radius;


};  // Sphere class





class Text : public VrmlNode
{

public:

  Text() { 
    set_node_name("Text"); 
    set_node_type(TextType);

    add_field("string", &string, exposedField);
    add_field("fontStyle", &fontStyle, exposedField);
    add_field("length", &length, exposedField);
    add_field("maxExtent", &maxExtent, exposedField);

    fontStyle.set_node_ptr(NULL);
    maxExtent.set_float(0);

  }  // constructor

  ~Text() {}


  void add_Ring();

  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return HeadInsertNode; }


private:

  MFString string;
  SFNode fontStyle;
  MFFloat length;
  SFFloat maxExtent;


};  // Text class


#endif
