//
// $Id: GroupingNodes.h,v 1.1 2000/01/17 18:47:59 min Exp min $
//

#ifndef __GroupingNodes_H
#define __GroupingNodes_H

#include "VrmlNode.h"
#include "Runtime.h"
#include "Fields.h"





class GroupingNode : public VrmlNode
{

public:

  GroupingNode() {
    add_field("addChildren", &addChildren, eventIn);
    add_field("removeChildren", &removeChildren, eventIn);
    add_field("children", &children, exposedField);
    add_field("bboxCenter", &bboxCenter, field);
    add_field("bboxSize", &bboxSize, field);

    bboxCenter.set_vector(0, 0, 0);
    bboxSize.set_vector(-1, -1, -1);
  }  // constructor

  ~GroupingNode() {}

  R3GroupNode *get_group_node_ptr() { return group_node_ptr; }
  void set_group_node_ptr(R3GroupNode *new_ptr) { 
    set_ring_node_ptr(new_ptr);
    group_node_ptr = new_ptr;
  }

  MFNode *get_children() { return &children; }

  void init_Ring();
  void add_Ring();

  void sendEventIn(int field_id, void *new_value);


private:

  MFNode addChildren;
  MFNode removeChildren;
  MFNode children;
  SFVec3f bboxCenter;
  SFVec3f bboxSize;

  R3GroupNode *group_node_ptr;

};   // GroupingNode class





//
// Anchor
//

class Anchor : public GroupingNode
{

public:

  Anchor() : GroupingNode() { 
    set_node_name("Anchor");
    set_node_type(AnchorType);

    add_field("description", &description, exposedField);
    add_field("parameter", &parameter, exposedField);
    add_field("url", &url, exposedField);

    description.set_string("");
  }  // constructor
  ~Anchor() {}


  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return Unknown; }


private:

  SFString description;
  MFString parameter;
  MFString url;

};  // Anchor class





//
// Billboard
//

class Billboard : public GroupingNode
{

public:

  Billboard() : GroupingNode() { 
    set_node_name("Billboard"); 
    set_node_type(BillboardType);

    add_field("axisOfRotation", &axisOfRotation, exposedField);

    axisOfRotation.set_vector(0, 1, 0);

  }  // constructor

  ~Billboard() {}


  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return Unknown; }


private:

  SFVec3f axisOfRotation;


};  // Billboard class





//
// Collision
//

class Collision : public GroupingNode
{

public:

  Collision() : GroupingNode() { 
    set_node_name("Collision"); 
    set_node_type(CollisionType);

    add_field("collide", &collide, exposedField);
    add_field("proxy", &proxy, field);
    add_field("collideTime", &collideTime, eventOut);

    collide.set_boolean(FALSE);
    proxy.set_node_ptr(NULL);

  }  // constructor

  ~Collision() {}


  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return HeadInsertGroupNode; }


private:

  SFBool collide;
  SFNode proxy;
  SFTime collideTime;


};  // Collision class





//
// DirectionalLight
//

class DirectionalLight : public GroupingNode
{

public:

  DirectionalLight() : GroupingNode() { 
    set_node_name("DirectionalLight"); 
    set_node_type(DirectionalLightType);

    add_field("ambientIntensity", &ambientIntensity, exposedField);
    add_field("color", &color, exposedField);
    add_field("direction", &direction, exposedField);
    add_field("intensity", &intensity, exposedField);
    add_field("on", &on, exposedField);

    ambientIntensity.set_float(0);
    color.set_colour(1, 1, 1);
    direction.set_vector(0, 0, -1);
    intensity.set_float(1);
    on.set_boolean(TRUE);

  }  // constructor


  void init_Ring() {}
  void add_Ring();
  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return CreateNodeReplaceHead; }


private:

  SFFloat ambientIntensity;
  SFColor color;
  SFVec3f direction;
  SFFloat intensity;
  SFBool on;


};  // DirectionalLight class





//
// Group
//

class Group : public GroupingNode
{

public:

  Group() : GroupingNode() { 
    set_node_name("Group"); 
    set_node_type(GroupType);
  }
  ~Group() {}

  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return HeadInsertGroupNode; }


private:


};  // Group class





//
// Inline
//

class Inline : public GroupingNode
{

public:

  Inline() : GroupingNode() { 
    set_node_name("Inline"); 
    set_node_type(InlineType);
    
    add_field("url", &url, exposedField);

  }  // constructor

  ~Inline() {}


  void parse();
  void init_Ring() {}  // because R3LoadVrml2File already does a push
  void add_Ring() {}   // and a pop

  USE_ActionType get_USE_action() { return ReturnVrmlNode; }

  void sendEventIn(int field_id, void *new_value);

  char *get_filespec() { return url[0]->get_string(); }


private:

  MFString url;


};  // Inline class





//
// PointLight
//

class PointLight : public GroupingNode
{

public:

  PointLight() : GroupingNode() { 
    set_node_name("PointLight"); 
    set_node_type(PointLightType); 

    add_field("ambientIntensity", &ambientIntensity, exposedField);
    add_field("attenuation", &attenuation, exposedField);
    add_field("color", &color, exposedField);
    add_field("intensity", &intensity, exposedField);
    add_field("location", &location, exposedField);
    add_field("on", &on, exposedField);
    add_field("radius", &radius, exposedField);

    ambientIntensity.set_float(0);
    attenuation.set_vector(1, 0, 0);
    color.set_colour(1, 1, 1);
    intensity.set_float(1);
    location.set_vector(0, 0, 0);
    on.set_boolean(TRUE);
    radius.set_float(100);

  }  // constructor

  ~PointLight() {}


  void init_Ring() {}
  void add_Ring();
  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return CreateNodeReplaceHead; }


private:

  SFFloat ambientIntensity;
  SFVec3f attenuation;
  SFColor color;
  SFFloat intensity;
  SFVec3f location;
  SFBool on;
  SFFloat radius;


};  // PointLight class





//
// SpotLight
//

class SpotLight : public GroupingNode
{

public:

  SpotLight() : GroupingNode() {
    set_node_name("SpotLight"); 
    set_node_type(SpotLightType);

    add_field("ambientIntensity", &ambientIntensity, exposedField);
    add_field("attenuation", &attenuation, exposedField);
    add_field("beamWidth", &beamWidth, exposedField);
    add_field("color", &color, exposedField);
    add_field("cutOffAngle", &cutOffAngle, exposedField);
    add_field("direction", &direction, exposedField);
    add_field("intensity", &intensity, exposedField);
    add_field("location", &location, exposedField);
    add_field("on", &on, exposedField);
    add_field("radius", &radius, exposedField);

    ambientIntensity.set_float(0);
    attenuation.set_vector(1, 0, 0);
    beamWidth.set_float(1.570796);
    color.set_colour(1, 1, 1);
    cutOffAngle.set_float(0.785398);
    direction.set_vector(0, 0, -1);
    intensity.set_float(1);
    location.set_vector(0, 0, 0);
    on.set_boolean(TRUE);
    radius.set_float(100);

  }  // constructor

  ~SpotLight() {}

  void init_Ring() {}
  void add_Ring();
  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return CreateNodeReplaceHead; }


private:

  SFFloat ambientIntensity;
  SFVec3f attenuation;
  SFFloat beamWidth;
  SFColor color;
  SFFloat cutOffAngle;
  SFVec3f direction;
  SFFloat intensity;
  SFVec3f location;
  SFBool on;
  SFFloat radius;


};  // SpotLight class





//
// Switch
//

class Switch : public GroupingNode
{

public:

  Switch() : GroupingNode() {
    set_node_name("Switch"); 
    set_node_type(SwitchType);

    add_field("choice", &choice, exposedField);
    add_field("whichChoice", &whichChoice, exposedField);

    whichChoice.set_integer(-1);

  }  // constructor

  ~Switch() {}


  void init_Ring();
  void add_Ring();

  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return Unknown; }


private:

  MFNode choice;
  SFInt32 whichChoice;

  R3SwitchNode *switch_node;


};  // Switch class





//
// LOD
//

class LOD : public Switch
{

public:

  LOD() : Switch() { 
    set_node_name("LOD");
    set_node_type(LODType);

    add_field("level", &level, exposedField);
    add_field("center", &center, field);
    add_field("range", &range, field);

    center.set_vector(0, 0, 0);

  }  // constructor

  ~LOD() {}


  void init_Ring();
  void add_Ring();

  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return ReturnVrmlNode; }


private:

  R3LodNode *lod_node;

  MFNode level;
  SFVec3f center;
  MFFloat range;


};  // LOD class





//
// Transform
//

class Transform : public GroupingNode
{

public:

  Transform() : GroupingNode() { 
    set_node_name("Transform"); 
    set_node_type(TransformType); 

    add_field("center", &center, exposedField);
    add_field("rotation", &rotation, exposedField);
    add_field("scale", &scale, exposedField);
    add_field("scaleOrientation", &scaleOrientation, exposedField);
    add_field("translation", &translation, exposedField);

    center.set_vector(0, 0, 0);
    rotation.set_axis(0, 0, 1);
    rotation.set_angle(0);
    scale.set_vector(1, 1, 1);
    scaleOrientation.set_axis(0, 0, 1);
    scaleOrientation.set_angle(0);
    translation.set_vector(0, 0, 0);

  }  // constructor

  ~Transform() {}


  void init_Ring();
  void add_Ring();
  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return HeadInsertGroupNode; }


private:

  SFVec3f center;
  SFRotation rotation;
  SFVec3f scale;
  SFRotation scaleOrientation;
  SFVec3f translation;

  R3TransformationNode *transform_node;
  R3Affine *affine;
  void construct_transform();


};  // Transform class


#endif




