//
// $Id: Interpolators.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include "Interpolators.h"



//
// Interpolator
//


//
// ScalarInterpolator
//

void
ScalarInterpolator::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

  switch(field_id) {
    case 0: {
      RNScalar f = set_fraction.get_float();
      // now compute the right value, and send it out
      int nr_entries = key.get_size();
      if (f < key[0]->get_float()) 
	value_changed.set_float(keyValue[0]->get_float());
      else if (f > key[nr_entries - 1]->get_float())
	value_changed.set_float(keyValue[nr_entries - 1]->get_float());
      else {
	for(int i=0; i<nr_entries; i++) {
	  if (f == key[i]->get_float()) {
	    value_changed.set_float(keyValue[i]->get_float());
	    break;
	  }
	  if (f < key[i]->get_float()) {
	    // first scale fraction between 0 and 1
	    f = (f - key[i-1]->get_float()) / 
	      (key[i]->get_float() - key[i-1]->get_float());
	    value_changed.set_float(keyValue[i-1]->get_float() + f *
				    (keyValue[i]->get_float() -
				     keyValue[i-1]->get_float()));
	    break;
	  }
	}  // for
      }
      value_changed.sendEventOut();
      break;
    }
    case 1: 
      //      key = (RNArray<RNScalar> *) new_value; break;
    case 2: 
      //      keyValue = (RNArray<RNScalar> *) new_value; break;
    case 3:
      event_error(3);  break;  // not an EventIn
    default:
      event_error(field_id);
  }  // switch
}  // ScalarInterpolator::eventIn





//
// OrientationInterpolator
//

void
OrientationInterpolator::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

  switch(field_id) {
    case 0: {
      RNScalar f = set_fraction.get_float();
#ifdef __DEBUG_ORIENT_INTERP__
      cout << "OrientationInterpolator: fraction is " << f << endl;
#endif
      // now compute the right value, and send it out
      int nr_entries = key.get_size();
      if (f < key[0]->get_float()) {
	value_changed.set_axis(keyValue[0]->get_axis());
	value_changed.set_angle(keyValue[0]->get_angle());
      }
      else if (f > key[nr_entries - 1]->get_float()) {
	value_changed.set_axis(keyValue[nr_entries - 1]->get_axis());
	value_changed.set_angle(keyValue[nr_entries - 1]->get_angle());
      }
      else {
	for(int i=0; i<nr_entries; i++) {
	  if (f == key[i]->get_float()) {
	    value_changed.set_axis(keyValue[i]->get_axis());
	    value_changed.set_angle(keyValue[i]->get_angle());
	    break;
	  }
	  if (f < key[i]->get_float()) {
	    // first scale fraction between 0 and 1
	    f = (f - key[i-1]->get_float()) / 
	      (key[i]->get_float() - key[i-1]->get_float());
	    // only reason we need a separate class... SFRotation...
	    value_changed.set_axis(keyValue[i-1]->get_axis() + 
				   (f * (keyValue[i]->get_axis() - 
					 keyValue[i-1]->get_axis())));
	    value_changed.set_angle(keyValue[i-1]->get_angle() + f *
				    (keyValue[i]->get_angle() - 
				     keyValue[i-1]->get_angle()));
	    break;
	  }
	}  // for
      }
      R3Vector axis = value_changed.get_axis();
#ifdef __DEBUG_ORIENT_INTERP__
      cout << "Axis: " << axis.X() << ", " << axis.Y() << ", " << axis.Z()
	<< " Angle: " << value_changed.get_angle() << endl;
#endif
      value_changed.sendEventOut();
      break;
    }
    case 1: 
      //      key = (RNArray<RNScalar> *) new_value; break;
    case 2: 
      //      keyValue = (RNArray<SFRotation *> *) new_value; break;
    case 3:
      event_error(3);  break;  // not an EventIn
    default:
      event_error(field_id);
  }  // switch
}  // OrientationInterpolator::eventIn





//
// SFVec3fInterpolator
//

void
SFVec3fInterpolator::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

  switch(field_id) {
    case 0: {
      RNScalar f = set_fraction.get_float();
      // cout << "SFVec3fInterpolator got a fraction " << f << endl;
      // now compute the right value, and send it out
      int nr_entries = key.get_size();
      if (f < key[0]->get_float())
	value_changed.set_vector(keyValue[0]->get_vector());
      else if (f > key[nr_entries - 1]->get_float()) 
	value_changed.set_vector(keyValue[nr_entries - 1]->get_vector());
      else {
	for(int i=0; i<nr_entries; i++) {
	  if (f == key[i]->get_float()) {
	    value_changed.set_vector(keyValue[i]->get_vector());
	    break;
	  }
	  if (f < key[i]->get_float()) {
	    // first scale fraction between 0 and 1
	    f = (f - key[i-1]->get_float()) / 
	      (key[i]->get_float() - key[i-1]->get_float());
	    R3Vector V = keyValue[i-1]->get_vector() + f *
	      (keyValue[i]->get_vector() - keyValue[i-1]->get_vector());
	    value_changed.set_vector(V);
	    break;
	  }
	}  // for
      }
      value_changed.sendEventOut();
      break;
    }
    case 1: 
      //      key = (RNArray<RNScalar> *) new_value; break;
    case 2: 
      //      keyValue = (RNArray<R3Point *> *) new_value; break;
    case 3:
      event_error(3);  break;  // not an EventIn
    default:
      event_error(field_id);
  }  // switch

}  // SF_Vec3fInterpolator::eventIn





//
// MFVec3fInterpolator
//

void
MFVec3fInterpolator::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

  switch(field_id) {
    case 0: {
      RNScalar f = set_fraction.get_float();
      // now compute the right value, and send it out
      int nr_entries = key.get_size();
      if (f < key[0]->get_float()) send_out_vectors(0);
      else if (f > key[nr_entries - 1]->get_float())
	send_out_vectors((nr_entries - 1) * nr_outgoing_vectors);
      else {
	for(int i=0; i<nr_entries; i++) {
	  if (f == key[i]->get_float()) {
	    send_out_vectors(i * nr_outgoing_vectors);
	    break;
	  }
	  if (f < key[i]->get_float()) {
	    // first scale fraction between 0 and 1
	    f = (f - key[i-1]->get_float()) / 
	      (key[i]->get_float() - key[i-1]->get_float());
	    // scale each outgoing vector appropriately...
	    value_changed.clear();
	    int start_index = (i-1) * nr_outgoing_vectors;
	    int end_index = i * nr_outgoing_vectors;
	    for(int j=start_index; j<end_index; j++) {
	      R3Vector V;
	      V = keyValue[j]->get_vector() + f *
		(keyValue[j + nr_outgoing_vectors]->get_vector() - 
		 keyValue[j]->get_vector());
	      SFVec3f *out_vector = new SFVec3f;
	      out_vector->set_vector(V);
	      value_changed.add(out_vector);
	    }  // for
	    break;
	  }
	}  // for
      }
      value_changed.sendEventOut();
      break;
    }
    case 1: 
      //      key = (RNArray<RNScalar> *) new_value; break;
    case 2: 
      //      keyValue = (RNArray<R3Point *> *) new_value; break;
    case 3:
      event_error(3);  break;  // not an EventIn
    default:
      event_error(field_id);
  }  // switch

}  // MFVec3fInterpolator::eventIn



void
MFVec3fInterpolator::send_out_vectors(int start_index)
{
  value_changed.clear();

  int end_index = start_index + nr_outgoing_vectors;
  for(int i=start_index; i < end_index; i++) {
    SFVec3f *out_vector = new SFVec3f;
    out_vector->set_vector(keyValue[i]->get_vector());
    value_changed.add(out_vector);
  }

  value_changed.sendEventOut();

}  // send_out_vectors


//
// PositionInterpolator
//

//
// ColorInterpolator
//

//
// CoordinateInterpolator
//

//
// NormalInterpolator
//

