//
// $Id: MatAppShape.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include "MatAppShape.h"
#include "Textures.h"
#include "Runtime.h"





//
// Material
//

void
Material::add_Ring()
{
  RNScalar a = ambientIntensity.get_float();
  R3Brdf *brdf = new R3Brdf(RNRgb(a, a, a),
			    diffuseColor.get_colour(), 
			    specularColor.get_colour(),
			    emissiveColor.get_colour(), 
			    shininess.get_float(), 
			    1.0 - transparency.get_float());

  brdf_ptr = brdf;
}  // Material::add_Ring



void
Material::sendEventIn(int field_id, void *new_value)
{
  //  const int ambientIntensity_index = 0;
  //  const int diffuseColor_index = 1;
  //  const int emissiveColor_index = 2;
  //  const int shininess_index = 3;
  //  const int specularColor_index = 4;
  //  const int transparency_index = 5;
  
  VrmlNode::sendEventIn(field_id, new_value);

  switch(field_id) {
    case 0:
      {
	RNScalar a = ambientIntensity.get_float();
	brdf_ptr->SetAmbient(RNRgb(a, a, a));
	break;
      }
    case 1:
      brdf_ptr->SetDiffuse(diffuseColor.get_colour());
      break;
      
    case 2:
      brdf_ptr->SetEmission(emissiveColor.get_colour());
      break;
      
    case 3:
      brdf_ptr->SetShininess(shininess.get_float());
      break;
      
    case 4:
      brdf_ptr->SetSpecular(specularColor.get_colour());
      break;
      
    case 5:
      brdf_ptr->SetOpacity(1 - transparency.get_float());
      break;
    default: event_error(field_id);
  }  // switch

}  // Material::eventIn





//
// Appearance
//

void
Appearance::init_Ring()
{
  Material *vrml_material = new Material;
  material.set_node_ptr(vrml_material);
  R3Material *ring_material = new R3Material(vrml_material->get_brdf_ptr());
  R3MaterialNode *material_node = new R3MaterialNode(ring_material);
  set_ring_node_ptr(material_node);
}  // Appearance::init_Ring



void
Appearance::add_Ring()
{
  R3Material *ring_material;
  if (texture.get_node_ptr() && 
      ((Texture *) texture.get_node_ptr())->have_read_image()) {
    // maybe we forgot the TextureTransform
    if (textureTransform.get_node_ptr() == NULL) 
      textureTransform.set_node_ptr(new TextureTransform);
    ring_material = 
      new R3Material(((Material *) material.get_node_ptr())->get_brdf_ptr(),
		     ((Texture *) texture.get_node_ptr())->get_ring_ptr(),
		     ((TextureTransform *) textureTransform.get_node_ptr())->get_ring_ptr());
  }
  else
    ring_material = new R3Material(((Material *) material.get_node_ptr())->get_brdf_ptr());

  R3MaterialNode *ring_node = new R3MaterialNode(ring_material);

  set_ring_node_ptr(ring_node);

}  // Appearance::add_Ring



void
Appearance::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // Appearance::eventIn





//
// Shape
//

void
Shape::add_Ring()
{
  // now put the check for missing appearance here (used to be in parse)
  if (appearance.get_node_ptr() == NULL) {
    Appearance *app_ptr = new Appearance;
    app_ptr->init_Ring();
    appearance.set_node_ptr(app_ptr);
  }
    
  // add the R3ShapeNode to the kids to the R3MaterialNode
  //  vrml_appearance->get_node_ptr()->InsertNode(vrml_shape_node->get_node_ptr());
  R3Node *temp_node_ptr = appearance.get_node_ptr()->get_ring_node_ptr();
  R3Node *node_ptr_2 = geometry.get_node_ptr()->get_ring_node_ptr();
  temp_node_ptr->InsertNode(node_ptr_2);

  // add the R3MaterialNode to the kids of the head of the group_stack
  group_stack.Head()->InsertNode(appearance.get_node_ptr()->get_ring_node_ptr());
  
  set_ring_node_ptr(appearance.get_node_ptr()->get_ring_node_ptr());

}  // Shape::add_Ring



void
Shape::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // Shape::eventIn



