//
// $Id: MiscNodes.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

//
// AudioClip, FontStyle, Sound, WorldInfo
//

#include "RNBasics/RNCompat.h"

#include <sys/types.h>
#include <math.h>
#if (RN_OS == RN_IRIX)
#include <sys/prctl.h>
#include <dmedia/audio.h>
#include <signal.h>
#endif

#include "MiscNodes.h"
#include "R3Io.h"





//
// AudioClip
//

void
AudioClip::add_Ring()
{
#if (RN_OS == RN_IRIX)  
  char *filespec = url[0]->get_string();

  soundbuffer = NULL;

  char *extension = strrchr(filespec, '.');
  if (!extension) {
    cur_vrml_file->print_error_line();
    cerr << "No extension in audio filespec [" << filespec << "]" << endl;
  }
  else {
    extension++;
    if (strcmp(extension, "raw") == 0) {
      FILE *audiofile = fopen(filespec, "rb");

      if (!audiofile) {
	cur_vrml_file->print_error_line();
	cerr << "Could not open audiofile [" 
	     << filespec << "]" << endl;
	return;
      }

      fseek(audiofile, 0, SEEK_END);
      long filesize = ftell(audiofile);
      nr_samples = filesize / 2;
      fseek(audiofile, 0, SEEK_SET);

      soundbuffer = new short[nr_samples];

      if (!soundbuffer) {
	cur_vrml_file->print_error_line();
	cerr << "Could not allocate memory for audiofile ["
	     << filespec << "]" << endl;
	return;
      }

      fread(soundbuffer, 1, filesize, audiofile);

      if (R3Vrml_get_info_output())
	cout << "AudioClip: read raw audiofile [" << filespec << "]" << endl;
    }
  }  // else, found an extension
#endif
}  // AudioClip::add_Ring



void
AudioClip::start_audio()
{
#if (RN_OS == RN_IRIX)  
  // if we don't have a raw audio file, just use sfplay
  if (!soundbuffer) {
    char command[128];
    strcpy(command, "/usr/sbin/sfplay ");
    strcat(command, url[0]->get_string());
    strcat(command, " &");
    system(command);
  }
  else
    audio_pid = sproc(play_audio, PR_SALL, this);
#endif  
}  // AudioClip::start_audio



void
AudioClip::actual_set_samplerate()
{
#if (RN_OS == RN_IRIX)
  long val_buffer[2];
  val_buffer[0] = AL_OUTPUT_RATE;
  val_buffer[1] = my_samplerate;
  ALsetparams(AL_DEFAULT_DEVICE, val_buffer, 2);
#endif
}  // AudioClip::actual_set_samplerate



void
AudioClip::set_samplerate(long new_value)
{
  my_samplerate = new_value;
}  // AudioClip::set_samplerate



void
AudioClip::stop_audio()
{
#if (RN_OS == RN_IRIX)
  // if we don't have a raw audio file, attempt to kill sfplay
  if (!soundbuffer) {
  // alright, don't ask... this is not the right way...
    system("kill `ps -e | grep sfplay`");
  }
  else {
    kill(audio_pid, 9);
    my_samplerate = previous_rate;
    actual_set_samplerate();
  }
#endif
}  // AudioClip::stop_audio



void
play_audio(void *clip_ptr)
{
#if (RN_OS == RN_IRIX)  
  AudioClip *audioclip_ptr = (AudioClip *) clip_ptr;

  ALconfig config;  // port configuration structure
  ALport outport;   // audio configuration structure

  prctl(PR_TERMCHILD);        // exit when parent does
  sigset(SIGHUP, SIG_DFL);    // exit when sent SIGHUP by TERMCHILD

  long val_buffer[2];
  val_buffer[0] = AL_OUTPUT_RATE;
  ALgetparams(AL_DEFAULT_DEVICE, val_buffer, 2);
  long previous_rate = val_buffer[1];
  // also set in AudioClip in case of kill
  audioclip_ptr->save_previous_samplerate(previous_rate);
  //  cout << "current AL_OUTPUT_RATE: " << val_buffer[1] << endl;
  audioclip_ptr->actual_set_samplerate();

  config = ALnewconfig();
  ALsetqueuesize(config, 11000);
  ALsetwidth(config, AL_SAMPLE_16);   // 16-bit signed samples
  if (audioclip_ptr->get_stereo())
    ALsetchannels(config, AL_STEREO);
  else
    ALsetchannels(config, AL_MONO);

  long nr_samples = audioclip_ptr->get_nr_samples();
  //  cout << "Playing " << nr_samples << " samples" << endl;
  outport = ALopenport("outport", "w", config);
  ALwritesamps(outport, audioclip_ptr->get_soundbuffer(), nr_samples);

  ALfreeconfig(config);
  ALcloseport(outport);

  // don't reset to previous samplerate: if it was low, you'll still
  // hear a small part of the sample at the low rate...
  //  audioclip_ptr->set_samplerate(previous_rate);
  //  audioclip_ptr->actual_set_samplerate();
#endif
}  // AudioClip::play_audio



void
AudioClip::sendEventIn(int field_id, void *new_value)
{
  //  const int description_index = 0;
  //  const int loop_index = 1;
  //  const int pitch_index = 2;
  //  const int startTime_index = 3;
  //  const int stopTime_index = 4;
  //  const int url_index = 5;
  //  const int duration_changed_index = 6;
  //  const int isActive_index = 7;
  
  VrmlNode::sendEventIn(field_id, new_value);

  //  cout << "AudioClip::sendEventIn, id " << field_id << endl;
  switch(field_id) {
    case 3:
      start_audio();
      break;

    case 4:
      stop_audio();
      break;

  }  // switch

}  // AudioClip::sendEventIn





//
// FontStyle
//

void
FontStyle::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // FontStyle::eventIn





//
// Sound
//

void
Sound::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // Sound::sendEventIn





//
// WorldInfo
//

void
WorldInfo::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // WorldInfo::eventIn



