//
// $Id: MiscNodes.h,v 1.1 2000/01/17 18:47:59 min Exp min $
//

#ifndef __MISCNODES_H
#define __MISCNODES_H

#include "VrmlNode.h"
#include "Fields.h"





void play_audio(void *clip_ptr);

class AudioClip : public VrmlNode
{

public:

  AudioClip() { 
    set_node_name("AudioClip");
    set_node_type(AudioClipType);

    add_field("description", &description, exposedField);
    add_field("loop", &loop, exposedField);
    add_field("pitch", &pitch, exposedField);
    add_field("startTime", &startTime, exposedField);
    add_field("stopTime", &stopTime, exposedField);
    add_field("url", &url, exposedField);
    add_field("duration_changed", &duration_changed, eventOut);
    add_field("isActive", &isActive, eventOut);

    description.set_string("");
    loop.set_boolean(FALSE);
    pitch.set_float(1.0);
    startTime.set_time(0);
    stopTime.set_time(0);
    isActive.set_boolean(FALSE);

    stereo = 0;

  }  // constructor

  ~AudioClip() {
    if (soundbuffer) delete soundbuffer;
    my_samplerate = previous_rate;
    actual_set_samplerate();
  }  // destructor


  void add_Ring();

  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return ReturnVrmlNode; }

  short *get_soundbuffer() { return soundbuffer; }
  long get_nr_samples() { return nr_samples; }

  void start_audio();
  void stop_audio();
  void set_samplerate(long new_value);
  void save_previous_samplerate(long new_value) { previous_rate = new_value; }
  void actual_set_samplerate();

  void set_stereo(int new_value) { stereo = new_value; }
  int get_stereo() { return stereo; }


private:

  SFString description;
  SFBool loop;
  SFFloat pitch;
  SFTime startTime;
  SFTime stopTime;
  MFString url;
  SFTime duration_changed;
  SFBool isActive;

  short *soundbuffer;
  long nr_samples;
  long my_samplerate;
  long previous_rate;

  int audio_pid;

  int stereo;


};  // AudioClip class





class FontStyle : public VrmlNode
{

public:

  FontStyle() { 
    set_node_name("FontStyle"); 
    set_node_type(FontStyleType);

    add_field("family", &family, field);
    add_field("horizontal", &horizontal, field);
    add_field("justify", &justify, field);
    add_field("language", &language, field);
    add_field("leftToRight", &leftToRight, field);
    add_field("size", &size, field);
    add_field("spacing", &spacing, field);
    add_field("style", &style, field);
    add_field("topToBottom", &topToBottom, field);

    family.set_string("SERIF");
    horizontal.set_boolean(TRUE);
    SFString *justify_string = new SFString("BEGIN");
    justify.add(justify_string);
    leftToRight.set_boolean(TRUE);
    size.set_float(1.0);
    spacing.set_float(1.0);
    style.set_string("PLAIN");
    topToBottom.set_boolean(TRUE);

  }  // constructor

  ~FontStyle() {}


  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return Unknown; }


private:

  SFString family;
  SFBool horizontal;
  MFString justify;
  SFString language;
  SFBool leftToRight;
  SFFloat size;
  SFFloat spacing;
  SFString style;
  SFBool topToBottom;


};  // FontStyle class





class Sound : public VrmlNode
{

public:

  Sound() { 
    set_node_name("Sound");
    set_node_type(SoundType);

    source = new SFNode;

    add_field("direction", &direction, exposedField);
    add_field("intensity", &intensity, exposedField);
    add_field("location", &location, exposedField);
    add_field("maxBack", &maxBack, exposedField);
    add_field("maxFront", &maxFront, exposedField);
    add_field("minBack", &minBack, exposedField);
    add_field("minFront", &minFront, exposedField);
    add_field("priority", &priority, exposedField);
    add_field("source", source, exposedField);
    add_field("spatialize", &spatialize, field);

    direction.set_vector(0, 0, 1);
    intensity.set_float(1);
    location.set_vector(0, 0, 0);
    maxBack.set_float(10);
    maxFront.set_float(10);
    minBack.set_float(1);
    minFront.set_float(1);
    priority.set_float(0);
    spatialize.set_boolean(TRUE);

  }  // constructor

  ~Sound() {}


  USE_ActionType get_USE_action() { return ReturnVrmlNode; }

  void sendEventIn(int field_id, void *new_value);


private:

  SFVec3f direction;
  SFFloat intensity;
  SFVec3f location;
  SFFloat maxBack;
  SFFloat maxFront;
  SFFloat minBack;
  SFFloat minFront;
  SFFloat priority;
  SFNode *source;
  SFBool spatialize;

  
};  // Sound class





class WorldInfo : public VrmlNode
{

public:

  WorldInfo() { 
    set_node_name("WorldInfo"); 
    set_node_type(WorldInfoType);

    add_field("info", &info, field);
    add_field("title", &title, field);

    title.set_string("");

  }  // constructor

  ~WorldInfo() {}


  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return ReturnVrmlNode; }


private:

  MFString info;
  SFString title;


};  // WorldInfo class


#endif

