//
// $Id: PROTONodes.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include "PROTONodes.h"
#include "parse_nodes.h"
#include "Fields.h"
#include "VsExprNodes.h"
#include "Runtime.h"

PROTO *current_proto = NULL;

VsSymbolTable PROTOVarTable;





//
// PROTO
//

PROTO::PROTO()
{
  set_node_type(PROTOType);
  default_fields = new VsSymbolTable;
}  // constructor



PROTO::PROTO(PROTO& other) : VrmlNode(other)
{
  // copy default fields
  default_fields = new VsSymbolTable(other.get_default_fields());
  
  // copy body text
  body_text = new char[strlen(other.get_body_text()) + 1];
  strcpy(body_text, other.get_body_text());
}  // copy constructor



PROTO::~PROTO()
{
#ifdef __DEBUG_DESTRUCTORS__
  cout << "PROTO destructor: " << get_node_name() << endl;
#endif
  if (default_fields) delete default_fields;
  if (body_text) delete body_text;
}  // destructor



void
PROTO::parse()
{
  char token[MAX_TOKEN_LENGTH];

  cur_vrml_file->get_token(token);
  set_node_name(token);

  cur_vrml_file->accept_token(OPEN_BRACKET);

  // now parse the fields
  parse_fields();

  // now just store the body of the PROTO
  parse_body();

}  // PROTO::parse



void
PROTO::parse_fields()
{
  char token[MAX_TOKEN_LENGTH];
  int done = 0;

  while (!done) {
    cur_vrml_file->get_token(token);
    if (token[0] == CLOSE_BRACKET) done = 1;
    else {
      VsVariableNode *var_ptr = new VsVariableNode;
      // EXTERNPROTO doesn't have default values for the fields...
      var_ptr->parse(token, get_node_type() == PROTOType);
      default_fields->add(var_ptr);

      // we still need this for the old ROUTEing stuff...
      //      add_field_var_name(var_ptr->get_field_var_name());
      //add_field_ptr(var_ptr->get_field_ptr());
      // now add to fields <<<
    }
  }  // while

}  // PROTO::parse_fields



void
PROTO::parse_body()
{
  RNArray<char> *body_array = new RNArray<char>;

  int k;
  do {
    k = cur_vrml_file->get_buffered_char();
  } while (k != OPEN_BRACE);

  body_array->Insert(k);
  int level = 1;

  while (level > 0) {
    k = cur_vrml_file->get_buffered_char();
    if (k == MY_EOF) {
      cerr << "Error: unexpected EOF when parsing PROTO body" << endl;
      level = 0;
    }
    body_array->Insert(k);
    if (k == OPEN_BRACE) level++;
    else if (k == CLOSE_BRACE) level--;
  }  // while
  cur_vrml_file->get_buffered_char();  // go past the closing brace

  int nr_chars = body_array->NEntries();
  body_text = new char[nr_chars + 1];

  for(int i=0; i<nr_chars; i++) {
    body_text[i] = body_array->Kth(i);
  }
  body_text[nr_chars] = '\0';

  delete body_array;

}  // PROTO::parse_body
    
    

void
PROTO::instantiate()
{
#ifdef __DEBUG__
  cout << endl << "Instantiating PROTO: " << get_name() << endl << endl;
#endif

  if (body_text == NULL) {
    cur_vrml_file->print_error_line();
    cerr << "PROTO has no definition (maybe its file wasn't found)"  << endl;
    exit(1);
  }

  char token[MAX_TOKEN_LENGTH];
  int done = 0;

  current_proto = this;
  my_DEF_table.store_state();

  // now here copy the default VsSymbolTable to the actual VsSymbolTable
  VsSymbolTable *actual_fields = get_var_table_ptr();
  actual_fields->store_state();
  int nr_fields = default_fields->get_size();
  for(int i=0; i<nr_fields; i++) {
    VsSymbolNode *var_ptr = (*default_fields)[i];
    Field *field_ptr = var_ptr->get_field_ptr();
    VsVariableNode *new_var_ptr =
      new VsVariableNode(var_ptr->get_name(), 
			 field_ptr->create_copy(),
			 var_ptr->get_field_type());
    actual_fields->add(new_var_ptr);
    // set_field_ptr(i, new_var_ptr->get_field_ptr());
  }  // for

  // then parse the actual parameters to get some actual values
  parse_parameters();

  // now add these VsVariableNodes to the PROTOVarTable
  PROTOVarTable.store_state();
  for(i=0; i<nr_fields; i++) PROTOVarTable.add((*actual_fields)[i]);

  VrmlFile *prev_vrml_file = cur_vrml_file;
  cur_vrml_file = new VrmlFile;

  // set the filespec to "PROTO <protoname>" in case there are errors
  char my_name[MAX_TOKEN_LENGTH + 6];
  strcpy(my_name, "PROTO ");
  strcat(my_name, get_node_name());
  cur_vrml_file->set_filespec(my_name);

  cur_vrml_file->set_parse_source(body_text);
  cur_vrml_file->accept_token(OPEN_BRACE);
  done = 0;
  first_node_ptr = NULL;
  while (!done) {
    cur_vrml_file->get_token(token);
    if (token[0] == CLOSE_BRACE) done = 1;
    else if (!first_node_ptr) first_node_ptr = parse_node(token);
    else parse_node(token);
  }  // while

  PROTOVarTable.restore_state();
  my_DEF_table.restore_state();
  // we don't want to restore at all, because of incoming events!
  //  actual_fields->restore_state();  // Script node doesn't use PROTOVarTable
  // yet, there may be a problem with the creation of events... (index)

  delete cur_vrml_file;
  cur_vrml_file = prev_vrml_file;

#ifdef __DEBUG__
  cout << endl << "PROTO instantiated" << endl << endl;
#endif

}  // PROTO::instantiate



void
PROTO::parse_parameters()
{
  char token[MAX_TOKEN_LENGTH];
  cur_vrml_file->accept_token(OPEN_BRACE);

  int done = 0;

  while (!done) {
    cur_vrml_file->get_token(token);
    if (token[0] == CLOSE_BRACE) done = 1;
    else {
      VsSymbolTable *actual_fields = get_var_table_ptr();
      VsVariableNode *var_ptr = 
	(VsVariableNode *) actual_fields->find_symbol(token);

      if (!var_ptr) {
	cur_vrml_file->print_error_line();
	cerr << "Unknown field [" << token << "], in PROTO, aborting" << endl;
	exit(1);
      }
      else {
	Field *vrml_var_ptr = var_ptr->get_field_ptr();
	FieldVarType proto_field_type = vrml_var_ptr->get_field_var_type();

	if (proto_field_type == SFNodeType)
	  ((SFNode *) vrml_var_ptr)->parse_and_store_string();
	else if (proto_field_type == MFNodeType)
	  ((MFNode *) vrml_var_ptr)->parse_and_store_string();
	else {
	  vrml_var_ptr->clear();  // first remove default value
	  vrml_var_ptr->parse();
	}
      }  // else, we know this parameter
    }
  }  // while
}  // PROTO::parse_parameters



void
PROTO::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

  // first set my local values
  VsVariableNode *var_ptr = 
    (VsVariableNode *)(*default_fields)[field_id];
  var_ptr->get_field_ptr()->sendEventIn(new_value);

  // see if anyone else wants to know
  // only if it's an exposedField!!!
  get_field_ptr(field_id)->sendEventOut();

}  // PROTO::eventIn





//
// EXTERNPROTO
//

void
EXTERNPROTO::parse()
{
  char token[MAX_TOKEN_LENGTH];

  cur_vrml_file->get_token(token);
  set_node_name(token);

  cur_vrml_file->accept_token(OPEN_BRACKET);
  parse_fields();
  int nr_fields_before = default_fields->get_size();

  // now get the URL/URLs
  url.parse();

  // now get the file and store it in body_text
  char *filespec = url[0]->get_string();

  VrmlFile *proto_file = new VrmlFile;
  
  if (!proto_file->open(filespec)) {
    cur_vrml_file->print_error_line();
    cerr << "Could not open EXTERNPROTO file: " << filespec << endl;
    return;
  }

  proto_file->accept_token("PROTO");
  VrmlFile *prev_vrml_file = cur_vrml_file;
  cur_vrml_file = proto_file;

  // use our parse function again
  set_node_type(PROTOType);  // make sure we're parsing a real PROTO now
  PROTO::parse();
  set_node_type(EXTERNPROTOType);

  delete proto_file;
  cur_vrml_file = prev_vrml_file;

  // now we should do a parameter check:
  // compare the first half of our fields with the second half
  // keep the second half (they have the default values)
  // NOTE: the spec says that the EXTERNPROTO declaration may
  //   less parameters than the PROTO declaration,
  //   - that the EXTERNPROTO name may be different
  //   - that the parameter type may be different, e.g.:
  //     - eventIn in the EXTERNPROTO
  //     - exposedField in the PROTO
  //
  int nr_fields_after = default_fields->get_size() - nr_fields_before;
  if (nr_fields_before != nr_fields_after) {
    cerr << "Error: EXTERNPROTO/PROTO parameter mismatch: "
	 << nr_fields_before << "/" << nr_fields_after
	 << " parameters respectively" << endl;
    exit(1);
  }

  for(int i=0; i<nr_fields_before; i++) default_fields->remove_head();

}  // EXTERNPROTO::parse





//
// PROTOList
//

PROTO*
PROTOList::find_PROTO(char *nodename)
{
  int nr_entries = PROTOs.NEntries();

  for(int i=0; i<nr_entries; i++) {
    if (strcmp(nodename, PROTOs.Kth(i)->get_node_name()) == 0)
      return PROTOs.Kth(i);
  }
  return NULL;
}  // PROTOList::find_PROTO



