/* Include file for the R3 group node class */

#include "R3Intn.h"

/* Initialization functions */

int R3InitGroupNode();
void R3StopGroupNode();



/* Class definition */

class R3GroupNode : public R3InteriorNode {
   public:
        // Constructor functions
        R3GroupNode(const char *name = NULL, 
		    const R3Box& bbox = R3null_box, 
		    unsigned long flags = RN_NULL_FLAGS);
	R3GroupNode(const R3GroupNode& node);
	~R3GroupNode(void);

	// Group property functions/operators
	const R3Box& BBox(void) const;

	// Node property functions/operators
        virtual const R3Shape& BShape(void) const;
	virtual const RNInterval NFacets(void) const;
	virtual const RNFlags Flags(void) const;

	// Hierarchy property functions/operators
	virtual R3Node *ChildNode(int k) const;
	virtual const int NChildren(void) const;

	// Hierarchy manipulation functions/operators
	virtual void InsertNode(R3Node *node);
	virtual void RemoveNode(R3Node *node);

	// Manipulation functions/operators
	virtual void SetBBox(const R3Box& box);
	virtual void SetFlags(RNFlags flags);
	virtual void AddFlags(RNFlags flags);
	virtual void RemoveFlags(RNFlags flags);

	// Upkeep functions
	virtual void UpdateBBox(R3Node *node, const R3Box& oldbox, const R3Box& newbox);
	virtual void UpdateNFacets(R3Node *node, const RNInterval& oldnfacets, const RNInterval& newnfacets);
	virtual void UpdateFlags(R3Node *node, RNFlags oldflags, RNFlags newflags);

	// Simulation functions/operators 
//	virtual void Draw(R3Session& session, R3TraversalState& state);
//	virtual int Simulate(R3Session& session, R3TraversalState& state);

	// Class type definitions
	//R3_NODE_CLASS_DECLARATIONS(R3GroupNode);

    protected:
	//RNVector<R3Node *> *children;
	RNArray<R3Node*> *children;
	RNFlags flags, base_flags;
	RNInterval nfacets;
	R3Box bbox;
};



