//
// $Id: Runtime.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include "Runtime.h"





//
// EventQueue
//

void
EventQueue::add_event(Event *new_event)
{
  queue.InsertTail(new_event);
}  // EventQueue::add_event



int
EventQueue::process_events()
{
  if (queue.IsEmpty()) return 0;

  // see if there's something in the queue
  while (!queue.IsEmpty()) {
    // get the event at the head of the queue
    Event *event = (Event *) queue.Head();  // was HeadEntry
    queue.RemoveHead();

#ifdef __DEBUG_EVENTS__
    cout << "There's an event, from a "
	 << event->get_source_ptr()->get_node_name() << " to a "
	 << event->get_dest_ptr()->get_node_name() << endl;
#endif
    
    VrmlNode *dest_node = event->get_dest_ptr();

    // get value from source node, send it to dest node
    VrmlNode *source_node = event->get_source_ptr();
    Field *source_field = source_node->get_field_ptr(event->get_source_id());
    dest_node->sendEventIn(event->get_dest_id(), source_field);
  }  // while
  return 1;

}  // EventQueue::process_events





//
// TimeSensorList
//

void
TimeSensorList::add_timesensor(TimeSensor *new_timesensor)
{
  timesensors.Insert(new_timesensor);
}  // TimeSensorList::add_timesensor



int
TimeSensorList::send_updates()
{
  int nr_sensors = timesensors.NEntries();
  int redraw = 0;

  global_timer.read_since_1970();

  for(int i=0; i<nr_sensors; i++) {
    redraw |= timesensors.Kth(i)->update();
  }  // for

  return redraw;

}  // TimeSensorList::send_updates





//
// TouchSensorList
//

void
TouchSensorList::add_touchsensor(TouchSensor *new_touchsensor)
{
  touchsensors.Insert(new_touchsensor);
}  // TouchSensorList::add_touchsensor



int
TouchSensorList::click(int down)
{
  int nr_sensors = touchsensors.NEntries();
  int redraw = 0;

  for(int i=0; i<nr_sensors; i++) {
    redraw |= touchsensors.Kth(i)->click(down);
  }  // for

  return redraw;

}  // TouchSensorList::send_updates





//
// Globals
//

RNStack<R3GroupNode *> group_stack;
EventQueue event_queue;
TimeSensorList timesensors;
TouchSensorList touchsensors;
DEFTable my_DEF_table;
WallTimer global_timer;


