//
// $Id: Script.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include "Script.h"
#include "parse_nodes.h"
#include "Fields.h"
#include "PROTONodes.h"
#include "Runtime.h"






void
Script::parse()
{
#ifdef __DEBUG__
  cout << "Script::parse" << endl;
#endif

  cur_symbol_table = get_var_table_ptr();
  cur_vrml_file->accept_token(OPEN_BRACE);

  char token[MAX_TOKEN_LENGTH];
  int done = 0;

  while (!done) {
    cur_vrml_file->get_token(token);

    if (strcmp(token, "url") == 0) {
      // store some info useful for error messages
      url_line_nr = cur_vrml_file->get_line_nr();
      char *cur_filename = cur_vrml_file->get_filespec();
      url_filename = new char[strlen(cur_filename) + 1];
      strcpy(url_filename, cur_filename);
      url.parse();
    }
    else if (strcmp(token, "directOutput") == 0) directOutput.parse();
    else if (strcmp(token, "mustEvaluate") == 0) mustEvaluate.parse();
    else if ((strcmp(token, "field") == 0) ||
	     (strcmp(token, "eventIn") == 0) ||
	     (strcmp(token, "eventOut") == 0)) parse_field(token);
    else if (token[0] == CLOSE_BRACE) done = 1;
    else {
      parse_error(UnknownField, token);
      cur_vrml_file->read_until(CLOSE_BRACE);
    }
  }  // while

}  // Script::parse



void
Script::add_Ring()
{
  // now parse the url
  parse_url();
}  // Script::add_Ring



void
Script::parse_field(char *field_type_string)
{
  // note that when we're in a PROTO there can be an "IS something"
  // after every field...
  // note that "field" has a default value

  VsVariableNode *var_ptr = new VsVariableNode;
  // don't accept default values, because there may be an IS instead of
  // a default value, and we handle that differently...
  var_ptr->parse(field_type_string, FALSE);
  add_field(var_ptr);

  char token[MAX_TOKEN_LENGTH];
  cur_vrml_file->lookahead_token(token);
  if (strcmp(token, "IS") == 0) {
    cur_vrml_file->get_token(token);  // get the IS
    cur_vrml_file->get_token(token);
    int PROTO_field_index = current_proto->get_field_var_index(token);
    if (PROTO_field_index == -1) {
      cur_vrml_file->print_error_line();
      cerr << "IS reference to unknown field" << endl;
    }
    else {
      int my_field_index = get_field_var_index(var_ptr->get_field_var_name());
      if (var_ptr->get_field_type() == eventOut) {
	Event *new_event = 
	  new Event(this, my_field_index, current_proto, PROTO_field_index);
	get_field_ptr(my_field_index)->add_route(new_event);
      }
      else if (var_ptr->get_field_type() == field) {
	// NO! this has to be the other way round! (i.e. the PROTO field ptr
	// points to the Script field)
	// why?
	var_ptr->set_field_ptr(current_proto->get_field_ptr(PROTO_field_index));
      }
      else {  // eventIn foo IS protofield
	Event *new_event =
	  new Event(current_proto, PROTO_field_index, this, my_field_index);
	current_proto->get_field_ptr(PROTO_field_index)->add_route(new_event);
      }
    }
  }
  // else get the default value if there is one (exposedField and field)
  else if ((var_ptr->get_field_type() == exposedField) ||
	   (var_ptr->get_field_type() == field)) {
    // code taken from VsVariableNode.C ...
    Field *field_ptr = var_ptr->get_field_ptr();
    FieldVarType vrml_var_type = field_ptr->get_field_var_type();
      
    if (vrml_var_type == SFNodeType)
      ((SFNode *) field_ptr)->parse_and_store_string();
    else if (vrml_var_type == MFNodeType)
      ((MFNode *) field_ptr)->parse_and_store_string();
    else
      field_ptr->parse();
  }

}  // Script::parse_field



void
Script::parse_url()
{
  if (strncmp(url[0]->get_string(), "vrmlscript:", 11) &&
      strncmp(url[0]->get_string(), "javascript:", 11)) {
    cur_vrml_file->print_error_line();
    cerr << "Script Error: only inline VrmlScript and JavaScript supported"
	 << endl;
    return;
  }

  char *script_source = url[0]->get_string() + 11;
  vrmlscript_ptr = new VrmlScript(script_source);
  vrmlscript_ptr->set_source_info(url_filename, url_line_nr);
  vrmlscript_ptr->parse();

  // run the initialize function
  vrmlscript_ptr->execute("initialize", 0, NULL);

}  // Script::parse_url



void
Script::sendEventIn(int field_id, void *new_value)
{
  //  const int url_index = 0;
  //  const int directOutput_index = 1;
  //  const int mustEvaluate_index = 2;
  
  VrmlNode::sendEventIn(field_id, new_value);

  VsVariableNode *var_ptr = 
    (VsVariableNode *)(*get_var_table_ptr())[field_id];

  // if it's one of Script's own fields return
  if (field_id <= 2) return;

  // point to my variables table
  cur_symbol_table = get_var_table_ptr();

  // execute corresponding vrmlscript function
  // create the (value, timestamp) arguments
  Field *argv[2];
  argv[0] = var_ptr->get_field_ptr();
  argv[1] = new SFTime(global_timer.last_read_since_1970());
  vrmlscript_ptr->execute(var_ptr->get_field_var_name(), 2, argv);

  // send eventOut(s)
  // loop through my_var_table, check if node is eventOut,
  // check if its send out flag is set, if so, send it
  int nr_vars = get_nr_fields();
  for(int i=0; i<nr_vars; i++) {
    VsSymbolTable *var_table_ptr = get_var_table_ptr();
    VsVariableNode *var_ptr = (VsVariableNode *)(*var_table_ptr)[i];
    if (var_ptr->get_changed()) {
      var_ptr->set_changed(0);
      var_ptr->get_field_ptr()->sendEventOut();
    }
  }  // for

}  // Script::eventIn


