//
// $Id: Sensors.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include "Sensors.h"
#include "parse_nodes.h"
#include "Runtime.h"





//
// CylinderSensor
//

void
CylinderSensor::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // CylinderSensor::eventIn





//
// PlaneSensor
//

void
PlaneSensor::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // PlaneSensor::eventIn





//
// ProximitySensor
//

void
ProximitySensor::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // ProximitySensor::eventIn





//
// SphereSensor
//

void
SphereSensor::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // SphereSensor::eventIn





//
// TimeSensor
//

int
TimeSensor::update()
{
  // first check if enabled
  if (!enabled.get_boolean()) return 0;

  // time contains number of secs since 1970
  // use last_read to save a gettimeofday call
  time.set_time(global_timer.last_read_since_1970());
  cycleTime.set_time(time.get_time());

  double start = startTime.get_time();

  // return 1 if an event was processed
  int redraw = 0;

  // check if we have to switch on
  if (!isActive.get_boolean()) {
    if ((time.get_time() < start) || 
	((time.get_time() > disable_time) && (mode != 0))) return 0;

#ifdef __DEBUG_TIMESENSOR__
    cout.precision(12);
    cout << "Starting TimeSensor, startT:" << startTime.get_time()
      << " stopT: " << stopTime.get_time() 
	<< " cycleInt: " << cycleInterval.get_time() 
	  << " currentT: " << time.get_time() << endl;
#endif
    isActive.set_boolean(TRUE);
    previous_modulo = 0;

    redraw |= isActive.sendEventOut();
    redraw |= cycleTime.sendEventOut();
  }
  else {
    // _first_ check if we have to stop...

    if (mode != 0) {  // mode 0 is loop forever
      if (time.get_time() > disable_time) {
#ifdef __DEBUG_TIMESENSOR__
	cout << "Disabling TimeSensor, cycleInt: "
	     << cycleInterval.get_time() << ", curtime:" << time.get_time()
	       << ", disable_time:" << disable_time << endl;
#endif
	isActive.set_boolean(FALSE);
	redraw |= isActive.sendEventOut();
      }
    }

    redraw |= set_fraction_changed();

    // these two are sent always if we're active
    redraw |= time.sendEventOut();
    redraw |= fraction_changed.sendEventOut();
  }  // else, we're active

  return redraw;
}  // TimeSensorCallback



int
TimeSensor::set_fraction_changed()
{
  // compute fraction_changed, code taken from the VRML 2.0 spec
  double temp = (time.get_time() - startTime.get_time()) /
    cycleInterval.get_time();
  double modulo = temp - floor(temp);

  if (modulo < previous_modulo) {
#ifdef __DEBUG_TIMESENSOR__
    cout << "TimeSensor, cycleInt: " << cycleInterval.get_time()
      << " has cycled, prev:" << previous_modulo
      << ", newmod:" << modulo << endl;
#endif
    fraction_changed.set_float(1.0);
    previous_modulo = modulo;
    return cycleTime.sendEventOut();
  }
  else fraction_changed.set_float(modulo);
  previous_modulo = modulo;
  return 0;

}  // TimeSensor::set_fraction_changed



void
TimeSensor::add_Ring()
{
  set_mode();
  set_disable_time();
  if (mode == 0) startTime.set_time(global_timer.read_since_1970());
}  // TimeSensor::add_Ring



void
TimeSensor::set_mode()
{
  //
  // mode 0: stopTime <= startTime, loop = TRUE
  // mode 1: stopTime <= startTime, loop = FALSE
  // mode 2: stopTime > startTime, loop = TRUE
  // mode 3: stopTime > startTime, loop = FALSE
  //
  if (stopTime.get_time() <= startTime.get_time()) {
    if (loop.get_boolean()) mode = 0;
    else mode = 1;
  }
  else {
    if (loop.get_boolean()) mode = 2;
    else mode = 3;
  }
}  // TimeSensor::set_mode



void
TimeSensor::set_disable_time()
{
  // set the disable time depending on the current mode
  switch(mode) {
    case 0: break;
    case 1: 
      disable_time = startTime.get_time() + cycleInterval.get_time();
      break;
    case 2: disable_time = stopTime.get_time();
    case 3: 
      if (stopTime.get_time() < 
	  (startTime.get_time() + cycleInterval.get_time()))
	disable_time = stopTime.get_time();
      else
	disable_time = startTime.get_time() + cycleInterval.get_time();
      break;
  }  // switch
}  // TimeSensor::set_disable_time



void
TimeSensor::sendEventIn(int field_id, void *new_value)
{
  //  const int cycleInterval_index = 0;
  //  const int enabled_index = 1;
  //  const int loop_index = 2;
  //  const int startTime_index = 3;
  //  const int stopTime_index = 4;
  //  const int cycleTime_index = 5;
  //  const int fraction_changed_index = 6;
  //  const int isActive_index = 7;
  //  const int time_index = 8;

#ifdef __DEBUG_TIMESENSOR__
  cout << "TimeSensor eventIn, field_id is " << field_id
       << ", cycleInt: " << cycleInterval.get_time() << endl;
#endif

  RNBoolean prev_enabled;
  // 1 = enabled
  if (field_id == 1) prev_enabled = enabled.get_boolean();

  // an active TimeSensor ignores set_cycleInterval and set_startTime
  // events, also ignores set_stopTime for stopTime <= startTime
  if (isActive.get_boolean()) {
    if ((field_id == 0) || (field_id == 3) || ((field_id == 4) &&
        (((SFTime *) new_value)->get_time() <= startTime.get_time()))) {
      return;
    }
  }

  VrmlNode::sendEventIn(field_id, new_value);

  switch(field_id) {
    case 0:
      set_mode();
      set_disable_time();
      cycleInterval.sendEventOut();
      break;

    case 1:
      // if enabled becomes FALSE, send out an isActive = FALSE
      if (!enabled.get_boolean() && prev_enabled) {
	isActive.set_boolean(FALSE);
	isActive.sendEventOut();
	// also send final time and fraction_changed events
	time.sendEventOut();
	fraction_changed.sendEventOut();
	// enabled is exposedField
	enabled.sendEventOut();
      }
      break;

    case 2:
      set_mode();
      set_disable_time();
      loop.sendEventOut();
      break;

    case 3:
#ifdef __DEBUG_TIMESENSOR__
      cout << "TimeSensor startTime now is: " << startTime.get_time() << endl;
#endif
      set_mode();
      set_disable_time();
      startTime.sendEventOut();
      break;

    case 4:
      set_mode();
      set_disable_time();
      stopTime.sendEventOut();
      break;

    default:
      // 5, 6, 7 and 8 are all eventOut's
      event_error(field_id);
  }  // switch
}  // TimeSensor::eventIn





//
// TouchSensor
//

int TouchSensorCallback(R3Session& session, R3TraversalState&, R3Node*,
			R3MessageType, const void* msg_buffer, int,
			R3CallbackNode *node, void* VrmlNode_ptr)
{
  int my_message = *(int *) msg_buffer;

  TouchSensor *sensor_ptr = (TouchSensor *) VrmlNode_ptr;

  // first check if we're enabled
  if (!sensor_ptr->get_enabled()) return 0;

  // then check if isOver is already true, or already false
  RNBoolean isOver = sensor_ptr->get_isOver();
  if ((isOver && (my_message == ISOVER_MESSAGE)) ||
      (!isOver && (my_message == NOTOVER_MESSAGE))) return 0;

  // do something depending on message type...
  SFBool new_value;
  switch (my_message) {
    case ISOVER_MESSAGE: new_value.set_boolean(TRUE); break;
    case NOTOVER_MESSAGE: new_value.set_boolean(FALSE); break;
    default:
      cerr << "Error: wrong message type in TouchSensorCallback" << endl;
  }  // switch
#ifdef __DEBUG_EVENTS__
  cout << "In TouchSensorCallback, setting isOver to " 
       << new_value.get_boolean() << endl;
#endif
  // no access to isOver_index in this callback
  sensor_ptr->sendEventIn(5, &new_value);

  return sensor_ptr->get_field_ptr(5)->sendEventOut();

}  // TouchSensorCallback



int
TouchSensor::click(int down)
{
  if (down) {
#ifdef __DEBUG_EVENTS__
    cout << "TouchSensor::click(DOWN)" << endl;
#endif
    if (!isOver.get_boolean() || isActive.get_boolean()) return 0;

    isActive.set_boolean(TRUE);
    isActive.sendEventOut();
  }
  else {  // button is released, deactivate
#ifdef __DEBUG_EVENTS__
    cout << "TouchSensor::click(UP)" << endl;
#endif
    isActive.set_boolean(FALSE);
    isActive.sendEventOut();

    if (isOver.get_boolean()) {  // if we were still over the geometry, send time ev.
      touchTime.set_time(global_timer.read_since_1970());
      touchTime.sendEventOut();
    }
  }
  return 1;
}  // TouchSensor::click
      
    


void
TouchSensor::add_Ring()
{
  // create a callbacknode, store pointer to it
  R3CallbackNode *callback_node = new R3CallbackNode();
  set_ring_node_ptr(callback_node);

  // add callback below to node, message mask: PICK
  // (pointer to me (this) is the application specific data)
  callback_node->InsertCallback(R3_USER_MESSAGE_MASK, TouchSensorCallback,
				this);

  // NOTE: what if the TouchSensor isn't the first of the kids: then
  // the kids that have been added before, need to become kids of the
  // R3CallbackNode...  (still to do)

  // insert node in the model, make it the new top of stack
  group_stack.Head()->InsertNode(callback_node);
  group_stack.RemoveHead();
  group_stack.Push(callback_node);

} // parse_touchsensor



void
TouchSensor::sendEventIn(int field_id, void *new_value)
{
  //  const int enabled_index = 0;
  //  const int hitNormal_changed_index = 1;
  //  const int hitPoint_changed_index = 2;
  //  const int hitTexCoord_changed_index = 3;
  //  const int isActive_index = 4;
  //  const int isOver_index = 5;
  //  const int touchTime_index = 6;
  
  VrmlNode::sendEventIn(field_id, new_value);
    
}  // TouchSensor::eventIn





//
// VisibilitySensor
//

void
VisibilitySensor::sendEventIn(int field_id, void *new_value)
{
  VrmlNode::sendEventIn(field_id, new_value);

}  // VisibilitySensor::eventIn



