//
// $Id: Sensors.h,v 1.1 2000/01/17 18:47:59 min Exp min $
//

#ifndef __SENSORS_H
#define __SENSORS_H

#include "VrmlNode.h"
#include "Fields.h"
#include "WallTimer.h"





class CylinderSensor : public VrmlNode
{

public:

  CylinderSensor() { 
    set_node_name((char*)"CylinderSensor"); 
    set_node_type(CylinderSensorType);

    add_field((char*)"autoOffset", &autoOffset, exposedField);
    add_field((char*)"diskAngle", &diskAngle, exposedField);
    add_field((char*)"enabled", &enabled, exposedField);
    add_field((char*)"maxAngle", &maxAngle, exposedField);
    add_field((char*)"minAngle", &minAngle, exposedField);
    add_field((char*)"offset", &offset, exposedField);
    add_field((char*)"isActive", &isActive, eventOut);
    add_field((char*)"rotation_changed", &rotation_changed, eventOut);
    add_field((char*)"trackPoint_changed", &trackPoint_changed, eventOut);

    autoOffset.set_boolean(TRUE);
    diskAngle.set_float(0.262);
    enabled.set_boolean(TRUE);
    maxAngle.set_float(-1);
    minAngle.set_float(0);
    offset.set_float(0);
    isActive.set_boolean(TRUE);
  }  // constructor
  ~CylinderSensor() {}


  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return Unknown; }


private:

  SFBool autoOffset;
  SFFloat diskAngle;
  SFBool enabled;
  SFFloat maxAngle;
  SFFloat minAngle;
  SFFloat offset;
  SFBool isActive;
  SFRotation rotation_changed;
  SFVec3f trackPoint_changed;

};  // CylinderSensor class





class PlaneSensor : public VrmlNode
{

public:

  PlaneSensor() { 
    set_node_name((char*)"PlaneSensor"); 
    set_node_type(PlaneSensorType);

    add_field((char*)"autoOffset", &autoOffset, exposedField);
    add_field((char*)"enabled", &enabled, exposedField);
    add_field((char*)"maxPosition", &maxPosition, exposedField);
    add_field((char*)"minPosition", &minPosition, exposedField);
    add_field((char*)"offset", &offset, exposedField);
    add_field((char*)"isActive", &isActive, eventOut);
    add_field((char*)"trackPoint_changed", &trackPoint_changed, eventOut);
    add_field((char*)"translation_changed", &translation_changed, eventOut);

    autoOffset.set_boolean(TRUE);
    enabled.set_boolean(TRUE);
    maxPosition.set_vector(-1, -1);
    minPosition.set_vector(0, 0);
    offset.set_vector(0, 0, 0);

  }  // constructor

  ~PlaneSensor() {}


  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return Unknown; }


private:

  SFBool autoOffset;
  SFBool enabled;
  SFVec2f maxPosition;
  SFVec2f minPosition;
  SFVec3f offset;
  SFBool isActive;
  SFVec3f trackPoint_changed;
  SFVec3f translation_changed;


};  // PlaneSensor class





class ProximitySensor : public VrmlNode
{

public:

  ProximitySensor() { 
    set_node_name((char*)"ProximitySensor"); 
    set_node_type(ProximitySensorType);

    add_field((char*)"center", &center, exposedField);
    add_field((char*)"size", &size, exposedField);
    add_field((char*)"enabled", &enabled, exposedField);
    add_field((char*)"isActive", &isActive, eventOut);
    add_field((char*)"position_changed", &position_changed, eventOut);
    add_field((char*)"orientation_changed", &orientation_changed, eventOut);
    add_field((char*)"enterTime", &enterTime, eventOut);
    add_field((char*)"exitTime", &exitTime, eventOut);

    center.set_vector(0, 0, 0);
    size.set_vector(0, 0, 0);
    enabled.set_boolean(TRUE);

  }  // constructor

  ~ProximitySensor() {}


  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return Unknown; }


private:

  SFVec3f center;
  SFVec3f size;
  SFBool enabled;
  SFBool isActive;
  SFVec3f position_changed;
  SFRotation orientation_changed;
  SFTime enterTime;
  SFTime exitTime;


};  // ProximitySensor class





class SphereSensor : public VrmlNode
{

public:

  SphereSensor() { 
    set_node_name((char*)"SphereSensor"); 
    set_node_type(SphereSensorType);

    add_field((char*)"autoOffset", &autoOffset, exposedField);
    add_field((char*)"enabled", &enabled, exposedField);
    add_field((char*)"offset", &offset, exposedField);
    add_field((char*)"isActive", &isActive, eventOut);
    add_field((char*)"rotation_changed", &rotation_changed, eventOut);
    add_field((char*)"trackPoint_changed", &trackPoint_changed, eventOut);

    autoOffset.set_boolean(TRUE);
    enabled.set_boolean(TRUE);
    offset.set_axis(0, 1, 0);
    offset.set_angle(0);

  }  // constructor

  ~SphereSensor() {}


  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return Unknown; }


private:

  SFBool autoOffset;
  SFBool enabled;
  SFRotation offset;
  SFBool isActive;
  SFRotation rotation_changed;
  SFVec3f trackPoint_changed;


};  // SphereSensor class





class TimeSensor : public VrmlNode
{

public:

  TimeSensor() {
    set_node_name((char*)"TimeSensor"); 
    set_node_type(TimeSensorType); 
  
    add_field((char*)"cycleInterval", &cycleInterval, exposedField);
    add_field((char*)"enabled", &enabled, exposedField);
    add_field((char*)"loop", &loop, exposedField);
    add_field((char*)"startTime", &startTime, exposedField);
    add_field((char*)"stopTime", &stopTime, exposedField);
    add_field((char*)"cycleTime", &cycleTime, eventOut);
    add_field((char*)"fraction_changed", &fraction_changed, eventOut);
    add_field((char*)"isActive", &isActive, eventOut);
    add_field((char*)"time", &time, eventOut);
  
    cycleInterval.set_time(1);
    enabled.set_boolean(TRUE);
    loop.set_boolean(FALSE);
    startTime.set_time(0);
    stopTime.set_time(0);
    isActive.set_boolean(FALSE);

    previous_modulo = 0;
  
  }  // constructor

  ~TimeSensor() {}

  void add_Ring();

  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return Unknown; }

  int update();


private:

  SFTime cycleInterval;
  SFBool enabled;
  SFBool loop;
  SFTime startTime;
  SFTime stopTime;
  SFTime cycleTime;
  SFFloat fraction_changed;
  SFBool isActive;
  SFTime time;

  //
  // mode 0: stopTime <= startTime, loop = TRUE
  // mode 1: stopTime <= startTime, loop = FALSE
  // mode 2: stopTime > startTime, loop = TRUE
  // mode 3: stopTime > startTime, loop = FALSE
  //
  int mode;
  void set_mode();

  // time to disable the sensor given the current mode
  double disable_time;
  void set_disable_time();

  // to see if we did a cycle
  double previous_modulo;

  int set_fraction_changed();


};  // TimeSensor class





static const int ISOVER_MESSAGE = 0;
static const int NOTOVER_MESSAGE = 1;

class TouchSensor : public VrmlNode
{

public:

  TouchSensor() { 
    set_node_name((char*)"TouchSensor");
    set_node_type(TouchSensorType);

    add_field((char*)"enabled", &enabled, exposedField);
    add_field((char*)"hitNormal_changed", &hitNormal_changed, eventOut);
    add_field((char*)"hitPoint_changed", &hitPoint_changed, eventOut);
    add_field((char*)"hitTexCoord_changed", &hitTexCoord_changed, eventOut);
    add_field((char*)"isActive", &isActive, eventOut);
    add_field((char*)"isOver", &isOver, eventOut);
    add_field((char*)"touchTime", &touchTime, eventOut);

    enabled.set_boolean(TRUE);
    isOver.set_boolean(FALSE);
    isActive.set_boolean(FALSE);

  }  // constructor

  ~TouchSensor() {}


  void add_Ring();

  void sendEventIn(int field_id, void *new_value);

  RNBoolean get_enabled() { return enabled.get_boolean(); }
  RNBoolean get_isOver() { return isOver.get_boolean(); }

  USE_ActionType get_USE_action() { return Unknown; }

  int click(int down);


private:

  SFBool enabled;
  SFVec3f hitNormal_changed;
  SFVec3f hitPoint_changed;
  SFVec3f hitTexCoord_changed;
  SFBool isActive;
  SFBool isOver;
  SFTime touchTime;


};  // TouchSensor class





class VisibilitySensor : public VrmlNode
{

public:

  VisibilitySensor() { 
    set_node_name((char*)"VisibilitySensor"); 
    set_node_type(VisibilitySensorType);

    add_field((char*)"center", &center, exposedField);
    add_field((char*)"enabled", &enabled, exposedField);
    add_field((char*)"size", &size, exposedField);
    add_field((char*)"enterTime", &enterTime, eventOut);
    add_field((char*)"exitTime", &exitTime, eventOut);
    add_field((char*)"isActive", &isActive, eventOut);

    center.set_vector(0, 0, 0);
    enabled.set_boolean(TRUE);
    size.set_vector(0, 0, 0);
    isActive.set_boolean(FALSE);

  }  // constructor

  ~VisibilitySensor() {}


  void sendEventIn(int field_id, void *new_value);

  USE_ActionType get_USE_action() { return Unknown; }


private:

  SFVec3f center;
  SFBool enabled;
  SFVec3f size;
  SFTime enterTime;
  SFTime exitTime;
  SFBool isActive;


};  // VisibilitySensor class


#endif


