//
// $Id: VrmlNode.h,v 1.1 2000/01/17 18:47:59 min Exp min $
//

#ifndef __VRMLNODE_H
#define __VRMLNODE_H

#include <iostream>
#include "R3Shapes/R3Shapes.h"
#include "R3Graphics/R3Graphics.h"
#include "Files.h"
#include "VsSymbolTable.h"
#include "VsExprNodes.h"
#include "Field.h"
#include "vrml_types.h"

typedef enum { UnknownField } ParseErrorType;

typedef enum { 
  HeadInsertNode, HeadInsertGroupNode, CreateNodeReplaceHead, 
  ReturnVrmlNode, ReturnFirstVrmlNode, Unknown
} USE_ActionType;





class VrmlNode
{

public:

  //
  // constructor/destructor
  //
  VrmlNode();
  VrmlNode(VrmlNode& other);
  virtual ~VrmlNode();

  //
  // virtual functions
  //

  virtual void parse();
  virtual USE_ActionType get_USE_action() { return Unknown; }
  virtual void init_Ring() {}
  virtual void add_Ring() {}
  virtual void sendEventIn(int field_id, void *new_value);

  //
  // node type
  //
  VrmlNodeType get_node_type() { return node_type; }
  void set_node_type(VrmlNodeType new_type) { node_type = new_type; }

  //
  // node name
  //
  char *get_node_name() { return node_name; }
  void set_node_name(char *newname) { strcpy(node_name, newname); }

  //
  // variable table
  //

  void add_field(VsVariableNode *var_ptr) { fields->add(var_ptr); }
  void add_field(char *field_var_name, Field *field_ptr, 
		 FieldType field_type) {
    fields->add(new VsVariableNode(field_var_name, field_ptr, field_type));
  }  // add_field

  int get_field_var_index(char *field_var_name) {
    return fields->get_symbol_index(field_var_name);
  }
  Field *get_field_ptr(int field_index) { 
    return (*fields)[field_index]->get_field_ptr();
  }
  int get_nr_fields() { return fields->get_size(); }
  VsSymbolTable *get_var_table_ptr() { return fields; }

  //
  // RING node ptr
  //
  R3Node *get_ring_node_ptr() { return node_ptr; }
  void set_ring_node_ptr(R3Node *new_ptr) { node_ptr = new_ptr; }

  //
  // error messages
  //
  void parse_error(ParseErrorType error_type, char *token);
  void event_error(int field_id);
  void no_eventIn_error(int field_id);


private:

  VrmlNodeType node_type;
  char *node_name;

  VsSymbolTable *fields;

  R3Node *node_ptr;


};  // VrmlNode class


#endif


