//
// $Id: VsExprNodes.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include "VsExprNodes.h"

//
// VsBinaryExprNode
//

#include <iostream.h>
#include <math.h>
#include "Fields.h"
#include "MathConstants.h"





Field *
VsBinaryExprNode::evaluate()
{
  //
  // NOTE: unary operators are evaluated here too, with a left node NULL
  //
#ifdef __DEBUG_VRMLSCRIPT__
  if (left == NULL)
    cout << "  evaluating unary operator: " << my_op;
  else
    cout << "  evaluating binary operator: " << my_op;
  if ((left) && (left->get_node_type() == VsVariableNodeType) &&
      (((VsVariableNode *) left)->get_field_ptr()))
    cout << ", left: "
	 << ((VsVariableNode *) left)->get_field_ptr()->get_field_var_type_string();
  if ((right) && (right->get_node_type() == VsVariableNodeType) &&
      (((VsVariableNode *) right)->get_field_ptr()))
    cout << ", right: "
	 << ((VsVariableNode *) right)->get_field_ptr()->get_field_var_type_string();
  cout << endl;
#endif

  if (strcmp(my_op, "=") == 0) {
    ((VsVariableNode *) left)->set_changed(1);
    // if left pointer is NULL, it needs to allocate memory for a new
    // Field with the same type as the right hand side
    if (((VsVariableNode *) left)->get_field_ptr() == NULL) {
      ((VsVariableNode *) left)->set_field_ptr(right->evaluate()->create_copy());
      return left->evaluate();
    }
    else
      return *(((VsVariableNode *) left)->get_field_ptr()) = right->evaluate();
  }
  else if (strcmp(my_op, "+") == 0) {
    return (*(left->evaluate()) + right->evaluate());
  }
  else if (strcmp(my_op, "-") == 0) {
    return (*(left->evaluate()) - right->evaluate());
  }
  else if (strcmp(my_op, "+=") == 0) {
    return (*(left->evaluate()) += right->evaluate());
  }
  else if (strcmp(my_op, "-=") == 0) {
    return (*(left->evaluate()) -= right->evaluate());
  }
  else if (strcmp(my_op, "*") == 0) {
    return (*(left->evaluate()) * right->evaluate());
  }
  else if (strcmp(my_op, "/") == 0) {
    return (*(left->evaluate()) / right->evaluate());
  }
  else if (strcmp(my_op, "!") == 0) {
    return !(*(right->evaluate()));
  }
  else if (strcmp(my_op, "<") == 0) {
    return (*(left->evaluate()) < right->evaluate());
  }
  else if (strcmp(my_op, ">") == 0) {
    return (*(left->evaluate()) > right->evaluate());
  }
  else if (strcmp(my_op, "++") == 0) {
    if (left->get_node_type() == VsConstantNodeType)  // then it's prefix
      return (++(*(right->evaluate())));
    else
      return ((*(left->evaluate()))++);  // postfix
  }
  else if (strcmp(my_op, "--") == 0) {
    if (left->get_node_type() == VsConstantNodeType) // prefix
      return (--(*(right->evaluate())));
    else
      return ((*(left->evaluate()))--);  // postfix
  }
  else if (strcmp(my_op, "==") == 0) {
    return (*(left->evaluate()) == right->evaluate());
  }
  else {
    cerr << "VsBinaryExprNode: Error: operator " << my_op 
	 << " not yet supported" << endl;
    return new SFInt32(-1);
  }
    
}  // VsBinaryExprNode::evaluate





//
// VsConstantNode
//

VsConstantNode::VsConstantNode(char *new_value_string)
{
  strcpy(value_string, new_value_string);
  if ((strcmp(value_string, "TRUE") == 0) ||
      (strcmp(value_string, "true") == 0))
    my_value.set_float(1);
  else
    my_value.set_float(atof(value_string));

}  // constructor





//
// VsStringConstantNode
//





//
// VsSymbolNode
//

VsSymbolNode::~VsSymbolNode()
{
#ifdef __DEBUG_DESTRUCTORS__
  cout << "VsSymbolNode destructor: " << name << endl;
#endif
  if (name) delete name;
}  // destructor





//
// VsFunctionNode
//

void
VsFunctionNode::parse(Tokenfile *tokenfile_ptr)
{
  if (!tokenfile_ptr->accept_token("function")) return;

  char id[MAX_TOKEN_LENGTH];
  tokenfile_ptr->get_token(id);
  set_name(id);

#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "VsFunctionNode::parse [" << id << "]" << endl;
#endif

  cur_symbol_table->store_state();

  parse_arguments(tokenfile_ptr);

  int prev_size = cur_symbol_table->get_size();
  statements.parse(tokenfile_ptr);

  // now get the local variables from the cur_var_table and
  // add them to local_vars
  int new_size = cur_symbol_table->get_size();
  for(int i=prev_size; i<new_size; i++) {
    local_vars.add((*cur_symbol_table)[i]);
  }  // for

  // now remove arguments and local vars from cur_var_table
  cur_symbol_table->restore_state();

}  // VsFunctionNode::parse



void
VsFunctionNode::parse_arguments(Tokenfile *tokenfile_ptr)
{
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "  parse_arguments" << endl;
#endif

  tokenfile_ptr->accept_token(OPEN_PAREN);

  char token[MAX_TOKEN_LENGTH];
  int done = 0;
  while (!done) {
    int k = tokenfile_ptr->skip_whitespace();
    if (k == ')') done = 1;
    else {
      if (k == ',') tokenfile_ptr->get_token(token);
      tokenfile_ptr->get_token(token);
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
      cout << "  got token [" << token << "]" << endl;
#endif
      VsVariableNode *var_ptr = new VsVariableNode(token);
      cur_symbol_table->add(var_ptr);
      arguments.add(var_ptr);
    }
  }  // while

  tokenfile_ptr->get_token(token);  // get the close paren

}  // VsFunctionNode::parse_arguments



Field *
VsFunctionNode::evaluate(int argc, Field **argv)
{
#ifdef __DEBUG_VRMLSCRIPT__
  cout << "VsFunctionNode::evaluate [" << get_name() << "]" << endl;
#endif

  cur_symbol_table->store_state();

  // set pointers to the argument values
  int nr_arguments = arguments.get_size();
  for(int i=0; i<nr_arguments; i++) {
    ((VsVariableNode *) arguments[i])->set_field_ptr(argv[i]);
    cur_symbol_table->add(arguments[i]);
  }  // for

  // execute the statements (a return statement returns a value)
  Field *return_value = statements.evaluate();

  cur_symbol_table->restore_state();
  return return_value;

}  // VsFunctionNode::evaluate





//
// VsVariableNode
//

VsVariableNode::VsVariableNode(char *new_id)
{
  set_name(new_id);
  // when it gets assigned something, it'll take its type from the r.h.s.
  field_ptr = NULL;
  changed = 0;
}  // constructor just with id name



VsVariableNode::VsVariableNode(char *new_id, 
			       Field *new_field_ptr, FieldType new_type)
{
  set_name(new_id);
  field_ptr = new_field_ptr;
  field_type = new_type;
  changed = 0;
}  // constructor with id, Field and FieldType (e.g. eventIn)



VsVariableNode::VsVariableNode(VsVariableNode& other)
{
  set_name(other.get_name());
  Field *other_field_ptr = other.get_field_ptr();
  if (other_field_ptr) field_ptr = other_field_ptr->create_copy();
  field_type = other.get_field_type();
  changed = 0;
}  // copy constructor



VsVariableNode::~VsVariableNode()
{
  delete field_ptr;
}  // VsVariableNode destructor



void
VsVariableNode::parse(char *field_type_string, int get_default)
{
  if (strcmp(field_type_string, "eventIn") == 0) field_type = eventIn;
  else if (strcmp(field_type_string, "eventOut") == 0) 
    field_type = eventOut;
  else if (strcmp(field_type_string, "exposedField") == 0) 
    field_type = exposedField;
  else if (strcmp(field_type_string, "field") == 0) field_type = field;
  else {
    cur_vrml_file->print_error_line();
    cerr << "Error: unknown field type [" << field_type_string << "]" << endl;
    field_type = field;
  }

  field_ptr = parse_field_type();

  char token[MAX_TOKEN_LENGTH];
  cur_vrml_file->get_token(token);
  set_name(token);

  if (get_default) {  // exposedField and field have a default value
    if ((field_type == exposedField) || (field_type == field)) {
      FieldVarType vrml_var_type = field_ptr->get_field_var_type();
      
      if (vrml_var_type == SFNodeType)
	((SFNode *) field_ptr)->parse_and_store_string();
      else if (vrml_var_type == MFNodeType)
	((MFNode *) field_ptr)->parse_and_store_string();
      else
	field_ptr->parse();  // get the default value
    }
  }  // if we want to get a default (EXTERNPROTO doesn't have defaults)

}  // VsVariableNode::parse



Field *
VsVariableNode::parse_field_type()
{
  char token[MAX_TOKEN_LENGTH];

  cur_vrml_file->get_token(token);

  if (strcmp(token, "SFBool") == 0) return new SFBool;
  else if (strcmp(token, "SFColor") == 0) return new SFColor;
  else if (strcmp(token, "MFColor") == 0) return new MFColor;
  else if (strcmp(token, "SFFloat") == 0) return new SFFloat;
  else if (strcmp(token, "MFFloat") == 0) return new MFFloat;
  else if (strcmp(token, "SFImage") == 0) return new SFImage;
  else if (strcmp(token, "SFInt32") == 0) return new SFInt32;
  else if (strcmp(token, "MFInt32") == 0) return new MFInt32;
  else if (strcmp(token, "SFNode") == 0) return new SFNode;
  else if (strcmp(token, "MFNode") == 0) return new MFNode;
  else if (strcmp(token, "SFRotation") == 0) return new SFRotation;
  else if (strcmp(token, "MFRotation") == 0) return new MFRotation;
  else if (strcmp(token, "SFString") == 0) return new SFString;
  else if (strcmp(token, "MFString") == 0) return new MFString;
  else if (strcmp(token, "SFTime") == 0) return new SFTime;
  else if (strcmp(token, "MFTime") == 0) return new MFTime;
  else if (strcmp(token, "SFVec2f") == 0) return new SFVec2f;
  else if (strcmp(token, "MFVec2f") == 0) return new MFVec2f;
  else if (strcmp(token, "SFVec3f") == 0) return new SFVec3f;
  else if (strcmp(token, "MFVec3f") == 0) return new MFVec3f;
  else {
    cur_vrml_file->print_error_line();
    cerr << "Unknown VRML type [" << token << "], aborting" << endl;
    exit(1);
  }
  return NULL;

}  // VsVariableNode::parse_field_type





//
// VsArrayNode
//

void
VsArrayNode::parse(Tokenfile *tokenfile_ptr)
{
  tokenfile_ptr->accept_token(OPEN_BRACKET);

  index.parse_entry(tokenfile_ptr);

  tokenfile_ptr->accept_token(CLOSE_BRACKET);

}  // VsArrayNode::parse



Field *
VsArrayNode::evaluate()
{
  // terrible hack... :-(
  if (node_ptr->get_node_type() == VsConstructorNodeType)
    set_field_ptr(((VsConstructorNode *) node_ptr)->get_field_ptr());
  else
    set_field_ptr(((VsVariableNode *) node_ptr)->get_field_ptr());

  int int_index = (int)(((SFFloat *) index.evaluate())->get_float());

  return ((MFField *) get_field_ptr())->get_field(int_index);

}  // VsArrayNode::evaluate





//
// VsFunctionCallNode
//

void
VsFunctionCallNode::parse(Tokenfile *tokenfile_ptr)
{
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "  VsFunctionCallNode::parse" << endl;
#endif

  tokenfile_ptr->accept_token(OPEN_PAREN);

  char token[MAX_TOKEN_LENGTH];
  int done = 0;
  while (!done) {
    int k = tokenfile_ptr->skip_whitespace();
    if (k == ')') done = 1;
    else {
      if (k == ',') tokenfile_ptr->get_token(token);
      VsExpression *expression_ptr = new VsExpression;
      expression_ptr->parse_entry(tokenfile_ptr);
      arguments.Insert(expression_ptr);
    }
  }  // while

  tokenfile_ptr->accept_token(CLOSE_PAREN);

}  // VsFunctionCallNode::parse



Field *
VsFunctionCallNode::evaluate()
{
  // first evaluate all parameters
  int nr_parameters = arguments.NEntries();

  actual_arguments = new Field *[nr_parameters];
  // store them in an array of field pointers
  for(int i=0; i<nr_parameters; i++) {
    actual_arguments[i] = arguments.Kth(i)->evaluate();
  }  // for

  // call my function node with that array
  Field *return_value =
    function_ptr->evaluate(nr_parameters, actual_arguments);

  delete actual_arguments;
  return return_value;

}  // VsFunctionCallNode::evaluate





//
// VsBrowserCallNode
//

void
VsBrowserCallNode::parse(Tokenfile *tokenfile_ptr)
{
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "  VsBrowserCallNode::parse" << endl;
#endif

  tokenfile_ptr->accept_token("Browser");
  tokenfile_ptr->accept_token(".");

  char token[MAX_TOKEN_LENGTH];
  tokenfile_ptr->get_token(token);

  function_index = 0;
  while (browser_methods[function_index][0] != '\0') {
    if (strcmp(browser_methods[function_index], token) == 0) break;
    function_index++;
  }  // while

  if (browser_methods[function_index][0] == '\0') {
    function_index = -1;
    tokenfile_ptr->print_error_line();
    cerr << "Unknown Browser method call [" << token << "]" << endl;
  }

  //
  // now get the arguments (maybe none)
  //
  VsFunctionCallNode::parse(tokenfile_ptr);

  // check here for correct number of arguments?
  // correct type?

  return;

}  // VsBrowserCallNode::parse



Field *
VsBrowserCallNode::evaluate()
{
//   const int getName_index = 0;
//   const int getVersion_index = 1;
//   const int getCurrentSpeed_index = 2;
//   const int getCurrentFrameRate_index = 3;
//   const int getWorldURL_index = 4;
//   const int replaceWorld_index = 5;
//   const int createVrmlFromString_index = 6;
//   const int createVrmlFromURL_index = 7;
//   const int addRoute_index = 8;
//   const int deleteRoute_index = 9;
//   const int loadURL_index = 10;
//   const int setDescription_index = 11;

  switch(function_index) {
    case 0:
      return &browser_name;
    case 1:
      return &browser_version;
      
      // get this stuff from cur_vrml_info?  include VrmlInfo.h
      // then it's responsibility of rg to set the values
      // with -ro parameter you just get default values
    case 2:
    case 3:
    case 4:
      // return current main filespec
    case 5:
      // delete the current world
      // loop through this MFNode, call init_Ring and add_Ring
      // on each of them...
    case 6:
      // create a VrmlFile, set the source to be this string
      // and parse it, storing the root nodes in an MFNode
      // return this MFNode
    case 7:
      // if it's a file, do the same thing for this file
    case 8:
      // find the events in the passed nodes, if found
      // add an eventOut to the source node field
    case 9:
      // find the source node field, delete this specific route
    case 10:
      // launch another rg on this URL?
    case 11:
      // just write this to stdout
    default:
      cerr << "VsBrowserCallNode::evaluate error" << endl;
  }  // switch

  return NULL;

}  // VsBrowserCallNode::evaluate


//
// VsMathCallNode
//

void
VsMathCallNode::parse(Tokenfile *tokenfile_ptr)
{
#ifdef __DEBUG_PARSE_VRMLSCRIPT_
  cout << "  VsMathCallNode::parse" << endl;
#endif

  tokenfile_ptr->accept_token("Math");
  tokenfile_ptr->accept_token(".");

  if (parse_function_name(tokenfile_ptr)) {
    // now function pointer is set
    // use parent to parse the arguments
    VsFunctionCallNode::parse(tokenfile_ptr);
  }
  else if (parse_constant_name(tokenfile_ptr)) {
    return;
  }
  else {
    tokenfile_ptr->print_error_line();
    cerr << "Unknown Math object function" << endl;
    my_value.set_float(0);
    return;
  }
}  // VsMathCallNode::parse
  


Field *
VsMathCallNode::evaluate()
{
  if (is_constant) return &my_value;

  if (function_ptr_zero_args)
    my_value.set_float(function_ptr_zero_args());
  else {
    double arg1 = ((SFFloat *) arguments[0]->evaluate())->get_float();
    if (function_ptr_one_arg)
      my_value.set_float(function_ptr_one_arg(arg1));
    else {
      double arg2 = ((SFFloat *) arguments[1]->evaluate())->get_float();
      my_value.set_float(function_ptr_two_args(arg1, arg2));
    }
  }

  return &my_value;

}  // VsMathCallNode::evaluate



int
VsMathCallNode::parse_function_name(Tokenfile *tokenfile_ptr)
{
  char token[MAX_TOKEN_LENGTH];

  tokenfile_ptr->lookahead_token(token);

  function_ptr_zero_args = NULL;
  function_ptr_one_arg = NULL;
  function_ptr_two_args = NULL;

  // ask the compiler why all these have to be cast...
  if (strcmp(token, "abs") == 0) 
    function_ptr_one_arg = (double(*)(double)) &my_abs;
  else if (strcmp(token, "acos") == 0) 
    function_ptr_one_arg = (double(*)(double)) &my_acos;
  else if (strcmp(token, "asin") == 0)
    function_ptr_one_arg = (double(*)(double)) &my_asin;
  else if (strcmp(token, "atan") == 0)
    function_ptr_one_arg = (double(*)(double)) &my_atan;
  else if (strcmp(token, "ceil") == 0)
    function_ptr_one_arg = (double(*)(double)) &my_ceil;
  else if (strcmp(token, "cos") == 0)
    function_ptr_one_arg = (double(*)(double)) &my_cos;
  else if (strcmp(token, "exp") == 0) 
    function_ptr_one_arg = (double(*)(double)) &my_exp;
  else if (strcmp(token, "floor") == 0)
    function_ptr_one_arg = (double(*)(double)) &my_floor;
  else if (strcmp(token, "log") == 0)
    function_ptr_one_arg = (double(*)(double)) &my_log;
  else if (strcmp(token, "max") == 0)
    function_ptr_two_args = (double(*)(double, double)) &my_max;
  else if (strcmp(token, "min") == 0)
    function_ptr_two_args = (double(*)(double, double)) &my_min;
  else if (strcmp(token, "pow") == 0)
    function_ptr_two_args = (double(*)(double, double)) &my_pow;
  else if (strcmp(token, "random") == 0)
    function_ptr_zero_args = (double(*)()) &my_random;
  else if (strcmp(token, "round") == 0)
    function_ptr_one_arg = (double(*)(double)) &my_round;
  else if (strcmp(token, "sin") == 0)
    function_ptr_one_arg = (double(*)(double)) &my_sin;
  else if (strcmp(token, "sqrt") == 0)
    function_ptr_one_arg = (double(*)(double)) &my_sqrt;
  else if (strcmp(token, "tan") == 0)
    function_ptr_one_arg = (double(*)(double)) &my_tan;
  else return 0;

#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "  Math function [" << token << "]" << endl;
#endif

  is_constant = 0;
  tokenfile_ptr->get_token(token);
  return 1;

}  // VsMathCallNode::parse_function_name



int
VsMathCallNode::parse_constant_name(Tokenfile *tokenfile_ptr)
{
  char token[MAX_TOKEN_LENGTH];
  double value;

  tokenfile_ptr->lookahead_token(token);
  if (strcmp(token, "E") == 0) value = M_E;
  else if (strcmp(token, "LN10") == 0) value = M_LN10;
  else if (strcmp(token, "LN2") == 0) value = M_LN2;
  else if (strcmp(token, "PI") == 0) value = M_PI;
  else if (strcmp(token, "SQRT1_2") == 0) value = M_SQRT1_2;
  else if (strcmp(token, "SQRT2") == 0) value = M_SQRT2;
  else return 0;

#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "  Math constant [" << token << "]" << endl;
#endif

  is_constant = 1;
  tokenfile_ptr->get_token(token);
  my_value.set_float(value);
  return 1;

}  // VsMathCallNode::parse_constant_name



double my_abs(double x)
{
  return abs(x);
}  // my_abs



double my_acos(double x)
{
  return acos(x);
}  // my_acos



double my_asin(double x)
{
  return asin(x);
}  // my_asin



double my_atan(double x)
{
  return atan(x);
}  // my_atan



double my_ceil(double x)
{
  return ceil(x);
}  // my_ceil



double my_cos(double x)
{
  return cos(x);
}  // my_cos



double my_exp(double x)
{
  return exp(x);
}  // my_exp



double my_floor(double x)
{
  return floor(x);
}  // my_floor



double my_log(double x)
{
  return log(x);
}  // my_log



double my_max(double x, double y)
{
  if (x > y) return x; else return y;
}  // my_max



double my_min(double x, double y)
{
  if (x < y) return x; else return y;
}  // my_min


double my_pow(double x, double y)
{
  return pow(x, y);
}  // my_pow



double my_random()
{
#if (RN_OS == RN_IRIX)  
  return drand48();
#else
  return (double) rand()/RAND_MAX;
#endif
}  // my_random



double my_round(double x)
{
  return floor(x + 0.5);
}  // my_round


double my_sin(double x)
{
  return sin(x);
}  // my_sin



double my_sqrt(double x)
{
  return sqrt(x);
}  // my_sqrt



double my_tan(double x)
{
  return tan(x);
}  // my_tan





//
// VsConstructorNode
//

void
VsConstructorNode::parse(Tokenfile *tokenfile_ptr)
{
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "VsConstructorNode::parse" << endl;
#endif

  char token[MAX_TOKEN_LENGTH];
  tokenfile_ptr->get_token(token);
  
  if (strcmp(token, "SFColor") == 0) field_var_type = SFColorType;
  else if (strcmp(token, "MFColor") == 0) field_var_type = MFColorType;
  else if (strcmp(token, "MFFloat") == 0) field_var_type = MFFloatType;
  else if (strcmp(token, "SFImage") == 0) field_var_type = SFImageType;
  else if (strcmp(token, "MFInt32") == 0) field_var_type = MFInt32Type;
  else if (strcmp(token, "SFNode") == 0) field_var_type = SFNodeType;
  else if (strcmp(token, "MFNode") == 0) field_var_type = MFNodeType;
  else if (strcmp(token, "SFRotation") == 0) field_var_type = SFRotationType;
  else if (strcmp(token, "MFRotation") == 0) field_var_type = MFRotationType;
  // String, not SFString...
  else if (strcmp(token, "String") == 0) field_var_type = SFStringType;
  else if (strcmp(token, "MFString") == 0) field_var_type = MFStringType;
  else if (strcmp(token, "SFVec2f") == 0) field_var_type = SFVec2fType;
  else if (strcmp(token, "MFVec2f") == 0) field_var_type = MFVec2fType;
  else if (strcmp(token, "SFVec3f") == 0) field_var_type = SFVec3fType;
  else if (strcmp(token, "MFVec3f") == 0) field_var_type = MFVec3fType;
  else {
    tokenfile_ptr->print_error_line();
    cerr << "constructor for type [" << token << "] is not available" << endl;
    field_ptr = new SFFloat;
  }

  // now get the arguments
  VsFunctionCallNode::parse(tokenfile_ptr);

  // check if the number of arguments is correct

}  // VsConstructorNode::parse



Field *
VsConstructorNode::evaluate()
{
  int nr_arguments = arguments.NEntries();

  switch (field_var_type) {
    case SFColorType: {
      SFColor *sfcolor_ptr= new SFColor;
      field_ptr = sfcolor_ptr;
      sfcolor_ptr->set_colour(((SFFloat *) arguments[0]->evaluate())->get_float(),
            ((SFFloat *) arguments[1]->evaluate())->get_float(),
            ((SFFloat *) arguments[2]->evaluate())->get_float());
      return sfcolor_ptr;
    }
    case MFColorType: {
      MFColor *mfcolor_ptr = new MFColor;
      field_ptr = mfcolor_ptr;
      // add each argument, evaluate as SFColor
      for(int i=0; i<nr_arguments; i++) {
	mfcolor_ptr->add((SFColor *) arguments[i]->evaluate());
      }
      return mfcolor_ptr;
    }
    case MFFloatType: {
      MFFloat *mffloat_ptr = new MFFloat;
      field_ptr = mffloat_ptr;
      for(int i=0; i<nr_arguments; i++) {
	mffloat_ptr->add((SFFloat *) arguments[i]->evaluate());
      }
      return mffloat_ptr;
    }
    case SFImageType: {
      // args: SFInt32 w, SFInt32 h, SFInt32 comp, MFInt32 array
      int width  = ((SFInt32 *) arguments[0]->evaluate())->get_integer();
      int height = ((SFInt32 *) arguments[1]->evaluate())->get_integer();
      int depth  = ((SFInt32 *) arguments[2]->evaluate())->get_integer();
      MFInt32 *mfint32_ptr = (MFInt32 *) arguments[3]->evaluate();
      // now create data array, and copy MFInt32 values in there...
      // terribly inefficient :-(
      int image_size = width * height * depth;
      unsigned char *data = new unsigned char[image_size];
      for(int i=0; i<image_size; i++) {
	data[i] = (unsigned char) (*mfint32_ptr)[i]->get_integer();
      }
      R2ColorImage *image_ptr = new R2ColorImage(width, height, depth, 
             (void *) data);
      SFImage *sfimage_ptr = new SFImage;
      field_ptr = sfimage_ptr;
      sfimage_ptr->set_image(image_ptr);
      return sfimage_ptr;
    }
    case MFInt32Type: {
      MFInt32 *mfint32_ptr = new MFInt32;
      field_ptr = mfint32_ptr;
      for(int i=0; i<nr_arguments; i++) {
	mfint32_ptr->add((SFInt32 *) arguments[i]->evaluate());
      }
      return mfint32_ptr;
    }
    case SFNodeType: {
      // args: SFString def, set the string of the SFNode and parse it
      char *string_ptr = ((SFString *) arguments[0]->evaluate())->get_string();
      SFNode *node_ptr = new SFNode;
      field_ptr = node_ptr;
      node_ptr->set_my_string(string_ptr);
      return node_ptr;
    }
    case MFNodeType: {
      // args: SFNode
      MFNode *mfnode_ptr = new MFNode;
      field_ptr = mfnode_ptr;
      for(int i=0; i<nr_arguments; i++) {
	char *string_ptr = 
	  ((SFString *) arguments[0]->evaluate())->get_string();
	SFNode *node_ptr = new SFNode;
	node_ptr->set_my_string(string_ptr);
	mfnode_ptr->add(node_ptr);
      }  // for
      return mfnode_ptr;
    }
    case SFRotationType: {
      //
      // args: SFFloat x, SFFloat y, SFFloat z, SFFloat angle
      //
      // NOTE: other constructors are possible:
      // SFVec3f axis, SFFloat angle
      // SFVec3f fromvector, SFVec3f tovector
      //
      double x = ((SFFloat *) arguments[0]->evaluate())->get_float();
      double y = ((SFFloat *) arguments[1]->evaluate())->get_float();
      double z = ((SFFloat *) arguments[2]->evaluate())->get_float();
      double angle = ((SFFloat *) arguments[3]->evaluate())->get_float();

      SFRotation *sfrotation_ptr = new SFRotation;
      field_ptr = sfrotation_ptr;
      sfrotation_ptr->set_axis(x, y, z);
      sfrotation_ptr->set_angle(angle);
      return sfrotation_ptr;
    }
    case MFRotationType: {
      MFRotation *mfrotation_ptr = new MFRotation;
      field_ptr = mfrotation_ptr;
      for(int i=0; i<nr_arguments; i++) {
	SFRotation *sfrotation_ptr = (SFRotation *) arguments[i]->evaluate();
	mfrotation_ptr->add(sfrotation_ptr);
      }
      return mfrotation_ptr;
    }
    case SFStringType: {
      // arg: either a number or a string...
      Field *arg_ptr = arguments[0]->evaluate();
      char value_string[MAX_TOKEN_LENGTH];
      SFString *sfstring_ptr = new SFString;
      field_ptr = sfstring_ptr;
      if (arg_ptr->get_field_var_type() == SFFloatType) {
	double value = ((SFFloat *) arg_ptr)->get_float();
	sprintf(value_string, "%f", value);
	sfstring_ptr->set_string(value_string);
	return sfstring_ptr;
      }
      else if (arg_ptr->get_field_var_type() == SFInt32Type) {
	int value = ((SFInt32 *) arg_ptr)->get_integer();
	sprintf(value_string, "%d", value);
	sfstring_ptr->set_string(value_string);
	return sfstring_ptr;
      }
      else {
	sfstring_ptr->set_string(((SFString *) arg_ptr)->get_string());
	return sfstring_ptr;
      }
    }
    case MFStringType: {
      MFString *mfstring_ptr = new MFString;
      field_ptr = mfstring_ptr;
      for(int i=0; i<nr_arguments; i++) {
	SFString *sfstring_ptr = (SFString *) arguments[i]->evaluate();
	mfstring_ptr->add(sfstring_ptr);
      }
      return mfstring_ptr;
    }
    case SFVec2fType: {
      // args: SFFloat x, SFFloat y
      Float x = ((SFFloat *) arguments[0]->evaluate())->get_float();
      Float y = ((SFFloat *) arguments[1]->evaluate())->get_float();
      SFVec2f *sfvec2f_ptr = new SFVec2f;
      field_ptr = sfvec2f_ptr;
      sfvec2f_ptr->set_vector(x, y);
      return sfvec2f_ptr;
    }
    case MFVec2fType: {
      MFVec2f *mfvec2f_ptr = new MFVec2f;
      field_ptr = mfvec2f_ptr;
      for(int i=0; i<nr_arguments; i++) {
	SFVec2f *sfvec2f_ptr = (SFVec2f *) arguments[i]->evaluate();
	mfvec2f_ptr->add(sfvec2f_ptr);
      }
      return mfvec2f_ptr;
    }
    case SFVec3fType: {
      // args: SFFloat x, SFFloat y, SFFloat z
      Float x = ((SFFloat *) arguments[0]->evaluate())->get_float();
      Float y = ((SFFloat *) arguments[1]->evaluate())->get_float();
      Float z = ((SFFloat *) arguments[2]->evaluate())->get_float();
      SFVec3f *sfvec3f_ptr = new SFVec3f;
      field_ptr = sfvec3f_ptr;
      sfvec3f_ptr->set_vector(x, y, z);
      return sfvec3f_ptr;
    }
    case MFVec3fType: {
      MFVec3f *mfvec3f_ptr = new MFVec3f;
      field_ptr = mfvec3f_ptr;
      for(int i=0; i<nr_arguments; i++) {
	SFVec3f *sfvec3f_ptr = (SFVec3f *) arguments[i]->evaluate();
	mfvec3f_ptr->add(sfvec3f_ptr);
      }
      return mfvec3f_ptr;
    }
  }  // switch

  cerr << "VsConstructorNode::evaluate: Runtime error: unsupported constructor"
       << endl;
  return NULL;

}  // VsConstructorNode::evaluate


